Package: stars
Title: Spatiotemporal Arrays, Raster and Vector Data Cubes
Version: 0.7-3
Authors@R: 
    c(person(given = "Edzer",
             family = "Pebesma",
             role = c("aut", "cre"),
             email = "edzer.pebesma@uni-muenster.de",
             comment = c(ORCID = "0000-0001-8049-7069")),
      person(given = "Michael",
             family = "Sumner",
             role = "ctb", 
             comment=c(ORCID = "0000-0002-2471-7511")),
      person(given = "Etienne",
             family = "Racine",
             role = "ctb"),
      person(given = "Adriano",
             family = "Fantini",
             role = "ctb"), 
      person(given = "David", 
             family = "Blodgett", 
             role = "ctb"), 
      person(given = "Krzysztof", 
             family = "Dyba", 
             role = "ctb",
             comment=c(ORCID = "0000-0002-8614-3816")))
Description: Reading, manipulating, writing and plotting
    spatiotemporal arrays (raster and vector data cubes) in 'R', using 'GDAL'
    bindings provided by 'sf', and 'NetCDF' bindings by 'ncmeta' and 'RNetCDF'.
License: Apache License
URL: https://r-spatial.github.io/stars/,
        https://github.com/r-spatial/stars/
BugReports: https://github.com/r-spatial/stars/issues/
Additional_repositories: https://cran.uni-muenster.de/pebesma/
LazyData: true
Depends: R (>= 4.1.0), abind, sf (>= 1.0-19)
Imports: methods, parallel, rlang, units
Suggests: Cairo, CFtime (>= 1.7.1), classInt (>= 0.4-1), clue, covr,
        cubble (>= 0.3.0), cubelyr, digest, dplyr (>= 0.7-0),
        exactextractr, FNN, future.apply, ggforce, ggplot2, ggthemes,
        gstat, httr, jsonlite, knitr, lwgeom, maps, mapdata, ncdfCF,
        ncdfgeom, ncmeta (>= 0.0.3), OpenStreetMap, pbapply, plm,
        randomForest, raster, rmarkdown, RNetCDF (>= 1.8-2), sp,
        spacetime, spatstat (>= 2.0-1), spatstat.geom, starsdata, terra
        (>= 1.4-22), testthat, tibble, tidyr, tsibble, viridis, xts,
        zoo
VignetteBuilder: knitr
Encoding: UTF-8
Collate: 'aggregate.R' 'cf.R' 'cubble.R' 'datasets.R' 'dimensions.R'
        'extract.R' 'factors.R' 'geom.R' 'gt.R' 'init.R' 'intervals.R'
        'mdim.R' 'mosaic.R' 'ncdf.R' 'ncproxy.R' 'OpenStreetMap.R'
        'ops.R' 'plot.R' 'prcomp.R' 'proxy.R' 'rasterize.R' 'raster.R'
        'read.R' 'rotate.R' 'sample.R' 'sf.R' 'spacetime.R'
        'spatstat.R' 'sp.R' 'stars.R' 'subset.R' 'tidyverse.R' 'tile.R'
        'transform.R' 'values.R' 'warp.R' 'write.R' 'xts.R'
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-07-20 14:44:20 UTC; edzer
Author: Edzer Pebesma [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-8049-7069>),
  Michael Sumner [ctb] (ORCID: <https://orcid.org/0000-0002-2471-7511>),
  Etienne Racine [ctb],
  Adriano Fantini [ctb],
  David Blodgett [ctb],
  Krzysztof Dyba [ctb] (ORCID: <https://orcid.org/0000-0002-8614-3816>)
Maintainer: Edzer Pebesma <edzer.pebesma@uni-muenster.de>
Repository: CRAN
Date/Publication: 2026-07-20 21:00:02 UTC
Built: R 4.5.2; ; 2026-07-20 21:59:58 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method("$<-",stars)
S3method("[",dimension)
S3method("[",dimensions)
S3method("[",intervals)
S3method("[",intervals_list)
S3method("[",mdim)
S3method("[",nc_proxy)
S3method("[",stars)
S3method("[",stars_proxy)
S3method("[<-",dimensions)
S3method("[<-",stars)
S3method("[<-",stars_proxy)
S3method("[[<-",stars)
S3method("[[<-",stars_proxy)
S3method("dimnames<-",stars)
S3method("st_crs<-",dimensions)
S3method("st_crs<-",stars)
S3method("st_dimensions<-",list)
S3method("st_dimensions<-",stars)
S3method("st_dimensions<-",stars_proxy)
S3method("st_geotransform<-",stars)
S3method(Math,stars)
S3method(Math,stars_proxy)
S3method(Ops,stars)
S3method(Ops,stars_proxy)
S3method(adrop,stars)
S3method(adrop,stars_proxy)
S3method(aggregate,stars)
S3method(aggregate,stars_proxy)
S3method(all.equal,dimensions)
S3method(aperm,stars)
S3method(aperm,stars_proxy)
S3method(as.Date,stars)
S3method(as.POSIXct,stars)
S3method(as.data.frame,dimensions)
S3method(as.data.frame,stars)
S3method(as.data.frame,stars_proxy)
S3method(as.list,intervals)
S3method(asub,factor)
S3method(c,intervals)
S3method(c,nc_proxy)
S3method(c,stars)
S3method(c,stars_proxy)
S3method(contour,stars)
S3method(cut,array)
S3method(cut,matrix)
S3method(cut,stars)
S3method(dim,dimensions)
S3method(dim,stars)
S3method(dim,stars_proxy)
S3method(dimnames,stars)
S3method(drop_units,dimension)
S3method(drop_units,stars)
S3method(droplevels,stars)
S3method(droplevels,stars_proxy)
S3method(expand_dimensions,dimensions)
S3method(expand_dimensions,stars)
S3method(format,intervals)
S3method(format,intervals_list)
S3method(head,intervals)
S3method(hist,stars)
S3method(hist,stars_proxy)
S3method(image,stars)
S3method(image,stars_proxy)
S3method(is.na,stars)
S3method(is.na,stars_proxy)
S3method(length,intervals)
S3method(merge,stars)
S3method(merge,stars_proxy)
S3method(plot,nc_proxy)
S3method(plot,stars)
S3method(plot,stars_proxy)
S3method(prcomp,stars)
S3method(prcomp,stars_proxy)
S3method(predict,stars)
S3method(predict,stars_proxy)
S3method(print,dimensions)
S3method(print,nc_proxy)
S3method(print,stars)
S3method(print,stars_proxy)
S3method(print,stars_raster)
S3method(range,dimension)
S3method(seq,dimension)
S3method(split,stars)
S3method(split,stars_proxy)
S3method(st_apply,stars)
S3method(st_apply,stars_proxy)
S3method(st_area,stars)
S3method(st_as_sf,dimensions)
S3method(st_as_sf,stars)
S3method(st_as_sf,stars_proxy)
S3method(st_as_sfc,stars)
S3method(st_as_stars,CFDataset)
S3method(st_as_stars,CFVariable)
S3method(st_as_stars,OpenStreetMap)
S3method(st_as_stars,Raster)
S3method(st_as_stars,STF)
S3method(st_as_stars,STFDF)
S3method(st_as_stars,STS)
S3method(st_as_stars,STSDF)
S3method(st_as_stars,SpatRaster)
S3method(st_as_stars,Spatial)
S3method(st_as_stars,bbox)
S3method(st_as_stars,cubble_df)
S3method(st_as_stars,data.frame)
S3method(st_as_stars,default)
S3method(st_as_stars,im)
S3method(st_as_stars,list)
S3method(st_as_stars,mdim)
S3method(st_as_stars,nc_proxy)
S3method(st_as_stars,ncdfgeom)
S3method(st_as_stars,sf)
S3method(st_as_stars,sfc)
S3method(st_as_stars,stars)
S3method(st_as_stars,stars_proxy)
S3method(st_as_stars,xts)
S3method(st_bbox,default)
S3method(st_bbox,dimensions)
S3method(st_bbox,stars)
S3method(st_coordinates,dimensions)
S3method(st_coordinates,stars)
S3method(st_crop,mdim)
S3method(st_crop,nc_proxy)
S3method(st_crop,stars)
S3method(st_crop,stars_proxy)
S3method(st_crs,dimensions)
S3method(st_crs,stars)
S3method(st_dimensions,array)
S3method(st_dimensions,default)
S3method(st_dimensions,dimensions)
S3method(st_dimensions,matrix)
S3method(st_dimensions,stars)
S3method(st_downsample,stars)
S3method(st_downsample,stars_proxy)
S3method(st_extract,stars)
S3method(st_geometry,stars)
S3method(st_geotransform,default)
S3method(st_geotransform,dimensions)
S3method(st_geotransform,stars)
S3method(st_interpolate_aw,stars)
S3method(st_intersects,bbox)
S3method(st_intersects,stars)
S3method(st_join,stars)
S3method(st_mosaic,character)
S3method(st_mosaic,nc_proxy)
S3method(st_mosaic,stars)
S3method(st_mosaic,stars_proxy)
S3method(st_normalize,stars)
S3method(st_normalize,stars_proxy)
S3method(st_redimension,nc_proxy)
S3method(st_redimension,stars)
S3method(st_redimension,stars_proxy)
S3method(st_rotate,sf)
S3method(st_rotate,sfc)
S3method(st_rotate,stars)
S3method(st_sample,stars)
S3method(st_sample,stars_proxy)
S3method(st_set_bbox,dimensions)
S3method(st_set_bbox,stars)
S3method(st_set_bbox,stars_proxy)
S3method(st_transform,stars)
S3method(st_write,stars)
S3method(tail,intervals)
S3method(time,stars)
S3method(units,dimension)
S3method(write_stars,stars)
S3method(write_stars,stars_proxy)
export("st_dimensions<-")
export("st_geotransform<-")
export(as.tbl_cube.stars)
export(detect.driver)
export(expand_dimensions)
export(geom_stars)
export(make_intervals)
export(read_mdim)
export(read_ncdf)
export(read_stars)
export(st_apply)
export(st_as_stars)
export(st_cells)
export(st_contour)
export(st_dim_to_attr)
export(st_dimensions)
export(st_downsample)
export(st_extract)
export(st_flip)
export(st_geotransform)
export(st_get_dimension_values)
export(st_mosaic)
export(st_raster_type)
export(st_rasterize)
export(st_redimension)
export(st_res)
export(st_rgb)
export(st_rotate)
export(st_set_bbox)
export(st_set_dimensions)
export(st_sfc2xy)
export(st_tile)
export(st_warp)
export(st_xy2sfc)
export(write_mdim)
export(write_stars)
exportMethods("%in%")
import(sf)
import(units)
importFrom(abind,abind)
importFrom(abind,adrop)
importFrom(abind,asub)
importFrom(grDevices,cm)
importFrom(grDevices,col2rgb)
importFrom(grDevices,dev.capabilities)
importFrom(grDevices,dev.size)
importFrom(grDevices,grey)
importFrom(grDevices,rgb)
importFrom(graphics,axis)
importFrom(graphics,box)
importFrom(graphics,contour)
importFrom(graphics,hist)
importFrom(graphics,image)
importFrom(graphics,image.default)
importFrom(graphics,layout)
importFrom(graphics,lcm)
importFrom(graphics,par)
importFrom(graphics,plot)
importFrom(graphics,plot.new)
importFrom(graphics,plot.window)
importFrom(graphics,rasterImage)
importFrom(graphics,strwidth)
importFrom(graphics,text)
importFrom(graphics,title)
importFrom(methods,as)
importFrom(methods,new)
importFrom(methods,slot)
importFrom(methods,slotNames)
importFrom(parallel,parApply)
importFrom(rlang,"%||%")
importFrom(stats,aggregate)
importFrom(stats,complete.cases)
importFrom(stats,na.omit)
importFrom(stats,na.pass)
importFrom(stats,prcomp)
importFrom(stats,predict)
importFrom(stats,quantile)
importFrom(stats,runif)
importFrom(stats,setNames)
importFrom(stats,time)
importFrom(stats,var)
importFrom(tools,file_ext)
importFrom(utils,head)
importFrom(utils,methods)
importFrom(utils,modifyList)
importFrom(utils,packageVersion)
importFrom(utils,setTxtProgressBar)
importFrom(utils,tail)
importFrom(utils,txtProgressBar)

# lc.tif and lc.tif.aux.xml were generated from
# https://github.com/mtennekes/tmap/issues/368#issuecomment-595727166
# https://s3-us-west-2.amazonaws.com/mrlc/PR_landcover_wimperv_10-28-08_se5.zip
# with:
gdalwarp -tr 3000 -3000 pr_landcover_wimperv_10-28-08_se5.img lc.tif

library(ncdf4)
nc = nc_open("reduced.nc", write=TRUE)
ncatt_put(nc, 'sst', 'units', 'degree_C')
ncatt_put(nc, 'anom', 'units', 'degree_C')
ncatt_put(nc, 'err', 'units', 'degree_C')
ncatt_put(nc, 'ice', 'units', 'percent')
nc_sync(nc)
nc_close(nc)

### The functions in this file are a work in progress and are ###
### not currently used by `stars`                             ###

units_or_null <- function(u) {
  un = try(suppressWarnings(as_units(u)), silent = TRUE)
  if (inherits(un, "try-error")) {
    #warning(paste("ignoring unrecognized unit:", u), call. = FALSE)
    NULL
  } else
    un
}
unit_izer_ncmeta <- function(attribute, variable, ...) {
  unit_df <- attribute[attribute$name == "units", ]
  if (is.null(unit_df) || nrow(unit_df) < 1) return(NULL)
  unit <- lapply(unit_df$value, function(x) units_or_null(x[[1]]))
  unit_df$unit <- unit
  unit_df$unit_ok <- !unlist(lapply(unit, is.null))
  variable$unit <- unit_df$unit[match(variable$name, unit_df$variable)]
  variable
}
# TODO
# - allow stars_proxy to be pulled
# - allow stars_proxy to apply select_var
# - 
# examples
# f <- system.file("nc/reduced.nc", package = "stars")
# read_stars_tidync(f)
# read_stars_tidync(f, select_var = "anom", proxy = FALSE) ## only works if proxy = FALSE
# read_stars_tidync(f, lon = index <= 10, lat = index <= 12, time = index < 2)
#' @importFrom units set_units
read_stars_tidync = function(.x, ..., select_var = NULL, proxy = TRUE, make_time = TRUE, make_units = TRUE) {
  if (!requireNamespace("tidync", quietly = TRUE))
    stop("package tidync required, please install it first") # nocov
  
  if (length(.x) > 1) {
    warning("only first source/file used")
    .x = .x[1L]
  }
  coord_var = ncmeta::nc_coord_var(.x)
  tnc =  tidync::tidync(.x) %>%  tidync::hyper_filter(...)
  if (!is.null(select_var)) tnc = tidync::activate(tnc, select_var[1])
  ## FIXME: get this stuff from tidync object
  variable = unit_izer_ncmeta(ncmeta::nc_atts(.x), ncmeta::nc_vars(.x))

  tt = tidync::hyper_transforms(tnc, all = FALSE)
  tt =  lapply(tt, function(tab) tab[tab$selected, , drop = FALSE])
  nms = names(tt)
  
  if (make_units) {
    for (idim in seq_along(tt)) {
      if (nms[idim] %in% variable$name) {
        uit <- variable$unit[match(names(tt)[idim], variable$name)][[1]]
        if (nms[idim] %in% coord_var$variable && !nms[idim] %in% coord_var[["T"]]) {
          
      ## we have a standard units thing
      if (!is.null(uit)) {
        tt[[idim]][[nms[idim]]] = units::as_units(tt[[idim]][[nms[idim]]], uit)
      }
      
        }
      }}}
  
  curvilinear = character(0)
  ## are we curvilinear?
  XY_curvi <- NULL
  if (all(c("X", "Y") %in% names(coord_var))) {
    cvar = coord_var[coord_var$variable == tidync::hyper_vars(tnc)$name[1], ]
    XY_curvi = unlist(cvar[1, c("X", "Y")])
   
    if (all(!is.na(XY_curvi)) && all(tnc$variable$ndims[match(XY_curvi, tnc$variable$name)] == 2)) {
      ## AAND these are both 2d vars
      curvilinear = XY_curvi
    } else {
      XY_curvi <- character(0)
    }
  }
  curvi_coords = NULL
  if (length(curvilinear == 2)) {
   #grid = tnc$grid %>% tidyr::unnest()
   grid = tnc$grid %>% tidyr::unnest(cols = c(variables))
   curvi_coords = tidync::hyper_array(tnc %>%
                                       ## FIXME ...
                                       tidync::activate((grid %>% 
                                                           dplyr::filter(variable == XY_curvi[1]) %>% 
                                                           dplyr::pull(.data$grid))[1L]), select_var = XY_curvi)
  
   names(curvi_coords)[1:2] <- curvilinear
  }
  ## can we create a raster?
  raster = NULL
  if (length(tt) > 1) {
    raster = get_raster(affine = c(0, 0),
                        dimensions = names(tt)[1:2], curvilinear = length(curvilinear) == 2)
  }
  dims = vector("list", length(nms))
  for (i in seq_along(dims)) {
    nm = nms[i]
    if (nrow(tt[[nm]]) > 1) { ## we are rectilinear, or degenerate rectilinear
      dims[[i]] = create_dimension(values = tt[[nm]][[nm]])
    } else { ## we are simple offset
      uval = tt[[nm]][[nm]]  ## so we can get an NA, whether units of not
      
      dims[[i]] = create_dimension(from = 1L, to = 1L, offset = uval, delta = uval[NA])
    }
  }
  dims = create_dimensions(setNames(dims, nms), raster = raster)
  if (!is.null(curvi_coords)) {

    dims[[names(tt)[1]]]$values = curvi_coords[[1]]
    dims[[names(tt)[1]]]$offset = NA
    dims[[names(tt)[1]]]$delta = NA
    
    dims[[names(tt)[2]]]$values = curvi_coords[[2]]
    dims[[names(tt)[2]]]$offset = NA
    dims[[names(tt)[2]]]$delta = NA
    
    
  }
  if (make_time) {
   for (idim in seq_along(tt)) {
     if (nms[idim] %in% variable$name) {
       uit <- variable$unit[match(names(tt)[idim], variable$name)][[1]]
       if (nms[idim] %in% coord_var$variable && nms[idim] %in% coord_var[["T"]]) {
         ## we have a time thing
         dims[[idim]]  = make_cal_time2(dims[[idim]], nms[idim], uit)
       } 
     }
   }
  }
    
  
  
  if (!proxy) {
    x = tnc %>% 
       tidync::hyper_array(select_var = select_var, drop = FALSE)  ## always keep degenerate dims
    attr(x, "transforms") = NULL
    attr(x, "source") = NULL
    class(x) = "list"
    for (ivar in seq_along(x)) {
      if (names(x)[ivar] %in% variable$name) {
        uit <- variable$unit[match(names(x)[ivar], variable$name)][[1]]
        if (!is.null(uit)) x[[ivar]] <- units::as_units(x[[ivar]], uit)
      }
    }
    out = st_stars(x, dims)
    out = st_as_stars(out, curvilinear = curvi_coords)
    
    
  } else {

    ## if we are proxy, defer to GDAL subdatasets
    hvars = if (is.null(select_var)) tidync::hyper_vars(tnc)$name else select_var
    sds = unlist(sf::gdal_subdatasets(.x))
    if (length(sds) > 0)  {
      sds = unlist(lapply(sprintf(":%s", hvars), function(p) grep(p, sds, value = TRUE)))
    } else {
      sds = .x
    }
    out = stats::setNames(as.list(sds), hvars)
    out = st_stars_proxy(out, dims, NA_value = NA)
    #class(out) = c("ncdf_proxy", class(out))
  }
  
  
  out
}

## FIXME: this function should return unit-ed times, not modify dimensions
## it should also live in the units package (or RNetCDF, or PCICt or ...)
make_cal_time <- function(dimensions, time_name, time_unit = NULL, cal = NULL) {
  td = which(names(dimensions) == time_name)
  if (length(td) == 1) {
    tm = get_dimension_values(dimensions[[time_name]], geotransform = NA)
    u = time_unit
    if (is.null(u) || inherits(u, "try-error")) {
      warning("ignoring units of time dimension, not able to interpret")
      return(dimensions)
    }
    
    if (! is.null(cal)  && cal %in% c("360_day", "365_day", "noleap")) {
      if (!requireNamespace("PCICt", quietly = TRUE))
        stop("package PCICt required, please install it first") # nocov
      t01 = set_units(0:1, u, mode = "standard")
      delta = if (grepl("months", u)) {
        if (cal == "360_day")
          set_units(30 * 24 * 3600, "s", mode = "standard")
        else
          set_units((365/12) * 24 * 3600, "s", mode = "standard")
      } else
        set_units(as_units(diff(as.POSIXct(t01))), "s", mode = "standard")
      origin = as.character(as.POSIXct(t01[1]))
      v.pcict = PCICt::as.PCICt(tm * as.numeric(delta), cal, origin)
      if (!is.null(dimensions[[td]]$values)) {
        v = dimensions[[td]]$values
        if (inherits(v, "intervals")) {
          start = PCICt::as.PCICt(v$start * as.numeric(delta), cal, origin)
          end =   PCICt::as.PCICt(v$end   * as.numeric(delta), cal, origin)
          dimensions[[td]]$values = make_intervals(start, end)
        } else
          dimensions[[td]]$values = v.pcict
      } else {
        dimensions[[td]]$offset = v.pcict[1]
        dimensions[[td]]$delta = diff(v.pcict[1:2])
      }
      dimensions[[td]]$refsys = "PCICt"
    } else { # Gregorian/Julian, POSIXct:
      if (!is.null(dimensions[[td]]$values)) {
        v = dimensions[[td]]$values
        if (inherits(v, "intervals")) {
          start = as.POSIXct(units::set_units(v$start, u, mode = "standard")) # or: RNetCDF::utcal.nc(u, tm, "c")
          end =   as.POSIXct(units::set_units(v$end,   u, mode = "standard")) # or: RNetCDF::utcal.nc(u, tm, "c")
          dimensions[[td]]$values = make_intervals(start, end)
        } else
          dimensions[[td]]$values = as.POSIXct(units::set_units(tm, u, mode = "standard")) # or: RNetCDF::utcal.nc(u, tm, "c")
      } else {
        t0 = dimensions[[td]]$offset
        t1 = dimensions[[td]]$offset + dimensions[[td]]$delta
        t.posix = as.POSIXct(units::set_units(c(t0, t1), u, mode = "standard")) # or: utcal.nc(u, c(t0,t1), "c")
        dimensions[[td]]$offset = t.posix[1]
        dimensions[[td]]$delta = diff(t.posix)
      }
      dimensions[[td]]$refsys = "POSIXct"
    }
  }
  dimensions
}

library(stars)

# wind:
library(gstat)
data(wind)
library(sp)
library(sf)

wind.loc$y = as.numeric(char2dms(as.character(wind.loc[["Latitude"]])))
wind.loc$x = as.numeric(char2dms(as.character(wind.loc[["Longitude"]])))
locs = st_as_sf(wind.loc, coords = c("x", "y"), crs = st_crs(4326))


# match order:

data = as.data.frame(t(wind[,-c(1:3)]))
data$geom = st_geometry(locs[match(names(wind)[-c(1:3)], locs$Code),])
wind_ = st_sf(data)
wind.st = merge(st_as_stars(wind_), times = ISOdate(wind$year+1900, wind$month, wind$day, 0))
plot(wind.st)

# Produc:
data(Produc, package = "plm")

library(maps)
states = st_as_sf(map('state', plot=FALSE, fill=TRUE))[-8,]
yrs = 1970:1986
#time = as.POSIXct(paste(yrs, "-01-01", sep=""), tz = "GMT")

Pr = lapply(Produc[order(Produc[2], Produc[1]),-(1:3)], matrix, nrow = nrow(states))
(Produc.st = st_as_stars(Pr, 
	dimensions = st_dimensions(sfc = st_geometry(states[-8]), times = yrs)))

plot(Produc.st[1], max.plot = 17)
plot(Produc.st[2], max.plot = 17)

# gridded example:
library(gstat)
example(krigeST)
library(stars)
s = st_as_stars(DE_kriged)
stplot(DE_kriged,scales=list(draw=T))
plot(s, axes = TRUE)
f = as(s, "STFDF")
# roundtrip:
summary(f@data - DE_kriged@data)

library(httr)     # GET
library(jsonlite) # base64_dec
library(stars)

url = "http://localhost:8000/data"

get_data = function(url, expr = NULL) {
	if (is.null(expr))
		fromJSON( content(GET(url), "text", encoding = "UTF-8"))
	else {
		url = paste0(url, "?expr=", expr)
		unserialize(base64_dec(fromJSON( content(GET(url), "text", encoding = "UTF-8"))))
	}
}

put_data = function(url, name, value) {
	value = toJSON(base64_enc(serialize(value, NULL)))
	PUT(url, body = list(name = name, value = value), encode = "json")
}

get_data(url)
get_data(url, "y")
get_data(url, "z")

xx = get_data(url, "x")

tif = system.file("tif/L7_ETMs.tif", package = "stars")
x = read_stars(tif)

all.equal(x, xx)

#get_data(url, "x[,,,1]")
#get_data(url, "adrop(x[,,,1])")

put_data(url, "foo", "bar")
get_data(url)
get_data(url, "foo")
put_data(url, "z3", "3 * z")
get_data(url, "z3")

#put_data(url, "tif", x)
#get_data(url)

put_data(url, "xx", matrix(1:4,2))
get_data(url)
get_data(url, "xx")

library(sf)
pt = st_sf(a = 1, geom = st_sfc(st_point(3:4)))
put_data(url, "xx", pt)
get_data(url)
get_data(url, "xx")

library(httr)     # GET, POST, PUT
library(jsonlite) # base64_dec, base64_enc, toJSON, fromJSON
library(tibble)   # print

get_data = function(url, expr = NULL) {
	if (is.null(expr))
		fromJSON( content(GET(url), "text", encoding = "UTF-8"))
	else
		unserialize(base64_dec(fromJSON( 
			content(POST(url, body = list(expr = expr), encode = "json"),
				"text", encoding = "UTF-8")
		)))
}

put_data = function(url, name, value) {
	value = toJSON(base64_enc(serialize(value, NULL)))
	PUT(url, body = list(name = name, value = value), encode = "json")
}

# from where?
url = "http://localhost:8000/data"

get_data(url) # list

library(stars)
md = get_data(url, "md")

# select a country:
nl = st_as_sf(raster::getData("GADM", country = "NLD", level = 0)) %>%
	st_transform(st_crs(md))

get_data(url) # any data there?
md = get_data(url, "md")

proxy = md[nl,]$proxy

plot(st_geometry(nl), axes = TRUE)
s = sapply(proxy, function(x) {
	r = st_as_stars(x, downsample = 100, url = url)
	image(r, add = TRUE, rgb = 1:3, maxColorValue = 30000)
})
plot(st_geometry(md), border = 'yellow', add = TRUE, lwd = .8)
plot(st_geometry(nl), add=TRUE, border = 'orange', lwd = 2)

ndvi = function(x) (x[4]-x[1])/(x[4]+x[1])
put_data(url, "ndvi", ndvi)
get_data(url, "ndvi")
plot(st_geometry(nl), axes = TRUE)

x = st_apply(proxy[[10]], 1:2, ndvi)
xs = st_as_stars(x, url = url, downsample = 50)
image(xs, add = TRUE)

s = sapply(proxy, function(x) {
	x = st_apply(x, 1:2, ndvi)
	r = st_as_stars(x, url = url, downsample = 100)
	image(r, add = TRUE, zlim = c(-1,1), col = colorRampPalette(c(grey(.1), grey(.5), 'green'))(10))
})
plot(st_geometry(nl), add = TRUE, border = 'red')

library(stars)
library(xts)
x.orig = read_stars(system.file("tif/L7_ETMs.tif", package = "stars"))
x = x.orig
(x = st_set_dimensions(x, 3, Sys.Date() + 1:6))
(x = st_set_dimensions(x, 3, names = c("x", "y", "time")))
x = as.xts(x)
x[1:3,1:5]
y = st_as_stars(x) # has other dimensions collapsed
y = st_as_stars(x, dimensions = st_dimensions(x.orig)[c("band","x","y")]) 
setNames(y, "L7_ETMs.tif") %>%
	aperm(c(2,3,1)) -> yy
all.equal(yy, x.orig)

# xts -> equal-sized xts (reverse order)
f1 = function(x) {
	xts(as.matrix(x)[rev(seq_len(nrow(x))),], index(x))
}

# xts -> single number (compute time mean)
f2 = function(x) {
	apply(x, 2, mean)
}

# xts -> multiple numbers (return mean and variance)
f3 = function(x) {
	apply(x, 2, function(x) c(mean = mean(x), var = var(x)))
}

# xts -> monthly maxima
f4 = function(x, by = "months") {
	a = aggregate(x, cut(index(x), breaks = by), max)
	xts(as.matrix(a), as.POSIXct(index(a)))
}

# univariate examples:
x = xts(1:5, Sys.Date() + 1:5)
f1(x)
f2(x)
f3(x)
x = xts(rnorm(50), Sys.Date() + 1:50)
f4(x, "weeks")

# multivariate examples, only working on time:
x = xts(cbind(1:5,5:1,runif(5)), Sys.Date() + 1:5)
f1(x)
f2(x)
f3(x)
x = xts(cbind(rnorm(50), runif(50)), Sys.Date() + 1:50)
f4(x)

# running them on a 4-d cube:
# create 4-dimensional cube: x/y/band/time
c(x.orig, 2*x.orig, 3*x.orig, 4*x.orig, along = "time") %>%
  st_set_dimensions(4, Sys.Date()+1:4) -> d4
d4

# apply a time reducer to a 4d cube, either dropping time, or adding a new dimension
reduce_4d_time_xts = function(x, f, time = "time", ...) {
	out = f(as.xts(x), ...)
	t = which(names(dim(x)) == time)
	if (!is.null(dim(out))) {
		out = t(out)
		dim(out) = c(dim(x)[-t], new = prod(dim(out))/prod(dim(x)[-t]))
		st_as_stars(list(x = out)) # but this looses all the dimension metadata
	} else { # f() returns a vector, not a matrix
		dim(out) = dim(x)[-t]
		st_as_stars(list(x = out), dimensions = st_dimensions(x)[-t]) 
	}
}

reduce_4d_time_xts(d4, f2)
reduce_4d_time_xts(d4, f1)
reduce_4d_time_xts(d4, f3)
# reduce_4d_time_xts(d4[,1:20,1:20], f4, by = "days") -->> takes ages!!

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(echo = TRUE, collapse = TRUE, dev = "png")
set.seed(13579)
knitr::opts_chunk$set(fig.height = 4.5)
knitr::opts_chunk$set(fig.width = 6)
EVAL = x = suppressWarnings(require(starsdata, quietly = TRUE))

## ----eval=FALSE---------------------------------------------------------------
# install.packages("starsdata", repos = "https://cran.uni-muenster.de/pebesma/")
# # possibly after: options(timeout = 100)
# # or from an alternative repository:
# # install.packages("starsdata", repos = "http://pebesma.staff.ifgi.de", type = "source")

## -----------------------------------------------------------------------------
library(stars)
tif = system.file("tif/L7_ETMs.tif", package = "stars")
rasterio = list(nXOff = 6, nYOff = 6, nXSize = 100, nYSize = 100, bands = c(1, 3, 4))
(x = read_stars(tif, RasterIO = rasterio))
dim(x)

## -----------------------------------------------------------------------------
st_dimensions(read_stars(tif))

## -----------------------------------------------------------------------------
rasterio = list(nXOff = 6, nYOff = 6, nXSize = 100, nYSize = 100,
                nBufXSize = 20, nBufYSize = 20, bands = c(1, 3, 4))
(x = read_stars(tif, RasterIO = rasterio))

## -----------------------------------------------------------------------------
rasterio = list(nXOff = 6, nYOff = 6, nXSize = 3, nYSize = 3,
   nBufXSize = 100, nBufYSize = 100, bands = 1)
x = read_stars(tif, RasterIO = rasterio)
dim(x)
plot(x)

## -----------------------------------------------------------------------------
rasterio = list(nXOff = 6, nYOff = 6, nXSize = 3, nYSize = 3,
   nBufXSize = 100, nBufYSize = 100, bands = 1, resample = "cubic_spline")
x = read_stars(tif, RasterIO = rasterio)
dim(x)
plot(x)

## ----eval=EVAL----------------------------------------------------------------
granule = system.file("sentinel/S2A_MSIL1C_20180220T105051_N0206_R051_T32ULE_20180221T134037.zip", package = "starsdata")
s2 = paste0("SENTINEL2_L1C:/vsizip/", granule, "/S2A_MSIL1C_20180220T105051_N0206_R051_T32ULE_20180221T134037.SAFE/MTD_MSIL1C.xml:10m:EPSG_32632")
(p = read_stars(s2, proxy = TRUE))

## ----eval=EVAL----------------------------------------------------------------
system.time(plot(p))

## ----eval = FALSE-------------------------------------------------------------
# p = read_stars(s2, proxy = FALSE)

## -----------------------------------------------------------------------------
methods(class = "stars_proxy")

## ----eval=EVAL----------------------------------------------------------------
x = c("avhrr-only-v2.19810901.nc",
"avhrr-only-v2.19810902.nc",
"avhrr-only-v2.19810903.nc",
"avhrr-only-v2.19810904.nc",
"avhrr-only-v2.19810905.nc",
"avhrr-only-v2.19810906.nc",
"avhrr-only-v2.19810907.nc",
"avhrr-only-v2.19810908.nc",
"avhrr-only-v2.19810909.nc")
file_list = system.file(paste0("netcdf/", x), package = "starsdata")
y = read_stars(file_list, quiet = TRUE, proxy = TRUE)
names(y)
y["sst"]

## ----eval=EVAL----------------------------------------------------------------
bb = st_bbox(c(xmin = 10.125, ymin = 0.125, xmax = 70.125, ymax = 70.125))
ysub = y[bb]
st_dimensions(ysub)
class(ysub) # still no data here!!
plot(ysub, reset = FALSE) # plot reads the data, at resolution that is relevant
plot(st_as_sfc(bb), add = TRUE, lwd = .5, border = 'red')

## ----eval=EVAL----------------------------------------------------------------
yy = adrop(y)
yyy = yy[,1:10,1:10,]
class(yyy) # still no data
st_dimensions(yyy) # and dimensions not adjusted
attr(yyy, "call_list") # the name of object in the call (y) is replaced with x:

## ----eval = FALSE-------------------------------------------------------------
# plot(st_apply(x, c("x", "y"), range))

## ----eval=EVAL----------------------------------------------------------------
(x = st_as_stars(yyy)) # read, adrop, subset

## ----eval=EVAL----------------------------------------------------------------
# S2 10m: band 4: near infrared, band 1: red.
#ndvi = function(x) (x[4] - x[1])/(x[4] + x[1])
ndvi = function(x1, x2, x3, x4) (x4 - x1)/(x4 + x1)
rm(x)
(s2.ndvi = st_apply(p, c("x", "y"), ndvi))
system.time(plot(s2.ndvi)) # read - compute ndvi - plot 

## -----------------------------------------------------------------------------
s1 = st_as_stars(matrix(1:16, 4))
s2 = st_as_stars(matrix(1:16, 4))
s3 = st_as_stars(matrix(1:16, 4))
attr(s1, "dimensions")$X1$offset = 0
attr(s1, "dimensions")$X2$offset = 4
attr(s2, "dimensions")$X1$offset = 0
attr(s2, "dimensions")$X2$offset = 4
attr(s3, "dimensions")$X1$offset = 0
attr(s3, "dimensions")$X2$offset = 4
attr(s1, "dimensions")$X1$delta =  1
attr(s1, "dimensions")$X2$delta = -1
attr(s2, "dimensions")$X1$delta =  2
attr(s2, "dimensions")$X2$delta = -2
attr(s3, "dimensions")$X1$delta =  3
attr(s3, "dimensions")$X2$delta = -3
plot(s1, axes = TRUE, text_values = TRUE, text_color = 'orange')
plot(s2, axes = TRUE, text_values = TRUE, text_color = 'orange')
plot(s3, axes = TRUE, text_values = TRUE, text_color = 'orange')

## ----eval=TRUE----------------------------------------------------------------
fn1 = paste0(tempdir(), .Platform$file.sep, "img1.tif")
fn2 = paste0(tempdir(), .Platform$file.sep, "img2.tif")
fn3 = paste0(tempdir(), .Platform$file.sep, "img3.tif")
write_stars(s1, fn1)
write_stars(s2, fn2)
write_stars(s3, fn3) 
(r1 = read_stars(c(fn1, fn2, fn3), proxy = TRUE))

## ----eval=TRUE----------------------------------------------------------------
st_as_stars(r1) |>
  merge() |>
  plot(breaks = "equal", text_values = TRUE, text_color = 'orange', axes = TRUE)

## ----eval=TRUE----------------------------------------------------------------
st_as_stars(r1[,2:4,2:4]) |>
  merge() |>
  plot(breaks = "equal", text_values = TRUE, text_color = 'orange', axes = TRUE)

## -----------------------------------------------------------------------------
s4 = st_as_stars(matrix(1: 16, 4))
s5 = st_as_stars(matrix(1: 64, 8))
s6 = st_as_stars(matrix(1:144,12))
attr(s4, "dimensions")$X1$offset = 0
attr(s4, "dimensions")$X2$offset = 4
attr(s5, "dimensions")$X1$offset = 0
attr(s5, "dimensions")$X2$offset = 4
attr(s6, "dimensions")$X1$offset = 0
attr(s6, "dimensions")$X2$offset = 4
attr(s4, "dimensions")$X1$delta =  1
attr(s4, "dimensions")$X2$delta = -1
attr(s5, "dimensions")$X1$delta =  1/2
attr(s5, "dimensions")$X2$delta = -1/2
attr(s6, "dimensions")$X1$delta =  1/3
attr(s6, "dimensions")$X2$delta = -1/3
plot(s4, axes = TRUE, text_values = TRUE, text_color = 'orange')
plot(s5, axes = TRUE, text_values = TRUE, text_color = 'orange')
plot(s6, axes = TRUE, text_values = TRUE, text_color = 'orange')

## ----eval=TRUE----------------------------------------------------------------
fn4 = paste0(tempdir(), .Platform$file.sep, "img4.tif")
fn5 = paste0(tempdir(), .Platform$file.sep, "img5.tif")
fn6 = paste0(tempdir(), .Platform$file.sep, "img6.tif")
write_stars(s4, fn4)
write_stars(s5, fn5)
write_stars(s6, fn6) 
(r2 = read_stars(c(fn4, fn5, fn6), proxy = TRUE))

st_as_stars(r2) |>
  merge() |>
  plot(breaks = "equal", text_values = TRUE, text_color = 'orange', axes = TRUE)
st_as_stars(r2[,2:4,2:4]) |>
  merge() |>
  plot(breaks = "equal", text_values = TRUE, text_color = 'orange', axes = TRUE)

## ----eval=TRUE----------------------------------------------------------------
(r3 = read_stars(c(fn6, fn5, fn4), proxy = TRUE))

st_as_stars(r3) |>
  merge() |>
  plot(breaks = "equal", text_values = TRUE, text_color = 'orange', axes = TRUE)
st_as_stars(r3[,2:6,3:6]) |>
  merge() |>
  plot(breaks = "equal", text_values = TRUE, text_color = 'orange', axes = TRUE)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(echo = FALSE, collapse = TRUE, dev = "png")
suppressPackageStartupMessages(library(dplyr))
knitr::opts_chunk$set(fig.height = 4.5)
knitr::opts_chunk$set(fig.width = 6)

## -----------------------------------------------------------------------------
tribble(
  ~raster, ~stars, ~`Note/comment`,
  "{raster, stack, brick} (read)", "read_stars or read_stars(along = ...)","",
  "{stack, brick} (concatenate layers)", "c or c(along = ...)","",
  "subset", "{[ ] , slice, filter}","",
  "addLayer", "c() or c(along = ...)","",
  "dropLayer", "{[ ] , slice, filter}","",
  "unstack", "combine lapply and {[ ] , slice, filter}",""
) |> 
  knitr::kable()

## -----------------------------------------------------------------------------
tribble(
  ~raster, ~stars, ~`Note/comment`,
  "merge", "c", "#, currently only works for adjacent objects",
  "mosaic", "st_mosaic", "these are not identical, read the docs carefully",
  "crop", "filter, st_crop", "",
  "setExtent", "", "# maybe use st_warp?",
  "trim", "", "#",
  "aggregate", "aggregate", "WIP; raster's aggregate with fact=2 will not work, use st_warp in that case?",
  "disaggregate", "", "# use st_warp(use_gdal = TRUE)?",
  "resample", "{st_transform, st_warp}", "",
  "projectRaster", "{st_transform, st_warp}", "",
  "shift", "", "#, now use st_set_dimensions",
  "flip", "[] with reversed index", "#",
  "rotate", "", "*",
  "t", "", "NA"
) |> 
  knitr::kable()

## -----------------------------------------------------------------------------
tribble(
  ~raster, ~stars, ~`Note/comment`,
  "calc", "st_apply", "",
  "overlay", "c(along = , ...) |> st_apply(...)", "",
  "cover", "[ ] <-", "",
  "mask", "[ ], st_crop", "when using an sf polygon as mask",
  "mask", "[ ]<- NA", "when using a stars object to mask",
  "cut", "cut", "",
  "subs", "", "",
  "reclassify", "mutate with case_when", "or forcats::fct_recode ?",
  "reclassify", "cut", "",
  "init", "[ ] <-", "",
  "stackApply", "{[ ] , slice, filter} |> st_apply(...)", "",
  "stackSelect", "", "",
) |> 
  knitr::kable()

## -----------------------------------------------------------------------------
tribble(
  ~raster, ~stars, ~`Note/comment`,
  "distance", "", "#",
  "gridDistance", "", "*",
  "distanceFromPoints", "", "#",
  "direction", "", "*",
  "focal", "f = st_apply(x1, 3, foc, w = matrix(1, 3, 3))", "See. [issue 176](https://github.com/r-spatial/stars/issues/176)",
  "localFun", "", "*",
  "boundaries", "st_as_sf(as_points=FALSE, merge=TRUE, connect8=TRUE)", "",
  "clump", "st_as_sf(r, merge = TRUE)", "`st_as_sf` returns polygons, `clump` a raster",
  "adjacent", "", "*",
  "area", "st_area", "",
  "terrain", "", "#",
  "Moran", "", ""
) |> 
  knitr::kable()

## -----------------------------------------------------------------------------
tribble(
  ~raster, ~stars, ~`Note/comment`,
  "predict", "predict", "",
  "interpolate", "gstat::idw, gstat::krige", "`st_warp` has raster-raster interpolations of `gdalwarp`"
) |> 
  knitr::kable()

## -----------------------------------------------------------------------------
tribble(
  ~raster, ~stars, ~`Note/comment`,
  "rasterize", "st_as_stars", "",
  "rasterToPoints", "st_as_sf(as_points=TRUE)", "",
  "rasterToPolygons", "st_as_sf(as_points=FALSE, ...), st_polygonize", "",
  "rasterToContour", "st_contour", "requires GDAL >= 2.4.0",
  "rasterFromXYZ", "", "",
  "rasterFromCells", "", "",
) |> 
  knitr::kable()

## -----------------------------------------------------------------------------
tribble(
  ~raster, ~stars, ~`Note/comment`,
  "cellStats", "st_apply", "",
  "summary", "print, summary(as.vector(. |> pull()))", "",
  "freq", "table", "*",
  "crosstab", "", "",
  "unique", "unique(as.vector(. |> pull()))", "",
  "zonal", "", "*"
) |> 
  knitr::kable()

## -----------------------------------------------------------------------------
tribble(
  ~raster, ~stars, ~`Note/comment`,
  "getValues", "{pull, [[ ]]}", "",
  "getValuesBlock", "{[ ] , slice, filter} |> pull()", "",
  "getValuesFocal", "{[ ] , slice, filter} |> pull()", "",
  "as.matrix", "[[ ]]", "currently behaves somewhat unexpectedly *", 
  "as.array", "[[ ]]", "currently behaves somewhat unexpectedly *", 
  "extract (by cell)", "{[ ] , slice, filter}", "",
  "extract (by polygon)", "x[sf_object]", "",
  "extract (by point)", "aggregate(stars_object, sf_object, function(x) x[1], as_points = FALSE)", "",
  "sampleRandom", "", "*", 
  "sampleRegular", "", "*",
  "minValue", "purrr::map(x, min)", "",
  "maxValue", "purrr::map(x, max)", "",
  "setMinMax", "", "," 
) |> 
  knitr::kable()

## -----------------------------------------------------------------------------
tribble(
  ~raster, ~stars, ~`Note/comment`,
  "plot", "plot, geom_stars", "",
  "plotRGB", "plot(x, rgb =...)", "",
  "spplot", "-", "",
  "image", "image", "",
  "persp", "-", "",
  "contour", "(st_contour, then sf::plot)", "",
  "filledContour", "(same)", "",
  "text", "text", "",
  "`hist`", "`hist(x[[1]])`", "",
  "barplot", "", "",
  "density", "", "",
  "pairs", "", "",
  "boxplot", "", ""
) |> 
  knitr::kable()

## -----------------------------------------------------------------------------
tribble(
  ~raster, ~stars, ~`Note/comment`,
  "ncol", "dim(x)[1]", "or use name instead of 1; cols may be the second dimension!", 
  "nrow", "dim(x)[2]", "or use name instead of 2; rows may be the first dimension!", 
  "ncell", "prod(dim(x))", "",
  "res", "st_dimensions", "can also not be a constant in case of rectilinear or curvilinear grids", 
  "nlayers", "-", "there is no concept of layers in stars", 
  "names", "names", "",
  "xres", "st_res(x)[1]", "may not be a constant in case of rectilinear or curvilinear grids", 
  "yres", "st_res(x)[2]", "may not be a constant in case of rectilinear or curvilinear grids", 
  "xmin", "st_bbox(x)[1]", "",
  "xmax", "st_bbox(x)[3]", "",
  "ymin", "st_bbox(x)[2]", "",
  "ymax", "st_bbox(x)[4]", "",
  "extent", "st_bbox(x)", "different ordering of numbers", 
  "origin", "-", "",
  "projection", "st_crs(x)", "",
  "isLonLat", "st_is_longlat(st_crs(x))", "",
  "filename", "", "stars_proxy objects carry file names where otherwise the array data is", 
  "bandnr", "", "stars has no general concept of bands", 
  "nbands", "dim(x)[3]", "may also be time; bands may also be in another dimension, or have another name", 
  "compareRaster", "all.equal(st_dimensions(x), st_dimensions(y))", "*", 
  "NAvalue", "", "-"
) |> 
  knitr::kable()

## -----------------------------------------------------------------------------
tribble(
  ~raster, ~stars, ~`Note/comment`,
  "xFromCol", "st_get_dimension_values(., 'x')[col]", "I am not sure how to generally get the x dimension name - here it is x",
  "yFromRow", "st_get_dimension_values(., 'y')[row]", "I am not sure how to generally get the y dimension name - here it is y", 
  "xFromCell", "", "",
  "yFromCell", "", "",
  "xyFromCell", "", "", 
  "colFromX", "", "*",
  "rowFromY", "", "*", 
  "rowColFromCell", "", "",
  "cellFromXY", "st_cells()", "",
  "cellFromRowCol", "", "",
  "cellsFromExtent", "", "",
  "coordinates", "st_coordinates", "",
  "validCell", "", "",
  "validCol", "col |> between(st_dimensions(.)$x$from, st_dimensions(.)$x$to)", "raster columns are not always named 'x'", 
  "validRow", "row |> between(st_dimensions(.)$y$from, st_dimensions(.)$y$to)", "raster columns are not always named 'y'", 
  "setValues", "[ ] <-", "",
  "writeRaster", "write_stars", "currently uses GDAL, somewhat limited", 
  "KML", "", ""
) |> 
  knitr::kable()

# script.R
library(jsonlite) # base64_enc

# load stars
library(stars)

# load some imagery meta data
# from a S2 .zip file, create a readable gdal source:
s2_expand_zip = function(s) {
	paste0("/vsizip/", s, "/", sub(".zip", ".SAFE", s), "/MTD_MSIL1C.xml")
}
lst = list.files(pattern = "*.zip")
lst = lst[c(grep("S2A_MSIL1C", lst), grep("S2B_MSIL1C", lst))]
time = strptime(substr(lst, 12, 26), "%Y%m%dT%H%M%S")
l = lapply(s2_expand_zip(lst), read_stars, sub = 1, proxy = TRUE)
bb = do.call(c, lapply(l, function(x) st_as_sfc(st_bbox(st_dimensions(x)))))
epsg = sapply(l, function(x) st_crs(x)$epsg)
library(tibble)
md = st_sf(tibble(proxy = l, time = as.POSIXct(time), epsg = epsg, geom = bb))

# create global database
data = list(md = md)
rm(md)

#* Log some information about the incoming request
#* @filter logger
function(req) {
  cat(as.character(Sys.time()), "-", 
    req$REQUEST_METHOD, req$PATH_INFO, "-", 
    req$HTTP_USER_AGENT, "@", req$REMOTE_ADDR, ":", req$postBody, "\n")
  plumber::forward()
}

# plumber REST end points /data:
#* @get  /data
#* @post /data
get_data <- function(req, expr = NULL) {
	print(expr)
	if (is.null(expr))
		names(data)
	else 
		base64_enc(serialize( eval(parse(text = expr), data), NULL)) # to char
}

#* @put /data
put_data <- function(req, name, value) {
	data[[name]] <<- unserialize(base64_dec(fromJSON(value)))
	NULL
}

# script.R
library(jsonlite) # base64_enc

# load stars
library(stars)

# load some imagery data
x = read_stars(system.file("tif/L7_ETMs.tif", package = "stars"))

# global database
data = list(x = x, y = "foo", z = 0:10)

#* @get /data
get_data <- function(expr = NULL) {
	if (is.null(expr))
		names(data)
	else 
		base64_enc(serialize( eval(parse(text = expr), data), NULL)) # to char
}

#* @put /data
put_data <- function(req, name, value) {
	data[[name]] <<- unserialize(base64_dec(fromJSON(value)))
	NULL
}

shorten_names = function(x, n) {
	x = head(x ,n)
	bn = basename(x)
	here = paste0(normalizePath("."), .Platform$file.sep, bn)
	clean = gsub("\"", "", bn)
	if (any(x != here))
		paste0("[...]", .Platform$file.sep, clean)
	else
		clean
}

#' @export
print.stars_proxy = function(x, ..., n = 1e5, nfiles = 10, simplify = TRUE) {
	if (!is.null(attr(x, "resolutions")))
		cat("multi-resolution ")
	cat("stars_proxy object with", length(x), 
		if (length(x) > 1) "attributes" else "attribute",
		"in", sum(lengths(x)), "file(s)")
	if (length(x[[1]]) > nfiles)
		cat("; showing the first", min(length(x[[1]]), nfiles), "filenames\n")
	else
		cat(":\n")
	names = lapply(x, function(nm) if (is.function(nm)) nm() else nm)
	if (simplify)
		print(lapply(names, shorten_names, n = nfiles))
	else
		print(lapply(names, head, n = nfiles))
	if (!is.null(attr(x, "NA_value")) && !is.na(attr(x, "NA_value")))
		cat("NA_value: ", attr(x, "NA_value"), "\n")
	cat("dimension(s):\n")
	print(st_dimensions(x), ...)
	if (!is.null(attr(x, "call_list"))) {
		cat("call_list:\n")
		l = attr(x, "call_list")
		print(lapply(l, `attr<-`, ".Environment", NULL))
		cat("This object has pending lazy operations: dimensions as printed may not reflect this.\n")
	}
}

#' @export
dim.stars_proxy = function(x) {
	dim(st_dimensions(x))
}

#' @export
as.data.frame.stars_proxy = function(x, ...) {
	as.data.frame(st_as_stars(x), ...)
}


#' @name plot
#' @export
#' @details when plotting a subsetted \code{stars_proxy} object, the default value for argument \code{downsample} will not be computed correctly, and has to be set manually.
plot.stars_proxy = function(x, y, ..., downsample = get_downsample(dim(x))) {
	plot(st_as_stars(x, downsample = downsample, ...), ..., downsample = 0)
}

st_stars_proxy = function(x, dimensions, ..., NA_value, resolutions, RasterIO = list(), file_dim = NULL, class = character(0)) {
	stopifnot(!missing(NA_value),
		!missing(resolutions),
		length(list(...)) == 0,
		is.list(x),
		inherits(dimensions, "dimensions"))

	if (length(RasterIO) == 0)
		RasterIO = NULL
	structure(x, dimensions = dimensions, NA_value = NA_value, resolutions = resolutions,
		RasterIO = RasterIO, file_dim = file_dim, class = c(class, "stars_proxy", "stars"))
}

add_resolution = function(lst) {
	n = length(lst)
	resolutions = data.frame(x = numeric(n), y = numeric(n))
	for (i in seq_along(lst)) {
		d = st_dimensions(lst[[i]])
		xy = attr(d, "raster")$dimensions
		resolutions[i,] = c(d[[ xy[1] ]]$delta, d[[ xy[2] ]]$delta)
	}
	rownames(resolutions) = sapply(lst, names)
	structure(lst, resolutions = resolutions)
}

#' @export
#' @param along_crs logical; if \code{TRUE}, combine arrays along a CRS dimension
#' @name c.stars
c.stars_proxy = function(..., along = NA_integer_, along_crs = FALSE, try_hard = FALSE, 
						 nms = names(list(...)), tolerance = sqrt(.Machine$double.eps)) {
	dots = list(...)
	get_file_dim = function(dots) {
			do.call(rbind, lapply(dots, attr, "file_dim"))
	}
	if (!all(sapply(dots, function(x) inherits(x, "stars_proxy"))))
		stop("all arguments to c() should be stars_proxy objects")
	rio = attr(dots[[1]], "RasterIO")

	# Case 1: merge attributes of several objects by simply putting them together in a single stars object;
	# dim does not change:
	if (length(dots) == 1 && length(along) == 1 && is.na(along)) # do nothing
		dots[[1]]
	else if (along_crs)
		combine_along_crs_proxy(dots)
	else if (identical(along, NA_integer_)) { 
		if (identical_dimensions(dots))
			st_stars_proxy(setNamesIfnn(do.call(c, lapply(dots, unclass)), nms),
						   dimensions = st_dimensions(dots[[1]]), 
						   NA_value = attr(dots[[1]], "NA_value"), 
						   resolutions = NULL,
						   file_dim = get_file_dim(dots),
						   RasterIO = rio)
		else if (identical_dimensions(dots, ignore_resolution = TRUE, tolerance = tolerance)) {
			dots = add_resolution(dots)
			st_stars_proxy(setNamesIfnn(do.call(c, lapply(dots, unclass)), nms),
						   dimensions = st_dimensions(dots[[1]]), 
						   resolutions = attr(dots, "resolutions"),
						   NA_value = attr(dots[[1]], "NA_value"), 
						   file_dim = get_file_dim(dots),
						   RasterIO = rio)
		} else {
			# currently catches only the special case of ... being a broken up time series:
			along = sort_out_along(dots)
			if (!is.na(along))
				do.call(c, c(dots, along = along))
			else if (!try_hard)
				stop("don't know how to merge arrays: please specify parameter along")
			else {
				d = lapply(dots, st_dimensions)
				ident = c(TRUE, sapply(d[-1], identical, d[[1]]))
				if (!all(ident))
					warning(paste(
					"ignored subdataset(s) with dimensions different from first subdataset:", 
					paste(which(!ident), collapse = ", "), 
					"\nuse gdal_subdatasets() to find all subdataset names"))
				if (!is.null(nms))
					nms = nms[ident]
				st_stars_proxy(setNamesIfnn(do.call(c, lapply(dots[ident], unclass)), nms),
							   dimensions = st_dimensions(dots[[1]]), 
							   NA_value = attr(dots[[1]], "NA_value"), 
							   resolutions = NULL, 
							   file_dim = get_file_dim(dots),
							   RasterIO = rio)
			}
		}
	} else { # arrange along "along" dimension:
		if (is.list(along)) { # custom ordering of ... over dimension(s) with values specified
			stop("for proxy objects, along argument as list is not implemented")
		} else { # loop over attributes, abind them:
			# along_dim: the number of the dimension along which we merge arrays
			d = st_dimensions(dots[[1]])
			along_dim = if (is.character(along)) {
				along_dim = which(along == names(d))
				if (length(along_dim) == 0)
					length(d) + 1
				else
					along_dim
			} else
				along
			# ret = propagate_units(mapply(abind, ..., along = along_dim, SIMPLIFY = FALSE), dots[[1]])
			ret = if (length(dots[[1]]) == 1)
					list(attr = unlist(do.call(c, lapply(dots, unclass))))
				else {
					m = mapply(c, lapply(dots, unclass)) # simplifies to list matrix
					setNames(lapply(seq_len(nrow(m)), function(i) unlist(m[i,])), names(dots[[1]]))
				}

			dims = combine_dimensions(dots, along_dim)
			if (along_dim == length(d) + 1)
				names(dims)[along_dim] = if (is.character(along)) along else "new_dim"
			st_stars_proxy(ret, dimensions = dims, NA_value = attr(dots[[1]], "NA_value"),
				resolutions = NULL,
				file_dim = get_file_dim(dots),
				RasterIO = rio)
		}
	}
}

combine_along_crs_proxy = function(dots) {
	crs = lapply(l, st_crs)
	l = lapply(dots, st_set_crs, value = NA)
	ret = do.call(c, c(l, along = "crs"))
	st_set_dimensions(ret, "crs", values = crs, point = TRUE)
}


#' @export
#' @name redimension
st_redimension.stars_proxy = function(x, new_dims = st_dimensions(x), along = list(new_dim = names(x)), ...) {

	d = st_dimensions(x)
	new_dim = create_dimension(values = along[[1]])
	dims = create_dimensions(c(d, new_dim = list(new_dim)), attr(d, "raster"))
	names(dims)[names(dims) == "new_dim"] = names(along)
	ret = list(unlist(do.call(c, lapply(x, unclass))))
	st_stars_proxy(setNames(ret, paste(names(x), collapse = ".")), dimensions = dims,
		NA_value = attr(x, "NA_value"), resolutions = attr(x, "resolutions"))
}

# fetch a stars object from a stars_proxy object, using downsampling
fetch = function(x, downsample = 0, ...) {
	stopifnot(inherits(x, "stars_proxy"))
	d = st_dimensions(x)
	xy = attr(d, "raster")$dimensions
	dx = d[[ xy[1] ]]
	dy = d[[ xy[2] ]]

	nBufXSize = nXSize = dx$to - dx$from + 1
	nBufYSize = nYSize = dy$to - dy$from + 1

	downsample = rep(downsample, length.out = 2)
	if (downsample[1] > 0)
		nBufXSize = ceiling(nBufXSize / (downsample[1] + 1))
	if (downsample[2] > 0)
		nBufYSize = ceiling(nBufYSize / (downsample[2] + 1))

	# issue #438:
	if (any(downsample > 0) && !is.null(attr(x, "RasterIO")))
		warning("with RasterIO defined, argument downsample is ignored")
	rasterio = attr(x, "RasterIO") %||% list(nXOff = dx$from, nYOff = dy$from, 
			nXSize = nXSize, nYSize = nYSize, nBufXSize = nBufXSize, nBufYSize = nBufYSize)

	# select bands?
	bands <- d[["band"]]
	if (!is.null(bands)) {
		if (!is.null(bands$values) && is.numeric(bands$values)) 
			rasterio$bands = bands$values
		else if (!is.na(bands$from) && !is.na(bands$to) 
				 # && (bands$to - bands$from + 1) < length(x[[1]])
			)
			rasterio$bands = seq(bands$from, bands$to)
		if (!is.null(rasterio$bands) && length(rasterio$bands) > 1 && 
				length(rasterio$bands) == length(x[[1]])) # one band in each file
			rasterio$bands = NULL # https://github.com/r-spatial/stars/issues/608
	}

	# do it:
	ret = vector("list", length(x))
	res <- attr(x, "resolutions")
	for (i in seq_along(ret)) {
		if (!is.null(res) && any(res[1,] != res[i,])) {
			mult = c(res[i,1] / res[1,1], res[i,2] / res[1,2])
			rasterio$nXOff = floor((dx$from - 1) / mult[1]) + 1
			rasterio$nYOff = floor((dy$from - 1) / mult[2]) + 1
			rasterio$nXSize = ceiling(dx$to / mult[1]) - floor((dx$from - 1) / mult[1])
			rasterio$nYSize = ceiling(dy$to / mult[2]) - floor((dy$from - 1) / mult[2])
			rasterio$nBufXSize = ceiling(rasterio$nXSize * mult[1] / (downsample[1] + 1))
			rasterio$nBufYSize = ceiling(rasterio$nXSize * mult[2] / (downsample[2] + 1))
			mod = function(a, n) { a - n * floor(a/n) }
			offset = round(c(mod(dx$from - 1, mult[1]), mod(dy$from - 1, mult[2])))
		} else
			offset = c(0,0)
		file_name = unclass(x)[[i]]
		if (is.function(file_name)) # realise/evaluate:
			file_name = file_name()
		ret[[i]] = read_stars(file_name, RasterIO = rasterio, 
			NA_value = attr(x, "NA_value") %||% NA_real_, normalize_path = FALSE,
			proxy = FALSE, ...)
		if (i == 1)
			dm1 = dim(ret[[1]])
		else {
			xrange = seq_len(dm1[1]) + offset[1]
			yrange = seq_len(dm1[2]) + offset[2]
			ret[[i]] = ret[[i]] [ , xrange, yrange ]
			st_dimensions(ret[[i]]) = st_dimensions(ret[[1]])
		}
	}

	along = if (length(dim(x)) > 3)
			setNames(list(st_get_dimension_values(x, 4)), tail(names(st_dimensions(x)), 1))
		else
			list(new_dim = names(x))
	
	ret = if (length(ret) == 1)
		st_redimension(ret[[1]], name = along)
	else
		do.call(c, lapply(ret, st_redimension, along = along))
	
	new_dim = st_dimensions(ret)
#	for (dm in setdiff(names(d), xy)) # copy over non x/y dimension values, if present:
#		if (dm %in% names(new_dim))
#			new_dim[[dm]] = d[[dm]]
	if (length(d) > 2)
		for (dm in 3:length(d)) {
			new_dim[[dm]] = d[[dm]] # copy all fields - what if this was downsampled?
			names(new_dim)[dm] = names(d)[dm]
		}

	ret = unclass(ret)
	for (i in seq_along(ret)) {
		file_dim = attr(x, "file_dim")
		if (is.null(bands) && !is.null(file_dim) && ncol(file_dim) == length(dim(new_dim)) 
				&& ncol(file_dim) == 3) { # https://github.com/r-spatial/stars/issues/596
			r = new_dim[[3]]$from:new_dim[[3]]$to # FIXME: or else use $values?
			ret[[i]] = ret[[i]][,,r]
		}
		dim(ret[[i]]) = dim(new_dim)
	}
	adrop(st_set_crs(st_stars(setNames(ret, names(x)), new_dim), st_crs(x)))
}

check_xy_warn = function(call, dimensions) {
	if (as.character(as.list(call)[[1]]) == "st_apply") {
		# check dims
		MARGIN = as.list(call)$MARGIN
		if (inherits(MARGIN, "call"))
			MARGIN = eval(MARGIN, environment(call))
		if (inherits(MARGIN, "name"))
			MARGIN = get("MARGIN", environment(call))
		xy = attr(dimensions, "raster")$dimensions
		ok = if (is.numeric(MARGIN))
				all(which(names(dimensions) %in% xy) %in% MARGIN)
			else
				all(xy %in% MARGIN)
		if (!ok)
			warning("st_apply on x/y dimensions applied to downsampled image(s)")
	}
}

#' @name st_as_stars
#' @param downsample integer: if larger than 0, downsample with this rate (number of pixels to skip in every row/column); if length 2, specifies downsampling rate in x and y.
#' @param url character; URL of the stars endpoint where the data reside 
#' @param envir environment to resolve objects in
#' @export
st_as_stars.stars_proxy = function(.x, ..., downsample = 0, url = attr(.x, "url"), 
		envir = parent.frame()) {
	if (! is.null(url)) { # execute/get remotely: # nocov start
		# if existing, convert call_list to character:
		attr(.x, "call_list") = lapply(attr(.x, "call_list"), deparse)
		# push the object to url, then st_as_stars() it there:
		tempnam = substr(tempfile(pattern = "Z", tmpdir = "", fileext = ""), 2, 15)
		put_data_url(url, tempnam, .x)
		expr = paste0("st_as_stars(", tempnam, ", url = NULL, downsample=", downsample, 
			", envir = data)") # evaluate in "data" first
		ret = get_data_url(url, expr)
		put_data_url(url, tempnam, NULL) # remove the temporary object
		ret # nocov end
	} else {
		cl = attr(.x, "call_list")
		# FIXME: this means we ALWAYS process after (possibly partial) reading; 
		# there are cases where this is not right. Hence:
		# TODO: only warn when there is a reason to warn.
		if (!all(downsample == 0))
			lapply(cl, check_xy_warn, dimensions = st_dimensions(.x))
		process_call_list(fetch(.x, ..., downsample = downsample), cl, envir = envir, downsample = downsample)
	}
}

st_as_stars_proxy = function(x, fname = tempfile(fileext = rep_len(".tif", length(x))),
		quiet = TRUE, NA_value = NA_real_) {
	stopifnot(inherits(x, "stars"))
	if (inherits(x, "stars_proxy"))
		return(x)
	for (i in seq_along(x))
		write_stars(x[i], fname[i], NA_value = NA_value)
	if (!quiet)
		cat(paste("writing to", fname, "\n"))
	st_stars_proxy(setNames(as.list(fname), names(x)), st_dimensions(x), 
		NA_value = NA_value, resolutions = NULL)
}

# execute the call list on a stars object
process_call_list = function(x, cl, envir = new.env(), downsample = 0) {
	for (i in seq_along(cl)) {
		if (is.character(cl[[i]]))
			cl[[i]] = parse(text = cl[[i]])[[1]]
		stopifnot(is.call(cl[[i]]))
		env = environment(cl[[i]])
		env [[ names(cl[[i]])[2] ]] = x # here, a stars_proxy may be replaced with the fetched stars object
		old_downsample = env$downsample # might be NULL
		if (!is.null(env$downsample) && any(env$downsample != downsample)) {
			cat(paste0("overriding downsample of (sub)expression to c(", paste(downsample, collapse = ","), ")\n"))
			env$downsample = downsample
		}
		# so we need to do that for other args too: https://github.com/r-spatial/stars/issues/390 :
		if ("e2" %in% names(env) && inherits(env$e2, "stars_proxy")) # binary ops: also fetch the second arg
			env$e2 = st_as_stars(env$e2, downsample = downsample)
		if ("i" %in% names(env) && inherits(env$i, "stars_proxy")) # binary ops: also fetch the second arg
			env$i = st_as_stars(env$i, downsample = downsample)
		x = eval(cl[[i]], env)
		env$downsample = old_downsample
	}
	x
}

# add a call to the call list, possibly replacing function name (fn) and first arg name
collect = function(x, call, fn, args = "x", env, ...) {
	call_list = attr(x, "call_list") %||% list()
	dots = list(...)
	nd = names(dots)
	# I would say now to do
	# env = as.environment(append(as.list(env), dots)) -> but that didn't work.
	# so we iterate over ... :
	for (i in seq_along(dots))
		env[[ nd[i] ]] = dots[[i]]
	args = c(args, nd)
	# set function to call:
	lst = as.list(call)
	if (!missing(fn))
		lst[[1]] = as.name(fn)
	# set argument names:
	if (!missing(fn) && fn == "[") {
		lst[[2]] = as.name(args[1])
		lst[[3]] = as.name(args[2])
		for (i in seq_along(args)[-(1:2)]) {
			if (!args[i] %in% names(lst))
				lst[[ args[i] ]] = as.name(args[i]) # appends
		}
	} else {
		for (i in seq_along(args)) {
			lst[[i+1]] = as.name(args[i])
			names(lst)[[i+1]] = args[i]
		}
	}
	call = as.call(lst)
	environment(call) = env
	structure(x, call_list = c(call_list, call))
}

#' @export
adrop.stars_proxy = function(x, drop = which(dim(x) == 1), ...) {
	collect(x, match.call(), "adrop", c("x", "drop"), env = environment(), ...)
}

#' @export
aperm.stars_proxy = function(a, perm = NULL, ...) {
	collect(a, match.call(), "aperm", c("a", "perm"), env = environment(), ...)
}

#' @export
is.na.stars_proxy = function(x) {
	collect(x, match.call(), "is.na", "x", env = environment())
}

#' @name stars_subset
#' @export
"[<-.stars_proxy" = function(x, i, ..., downsample = 0, value) {
	# https://stackoverflow.com/questions/9965577/copy-move-one-environment-to-another
	# deep copy the environment, to avoid side effect later on:
	# FIXME: to investigate - should this be done to env in every call to collect()?
	e = environment()
	# parent not the calling frame but the package namespace:
	env = list2env(as.list(e, all.names = TRUE), parent = parent.env(e))
	collect(x, match.call(), "[<-", c("x", "i", "value"), env)
}

#' @export
split.stars_proxy = function(x, ...) {
	collect(x, match.call(), "split", env = environment())
}

#' @export
merge.stars_proxy = function(x, y, ..., name = "attributes") {
	if (!missing(y))
		stop("argument y needs to be missing: merging attributes of x")
	if (!is.null(attr(x, "call_list")) || !is.null(attr(x, "resolutions"))) # postpone:
		collect(x, match.call(), "merge", c("x", "y", "name"), env = environment(), ...)
	else {
		if (length(x) > 1) { 
			cl = class(x)
			x = unclass(x)
			x[[1]] = unlist(x)
			x[2:length(x)] = NULL
			class(x) = cl
		}
		st_stars_proxy(x, dimensions = create_dimensions(append(st_dimensions(x), 
			list(band = create_dimension(values = names(x[[1]])))), 
			raster = attr(st_dimensions(x), "raster")), 
			NA_value = attr(x, "NA_value"),
			resolutions = attr(x, "resolutions"))
	}
}

#' @export
"[.stars_proxy" = function(x, i = TRUE, ..., drop = FALSE, crop = TRUE) {
	get_range = function(expr) {
		v = try(eval(expr, parent.frame(2)), silent = TRUE)
		if (is.numeric(v) && all(diff(v) == 1))
			range(v)
		else
			NULL
	}
	rio = attr(x, "RasterIO")
	dim_orig = dim(x)
	mc = match.call()
	lst = as.list(mc)
	cl = attr(x, "call_list")
	if (length(lst) < 3)
		return(x) # 
	if (missing(i) | !"i" %in% names(lst)) # insert:
		lst = c(lst[1:2], i = TRUE, lst[-(1:2)])
	if (inherits(i, c("character", "logical", "numeric")) && is.null(cl)) {
		if (!is.null(unclass(x)[[i]])) { # can/should be selected now:
			if (!is.null(resolutions <- attr(x, "resolutions")))
				resolutions = resolutions[i, ]
			x = st_stars_proxy(unclass(x)[i], st_dimensions(x), NA_value = attr(x, "NA_value"),
				resolutions = resolutions, file_dim = attr(x, "file_dim"),
				RasterIO = rio)
			lst[["i"]] = TRUE # this one has been handled now
		}
		ix = 1
		while (length(lst) >= 4) { # https://github.com/r-spatial/stars/issues/496
			if (!is.null(r <- get_range(lst[[4]]))) {
				attr(x, "dimensions")[[ix]]$from = r[1]
				attr(x, "dimensions")[[ix]]$to = r[2]
				if (inherits(attr(x, "dimensions")[[ix]]$values, "CFTime")) {
					idx = CFtime::indexOf(lst[[4]], attr(x, "dimensions")[[ix]]$values)
					attr(x, "dimensions")[[ix]]$values = attr(idx, "CFTime")
				} else if(!is.null(attr(x, "dimensions")[[ix]]$values)) {
					attr(x, "dimensions")[[ix]]$values <- 
						attr(x, "dimensions")[[ix]]$values[r[1]:r[2]]
				}
			}
			ix = ix + 1
			lst[[4]] = NULL # eat/remove
		}
	} else if (crop && inherits(i, c("sf", "sfc", "stars", "bbox"))) {
		x = st_crop(x, i, ..., collect = FALSE) # does bounding box cropping only
		if (inherits(i, c("stars", "bbox")))
			lst[["i"]] = TRUE # this one has been handled now
	}

	# return or collect?
	file_dim = attr(x, "file_dim") %||% matrix(dim(x)[1:2], 1)
	n_file_dim = ncol(file_dim)
	if (length(lst) == 3 && isTRUE(lst[["i"]]) && is.null(cl) &&  # drop a number of files in the lists of files?
			length(x) && length(dim(x)) > n_file_dim && 
			length(x[[1]]) == prod(dim_orig[-(seq_len(ncol(file_dim)))])) { # https://github.com/r-spatial/stars/issues/561
		# select from the vectors of proxy object names:
		get_ix = function(d) {
			stopifnot(inherits(d, "dimension"))
			if (!is.na(d$from))
				seq(d$from, d$to)
			else
				d$values
		}
		d = st_dimensions(x)[-seq_len(n_file_dim)] # dimensions not in the file(s)
		e = do.call(expand.grid, lapply(dim_orig[-seq_len(n_file_dim)], seq_len)) # all combinations
		e$rn = seq_len(nrow(e)) # their index
		f = do.call(expand.grid, lapply(d, get_ix)) # the ones we want
		if (!requireNamespace("dplyr", quietly = TRUE))
			stop("package dplyr required, please install it first") # nocov
		sel = dplyr::inner_join(e, f, by = colnames(f))$rn
		for (i in seq_along(x)) # select:
			x[[i]] = x[[i]][sel]
		x
	} else { # still processing the geometries inside the bbox:
		if (length(lst) == 3 && isTRUE(lst[["i"]]) && is.null(cl))
			x
		else
			collect(x, as.call(lst), "[", c("x", "i", "drop", "crop"), 
				env = environment()) # postpone every arguments > 3 to after reading cells
	}
}

# shrink bbox with e * width in each direction
bb_shrink = function(bb, e) {
	stopifnot(inherits(bb, "bbox"))
	dx = diff(bb[c("xmin", "xmax")])
	dy = diff(bb[c("ymin", "ymax")])
	st_bbox(setNames(c(bb["xmin"] + e * dx, 
		bb["ymin"] + e * dy, 
		bb["xmax"] - e * dx, 
		bb["ymax"] - e * dy), c("xmin", "ymin", "xmax", "ymax")),
		crs = st_crs(bb))
}

#' @name st_crop
#' @param collect logical; if \code{TRUE}, repeat cropping on \code{stars} object, i.e. after data has been read
#' @export
st_crop.stars_proxy = function(x, y, ..., crop = TRUE, epsilon = sqrt(.Machine$double.eps), collect = TRUE) {
	dm = st_dimensions(x)
	d_max = dim(x) + sapply(dm, function(x) x$from) - 1
	if (st_crs(x) != st_crs(y))
		stop("for cropping, the CRS of both objects has to be identical")
	if (crop && has_raster(x)) {
		rast = attr(dm, "raster")$dimensions
		xd = rast[1]
		yd = rast[2]
		bb = st_bbox(y)
		if (epsilon != 0)
			bb = bb_shrink(bb, epsilon)
		# FIXME: document how EXACTLY cropping works; https://github.com/hypertidy/tidync/issues/73
		cr = colrow_from_xy(matrix(bb, 2, byrow=TRUE), dm)
		# crop x:
		dm[[ xd ]]$from = max(1, cr[1, 1], na.rm = TRUE)
		dm[[ xd ]]$to = min(d_max[xd], cr[2, 1], na.rm = TRUE)
		if(!is.null(dm[[ xd ]]$values))
			dm[[ xd ]]$values = dm[[ xd ]]$values[dm[[ xd ]]$from:dm[[ xd ]]$to]
		
		# crop y:
		if (!is.na(dm[[ yd ]]$delta) && dm[[ yd ]]$delta < 0) # FIXME: just subtract values to avoid NA miss?
			cr[1:2, 2] = cr[2:1, 2]
		dm[[ yd ]]$from = max(1, cr[1, 2], na.rm = TRUE)
		dm[[ yd ]]$to = min(d_max[yd], cr[2, 2], na.rm = TRUE)
		if(!is.null(dm[[ yd ]]$values))
			dm[[ yd ]]$values = dm[[ yd ]]$values[dm[[ yd ]]$from:dm[[ yd ]]$to]
	}
	x = st_stars_proxy(x, dm, NA_value = attr(x, "NA_value"), resolutions = attr(x, "resolutions"),
		file_dim = attr(x, "file_dim")) # crop to bb
	if (collect)
		collect(x, match.call(), "st_crop", c("x", "y", "crop", "epsilon"),
			env = environment(), ...) # crops further when realised
	else
		x
}

#' @export
st_apply.stars_proxy = function(X, MARGIN, FUN, ..., CLUSTER = NULL, PROGRESS = FALSE, 
	FUTURE = FALSE, rename = TRUE, .fname) {
	mc = match.call()
	if (missing(.fname))
		.fname = as.character(mc[["FUN"]])[1]
	collect(X, mc, "st_apply", c("X", "MARGIN", "FUN", "CLUSTER", "PROGRESS", "FUTURE", 
		"rename", ".fname"), env = environment(), ...)
}

#' @export
#' @name predict.stars
predict.stars_proxy = function(object, model, ...) {
	collect(object, match.call(), "predict", c("object", "model"), env = environment(), ...)
}

#' @export
"[[<-.stars_proxy" = function(x, i, value) {
	y = unclass(x)
	y[[i]] = value
	structure(y, class = class(x))
}

#' @name st_normalize
#' @param domain ignored
#' @param x object of class `stars` or `stars_proxy`
#' @param ... ignored
#' @details the `stars_proxy` method does nothing, but raises an error when a raster does not start at cell (1,1)
#' @export
st_normalize.stars_proxy = function(x, domain = c(0, 0, 1, 1), ...) {
	stopifnot(all(domain == c(0,0,1,1)))
	d = st_dimensions(x)
	stopifnot(d[[1]]$from == 1, d[[2]]$from == 1)
	x
}

#' @export
image.stars_proxy <- function(x, ..., downsample = get_downsample(dim(x))) {
    image(st_as_stars(x, downsample = downsample), ...)
}

#nocov start
get_data_url = function(url, expr = NULL) {
	if (!requireNamespace("httr", quietly = TRUE)) # GET, POST, PUT
		stop("package httr required, please install it first") # nocov
	if (!requireNamespace("jsonlite", quietly = TRUE)) # base64_dec, base64_enc, toJSON, fromJSON
		stop("package jsonlite required, please install it first") # nocov

    if (is.null(expr))
        jsonlite::fromJSON( httr::content(httr::GET(url), "text", encoding = "UTF-8"))
    else {
        js = jsonlite::fromJSON(
            httr::content(httr::POST(url, body = list(expr = expr), encode = "json"),
                "text", encoding = "UTF-8"))
		if (is.list(js) && !is.null(js$error))
			stop(paste(js$error, ":", js$message))
        unserialize(jsonlite::base64_dec(js))
	}
}

put_data_url = function(url, name, value) {
	if (!requireNamespace("httr", quietly = TRUE)) # GET, POST, PUT
		stop("package httr required, please install it first") # nocov
	if (!requireNamespace("jsonlite", quietly = TRUE)) # base64_dec, base64_enc, toJSON, fromJSON
		stop("package jsonlite required, please install it first") # nocov

    value = jsonlite::toJSON(jsonlite::base64_enc(serialize(value, NULL)))
    httr::PUT(url, body = list(name = name, value = value), encode = "json")
}
#nocov end

# reset the offset when x is a sub-raster (i.e, starts at index larger than 1)
reset_sub = function(x) {
	d = st_dimensions(x)
	xy = attr(d, "raster")$dimensions
	if (all(is.na(xy)))
		return(x)
	
	for (i in xy) {
		if (d[[ i ]]$from > 1) {
			if (!is_regular_grid(x))
				stop("can only write sub-rasters for regular grids")
			ioff = d[[ i ]]$from - 1
			d[[ i ]]$offset = d[[ i ]]$offset + ioff * d[[ i ]]$delta
			d[[ i ]]$to = d[[ i ]]$to - ioff
			d[[ i ]]$from = 1
		}
	}
	structure(x, dimensions = d)
}

#' @export
st_write.stars = function(obj, dsn, layer, ...) {
	.Deprecated("write_stars") # nocov
}


#' @name write_stars
#' @export
write_stars = function(obj, dsn, layer, ...) UseMethod("write_stars")


#' write stars object to gdal dataset (typically: to file)
#' 
#' @param obj object of class \code{stars}
#' @param dsn gdal dataset (file) name
#' @param layer attribute name; if missing, the first attribute is written
#' @param ... passed on to \link[sf:gdal]{gdal_write}
#' @param driver driver driver name; see \link[sf]{st_drivers}
#' @param options character vector with dataset creation options, passed on to GDAL
#' @param type character; output binary type, one of: \code{Byte} for eight bit unsigned integer, \code{UInt16} for sixteen bit unsigned integer, \code{Int16} for sixteen bit signed integer, \code{UInt32} for thirty two bit unsigned integer, \code{Int32} for thirty two bit signed integer, \code{Float32} for thirty two bit floating point, \code{Float64} for sixty four bit floating point.
#' @param NA_value non-NA value that should represent R's \code{NA} value in the target raster file; if set to \code{NA}, it will be ignored.
#' @param update logical; if \code{TRUE}, an existing file is being updated
#' @param normalize_path logical; see \link{read_stars}
#' @param scale_offset length 2 numeric vector with scale, offset values: raw values computed by raw = (value - offset) / scale are written to dsn; scale and offset values are written to dsn or else a warning is raised
#' @name write_stars
#' @export
write_stars.stars = function(obj, dsn, layer = 1, ..., driver = detect.driver(dsn), 
		options = character(0), 
		type = if (is.factor(obj[[1]]) && length(levels(obj[[1]])) < 256) "Byte" else "Float32", 
		NA_value = NA_real_, update = FALSE, normalize_path = TRUE, scale_offset = c(1.0, 0.0)) {

	if (missing(layer) && length(obj) > 1)
		warning("all but first attribute are ignored")
	if (length(layer) > 1)
		stop("layer should have length 1; for writing multi-band images use merge() to merge layers into a dimension")
	obj = st_upfront(obj[layer])
	if (! update) # new file: should not be a sub-array
		obj = reset_sub(obj)
	if (!update && !is.null(attr(obj[[1]], "colors"))) # add r g b alpha table from colors:
		obj[[1]] = structure(obj[[1]], rgba = t(col2rgb(attr(obj[[1]], "colors"), alpha = TRUE)))
	if (normalize_path)
		dsn = enc2utf8(maybe_normalizePath(dsn, TRUE))
	sf::gdal_write(obj, ..., file = dsn, driver = driver, options = options, 
		type = type, NA_value = NA_value, geotransform = st_geotransform(obj), 
		update = update, scale_offset = scale_offset)
	invisible(obj)
}

#' @name write_stars
#' @param chunk_size length two integer vector with the number of pixels (x, y) used in the read/write loop; see details.
#' @param progress logical; if \code{TRUE}, a progress bar is shown
#' @details \code{write_stars} first creates the target file, then updates it sequentially by writing blocks of \code{chunk_size}.
#' @details in case \code{obj} is a multi-file \code{stars_proxy} object, all files are written as layers into the output file \code{dsn}
#' @export
write_stars.stars_proxy = function(obj, dsn, layer = 1, ..., driver = detect.driver(dsn), 
		options = character(0), scale_offset = c(1.0, 0.0), type = "Float32", NA_value = NA_real_, 
		chunk_size = c(dim(obj)[1], floor(25e6 / dim(obj)[1])), progress = TRUE) {

	if (!missing(layer))
		obj = obj[layer]

	cl = attr(obj, "call_list")
	if (is.null(cl) && (length(obj[[1]]) > 1 || length(obj) > 1)) { # collapse bands:
		out_file = tempfile(fileext = ".vrt")
		gdal_utils("buildvrt", unlist(obj), out_file, options = "-separate")
		obj[[1]] = out_file
	}

	if (progress) {
		pb = txtProgressBar()
		setTxtProgressBar(pb, 0)
	}

	dim_obj = dim(obj)
	if (prod(chunk_size) > prod(dim_obj[1:2]))
		write_stars(st_as_stars(obj), dsn, layer, ..., driver = driver, options = options,
			scale_offset = scale_offset, type = type, NA_value = NA_value)
	else { # write chunked: https://github.com/r-spatial/stars/pull/291/files

		di_write = di_read = st_dimensions(obj)
		di_from = c(di_read[[1]]$from, di_read[[2]]$from)
		created = FALSE
		ncol = ceiling(dim_obj[1] / chunk_size[1])
		nrow = ceiling(dim_obj[2] / chunk_size[2])
		for (col in 1:ncol) {
			di_write[[1]]$from = 1 + (col - 1) * chunk_size[1]
			di_write[[1]]$to   = min(col * chunk_size[1], dim_obj[1])
			di_read[[1]]$from = di_write[[1]]$from + di_from[1] - 1
			di_read[[1]]$to   = di_write[[1]]$to + di_from[1] - 1
			di_write[[1]]$offset = with(di_read[[1]], offset + delta * (from - 1))
			for (row in 1:nrow) {
				di_write[[2]]$from = 1 + (row - 1) * chunk_size[2]
				di_write[[2]]$to   = min(row * chunk_size[2], dim_obj[2])
				di_read[[2]]$from = di_write[[2]]$from + di_from[2] - 1
				di_read[[2]]$to   = di_write[[2]]$to + di_from[2] -1
				di_write[[2]]$offset = with(di_read[[2]], offset + delta * (from - 1))
				chunk = st_as_stars(structure(obj, dimensions = di_read))
				attr(chunk, "dimensions")[[1]] <- di_write[[1]] # x
				attr(chunk, "dimensions")[[2]] <- di_write[[2]] # y

				if (! created) { # create:
					d = st_dimensions(chunk)
					d_obj = st_dimensions(obj)
					d[[1]]$from = d[[2]]$from = 1
					d[[1]]$to = d[[1]]$from + dim_obj[1] - 1
					d[[2]]$to = d[[2]]$from + dim_obj[2] - 1
					gt = st_geotransform(structure(obj, dimensions = d))
					sf::gdal_write(structure(obj, dimensions = d), ..., file = dsn, driver = driver, options = options,
						type = type, scale_offset = scale_offset, NA_value = NA_value, geotransform = gt) # branches on stars_proxy
				}
				created = TRUE
				write_stars(chunk, dsn = dsn, layer = layer, driver = driver,
					options = options, type = type, scale_offset = scale_offset,
					update = TRUE, NA_value = NA_value)
				if (progress)
					setTxtProgressBar(pb, ((col-1) * nrow + row) / (ncol * nrow))
			}
		}
	}
	if (progress)
		close(pb)
	invisible(obj)
}

#' @name write_stars
#' @export
#' @param filename character; used for guessing driver short name based on file 
#' extension; see examples
#' @examples 
#' detect.driver("L7_ETMs.tif")
detect.driver = function(filename) { #nocov start
	# from raster::.getFormat:
	ext <- tolower(tools::file_ext(filename))
	if (nchar(ext) < 2) {
		warning("file without extension, using driver GTiff", call. = FALSE)
		'GTiff'
	} else {
		switch(ext, 
			tif  = ,
			tiff = 'GTiff',
			grd  = 'rraster',
			nc   = ,
			cdf  = ,
			ncdf = 'NetCDF',
			zarr = 'ZARR',
			kml  = ,
			kmz  = 'KML',
			big  = 'big.matrix',
			sgrd = ,
			sdat = 'SAGA',
			bil  = 'BIL',
			bsq  = 'BSQ',
			bip  = 'BIP',
			bmp  = 'BMP',
			gen  = 'ADRG',
			bt   = 'BT',
			envi = 'ENVI',
			ers  = 'ERS',
			img  = 'HFA',
			rst  = 'RST',
			mpr  = 'ILWIS',
			rsw  = 'RMF',
			flt  = 'EHdr',
			gpkg = 'GPKG',
			{
				warning("extension ", ext, " is unknown. Using default driver GTiff.")
				'GTiff'
			}
		)
		}
}
#nocov end

#' @name st_as_stars
#' @export
#' @param check_times logical; should we check that the time stamps of all time series are identical?
st_as_stars.cubble_df = function(.x, ..., check_times = FALSE) {
    if (!requireNamespace("tsibble", quietly = TRUE))
        stop("package cubble required, please install it first") #nocov
    if (!requireNamespace("cubble", quietly = TRUE))
        stop("package cubble required, please install it first") #nocov
    if (!requireNamespace("dplyr", quietly = TRUE))
        stop("package dplyr required, please install it first") #nocov
	# time_column = cubble::index(.x)
	nr = sapply(.x$ts, nrow)
	stopifnot(length(unique(nr)) == 1)
	ts1 = .x$ts[[1]]
	dt = which(sapply(ts1, inherits, c("Date", "POSIXct", "units")))
	if (length(dt) > 1) {
		message("using only first time column for time index")
		dt = dt[1]
	}
	if (length(dt) == 0)
		stop("no time column found")
	times = ts1[[dt]]
	if (check_times)
		stopifnot(all(sapply(.x$ts, function(ts) identical(ts[[dt]], times))))
	else
		message("assuming times are identical for all elements in in the ts list column")
	m = do.call(rbind, lapply(.x$ts, function(df) df[-dt][[1]]))
	if (inherits(.x, "sf")) { # vector data cube
		sfc = st_geometry(.x)
		d = create_dimensions(list(
			   geometry = create_dimension(values = sfc),
			   time = create_dimension(values = times))
		)
		st_stars(list(values = m), d)
	} else { # raster:
		kv = tsibble::key_vars(.x)
		.x = dplyr::as_tibble(.x)
		for (k in kv)
			.x[[k]] = NULL
		.x$ts = NULL # remove time series payload
		.x = cbind(.x, m)
		merge(st_as_stars(.x))
	}
}

maybe_normalizePath = function(.x, np = FALSE) {
	prefixes = c("ZARR:", "NETCDF:", "HDF5:", "HDF4:", "HDF4_EOS:", "SENTINEL2_L1", "SENTINEL2_L2", "GPKG:", "/vsi", "http://", "https://")
	has_prefix = function(pf, x) substr(x, 1, nchar(pf)) == pf
	if (is.function(.x) || !np || any(sapply(prefixes, has_prefix, x = .x)))
		.x
	else
		setNames(normalizePath(.x, mustWork = FALSE), names(.x))
}

enc2utf8char = function(x) {
	if (is.character(x))
		enc2utf8(x)
	else
		x
}

get_names = function(x) {
	if (is.function(x))
		x()
	else
		x
}

is_functions = function(x) {
	is.function(x) || all(sapply(x, is.function))
}

is_big = function(x, ..., sub = sub, n_proxy) {
	prod(dim(read_stars(x, ..., sub = sub, proxy = TRUE, quiet = TRUE))) > n_proxy
}

# strip common start & end parts of x, if remainder != ""
unique_part = function(x, prefix) {
	stopifnot(is.character(x), length(x) > 1)
	original = x
	while (nchar(x[1]) && length(unique(substr(x, 1, 1))) == 1)
		x = sapply(x, substring, 2)
	while ((n <- nchar(x[1])) && length(unique(substr(x, n, n))) == 1)
		x = sapply(x, substring, 1, n - 1)
	if (any(x == ""))
		original
	else if (is.character(prefix))
		paste0(prefix, x)
	else
		x
}

geoloc_is_2D = function(geolocation, driver) { # the thing has 2-D x and y arrays:
	# read coordinate array dimensions only:
	gxy = c(geolocation$X_DATASET, geolocation$Y_DATASET)
	lon = adrop(read_stars(gxy[1], driver = driver, quiet = TRUE,
		proxy = TRUE, curvilinear = character(0)))
	lat = adrop(read_stars(gxy[2], driver = driver, quiet = TRUE,
		proxy = TRUE, curvilinear = character(0)))
	dim(lon)[2] > 1 || dim(lat)[2] > 1
}


#' read raster/array dataset from file or connection
#'
#' read raster/array dataset from file or connection
#' @param .x character vector with name(s) of file(s) or data source(s) to be read, or a function that returns such a vector
#' @param options character; opening options
#' @param driver character; driver to use for opening file. To override fixing for subdatasets and autodetect them as well, use \code{NULL}.
#' @param sub character, integer or logical; name, index or indicator of sub-dataset(s) to be read
#' @param quiet logical; print progress output?
#' @param NA_value numeric value to be used for conversion into NA values; by default this is read from the input file
#' @param along length-one character or integer, or list; determines how several arrays are combined, see Details.
#' @param RasterIO list with named parameters for GDAL's RasterIO, to further control the extent, resolution and bands to be read from the data source; see details.
#' @param proxy logical; if \code{TRUE}, an object of class \code{stars_proxy} is read which contains array
#' metadata only; if \code{FALSE} the full array data is read in memory. Always \code{FALSE} for curvilinear girds.
#' If set to a number, defaults to \code{TRUE} when the number of cells to be read is larger than that number.
#' @param curvilinear length two character vector with names of subdatasets holding longitude and latitude values for all raster cells, or named length 2 list holding longitude and latitude matrices; the names of this list should correspond to raster dimensions referred to
#' @param normalize_path logical; if \code{FALSE}, suppress a call to \link{normalizePath} on \code{.x}
#' @param RAT character; raster attribute table column name to use as factor levels
#' @param tolerance numeric; passed on to \link{all.equal} for comparing dimension parameters.
#' @param ... passed on to \link{st_as_stars} if \code{curvilinear} was set
#' @param exclude character; vector with category value(s) to exclude
#' @param shorten logical or character; if \code{TRUE} and \code{length(.x) > 1}, remove common start and end parts of array names; if character a new prefix
#' @return object of class \code{stars}
#' @details In case \code{.x} contains multiple files, they will all be read and combined with \link{c.stars}. Along which dimension, or how should objects be merged? If \code{along} is set to \code{NA} it will merge arrays as new attributes if all objects have identical dimensions, or else try to merge along time if a dimension called \code{time} indicates different time stamps. A single name (or positive value) for \code{along} will merge along that dimension, or create a new one if it does not already exist. If the arrays should be arranged along one of more dimensions with values (e.g. time stamps), a named list can passed to \code{along} to specify them; see example.
#'
#' \code{RasterIO} is a list with zero or more of the following named arguments:
#' \code{nXOff}, \code{nYOff} (both 1-based: the first row/col has offset value 1),
#' \code{nXSize}, \code{nYSize}, \code{nBufXSize}, \code{nBufYSize}, \code{bands}, \code{resample}.
#' See \url{https://gdal.org/en/latest/doxygen/classGDALDataset.html} for their meaning;
#' \code{bands} is an integer vector containing the band numbers to be read (1-based: first band is 1).
#' Note that if \code{nBufXSize} or \code{nBufYSize} are specified for downsampling an image,
#' resulting in an adjusted geotransform. \code{resample} reflects the resampling method and
#' has to be one of: "nearest_neighbour" (the default),
#' "bilinear", "cubic", "cubic_spline", "lanczos", "average", "mode", or "Gauss".
#'
#' Data that are read into memory (\code{proxy=FALSE}) are read into a numeric (double) array, except for categorical variables which are read into an numeric (integer) array of class \code{factor}.
#' @export
#' @examples
#' tif = system.file("tif/L7_ETMs.tif", package = "stars")
#' (x1 = read_stars(tif))
#' (x2 = read_stars(c(tif, tif)))
#' (x3 = read_stars(c(tif, tif), along = "band"))
#' (x4 = read_stars(c(tif, tif), along = "new_dimensions")) # create 4-dimensional array
#' x1o = read_stars(tif, options = "OVERVIEW_LEVEL=1")
#' t1 = as.Date("2018-07-31")
#' # along is a named list indicating two dimensions:
#' read_stars(c(tif, tif, tif, tif), along = list(foo = c("bar1", "bar2"), time = c(t1, t1+2)))
#'
#' m = matrix(1:120, nrow = 12, ncol = 10)
#' dim(m) = c(x = 10, y = 12) # named dim
#' st = st_as_stars(m)
#' attr(st, "dimensions")$y$delta = -1
#' attr(st, "dimensions")$y$offset = 12
#' st
#' tmp = tempfile(fileext = ".tif")
#' write_stars(st, tmp)
#' (red <- read_stars(tmp))
#' read_stars(tmp, RasterIO = list(nXOff = 1, nYOff = 1, nXSize = 10, nYSize = 12,
#'    nBufXSize = 2, nBufYSize = 2))[[1]]
#' (red <- read_stars(tmp, RasterIO = list(nXOff = 1, nYOff = 1, nXSize = 10, nYSize = 12,
#'    nBufXSize = 2, nBufYSize = 2)))
#' red[[1]] # cell values of subsample grid:
#' \dontrun{
#'   plot(st, reset = FALSE, axes = TRUE, ylim = c(-.1,12.1), xlim = c(-.1,10.1),
#'     main = "nBufXSize & nBufYSize demo", text_values = TRUE)
#'   plot(st_as_sfc(red, as_points = TRUE), add = TRUE, col = 'red', pch = 16)
#'   plot(st_as_sfc(st_as_stars(st), as_points = FALSE), add = TRUE, border = 'grey')
#'   plot(st_as_sfc(red, as_points = FALSE), add = TRUE, border = 'green', lwd = 2)
#' }
#' file.remove(tmp)
read_stars = function(.x, sub = TRUE, ..., options = character(0), 
		driver = character(0), quiet = FALSE, NA_value = NA_real_, along = NA_integer_,
		RasterIO = list(), proxy = getOption("stars.n_proxy") %||% 1.e8,
		curvilinear = character(0), normalize_path = TRUE, RAT = character(0),
		tolerance = 1e-10, exclude = "", shorten = TRUE) {

	is_false = function(x) { length(x) == 1 && is.logical(x) && !x }
	if (!is.list(.x) && length(.x) > 1 && is.null(names(.x)) && !is_false(shorten) &&
				any((short_names <- unique_part(.x, shorten)) != .x))
		names(.x) = short_names
	if (is.numeric(proxy)) {
		if (is_functions(.x))
			proxy = TRUE
		else if (length(curvilinear))
			proxy = FALSE
	}

	x = if (is.list(.x)) {
			f = function(y, np) enc2utf8char(maybe_normalizePath(y, np))
			rapply(.x, f, classes = c("character", "function"), how = "replace", np = normalize_path)
		} else
			enc2utf8char(maybe_normalizePath(.x, np = normalize_path))

	if (missing(curvilinear)) { # see if we can get it from metadata item "GEOLOCATION":
		geolocation <- try(gdal_metadata(.x, "GEOLOCATION"), silent = TRUE)
		if (!inherits(geolocation, "try-error") && geoloc_is_2D(geolocation, driver)) { # the thing has 2-D x and y arrays:
			# read coordinate arrays:
			gxy = c(geolocation$X_DATASET, geolocation$Y_DATASET)
			lon = adrop(read_stars(gxy[1], driver = driver, quiet = quiet, NA_value = NA_value,
				RasterIO = RasterIO, proxy = FALSE, curvilinear = character(0), ...))
			lat = adrop(read_stars(gxy[2], driver = driver, quiet = quiet, NA_value = NA_value,
				RasterIO = RasterIO, proxy = FALSE, curvilinear = character(0), ...))
			if (all(dim(lon) > 1) && all(dim(lat) > 1)) # both are 2-D matrices:
				curvilinear = setNames(c(st_set_dimensions(lon, names = c("x", "y")),
					st_set_dimensions(lat, names = c("x", "y"))), c("x", "y"))
			else {
				ulon = unique(diff(as.vector(lon[[1]])))
				ulat = unique(diff(as.vector(lat[[1]])))
				if (length(ulon) > 1 || length(ulat) > 1)
					warning("ignoring irregular (rectilinear) x and/or y coordinates!")
			}
		}
	} else if (length(curvilinear) == 2 && is.character(curvilinear)) { # user-set, read them
		lon = adrop(read_stars(.x, sub = curvilinear[1], driver = driver, quiet = quiet, NA_value = NA_value,
			RasterIO = RasterIO, proxy = FALSE, ..., sub_required = TRUE))
		lat = adrop(read_stars(.x, sub = curvilinear[2], driver = driver, quiet = quiet, NA_value = NA_value,
			RasterIO = RasterIO, proxy = FALSE, ..., sub_required = TRUE))
		curvilinear = setNames(c(st_set_dimensions(lon, names = c("x", "y")),
			st_set_dimensions(lat, names = c("x", "y"))), c("x", "y"))
	}
	if (length(curvilinear) && isTRUE(proxy))
		warning("proxy = TRUE may not work for curvilinear rasters")

	if (length(x) > 1) { # loop over data sources, read, c() all, and RETURN:
		ret = lapply(x, read_stars, options = options, driver = driver, sub = sub, quiet = quiet,
			NA_value = NA_value, RasterIO = as.list(RasterIO), proxy = proxy, curvilinear = curvilinear,
			along = if (length(along) > 1) along[-1] else NA_integer_, normalize_path = normalize_path)
		return(do.call(c, append(ret, list(along = along, tolerance = tolerance))))
	}

	# else: do it --
	data = sf::gdal_read(get_names(x),
		options = options, driver = driver, read_data = ifelse(is.logical(proxy), !proxy, proxy),
		NA_value = NA_value, RasterIO_parameters = as.list(RasterIO))
	if (!is.null(data$default_geotransform) && data$default_geotransform == 1) {
		## we have the 0 1 0 0 0 1 transform indicated
		## so stars policy is flip-y and shift to be in 0, ncol, 0, nrow
		data$geotransform <- c(0, 1, 0, data$rows[2L], 0, -1)
	}
	if (length(data$bands) == 0) { # read sub-datasets: different attributes
		sub_names = split_strings(data$sub) # get named list
		sub_datasets = sub_names[seq(1, length(sub_names), by = 2)]
		if (length(sub_datasets) == 1 && is.na(sub_datasets[1]))
			stop("cannot derive subdataset names from data source: please provide manually")
		# sub_datasets = gdal_subdatasets(x, options)[sub] # -> would open x twice

		# FIXME: only tested for NetCDF:
		nms = sapply(strsplit(unlist(sub_datasets), ":"), tail, 1)
		names(sub_datasets) = nms
		if (any(sapply(sub_datasets[sub], is.null)))
			sub = sub("^//", "/", sub) # GDAL2->3, HDF5, double to single slash?
		sub_datasets = sub_datasets[sub]
		nms = names(sub_datasets)

		.read_stars = function(x, options, driver, quiet, proxy, curvilinear) {
			if (! quiet)
				cat(paste0(tail(strsplit(x, ":")[[1]], 1), ", "))
			read_stars(x, options = options, driver = driver, NA_value = NA_value,
				RasterIO = as.list(RasterIO), proxy = proxy, curvilinear = curvilinear)
		}

		driver = if (is.null(driver) || data$driver[1] %in% c("HDF5", "HDF4"))
				character(0)
			else # override auto-detection:
				data$driver[1]

		ret = lapply(sub_datasets, .read_stars, options = options,
			driver = driver, quiet = quiet, proxy = proxy, curvilinear = curvilinear)
		if (! quiet)
			cat("\n")
		# return:
		if (length(ret) == 1)
			ret[[1]]
		else {
			ret = do.call(c, append(ret, list(try_hard = TRUE, nms = nms)))
			if (length(nms) == length(ret)) # lost my ability to solve this here...
				setNames(ret, nms)
			else
				ret
		}
	} else { # we have one single array: read it
		if (!isTRUE(sub)) {
			sub_required = if (is.null(list(...)$sub_required))
					FALSE
				else
					list(...)$sub_required
			if (sub_required)
				stop(paste("only one array present in", .x, ": cannot resolve subdataset", sub))
			else
				warning("only one array present: argument 'sub' will be ignored")
		}
		meta_data = structure(data, data = NULL) # take meta_data only
		if (is.numeric(proxy) && !is.null(data$proxy))
			proxy = data$proxy
		data = if (proxy)
				.x
			else
				get_data_units(attr(data, "data")) # extract data array; sets units if present
		if (meta_data$driver[1] == "netCDF")
			meta_data = parse_netcdf_meta(meta_data, get_names(x)) # sets all kind of units
		if (! proxy && !is.null(meta_data$units) && !is.na(meta_data$units)
				&& !inherits(data, "units")) # set units
			units(data) = try_as_units(meta_data$units)
		meta_data = parse_gdal_meta(meta_data)

		newdims = lengths(meta_data$dim_extra)
		if (length(newdims) && !proxy)
			dim(data) = c(dim(data)[1:2], newdims)

		# handle color table and/or attribute table
		ct = meta_data$color_tables
		at = meta_data$attribute_tables
		if (meta_data$driver[1] == "AIG" && identical(names(at[[1]]), c("VALUE", "COUNT")))
			at = list() # skip it: https://github.com/r-spatial/stars/issues/435
		# FIXME: how to handle multiple color, category or attribute tables?
		if (!proxy && (any(lengths(ct) > 0) || any(lengths(at) > 0))) {
			if (any(meta_data$ranges[1, c(2,4)] == 1)) { # adjust data to min/max values from image metadata
				r = range(data, na.rm = TRUE)
				min_value = if (meta_data$ranges[1,2] == 1)
						meta_data$ranges[1,1]
					else
						r[1]
				max_value = if (meta_data$ranges[1,4] == 1)
						meta_data$ranges[1,3]
					else
						r[2]
				data[data < min_value | data > max_value] = NA
			}

			# convert color table ct to a vector of R colors:
			co = if (any(lengths(ct) > 0)) {
				ct = ct[[ which(length(ct) > 0)[1] ]]
				apply(ct, 1, function(x) rgb(x[1], x[2], x[3], x[4], maxColorValue = 255))
			} else
				NULL

			if (any(lengths(at) > 0)) {
				# select attribute table:
				which.at = which(lengths(at) > 0)[1]
				which.column = if (length(RAT))
						RAT[1]
					else
						which(sapply(at[[which.at]], class) == "character")[1]
				if (length(which.column) == 0)
					which.column = 1
			} else
				which.column = NA
			if (!(all(is.na(which.column)))) {
				at = at[[ which.at ]]
				labels = at[[ which.column ]]
				levels = if (!is.na(v <-  match("value", tolower(names(at)))))
						at[[ v ]]
					else 
						0:(length(labels) - 1) # why not 1:length(labels)? renumbered in match()
				if (length(exclude)) {
					ex = labels %in% exclude
					labels = labels[!ex]
					levels = levels[!ex]
					if (!is.null(co))
						co = co[!ex]
				} else
					ex = rep(FALSE, length(levels))
			} else {
				levels = sort(unique(as.vector(data)))
				labels = as.character(levels)
				if (!is.null(co))
					co = co[levels + 1]
				ex = rep(FALSE, length(levels))
			}
			# f = factor(as.vector(data), levels = levels, labels = labels) # is too costly;
			# see https://github.com/r-spatial/stars/issues/565:
			# construct factor array manually:
			data = if (any(du <- duplicated(labels))) {
				ul = labels[!du] # unique, non-duplicated
				m = match(labels, ul) # renumber table
				dm = match(as.vector(data), levels)
				structure(m[dm], levels = ul, dim = dim(data), 
							 colors = co, exclude = ex, class = "factor")
			} else
				structure(match(as.integer(as.vector(data)), levels),
							 levels = labels, dim = dim(data), 
							 colors = co, exclude = ex, class = "factor")
		}

		dims = if (proxy) {
				if (length(meta_data$bands) > 1)
					c(x = meta_data$cols[2],
					  y = meta_data$rows[2],
					  band = length(meta_data$bands),
					  lengths(meta_data$dim_extra))
				else
					c(x = meta_data$cols[2],
					  y = meta_data$rows[2],
					  lengths(meta_data$dim_extra))
			} else
				NULL

		# return:
		name_x = if (!is.null(meta_data$long_name) && !is.na(meta_data$long_name))
				meta_data$long_name
			else if (is.function(.x))
				names(.x()) %||% .x()
			else
				x
		name_x = tail(strsplit(name_x, '[\\\\/:]+')[[1]], 1)
		ret = if (proxy) { # no data present, subclass of "stars":
				dims = create_dimensions_from_gdal_meta(dims, meta_data)
				st_stars_proxy(setNames(list(x), names(.x) %||% name_x), dims,
					NA_value = NA_value, resolutions = NULL, RasterIO = RasterIO, 
					file_dim = matrix(dim(dims), 1, dimnames = list(NULL, names(dims))))
			} else
				st_stars(setNames(list(data), names(.x) %||% name_x),
					create_dimensions_from_gdal_meta(dim(data), meta_data))

		if (is.list(curvilinear))
			st_as_stars(ret, curvilinear = curvilinear, ...)
		else
			ret
	}
}

get_data_units = function(data) {
	units = unique(attr(data, "units")) # will fail parsing in as_units() when more than one
	if (length(units) > 1) {
		warning(paste("more than one unit available for array: ignoring all")) # nocov
		structure(data, units = NULL)
	} else {
		if (!is.null(units) && nzchar(units)) {
			if (units == "arc-second") # datum grids may have this
				units = "arcsecond"
			units = try_as_units(units)
		}
		if (inherits(units, "units"))
			units::set_units(structure(data, units = NULL), units, mode = "standard")
		else
			structure(data, units = NULL)
	}
}

test_that("cubble", {
  skip_if_not_installed("cubble")
  skip_if_not_installed("dplyr")
  library(dplyr)
  library(cubble)
  skip_if(packageVersion("cubble") == "0.3.1") # https://github.com/huizezhang-sherry/cubble/issues/30
  library(units)
  tm = set_units(1:6, "days since 1985-01-01")
  # l7 = st_as_stars(L7_ETMs) |> st_set_dimensions(3, values = Sys.time() + 1:6)
  l7 = st_as_stars(L7_ETMs) |> st_set_dimensions(3, values = tm)

  l7sub = l7[,1:30,1:30]
  cu = as_cubble(l7sub)
  s2 = st_as_stars(cu)
  # expect_equal(l7sub, s2)
  expect_true(all.equal(l7sub, s2, check.attributes=FALSE))

  prec_file = system.file("nc/test_stageiv_xyt.nc", package = "stars")
  #(prec = read_stars(gdal_subdatasets(prec_file)[[1]]))
  (prec = read_ncdf(prec_file))
  sf::read_sf(system.file("gpkg/nc.gpkg", package = "sf"), "nc.gpkg") %>%
    st_transform(st_crs(prec)) -> nc # transform from NAD27 to WGS84
  nc_outline = st_union(st_geometry(nc))
  plot_hook = function() plot(nc_outline, border = 'red', add = TRUE)
  prec %>%
    slice(index = 1:12, along = "time") %>%
    plot(downsample = c(3, 3, 1), hook = plot_hook)
  a = aggregate(prec, by = nc, FUN = max)
  a.cb = as_cubble(a, key = id, index = time)
  a2 = st_as_stars(a.cb)
  # expect_equal(a, a2)
  if (packageVersion("units") < '0.8.6')
    a = drop_units(a)
  expect_true(all.equal(a, a2, check.attributes = FALSE))
})

# convert arrays to data.frame, in long form
to_df = function(x) {
	out = dplyr::as_tibble(lapply(x, function(y) structure(y, dim = NULL)))
	if(any(sapply(out, inherits, "sfc"))) sf::st_as_sf(out) else out
}

set_dim = function(x, d) {
	lapply(x, function(y, dims) structure(y, dim = dims), dims = d)
}

get_dims = function(d_cube, d_stars) {
	xy = attr(d_stars, "raster")$dimensions
	d_stars = d_stars[names(d_cube)]
	for (i in seq_along(d_cube)) {
		d_stars[[i]]$values = if (inherits(d_stars[[i]]$values, "intervals")) {
				v = d_stars[[i]]$values
				d_stars[[i]]$values = v[ na.omit(find_interval(d_cube[[i]], v)) ]
			} else if (is.list(d_stars[[i]]$values)) {
				d_stars[[i]]$values[ d_cube[[i]] ]
			} else
				d_cube[[i]]
		d_stars[[i]] = create_dimension(values = d_stars[[i]]$values, point = d_stars[[i]]$point, 
			refsys = d_stars[[i]]$refsys, is_raster = names(d_stars)[i] %in% xy)
	}
	d_stars
}

#' dplyr verbs for stars objects
#' 
#' dplyr verbs for stars objects; package dplyr needs to be loaded before these methods can be used for stars objects.
#' @param .data object of class \code{stars}
#' @param ... see \link[dplyr]{filter}
#' @name dplyr
filter.stars <- function(.data, ...) {
    if (!requireNamespace("dplyr", quietly = TRUE))
        stop("package dplyr required, please install it first") # nocov
    if (!requireNamespace("cubelyr", quietly = TRUE))
        stop("package cubelyr required, please install it first") # nocov
	cb = cubelyr::as.tbl_cube(.data)
	cb = dplyr::filter(cb, ...)
	st_as_stars(cb$mets, dimensions = get_dims(cb$dims, st_dimensions(.data)))
}

#' @name dplyr
filter.stars_proxy = function(.data, ...) {
	collect(.data, match.call(), "filter", ".data", env = environment())
}


#' @name dplyr
mutate.stars <- function(.data, ...) {
	ret = dplyr::mutate(to_df(.data), ...)
	st_as_stars(set_dim(ret, dim(.data)), dimensions = st_dimensions(.data))
}

#' @name dplyr
mutate.stars_proxy = function(.data, ...) {
	collect(.data, match.call(), "mutate", ".data", env = parent.frame())
}

#' @name dplyr
transmute.stars <- function(.data, ...) {
	ret = dplyr::transmute(to_df(.data), ...)
	st_as_stars(set_dim(ret, dim(.data)), dimensions = st_dimensions(.data))
}

#' @name dplyr
transmute.stars_proxy = function(.data, ...) {
	collect(.data, match.call(), "transmute", ".data", env = environment())
}


#' @name dplyr
select.stars <- function(.data, ...) {
    if (!requireNamespace("dplyr", quietly = TRUE))
        stop("package dplyr required, please install it first") # nocov
    ret <- dplyr::select(to_df(.data), ...)
	st_as_stars(set_dim(ret, dim(.data)), dimensions = st_dimensions(.data))
}


#' @name dplyr
select.stars_proxy = function(.data, ...) {
	collect(.data, match.call(), "select", ".data", env = environment())
}

#' @name dplyr
rename.stars <- function(.data, ...) {
    if (!requireNamespace("dplyr", quietly = TRUE))
        stop("package dplyr required, please install it first") # nocov
    ret <- dplyr::rename(to_df(.data), ...)
	st_as_stars(set_dim(ret, dim(.data)), dimensions = st_dimensions(.data))
}

#' @name dplyr
rename.stars_proxy = function(.data, ...) {
	collect(.data, match.call(), "rename", ".data", env = environment())
}

#' @param var see \link[dplyr]{pull}
#' @name dplyr
pull.stars = function (.data, var = -1) {
    if (!requireNamespace("dplyr", quietly = TRUE))
        stop("package dplyr required, please install it first") # nocov
    if (!requireNamespace("rlang", quietly = TRUE))
        stop("package rlang required, please install it first") # nocov
	var = rlang::enquo(var)
	structure(dplyr::pull(to_df(.data), !!var), dim = dim(.data))
}

#' @name dplyr
pull.stars_proxy = function(.data, ...) {
	collect(.data, match.call(), "pull", ".data", env = environment())
}

#' @name dplyr
#' @param x object of class \code{stars}
#' @export
as.tbl_cube.stars = function(x, ...) {
    if (!requireNamespace("cubelyr", quietly = TRUE))
        stop("package cubelyr required, please install it first") # nocov
	cleanup = function(y) {
		if (is.list(y))
			seq_along(y)
		else
			y
	}
	dims = lapply(expand_dimensions(x), cleanup)
	cubelyr::tbl_cube(dims, c(unclass(x)))
}

#' @name dplyr
#' @param along name or index of dimension to which the slice should be applied
#' @param index integer value(s) for this index
#' @param drop logical; drop dimensions that only have a single index?
#' @examples
#' tif = system.file("tif/L7_ETMs.tif", package = "stars")
#' x1 = read_stars(tif)
#' if (require(dplyr, quietly = TRUE)) {
#'  x1 |> slice("band", 2:3)
#'  x1 |> slice("x", 50:100)
#' }
slice.stars <- function(.data, along, index, ..., drop = length(index) == 1) {
  #stopifnot(length(index) == 1)
  if (!requireNamespace("rlang", quietly = TRUE))
    stop("package rlang required, please install it first") # nocov
    
  nd <- length(dim(.data))
  indices <- rep(list(rlang::missing_arg()), nd + 1)
  along = rlang::expr_text(rlang::ensym(along))
  ix = which(along == names(st_dimensions(.data)))[1]
  indices[[ix + 1]] <- index
  indices[["drop"]] <- drop
  
  eval(rlang::expr(.data[!!!indices]))
}

#' @name dplyr
slice.stars_proxy <- function(.data, along, index, ...) {
  # TODO: add adrop argument, this requires an eager implementation of
  # adrop.stars_proxy

  # If there are already operations queued, just add to the queue
  if (!is.null(attr(.data, "call_list")) || 
	  	(length(.data) == 1 && is.character(.data[[1]]))) # #527
    return(collect(.data, match.call(), "slice", ".data",
                   env = parent.frame(), ...))

  # figure out which dimensions are part of the files
  vecsize <- rev(cumprod(rev(dim(.data))))

  # NOTE: The first set of dimensions corresponds to the dimensions in the
  # files. The second set of dimensions corresponds to the list of files. It may
  # be undecided where exactly the break is (at least without reading in the
  # files) if there are a singleton dimensions, I am not sure if this matters,
  # for now just assume the maximum index, this should be the safe choice.
  # Singleton dimensions that are part of files probably need some logic
  # somewhere else and cannot just be ignored.

  # Can we assume, that all elements of .data are the same?
  first_concat_dim <- max(which(vecsize == length(.data[[1]])))
  stopifnot(first_concat_dim > 0)
  all_dims <- st_dimensions(.data)
  file_dims <- all_dims[seq_len(first_concat_dim - 1)]
  concat_dims <- all_dims[first_concat_dim:length(dim(.data))]
  d_concat_dims <- dim(concat_dims)
  l_concat_vec <- prod(d_concat_dims)

  # what is the dimension we have to subset
  ix <- which(names(all_dims) == along) - length(file_dims)
  stopifnot(length(ix) == 1)

  # if the slice is on file dimensions we have to queue the operation
  if (ix <= 0)
    return(collect(.data, match.call(), "slice", ".data",
                   env = parent.frame(), ...))

  # subset indices for the files, it may be faster to calculate these and not
  # take them from an array.
  d <- array(seq_len(l_concat_vec), d_concat_dims)
  idx <- rep(list(quote(expr = )), length(d_concat_dims))
  idx[[ix]] <- index
  vidx <- as.vector(do.call(`[`, c(list(d), idx)))

  # The actual subsetting of files and dimensions
  file_list_new <- lapply(.data, function(x) x[vidx])
  all_dims[[along]] <- all_dims[[along]][index]

  # construct stars_proxy
  st_stars_proxy(as.list(file_list_new), all_dims,
                 NA_value = attr(.data, "NA_value"),
                 resolutions = attr(.data, "resolutions"),
                 RasterIO = attr(.data, "RasterIO"))
}

#' @name st_coordinates
#' @param .x object to be converted to a tibble
as_tibble.stars = function(.x, ..., add_max = FALSE, center = NA) {
    if (!requireNamespace("tibble", quietly = TRUE))
        stop("package tibble required, please install it first") # nocov

	tibble::as_tibble(append(
			st_coordinates(.x, add_max = add_max, center = center),
			lapply(.x, function(y) structure(y, dim = NULL))
		)
	)
}

#' @name dplyr
#' @param data data set to work on
#' @param replace see \link[tidyr]{replace_na}: list with variable=value pairs, where value is the replacement value for NA's
replace_na.stars = function(data, replace, ...) {
    if (!requireNamespace("tidyr", quietly = TRUE))
        stop("package tidyr required, please install it first") # nocov
    if (!requireNamespace("cubelyr", quietly = TRUE))
        stop("package cubelyr required, please install it first") # nocov
	cb = cubelyr::as.tbl_cube(data)
	d = dim(cb$mets[[1]])
	cb$mets = as.data.frame(lapply(cb$mets, as.vector))
	cb$mets = unclass(tidyr::replace_na(cb$mets, replace, ...))
	for (i in seq_along(cb$mets))
		cb$mets[[i]] = structure(cb$mets[[i]], dim = d)
	st_as_stars(cb$mets, dimensions = get_dims(cb$dims, st_dimensions(data)))
}

#' @name dplyr
replace_na.stars_proxy = function(data, ...) {
	collect(data, match.call(), "replace_na", "data", env = environment())
}


#' ggplot geom for stars objects
#' 
#' ggplot geom for stars objects
#' @name geom_stars
#' @param mapping see \link[ggplot2:geom_tile]{geom_raster}
#' @param data see \link[ggplot2:geom_tile]{geom_raster}
#' @param ... see \link[ggplot2:geom_tile]{geom_raster}
#' @param downsample downsampling rate: e.g. 3 keeps rows and cols 1, 4, 7, 10 etc.; a value of 0 does not downsample; can be specified for each dimension, e.g. \code{c(5,5,0)} to downsample the first two dimensions but not the third.
#' @param sf logical; if \code{TRUE} rasters will be converted to polygons and plotted using \link[ggplot2:ggsf]{geom_sf}.
#' @param na.action function; if \code{NA} values need to be removed before plotting use the value \code{na.omit} here (only applies to objects with raster dimensions)
#' @details \code{geom_stars} returns (a call to) either \link[ggplot2:geom_tile]{geom_raster}, \link[ggplot2]{geom_tile}, or \link[ggplot2:ggsf]{geom_sf}, depending on the raster or vector geometry; for the first to, an \link[ggplot2]{aes} call is constructed with the raster dimension names and the first array as fill variable. Further calls to \link[ggplot2:coord_fixed]{coord_equal} and \link[ggplot2]{facet_wrap} are needed to control aspect ratio and the layers to be plotted; see examples. If a \code{stars} array contains hex color values, and no \code{fill} parameter is given, the color values are used as fill color; see the example below.
#' 
#' If visual artefacts occur (Moiré-Effekt), then see the details section of \link{plot.stars}
#' @export
#' @examples
#' system.file("tif/L7_ETMs.tif", package = "stars") |> read_stars() -> x
#' if (require(ggplot2, quietly = TRUE)) {
#'   ggplot() + geom_stars(data = x) +
#'     coord_equal() +
#'     facet_wrap(~band) +
#'     theme_void() +
#'     scale_x_discrete(expand=c(0,0))+
#'     scale_y_discrete(expand=c(0,0))
#'   # plot rgb composite:
#'   st_as_stars(L7_ETMs)[,,,1:3] |> st_rgb() -> x # x contains colors as pixel values
#'   ggplot() + geom_stars(data = x)
#' }
geom_stars = function(mapping = NULL, data = NULL, ..., downsample = 0, sf = FALSE, 
					  na.action = na.pass) {

	if (!requireNamespace("ggplot2", quietly = TRUE))
		stop("package ggplot2 required, please install it first") # nocov
	if (!requireNamespace("tibble", quietly = TRUE))
		stop("package tibble required, please install it first") # nocov

	if (is.null(data)) stop("argument data should be a stars or stars_proxy object")

	for (i in seq_along(data)) {
		if (inherits(data[[i]], "units"))
			data[[i]] = units::drop_units(data[[i]])
	}
	if (inherits(data, "stars_proxy"))
		data = st_as_stars(data, downsample = downsample) # fetches data
	else if (any(downsample > 0))
		data = st_downsample(data, downsample)
	
	all_colors = function (x) {
		is.character(x) && all(nchar(x) %in% c(7, 9) & substr(x, 1, 1) == "#", na.rm = TRUE)
	}
	if (is.null(list(...)$fill) && all_colors(fill <- as.vector(data[[1]])))
		return(geom_stars(mapping = mapping, data = data, sf = sf, na.action = na.action, 
						  ..., fill = fill)) # RETURNS/recurses

	if (is_curvilinear(data) || sf)
		data = st_xy2sfc(data, as_points = FALSE) # removes NA's by default

	d = st_dimensions(data)

	if (has_raster(d) && (is_regular_grid(d) || is_rectilinear(d))) {
		xy = attr(d, "raster")$dimensions
		if (is_regular_grid(d)) {
			mapping = if (is.null(mapping))
					ggplot2::aes(x = !!rlang::sym(xy[1]), y = !!rlang::sym(xy[2]),
						fill = !!rlang::sym(names(data)[1])) 
				else 
           			modifyList( ggplot2::aes(x = !!rlang::sym(xy[1]), y = !!rlang::sym(xy[2]),
						fill = !!rlang::sym(names(data)[1])), mapping) 
			data = na.action(tibble::as_tibble(data))
			ggplot2::geom_raster(mapping = mapping, data = data, ...)
		} else {  # rectilinear: use geom_rect, passing on cell boundaries
			xy_max = paste0(xy, "_max")
			mapping = if (is.null(mapping))
					ggplot2::aes(xmin = !!rlang::sym(xy[1]), ymin = !!rlang::sym(xy[2]),
						xmax = !!rlang::sym(xy_max[1]), ymax = !!rlang::sym(xy_max[2]),
						fill = !!rlang::sym(names(data)[1]))
				else 
					modifyList(ggplot2::aes(xmin = !!rlang::sym(xy[1]), ymin = !!rlang::sym(xy[2]),
						xmax = !!rlang::sym(xy_max[1]), ymax = !!rlang::sym(xy_max[2]),
						fill = !!rlang::sym(names(data)[1])), mapping) 
			data = na.action(tibble::as_tibble(data, add_max = TRUE))
			ggplot2::geom_rect(mapping = mapping, data = data, ...)
		}
	} else if (has_sfc(d)) {
		if (is.null(mapping)) {
			mapping = ggplot2::aes(fill = !!rlang::sym(names(data)[1])) } else {
		        mapping = modifyList( ggplot2::aes(fill = !!rlang::sym(names(data)[1])), mapping) }
		ggplot2::geom_sf(data = st_as_sf(data, long = TRUE), mapping = mapping, ...)
	} else
		stop("geom_stars only works for objects with raster or vector geometries")
						     
}

#' @name geom_stars
theme_stars = function(...) {

	if (!requireNamespace("ggplot2", quietly = TRUE))
		stop("package ggplot2 required, please install it first") # nocov

	# coord_equal() +
    # scale_fill_viridis() +
    # scale_x_discrete(expand=c(0,0)) +
    # scale_y_discrete(expand=c(0,0)) +
    ggplot2::theme_void()
}

register_all_s3_methods = function() {
	register_s3_method("cubelyr", "as.tbl_cube", "stars") # nocov start
	register_s3_method("dplyr", "filter", "stars") 
	register_s3_method("dplyr", "filter", "stars_proxy") 
	register_s3_method("dplyr", "select", "stars")
	register_s3_method("dplyr", "select", "stars_proxy")
	register_s3_method("dplyr", "mutate", "stars")
	register_s3_method("dplyr", "mutate", "stars_proxy")
	register_s3_method("dplyr", "pull", "stars")
	register_s3_method("dplyr", "pull", "stars_proxy")
	register_s3_method("dplyr", "rename", "stars")
	register_s3_method("dplyr", "rename", "stars_proxy")
	register_s3_method("dplyr", "slice", "stars")
	register_s3_method("dplyr", "slice", "stars_proxy")
	register_s3_method("dplyr", "transmute", "stars")
	register_s3_method("dplyr", "transmute", "stars_proxy")
	register_s3_method("tidyr", "replace_na", "stars")
	register_s3_method("tidyr", "replace_na", "stars_proxy")
	register_s3_method("lwgeom", "st_transform_proj", "stars")
	register_s3_method("sf", "st_join", "stars")
	register_s3_method("spatstat.geom", "as.owin", "stars")
	register_s3_method("spatstat.geom", "as.im", "stars")
	register_s3_method("tibble", "as_tibble", "stars")
	register_s3_method("xts", "as.xts", "stars") # nocov end
}

# from: https://github.com/tidyverse/hms/blob/master/R/zzz.R
# Thu Apr 19 10:53:24 CEST 2018
#nocov start
register_s3_method <- function(pkg, generic, class, fun = NULL) {
  stopifnot(is.character(pkg), length(pkg) == 1)
  stopifnot(is.character(generic), length(generic) == 1)
  stopifnot(is.character(class), length(class) == 1)

  if (is.null(fun)) {
    fun <- get(paste0(generic, ".", class), envir = parent.frame())
  } else {
    stopifnot(is.function(fun))
  }

  if (pkg %in% loadedNamespaces()) {
    registerS3method(generic, class, fun, envir = asNamespace(pkg))
  }

  # Always register hook in case package is later unloaded & reloaded
  setHook(
    packageEvent(pkg, "onLoad"),
    function(...) {
      registerS3method(generic, class, fun, envir = asNamespace(pkg))
    }
  )
}
#nocov end

suppressPackageStartupMessages(library(stars))
set.seed(1331)
st_sample(st_as_sfc(st_bbox(L7_ETMs)), 1)
b = st_buffer(st_sample(st_as_sfc(st_bbox(L7_ETMs)), 1), 335)
bba = stars:::st_align(st_bbox(b), st_dimensions(L7_ETMs))

plot(L7_ETMs[,,,1], extent = st_buffer(b, 20), reset = FALSE, key.pos = NULL)
plot(b, add = TRUE, col = NA, border = 'red')
plot(st_as_sfc(st_bbox(b)), col = NA, border = 'red', add = TRUE)
plot(st_as_sfc(bba), col = NA, border = 'blue', add = TRUE)
st_bbox(b)
bba

plot(L7_ETMs[,,,1], reset = FALSE, key.pos = NULL)
plot(b, col = 'red', add = TRUE)
plot(st_as_sfc(bba), col = NA, border = 'blue', add = TRUE)

b0 = st_buffer(b, -100)
L7 = st_as_stars(L7_ETMs)[b0][,,,1] |> st_normalize()
bbb = stars:::st_align(st_bbox(b), st_dimensions(L7))
plot(L7, extent = st_buffer(b, 120), reset = FALSE, key.pos = NULL)
plot(st_as_sfc(st_bbox(L7)), col = NA, border = 'black', add = TRUE)
plot(st_as_sfc(bbb), col = NA, border = 'blue', add = TRUE)
plot(b, col = NA, border = 'red', add = TRUE)

all.equal(bbb, bba)

suppressPackageStartupMessages(library(stars))

tif = system.file("tif/L7_ETMs.tif", package = "stars")
(x = read_stars(tif))

# write stars object:
write_stars(x, tempfile(), driver = "GTiff")

# write stars_proxy object:
(x = read_stars(tif, proxy = TRUE))
write_stars(x, tempfile(), driver = "GTiff")

# show that extract with exact=TRUE equals st_interpolate_aw(), only much faster
options(digits=15)
suppressPackageStartupMessages(library(stars))
L = st_as_stars(L7_ETMs)[,,,1]
set.seed(123432)
st_bbox(L) |> 
  st_as_sfc() |> 
  st_centroid() |> 
  st_buffer(50) |>
  st_sample(10) |> 
  st_combine() |> 
  st_convex_hull() -> ch
if (interactive()) {
 plot(L, reset = FALSE)
 plot(ch, add = TRUE, border = 'red', col = NA)
}
if (FALSE) {
e = st_extract(L, ch)[[1]]
a = aggregate(L, ch, mean)[[1]]
identical(e, a) # same code path
ee = aggregate(L, ch, mean, exact = TRUE)[[1]]
all.equal(e, ee)
aw = suppressWarnings(st_interpolate_aw(L, ch, extensive = FALSE)$V1)
all.equal(as.vector(ee), aw)
ee0 = st_extract(L, ch, exact = TRUE)[[1]]
all.equal(ee0, ee) # ??
}

# things related to the CFtime and ncdfCF packages

CF_calendar_regular = c("standard", "gregorian", "proleptic_gregorian", "utc", "tai")
CF_calendar_model = c("360_day", "365_day", "noleap", "366_day", "all_leap")

is_CFTime = function(x) {
	# catch things like "CFTime_360_day" as a CFTime refsys
	length(x) == 1 && !is.na(x) && is.character(x) && substr(x, 1, 6) == "CFTime"
}

#' @param sf_geometry sf data.frame with geometry and attributes to be added to stars object.
#' Must have same number of rows as timeseries instances.
#' @details For the \code{ncdfgeom} method: objects are point-timeseries with optional line or polygon geometry for each timeseries specified with the \code{sf_geometry} parameter. See \pkg{ncdfgeom} for more about this NetCDF-based format for geometry and timeseries.
#' @name st_as_stars
#' @export
st_as_stars.ncdfgeom <- function(.x, ..., sf_geometry = NA) {

  crs <- sf::st_crs('OGC:CRS84')

  if(length(.x$alts) == 0) {
    ts_points <- data.frame(X = .x$lons, Y = .x$lats)
    ts_points <- sf::st_as_sf(ts_points, coords = c("X", "Y"), crs = crs)

  } else {
    ts_points <- data.frame(X = .x$lons, Y = .x$lats, Z = .x$alts)
    ts_points <- sf::st_as_sf(ts_points, coords = c("X", "Y", "Z"), crs = crs)
  }


  data <- .x$data_frames[[1]]

  gdim <- create_dimension(from = 1, to = length(.x$lats),
                           refsys = crs, point = TRUE,
                           values = ts_points$geometry)
  tdim <- create_dimension(from = 1, to = length(.x$time),
                           refsys = "POSIXct", point = FALSE,
                           values = as.POSIXct(.x$time))
  dim <- list(time = tdim, points = gdim)

  if(inherits(sf_geometry, "sf")) {
    if(length(gdim$values) != length(st_geometry(sf_geometry)))
      stop("geometry must be same length as instance dimension of timeseries")

    is_point <- any(grepl("point", class(st_geometry(sf_geometry)), ignore.case = TRUE))

    sf_dim <- create_dimension(from = 1, to = length(gdim$values),
                               refsys = st_crs(sf_geometry),
                               point = is_point, is_raster = FALSE,
                               values = st_geometry(sf_geometry))

    dim <- c(dim, list(geometry = sf_dim))
  }

  st_stars(x = setNames(list(as.matrix(.x$data_frames[[1]])),
                        .x$varmeta[[1]]$name),
           dimensions =  create_dimensions(dim))
}

# Take a CFVariable `var` and convert its axes into a dimensions structure
.dimensions_from_CF <- function(var) {
	axes = var$axes
	axis_order = match(sapply(axes, function(ax) ax$orientation), c("X", "Y", "Z", "T"))
	if (any(is.na(axis_order[1:2])))
		stop("Data variable does not have X and/or Y axes")
	if (axes[[ axis_order[1] ]]$length == 1 || axes[[ axis_order[2] ]]$length == 1)
		stop("Raster dimensions are degenerate")
	
	crs = if (is.null(var$crs))
		"OGC:CRS84" # If no grid_mapping, then data is lat-long, assuming WGS84 datum here,
	else {      # otherwise, tease the AUTHORITY out of the WKT2 string of the grid_mapping
				# This will often fail to identify an AUTHORITY because CF grid_mappings
				# are incomplete and a WKT2 thus has many "unknown"s.
		wkt = var$crs_wkt2
		auth = regexec('([a-zA-Z0-9]*)",["]?([a-zA-Z0-9]*)["]?\\]\\]$', wkt)
		auth = regmatches(wkt, auth)[[1]]
		if (!length(auth))
			NA_character_
		else
			paste(auth[-1], collapse = ":")
	}
	
	raster = get_raster(dimensions = c(axes[[ axis_order[1] ]]$name, axes[[ axis_order[2] ]]$name))
	dimensions = lapply(axes, function(ax) {
		if (ax$length == 1) 
			return (NULL)
			
		switch(ax$orientation,
			   X = create_dimension(values = ax$coordinates, refsys = crs, is_raster = TRUE), 
			   Y = create_dimension(values = ax$coordinates, refsys = crs, is_raster = TRUE),
			   T = create_dimension(to = length(time), refsys = "CFtime", values = ax$time),
		   		{
					v = ax$coordinates
		   			u = ax$attribute("units")
					if (!is.na(u))
						units(v) = try_as_units(u)
		   			create_dimension(values = v)
		   		}
			)
		}
	)
	dimensions = dimensions[lengths(dimensions) > 0]
	create_dimensions(dimensions, raster)
}

#' @param .var Character vector with names of netCDF data variables to convert
#'   into stars attributes. All data variables have to use the same set of axes.
#'   If omitted, the first data variable will be used.
#' @param all_compatible Logical flag that indicates if all data variables with
#'   the same axes as argument `var` uses should be converted into attributes.
#'   Default is `TRUE` when only one data variable is specified, `FALSE`
#'   otherwise.
#' @name st_as_stars
#' @export
st_as_stars.CFDataset <- function(.x, ..., .var = .x$var_names[1], all_compatible = length(.var) == 1) {
	if (!requireNamespace("ncdfCF", quietly = TRUE))
		stop("Please install package 'ncdfCF' before using this functionality")
	
	all_var_names = .x$var_names
	if (!is.character(.var) || !length(.var) || !all(.var %in% all_var_names))
		stop("Argument `.var` must be a character vector of one of more data variable names in the data set.")
	all_compatible = is.logical(all_compatible) && all_compatible[1]
	
	# Grab all CFVariables from the passed data variable names - check that all are compatible on dimensions
	axis_names = sapply(.x[[ .var[1] ]]$axes, function(ax) ax$fullname)
	num_axes = length(axis_names)
	if (length(.var) > 1) {
		vars = lapply(.var, function(v_name) {
			v = .x[[v_name]]
			v_ax_names = sapply(v$axes, function(ax) ax$fullname)
			if (length(v_ax_names) == num_axes && all(v_ax_names == axis_names)) 
				v
			else
				NULL
		})
		if (any(is.null(vars)))
			stop("Data variables in argument `.var` have different axes.")
		names(vars) <- sapply(vars, function(v) v$name)
	} else if (all_compatible) {
		vars = lapply(all_var_names, function(nm) {
			v = .x[[nm]]
			if (all(sapply(v$axes, function(ax) ax$fullname) == axis_names))
				v
			else
				NULL
		})
		names(vars) <- sapply(vars, function(v) v$name)
		vars = vars[lengths(vars) > 0]
		if (!all(.var %in% names(vars)))
			stop("Data variables in argument `var` have different axes.")
	} else
		vars = setNames(list(.x[[ .var[1] ]]), .var[1])
	
	# Turn the list of CFVariables into stars attributes
	attributes = lapply(vars, function(v) {
		att = v$raw()
		dimnames(att) <- NULL
		un = v$attribute("units")
		if (!is.na(un))
			units(att) = try_as_units(un)
		att
	})
	names(attributes) = names(vars)
	
	dimensions = .dimensions_from_CF(vars[[1]])	
	st_stars(attributes, dimensions = dimensions)
}

#' @name st_as_stars
#' @export
st_as_stars.CFVariable <- function(.x, ...) {
	if (!requireNamespace("ncdfCF", quietly = TRUE))
		stop("Please install package 'ncdfCF' before using this functionality")
	
	dimensions = .dimensions_from_CF(.x)
	att <- .x$raw()
	dimnames(att) <- NULL
	attributes = setNames(list(att), .x$name)
	un = .x$attribute("units")
	if (!is.na(un))
		units(attributes[[1]]) = try_as_units(un)
	
	st_stars(attributes, dimensions = dimensions)
}

suppressPackageStartupMessages(library(stars))
library(sf)
#jp2 = system.file("jp2/B01.jp2", package = "stars")
#(x = read_stars(jp2, options = c("OVERVIEW_LEVEL=3")))
tif = system.file("tif/L7_ETMs.tif", package = "stars")
#(x = read_stars(tif, options = c("OVERVIEW_LEVEL=3"))) # FIXME: breaks on windows
(x = read_stars(tif)[,1:20,1:20])
# names(x) = "L7_ETM"

# library(abind)
# x = adrop(x)
image(x)
(sfc = st_as_sfc(x, as_points = FALSE))
plot(sfc, add  =TRUE)
(sfc = st_as_sfc(x, as_points = TRUE))
plot(sfc, add = TRUE)

sf = st_as_sf(x, as_points = FALSE)
plot(sf, border = NA)
sf = st_as_sf(x[,,,1], as_points = FALSE, merge = TRUE)
plot(sf, border = NA)

sfc1 <- st_as_sfc(x, as_points = TRUE, use_cpp = TRUE, na.rm = FALSE)
sfc2 <- st_as_sfc(x, as_points = TRUE, use_cpp = FALSE)
identical(sfc1, sfc2)

sfc1 <- st_as_sfc(x, as_points = FALSE, use_cpp = TRUE, na.rm = FALSE)
sfc2 <- st_as_sfc(x, as_points = FALSE, use_cpp = FALSE)
identical(sfc1, sfc2)

# sf -> stars -> sf
x = st_sfc(st_point(0:1), st_point(1:2), st_point(2:3))
m = matrix(1:9,3)
colnames(m) = c("a", "b", "c")
foo = st_sf(m, geom = x)
x = st_as_stars(foo)
c(x)
st_redimension(x)
c(x, x)
(y = st_redimension(x, along = list(time = as.Date("2017-11-27") + 0:2)))
st_as_sf(y)
st_as_sf(st_as_stars(foo))

c(x,x,along=1) # NEEDS FIXING!
# st_as_stars.sfc
st_as_stars(st_geometry(st_as_sf(y)), nx = 10, ny = 10, as_points = FALSE)

# https://github.com/r-spatial/stars/issues/295
suppressPackageStartupMessages(library(stars))
raw <- read_stars(system.file("nc/bcsd_obs_1999.nc", package = "stars"))
foo <- function(x, idx) stats::lowess(idx, x)$y

timeline <- st_get_dimension_values(raw, "time")
smooth = st_apply(raw,
	MARGIN = c("x", "y"),
	FUN = foo,
	idx = st_get_dimension_values(raw, "time")
)

st_set_dimensions(smooth,
	which = "foo",
	values = st_get_dimension_values(raw, "time"),
	names = "time"
)

raw |>
  st_apply(MARGIN = c("x", "y"), FUN = foo, idx = timeline) |>
  st_set_dimensions("foo", st_dimensions(raw)["time"])

suppressPackageStartupMessages(library(stars))
#suppressPackageStartupMessages(library(lwgeom))
geomatrix = system.file("tif/geomatrix.tif", package = "stars")
(x = read_stars(geomatrix))
new = st_crs(4326)
y = st_transform(x, new)
plot(st_transform(st_as_sfc(st_bbox(x)), new), col = NA, border = 'red')
plot(st_as_sfc(y, as_points=FALSE), col = NA, border = 'green', axes = TRUE, add = TRUE)
image(y, add = TRUE)
plot(st_as_sfc(y, as_points=TRUE), pch=3, cex=.5, col = 'blue', add = TRUE)
plot(st_transform(st_as_sfc(x, as_points=FALSE), new), add = TRUE)

tif = system.file("tif/L7_ETMs.tif", package = "stars")
x = read_stars(tif)[,1:10,1:10,1:3]
x_ = st_transform(x, st_crs(4326))
## IGNORE_RDIFF_BEGIN
if (require(lwgeom, quietly = TRUE)) {
  x__ = st_transform_proj(x, st_crs(4326)$proj4string)
  st_crs(x_) = st_crs(x__)
  print(all.equal(x_,x__))

# x__ = st_transform(x, x_) #now in st_warp 
# all.equal(x_, x__)

  x = st_xy2sfc(x, as_points = FALSE)
  (x_ = st_transform(x, st_crs(4326)))
  (x__ = st_transform_proj(x, st_crs(4326)))
  print(all.equal(x_,x__,check.attributes = FALSE))
}
## IGNORE_RDIFF_END

# nothing to transform: now errs:
try(st_transform(st_as_stars(list(m = matrix(1,10,10))), st_crs(4326)))
try(st_transform_proj(st_as_stars(list(m = matrix(1,10,10))), st_crs(4326)))

#' cut methods for stars objects
#' 
#' cut methods for stars objects
#' @name cut_stars
#' @param x see \link[base]{cut}
#' @param breaks see \link[base]{cut}
#' @param ... see \link[base]{cut}
#' @return an array or matrix with a \code{levels} attribute; see details
#' @details R's \code{factor} only works for vectors, not for arrays or matrices. This is a work-around (or hack?) to keep the factor levels generated by \code{cut} and use them in plots.
#' @export
#' @examples
#' tif = system.file("tif/L7_ETMs.tif", package = "stars")
#' x = read_stars(tif)
#' cut(x, c(0, 50, 100, 255))
#' cut(x[,,,1], c(0, 50, 100, 255))
#' plot(cut(x[,,,1], c(0, 50, 100, 255)))
cut.array = function(x, breaks, ...) { 
	structure(cut(as.vector(x), breaks, ...), dim = dim(x))
}

#' @name cut_stars
#' @export
cut.matrix = cut.array

#' @name cut_stars
#' @export
#' @examples
#' tif = system.file("tif/L7_ETMs.tif", package = "stars")
#' x1 = read_stars(tif)
#' (x1_cut = cut(x1, breaks = c(0, 50, 100, Inf)))  # shows factor in summary
#' plot(x1_cut[,,,c(3,6)]) # propagates through [ and plot
cut.stars = function(x, breaks, ...) {
	my_cut = function(x, breaks, ...) structure(cut(x, breaks, ...), dim = dim(x))
	st_stars(lapply(x, my_cut, breaks = breaks, ...), st_dimensions(x))
}

#' @export
droplevels.stars = function(x, ...) {
	drop_level = function(x, ...) {
		d = dim(x)
		l = levels(x)
		co = attr(x, "colors")
		dim(x) = NULL
		x = droplevels(x, ...)
		sel = match(levels(x), l)
		structure(x, dim = d, colors = co[sel])
	}
	for (i in seq_along(x)) {
		if (inherits(x[[i]], "factor"))
			x[[i]] = drop_level(x[[i]], ...)
	}
	x
}

#' @export
droplevels.stars_proxy = function(x, ...) {
	collect(x, match.call(), "droplevels", env = environment())
}

suppressPackageStartupMessages(library(stars))
# predict:
tif = system.file("tif/L7_ETMs.tif", package = "stars")
(x = read_stars(tif))
st_crs(x) = st_crs(x) # cleans trailing space
model = lm(x~L7_ETMs.tif, head(as.data.frame(x), 50))
x = predict(x, model)

(y = read_stars(tif, proxy = TRUE))
(yp = predict(y, model))
all.equal(x, st_as_stars(yp))

yy = st_set_dimensions(y, 3, values = paste0("band", 1:6))
(y_spl = split(yy, 3))
as.data.frame(st_as_stars(split(yy, 3)))[1:3,]

# sf conversion things

#' @export
#' @name st_as_sf
st_as_sfc.stars = function(x, ..., as_points, which = seq_len(prod(dim(x)[1:2]))) {

	r = attr(st_dimensions(x), "raster")
	gt = st_geotransform(x)
	d = st_dimensions(x)[r$dimensions]
	if (xor(is.null(d[[1]]$values), is.null(d[[2]]$values))) {
		# mixed regular/rectilinear dimensions: make rectilinear; https://github.com/r-spatial/stars/issues/458
		e = expand_dimensions(d)
		if (is.null(d[[1]]$values)) {
			d[[1]]$from = d[[1]]$to = NA
			d[[1]]$values = e[[1]]
		}
		if (is.null(d[[2]]$values)) {
			d[[2]]$from = d[[2]]$to = NA
			d[[2]]$values = e[[2]]
		}
	}
	sfc = st_as_sfc(d, ..., as_points = as_points, which = which, geotransform = gt) 
	# swap axes?
	if (st_axis_order() && isTRUE(st_crs(x, parameters = TRUE)$yx))
		st_transform(sfc, pipeline = "+proj=pipeline +step +proj=axisswap +order=2,1")
	else
		sfc
}

#' replace x y raster dimensions with simple feature geometry list (points, or polygons = rasterize) and vice versa
#' @param x object of class \code{stars}
#' @param as_points logical; if \code{TRUE}, generate points at cell centers, else generate polygons
#' @param ... for `st_xy2sfc`: arguments passed on to \code{st_as_sfc}, for `st_sfc2xy` arguments passed on to \link{st_as_stars.data.frame}
#' @param na.rm logical; omit (remove) cells which are entirely missing valued (across other dimensions)?
#' @details `st_xy2sfc` replaces x y raster dimensions with simple feature geometry list (points, or polygons = rasterize)
#' @return `st_xy2sfc` returns an object of class \code{stars} with x and y raster dimensions replaced by a single sfc geometry list column containing either points, or polygons. Adjacent cells with identical values are not merged; see \code{st_rasterize} for this.
#' @export
st_xy2sfc = function(x, as_points, ..., na.rm = TRUE) {

	if (inherits(x, "stars_proxy"))
		x = st_as_stars(x)

	d = st_dimensions(x)
	olddim = dim(x)

	if (! has_raster(x))
		stop("x and/or y not among dimensions")

	x = st_upfront(x)

	dxy = attr(d, "raster")$dimensions
	xy_pos = match(dxy, names(d))
	stopifnot(all(xy_pos == 1:2))
	if (missing(as_points) && isTRUE(d[[dxy[1]]]$point) && isTRUE(d[[dxy[2]]]$point))
		as_points = TRUE

	# find which records are NA for all attributes:
	a = abind(x, along = length(dim(x)) + 1)
	keep = if (na.rm)
			as.vector(apply(a, c(1,2), function(x) !all(is.na(x))))
		else
			rep(TRUE, prod(dim(x)[1:2]))

	# flatten two dims x,y to one dim sfc (replacing "x")
	sfc = st_as_sfc(x, as_points = as_points, ..., which = which(keep))
	# overwrite raster-x with sfc:
	d[[ dxy[1] ]] = create_dimension(from = 1, to = length(sfc), values = sfc)
	# rename raster-x to sfc:
	names(d)[names(d) == dxy[1] ] = "geometry"
	# remove y:
	d[[ dxy[2] ]] = NULL
	attr(d, "raster") = get_raster(dimensions = rep(NA_character_, 2))
	# flatten arrays:
	x = unclass(x) # to omit using [[<-.stars method:
	for (i in seq_along(x))
		dim(x[[i]]) = c(geometry = length(keep), olddim[-xy_pos]) 
	# reduce arrays to non-NA cells:
	if (na.rm) {
		args = rep(list(rlang::missing_arg()), length(dim(x[[1]])))
		args[[1]] = which(keep)
		args[["drop"]] = FALSE
		for (i in seq_along(x))
			x[[i]] = structure(eval(rlang::expr(x[[i]][ !!!args ])), 
				levels = attr(x[[i]], "levels"))
	}
	st_stars(x, dimensions = d)
}

#' @details `st_sfc2xy` replaces POINT simple feature geometry list with an x y raster
#' @param x object of class \code{stars}, or of class \code{sf}
#' @return `st_sfc2xy` returns an object of class \code{stars} with a POINT list replaced by X and Y raster dimensions. This only works when the points are distributed over a regular or rectilinear grid.
#' @export
#' @name st_xy2sfc
#' @examples
#' (reduced_nc = read_ncdf(system.file("nc/reduced.nc", package = "stars")))
#' (x = stars::st_xy2sfc(reduced_nc, as_points = TRUE, na.rm = FALSE))
#' # roundtrip:
#' st_sfc2xy(x, dims=c("X", "Y","zlev","time"))
st_sfc2xy = function(x, ...) {
	if (inherits(x, "sf"))
		x = st_as_stars(x)
	i = which_sfc(x)
	if (length(i) == 0)
		stop("x does not contain a geometry dimension")
	if (length(i) > 1) {
		stop("using first geometry dimension only")
		i = i[1]
	}
	d = st_dimensions(x)
	if (!inherits(d[[i]]$values, "sfc_POINT"))
		stop("point geometries expected")
	cc = st_coordinates(d[[i]]$values)
	df = as.data.frame(x)
	df$geometry = NULL
	s = st_as_stars(cbind(cc, df), ...)
	st_set_crs(s, st_crs(d))
}

#' @export
st_as_sf.dimensions = function(x, ...) {
	ix = which_sfc(x)[1]
	st_sf(setNames(list(x[[ ix ]]$values), names(x)[ix]), crs = st_crs(x), ...)
}


#' Convert stars object into an sf object
#' 
#' Convert stars object into an sf object
#' @name st_as_sf
#' @param x object of class \code{stars}
#' @param as_points logical; should cells be converted to points or to polygons? See details.
#' @param which linear index of cells to keep (this argument is not recommended to be used)
#' @param na.rm logical; should missing valued cells be removed, or also be converted to features?
#' @param merge logical; if \code{TRUE}, cells with identical values are merged (using \code{GDAL_Polygonize} or \code{GDAL_FPolygonize}); if \code{FALSE}, a polygon for each raster cell is returned; see details
#' @param use_integer (relevant only if \code{merge} is \code{TRUE}): if \code{TRUE}, before polygonizing values are rounded to 32-bits signed integer values (GDALPolygonize), otherwise they are converted to 32-bit floating point values (GDALFPolygonize).
#' @param long logical; if \code{TRUE}, return a long table form \code{sf}, with geometries and other dimensions recycled
#' @param connect8 logical; if \code{TRUE}, use 8 connectedness. Otherwise the 4 connectedness algorithm will be applied.
#' @param ... ignored
#' @details If \code{merge} is \code{TRUE}, only the first attribute is converted into an \code{sf} object. If \code{na.rm} is \code{FALSE}, areas with \code{NA} values are also written out as polygons. Note that the resulting polygons are typically invalid, and use \link[sf:valid]{st_make_valid} to create valid polygons out of them.
#' @export
#' @examples
#' tif = system.file("tif/L7_ETMs.tif", package = "stars")
#' x = read_stars(tif)
#' x = x[,1:100,1:100,6] # subset of a band with lower values in it
#' x[[1]][x[[1]] < 30] = NA # set lower values to NA
#' x[[1]] = x[[1]] < 100 # make the rest binary
#' x
#' (p = st_as_sf(x)) # removes NA areas
#' (p = st_as_sf(x[,,,1], merge = TRUE)) # glues polygons together
#' all(st_is_valid(p)) # not all valid, see details
#' plot(p, axes = TRUE)
#' (p = st_as_sf(x, na.rm = FALSE, merge = TRUE)) # includes polygons with NA values
#' plot(p, axes = TRUE)
st_as_sf.stars = function(x, ..., as_points = FALSE, merge = FALSE, na.rm = TRUE, 
		use_integer = is.logical(x[[1]]) || is.integer(x[[1]]), long = FALSE, connect8 = FALSE) { 

	if (length(x) == 0)
		return(st_as_sf.dimensions(st_dimensions(st_xy2sfc(x, as_points = as_points, 
			na.rm = FALSE))))

	crs = st_crs(x)
	d = st_dimensions(x)
	if (merge && !as_points && has_raster(x) && !any(is.na(st_geotransform(x)))) { # uses GDAL polygonize path:
		x = st_normalize(x)
		mask = if (na.rm) {
				mask = x[1]
				mask[[1]] = !is.na(mask[[1]])
				mask
			} else
				NULL

		ret = gdal_polygonize(x, mask, use_integer = use_integer,
				geotransform = st_geotransform(x), use_contours = FALSE, connect8 = connect8, ...)

		# factor levels?
		if (!is.null(lev <- attr(x[[1]], "levels"))) {
			ex = attr(x[[1]], "exclude")
			if(any(ex)){
				id = (seq_along(ex) - 1)[!ex] # index-values matching to levels
				ret[[1]] = structure(match(ret[[1]], id), class = "factor", levels = lev) # ordered?
			} else {
				ret[[1]] = structure(ret[[1]], class = "factor", levels = lev) # ordered?
			}
		}
		st_set_crs(ret, crs)
	} else {
		if (merge)
			stop("merge=TRUE and as_points=TRUE unsupported; consider using st_contour for generating contour lines")

		if (has_raster(x))
			x = st_xy2sfc(st_upfront(x), as_points = as_points, ..., na.rm = na.rm)

		if (! has_sfc(x))
			stop("no feature geometry column found")
	
		if (long) {
			st_as_sf(as.data.frame(x), crs = crs)
		} else {
			ix = which_sfc(x)
			if (length(ix) > 1)	
				warning("working on the first sfc dimension only") # FIXME: this probably only works for 2D arrays, now
			other_dim = setdiff(seq_along(dim(x)), ix[1])
			sfc = st_dimensions(x)[[ ix[1] ]]$values
			# other_values = st_dimensions(x)[[ other_dim[1] ]]$values
			other_values = lapply(st_dimensions(x)[other_dim], function(x) x$values)
			for (i in seq_along(other_values))
				if (inherits(other_values[[i]], "intervals"))
					other_values[[i]] = format(other_values[[i]])
			varnames = apply(do.call(expand.grid, other_values), 1, paste, collapse = ".")
			un_dim = function(x) { # remove a dim attribute from data.frame columns
				for (i in seq_along(x))
					x[[i]] = structure(x[[i]], dim = NULL)
				x
			}
			dfs = lapply(x, function(y) un_dim(as.data.frame(y)))
			nc = sapply(dfs, ncol)
			df = do.call(cbind, dfs)
	
			if (length(dim(x)) == 1) # one-dimensional cube...
				names(df) = names(x)
			else if (length(varnames) == ncol(df))
				names(df) = varnames
			else if (length(unique(names(df))) < ncol(df) && length(names(dfs)) == ncol(df)) # I hate this
				names(df) = names(dfs)
			else { # another exception... time as second dimension
				e = expand_dimensions(x)
				if (length(e[-ix]) == 1 && inherits(e[-ix][[1]], c("Date", "POSIXt"))) {
					names(df) = if (length(nc) > 1) {
							nm = expand.grid(e[-ix][[1]], names(x))
							paste(nm[[2]], nm[[1]], sep = ".")
						} else
							as.character(e[-ix][[1]])
				}
			}
	
			df[[ names(st_dimensions(x))[ ix[1] ] ]] = sfc # keep dimension name
			st_sf(df, crs = crs)
		}
	}
}


#' @export
#' @name st_as_sf
#' @param downsample see \link{st_as_stars}
st_as_sf.stars_proxy = function(x, ..., downsample = 0) {
	st_as_sf(st_as_stars(x, downsample = downsample), ...)
}

#' Compute or plot contour lines or sets
#' 
#' Compute contour lines or sets
#' @param x object of class \code{stars}
#' @param na.rm logical; should missing valued cells be removed, or also be converted to features?
#' @param contour_lines logical; if \code{FALSE}, polygons are returned (contour sets), otherwise contour lines
#' @param breaks numerical; values at which to "draw" contour levels
#' @details this function requires GDAL >= 2.4.0
#' @seealso for polygonizing rasters following grid boundaries, see \link{st_as_sf} with arguments \code{as_points=FALSE} and \code{merge=TRUE}; \link{contour} plots contour lines using R's native algorithm (which also plots contour levels)
#' @export
st_contour = function(x, na.rm = TRUE, contour_lines = FALSE, 
		breaks = pretty(na.omit(as.vector(x[[1]])))) {
#nocov start
	x = st_normalize(x)
	mask = if (na.rm) { 
			mask = x[1]
			mask[[1]] = !is.na(mask[[1]])
			mask
		} else
			NULL
	ret = gdal_polygonize(x, mask, use_integer = FALSE, geotransform = st_geotransform(x),
			use_contours = TRUE, contour_lines = contour_lines, breaks = breaks)
	# factor levels?
	if (!is.null(lev <- attr(x[[1]], "levels")))
		ret[[1]] = structure(ret[[1]], class = "factor", levels = lev)
	st_set_crs(ret, st_crs(x))
#nocov end
}

#' @export
st_as_stars.sfc = function(.x, ..., FUN = length, as_points = TRUE) {
	st = st_as_stars(st_bbox(.x), ...)
	sfc = st_as_sfc(st, as_points = as_points)
	i = st_intersects(sfc, .x)
	vals = sapply(i, FUN)
	st[[1]] = array(vals, dim(st[[1]]))
	st
}

#' @name st_as_stars
#' @details The \code{st_as_stars} method for \code{sf} objects without any additional
#' arguments returns a one-dimensional data cube with a dimension for the simple features
#' geometries, and all remaining attributes as data cube attributes. When used with
#' further arguments, the method for \code{data.frame}s is called.
#' @examples
#' nc = st_read(system.file("gpkg/nc.gpkg", package="sf"))
#' st_as_stars(nc)
#' @export
st_as_stars.sf = function(.x, ..., dims = attr(.x, "sf_column")) {
	.x = as.data.frame(.x)
	NextMethod(dims = dims, ...)
}

suppressPackageStartupMessages(library(stars))
tif = system.file("tif/L7_ETMs.tif", package = "stars")
x = read_stars(tif)
xy = structure(list(x = c(293253.999046018, 296400.196497684), y = c(9113801.64775462, 
9111328.49619133)), names = c("x", "y"))
pts = st_as_sf(data.frame(do.call(cbind, xy)), coords = c("x", "y"), crs = st_crs(x))
image(x, axes = TRUE)
plot(st_as_sfc(st_bbox(pts)), col = NA, add = TRUE)

bb = st_bbox(pts)
(xx = x[bb])
image(xx)
plot(st_as_sfc(bb), add = TRUE, col = NA)
st_contains(st_as_sfc(st_bbox(xx)), st_as_sfc(st_bbox(pts)))

image(x)
pt = st_point(c(x = 290462.103109179, y = 9114202.32594085))
buf = st_buffer(st_sfc(pt, crs = st_crs(x)), 1500)
plot(buf, add = TRUE)

buf = st_sfc(st_polygon(list(st_buffer(pt, 1500)[[1]], st_buffer(pt, 1000)[[1]])),
	crs = st_crs(x))
image(x[buf])
plot(buf, add = TRUE, col = NA)
image(x[buf,crop=FALSE])
plot(buf, add = TRUE, col = NA)

plot(x, rgb = 1:3)

# #638:
l = st_as_stars(L7_ETMs)
st_bbox(l) |> st_as_sfc() -> bb
(bb + 1e6) |> st_set_crs(st_crs(l)) -> b
st_crop(l, b)
st_intersects(b, st_as_sfc(st_bbox(l))) |> as.matrix()
(bb - 1e6) |> st_set_crs(st_crs(l)) -> b
st_crop(l, b)

#' build mosaic (composite) of several spatially disjoint stars objects
#' 
#' build mosaic (composite) of several spatially disjoint stars objects
#' @param .x object of class stars, or character vector with input dataset names
#' @param ... further input stars objects
#' @param dst character; destination file name; this will be a VRT file with references to the source file(s), see details
#' @param options character; options to the gdalbuildvrt command
#' @param file_ext character; file extension, determining the format used to write to (".tif" implies GeoTIFF)
#' @return the stars method returns a stars object with the composite of the input; the \code{character} method returns the file name of the file with the mosaic; see also the GDAL documentation of \code{gdalbuildvrt}
#' @details the gdal function buildvrt builds a mosaic of input images; these input images can be multi-band, but not higher-dimensional data cubes or stars objects with multiple attributes; note that for the `stars` method, the `dst` file may contain references to temporary files that are going to be removed at termination of the R session.
#' 
#' uses \link[sf]{gdal_utils} to internally call \code{buildvrt}; no executables external to R are called.
#' @export
#' @examples
#' x = read_stars(system.file("tif/L7_ETMs.tif", package = "stars"))
#' x1 = x[,100:200,100:200,]
#' x2 = x[,150:300,150:300,]
#' plot(st_mosaic(x1, x2))
st_mosaic = function(.x, ...) UseMethod("st_mosaic")

#' @export
#' @name st_mosaic
st_mosaic.stars = function(.x, ..., dst = tempfile(fileext = file_ext), 
		options = c("-vrtnodata", "-9999", "-srcnodata", "nan"),
# -srcnodata "nan": see https://github.com/r-spatial/stars/issues/274
		file_ext = ".tif") {
	lst_write = function(obj) {
		fname = tempfile(fileext = file_ext)
		write_stars(obj, fname)
		fname
	}
	missing_dst = missing(dst)
	objs = if (missing(.x)) list(...) else append(list(.x), list(...))
	src = sapply(objs, lst_write)
	st_mosaic(src, dst = dst, file_ext = file_ext, options = options)
	ret = setNames(read_stars(dst), names(objs[[1]])[1])
	if (!inherits(ret, "stars_proxy") && missing_dst) {
		unlink(src)
		unlink(dst)
	}
	ret
}

#' @export
#' @name st_mosaic
st_mosaic.character = function(.x, ...,  dst = tempfile(fileext = file_ext), 
		options = c("-vrtnodata", "-9999"), file_ext = ".tif") {
	sf::gdal_utils("buildvrt", .x, dst, options = options, ...)
	dst
}

#' @export
#' @name st_mosaic
st_mosaic.stars_proxy = function(.x, ..., dst = tempfile(fileext = file_ext),
		options = c("-vrtnodata", "-9999"), file_ext = ".tif") {
	if (length(.x) > 1 || length(.x[[1]]) > 1)
		stop("st_mosaic.stars_proxy only implemented for single-file proxy objects")
	objs = if (missing(.x)) list(...) else append(list(.x), list(...))
	files = sapply(objs, function(sp) sp[[1]])
	read_stars(st_mosaic(files, dst = dst, options = options, file_ext = file_ext), proxy = TRUE)
}

get_index_ranges = function(d, n, offset) {
	nd = (d - offset) %/% (n + 1)
	g = do.call(expand.grid, lapply(nd, seq_len))
	g = t(apply(g, 1, function(x) offset + 1 + (x - 1) * (n + 1)))
	dimnames(g) = NULL
	f = function(x) mapply(seq, x, length.out = n + 1, SIMPLIFY = FALSE)
	apply(g, 1, f, simplify = FALSE)
}

# reduce resolution of x, keeping (most of) extent
#' @export
#' @name st_downsample
st_downsample = function(x, n, ...) UseMethod("st_downsample")

#' downsample stars or stars_proxy objects
#'
#' downsample a stars or stars_proxy object either by skipping rows, columns and bands,
#' or by computing a single value (e.g. the mean) from the sub-tiles involved
#' 
#' @param x object of class stars or stars_proxy
#' @param n integer; for each dimension the number of pixels/lines/bands etc that will be skipped; see Details.
#' @param offset integer; offset(s) for downsampling, in pixels, starting at the offset of 
#' each dimension; should be smaller or equal to \code{n}
#' @param FUN function; if given, downsampling will apply FUN to each of the the subtiles
#' @param ... arguments passed on to \code{FUN} (e.g., \code{na.rm = TRUE} to ignore missing values if FUN is \code{mean})
#' @details If all n == 0, no downsampling takes place; if it is 1, every second row/column/band
#' is skipped, if it is 2, every second+third row/column/band are skipped, etc.
#' 
#' Downsampling a \code{stars_proxy} object returns a \code{stars} object, is
#' equivalent to calling \code{st_as_stars(x, downsample = 2)}, and only downsamples
#' the first two (x and y) dimensions.
#'
#' Downsampled regular rasters keep their dimension offsets, have a cell size (delta) that
#' is n[i]+1 times larger, and may result in a (slightly) different extent.
#'
#' Note that terra's \link[terra]{aggregate} with \code{fact=2} corresponds to
#' \code{st_downsample(x, n = 1, FUN = mean)}: \code{fact} is one larger than \code{n}.
#' @name st_downsample
#' @export
#' @examples
#' (m = matrix(1:121, 11, 11))
#' (s = st_as_stars(m))
#' st_downsample(s, 1)
#' st_downsample(s, 1)[[1]]
#' st_downsample(s, 1, offset = 1)
#' st_downsample(s, 1, offset = 1)[[1]]
#' st_downsample(s, 1, offset = c(0,1))
#' st_downsample(s, 1, offset = c(0,1))[[1]]
#' st_downsample(s, 1, FUN = mean)
#' st_downsample(s, 1, FUN = mean)[[1]]
#' st_downsample(s, 1, offset = 1, FUN = mean)
#' st_downsample(s, 1, offset = c(0,1), FUN = mean)[[1]]
st_downsample.stars = function(x, n, ..., offset = 0, FUN) {

	if (all(n == 0))
		return(x)

	x = st_normalize(x)
	d = dim(x)
	n = rep_len(as.integer(n), length(d))
	offset = rep_len(as.integer(offset), length(d))
	stopifnot(all(n >= 0), all(offset <= n), all(offset < d))
	new_dim = if (! missing(FUN))
			(d - offset) %/% (n + 1)
		else
			ceiling((d - offset) / (n + 1)) # include last, incomplete tile

	dims = st_dimensions(x)
	ix = setNames(vector("list", length(dims)), names(dims))
	for (i in seq_along(d)) # need to precompute:
		ix[[i]] = seq(1 + offset[i], d[i], n[i] + 1)
	xy = attr(dims, "raster")$dimensions
	for (i in seq_along(d)) {
		if (is_CFTime(dims[[i]]$refsys)) {
			dims[[i]]$to = unname(new_dim[i])
			time = dims[[i]]$values
			bnds = CFtime::bounds(time)
			time = CFtime::CFtime(CFtime::definition(time), 
								  CFtime::calendar(time), 
								  CFtime::offsets(time)[ ix[[i]] ])
			if (!is.null(bnds))
				CFtime::bounds(time) <- bnds[, ix[[i]] ]
			dims[[i]]$values = time
		} else {
			if (offset[i] != 0)
				dims[[i]]$offset = dims[[i]]$offset + offset[i] * dims[[i]]$delta
			dims[[i]]$delta = dims[[i]]$delta * (n[i] + 1)
			dims[[i]]$to = unname(new_dim[i])
			if (!is.null(dims[[i]]$values)) {
				if (is.matrix(dims[[i]]$values) && names(ix)[i] %in% xy)
					dims[[i]]$values = dims[[i]]$values[ ix[[ xy[1] ]], ix[[ xy[2] ]] ] # speaks for itself 
				else
					dims[[i]]$values = dims[[i]]$values[ ix[[i]] ]
			}
		}
	}
	if (!all(attr(dims, "raster")$affine == 0.0)) {
		if (!all(xy %in% names(d)[1:2]))
			stop("downsampling an affine raster needs to have x/y in dims 1 and 2")
		attr(dims, "raster")$affine = attr(dims, "raster")$affine * (n[1:2] + 1)
	}

	if (!missing(FUN)) { # compute FUN() values over the subtiles:
		stopifnot(is.function(FUN))
		l = get_index_ranges(d, n, offset)
		x = unclass(x) # so that `[[<-` doesn't go into the stars method, which recycles
		for (i in seq_along(x))
			x[[i]] = structure(sapply(l, function(y) FUN(as.vector(asub(x[[i]], y)), ...)),
							   dim = new_dim)
	} else { # downsample using `[`:
		args = rep(list(rlang::missing_arg()), length(d)+1)
		for (i in seq_along(d))
			if (n[i] > 0)
				args[[i+1]] = ix[[i]] # i==1 skipped: retain all attributes
		x = eval(rlang::expr(x[!!!args]))
	}
	st_stars(x, dims)
}

#' @export
#' @name st_downsample
st_downsample.stars_proxy = function(x, n, ...) {
	if (length(n) != 2)
		message("for stars_proxy objects, downsampling only happens for dimensions x and y")
	st_as_stars(x, downsample = n)
}


#' @export
st_sample.stars = function(x, size, ..., type = "random", replace = FALSE) {
	if (length(x) > 1)
		warning("only sampling the first attribute")
	if (type != "random")
		warning("only type 'random' supported")
	v = structure(x[[1]], dim = NULL)
	if (missing(replace))
		replace = size > length(v)
	v[sample(length(v), size, replace = replace, ...)]
}

#' @export
st_sample.stars_proxy = function(x, size, ..., type = "regular", quiet = TRUE) {
	if (type != "regular")
		stop("only type 'regular' for stars_proxy objects supported") # FIXME: tbd
	d = dim(x)
	downsampling_rate = c(floor(d[1:2] / sqrt(size)), d[-(1:2)])
	if (!quiet)
		print(downsampling_rate) # nocov
	st_as_stars(x, downsample = downsampling_rate)
	# st_sample(st, size, ...)
}

Sys.setenv(TZ="UTC")
suppressPackageStartupMessages(library(stars))
set.seed(13521) # runif
tif = system.file("tif/L7_ETMs.tif", package = "stars")
(x_ = read_stars(c(tif,tif))) # FIXME: not what you'd expect
(x = read_stars(tif))
# image(x)
gdal_crs(tif)
plot(x)
plot(x, join_zlim = FALSE)
x |> st_set_dimensions(names = c('a', 'b', 'c'))
st_get_dimension_values(x, 3)

(x1 = st_set_dimensions(x, "band", values = c(1,2,3,4,5,7), names = "band_number", point = TRUE))
rbind(c(0.45,0.515), c(0.525,0.605), c(0.63,0.69), c(0.775,0.90), c(1.55,1.75), c(2.08,2.35)) |>
	units::set_units(um) -> bw # units::set_units(µm) -> bw
# set bandwidth midpoint:
(x2 = st_set_dimensions(x, "band", values = 0.5 * (bw[,1]+bw[,2]), 
   names = "bandwidth_midpoint", point = TRUE))
# set bandwidth intervals:
(x3 = st_set_dimensions(x, "band", values = make_intervals(bw), names = "bandwidth"))

x + x
x * x
x[,,,1:3]
x[,1:100,100:200,]
sqrt(x)
st_apply(x, 3, min)
st_apply(x, 1:2, max)
st_apply(x, 1:2, range)

geomatrix = system.file("tif/geomatrix.tif", package = "stars")
x = read_stars(geomatrix)
y = st_transform(x, st_crs(4326))
st_coordinates(x)[1:2,]

nc = system.file("nc/tos_O1_2001-2002.nc", package = "stars")
if (nc != "") {
  print(x <- read_stars(nc))
  print(st_bbox(x))
  s = st_as_stars(st_bbox(x)) # inside = NA
  print(st_bbox(s))
  s = st_as_stars(st_bbox(x), inside = TRUE)
  print(st_bbox(s))
  s = st_as_stars(st_bbox(x), inside = FALSE)
  print(st_bbox(s))
  (s = st_as_stars(st_bbox(x), dy = 1))
  print(st_bbox(s))
  print(identical(st_as_stars(st_bbox(x), dx = 1), st_as_stars(st_bbox(x), dy = 1)))
  s = st_as_stars(st_bbox(x), dx = 10)
  print(st_bbox(s))
  s = st_as_stars(st_bbox(x), dx = 20)
  print(st_bbox(s))
  x1 = x
  st_crs(x1) = "OGC:CRS84"
  print(identical(st_as_stars(st_bbox(x1), dx = 1), st_as_stars(st_bbox(x1), dx = units::set_units(1, degree))))

  df = as.data.frame(x)
  if (suppressPackageStartupMessages(require(units, quietly = TRUE)))
    print(units::drop_units(x))

  print(dimnames(x))
  dimnames(x) <- letters[1:3]
  print(dimnames(x))
} # CFtime

st_as_stars()

# multiple sub-datasets:
nc_red = system.file("nc/reduced.nc", package = "stars")
red = read_stars(nc_red)
red
plot(red)

x = st_xy2sfc(read_stars(tif)[,1:10,1:10,], as_points = FALSE)
st_bbox(x)
x = read_stars(tif)
merge(split(x, "band"))

read_stars(c(tif,tif)) # merges as attributes
read_stars(c(tif,tif), along = "sensor")
read_stars(c(tif,tif), along = 4)
read_stars(c(tif,tif), along = "band")
read_stars(c(tif,tif), along = 3)

# cut:
tif = system.file("tif/L7_ETMs.tif", package = "stars")
x = read_stars(tif)
cut(x, c(0, 50, 100, 255))
cut(x[,,,1,drop=TRUE], c(0, 50, 100, 255))
plot(cut(x[,,,1,drop=TRUE], c(0, 50, 100, 255)))

st_bbox(st_dimensions(x))
x[x < 0] = NA
x[is.na(x)] = 0

# c:
f = system.file("netcdf/avhrr-only-v2.19810902.nc", package = "starsdata")
if (FALSE && f != "") {
  files = c("avhrr-only-v2.19810901.nc",
  "avhrr-only-v2.19810902.nc",
  "avhrr-only-v2.19810903.nc",
  "avhrr-only-v2.19810904.nc",
  "avhrr-only-v2.19810905.nc",
  "avhrr-only-v2.19810906.nc",
  "avhrr-only-v2.19810907.nc",
  "avhrr-only-v2.19810908.nc",
  "avhrr-only-v2.19810909.nc")
  l = list()
  for (f in files) {
	from = system.file(paste0("netcdf/", f), package = "starsdata")
  	l[[f]] = read_stars(from, sub = c("sst", "anom"))
  }
  ret = do.call(c, l)
  print(ret)
  ret = adrop(c(l[[1]], l[[2]], l[[3]], along = list(times = as.Date("1981-09-01") + 0:2)))
  print(ret)
  #ret = adrop(adrop(c(l[[1]], l[[2]], l[[3]], along = "times")))
  #print(ret)
}

st_dimensions(list(matrix(1, 4, 4))) # st_dimensions.default

if (FALSE && require("starsdata", quietly = TRUE)) {
  # curvilinear:
  s5p = system.file(
      "sentinel5p/S5P_NRTI_L2__NO2____20180717T120113_20180717T120613_03932_01_010002_20180717T125231.nc",
      package = "starsdata")
  print(s5p)
  lat_ds = paste0("HDF5:\"", s5p, "\"://PRODUCT/latitude")
  lon_ds = paste0("HDF5:\"", s5p, "\"://PRODUCT/longitude")
  nit_ds = paste0("HDF5:\"", s5p, "\"://PRODUCT/SUPPORT_DATA/DETAILED_RESULTS/nitrogendioxide_summed_total_column")
  lat = read_stars(lat_ds)
  lon = read_stars(lon_ds)
  nit = read_stars(nit_ds)
  nit[[1]][nit[[1]] > 9e+36] = NA
  
  ll = setNames(c(lon, lat), c("x", "y"))
  nit.c = st_as_stars(nit, curvilinear = ll)
  print(nit.c)

  s5p = system.file(
      "sentinel5p/S5P_NRTI_L2__NO2____20180717T120113_20180717T120613_03932_01_010002_20180717T125231.nc",
      package = "starsdata")
  nit.c2 = read_stars(s5p, 
  	sub = "//PRODUCT/SUPPORT_DATA/DETAILED_RESULTS/nitrogendioxide_summed_total_column",
    curvilinear = c("//PRODUCT/latitude", "//PRODUCT/longitude"))
  print(all.equal(nit.c, nit.c2))
}

suppressPackageStartupMessages(library(stars))

# regular, but not spatial:
d = st_dimensions(a = 1:3, b = 1:3, band = c("foo", "bar"))
(st = st_as_stars(array(1:18, c(3,3,2)), dimension = d))
as.data.frame(st, add_max=FALSE)[1:4,]
try(x <- st_bbox(st)) # error

# regular, geotransform:
d = st_dimensions(x = 1:3, y = 1:3, band = c("foo", "bar"), .raster = c("x", "y"))
(st = st_as_stars(array(1:18, c(3,3,2)), dimension = d))
as.data.frame(st, add_max=FALSE)[1:4,]
as.data.frame(st, add_max=TRUE)[1:4,]
st_bbox(st)

# rectilinear with offset given:
xd = c(1, 2, 4)
d = st_dimensions(x = xd, y = 1:3, band = c("foo", "bar"), .raster = c("x", "y"))
(st = st_as_stars(array(1:18, c(3,3,2)), dimension = d))
as.data.frame(st, add_max=FALSE)[1:4,]
as.data.frame(st, add_max=TRUE)[1:4,]
st_bbox(st)

# rectilinear with midpoints given:
xd = c(1, 2, 4)
d = st_dimensions(x = xd, y = 1:3, band = c("foo", "bar"), .raster = c("x", "y"), cell_midpoints = TRUE)
(st = st_as_stars(array(1:18, c(3,3,2)), dimension = d))
as.data.frame(st, add_max=FALSE)[1:4,]
as.data.frame(st, add_max=TRUE)[1:4,]
st_bbox(st)

# rectilinear with midpoints given, point support
xd = c(1, 2, 4)
d = st_dimensions(x = xd, y = 1:3, band = c("foo", "bar"), .raster = c("x", "y"), cell_midpoints = TRUE, point = TRUE)
(st = st_as_stars(array(1:18, c(3,3,2)), dimension = d))
as.data.frame(st, add_max=FALSE)[1:4,]
as.data.frame(st, add_max=TRUE)[1:4,]
st_bbox(st)

# rectilinear with start/end given:
#xd = stars:::make_intervals(start = c(1,2,4), end = c(2, 4, 8))
xd = c(1, 2, 4, 8) # one more than dim of the data array
d = st_dimensions(x = xd, y = 1:3, band = c("foo", "bar"), .raster = c("x", "y"))
(st = st_as_stars(array(1:18, c(3,3,2)), dimension = d))
as.data.frame(st, add_max=FALSE)[1:4,]
as.data.frame(st, add_max=TRUE)[1:4,]
st_bbox(st)

# with sfc:
sfc = st_sfc(st_point(0:1), st_point(2:1), st_point(4:3))
d = st_dimensions(x = sfc, y = 1:3, band = c("foo", "bar"))
(st = st_as_stars(array(1:18, c(3,3,2)), dimension = d))
as.data.frame(st, add_max=FALSE)[1:4,]
st_as_sf(st, long = TRUE)[1:4,]
st_bbox(st)

# rotated/sheared:
d = st_dimensions(x = 1:3, y = 1:3, band = c("foo", "bar"), .raster = c("x", "y"), affine = c(0.2, -0.2))
(st = st_as_stars(array(1:18, c(3,3,2)), dimension = d))
as.data.frame(st, add_max=FALSE)[1:4,]
try(as.data.frame(st, add_max=TRUE)[1:4,]) # errors
st_bbox(st)

# curvilinear:
set.seed(13531)
lon = st_as_stars(matrix(signif(runif(9), 2), 3, 3))
lat = st_as_stars(matrix(signif(runif(9), 2), 3, 3))
ll = setNames(c(lon, lat), c("lon", "lat"))
d = st_dimensions(lon = 1:3, lat = 1:3)
a = st_as_stars(list(X = array(1:9, c(3,3))), dimensions = d)
(st = st_as_stars(a, curvilinear = ll))
as.data.frame(st)
st_bbox(st)

# time regular:
t = as.Date("2019-01-20") + 0:3
d = st_dimensions(t, sfc)
(st = st_as_stars(list(a = matrix(1:12, 4, 3)), dimensions = d))
as.data.frame(st)

# time, rectilinear
t = as.Date("2019-01-20") + c(0, 2, 3, 10)
d = st_dimensions(t, sfc)
(st = st_as_stars(list(a = matrix(1:12, 4, 3)), dimensions = d))
as.data.frame(st)

stars_to_Spatial = function(from) { 
    if (!requireNamespace("sp", quietly = TRUE))
        stop("package sp required, please install it first") #nocov
	geom = if (has_raster(from)) {
		from = st_normalize(from)
		if (length(dim(from)) > 2)
			stop("stars object must have two (raster: x, y) dimensions")
		if (!is_regular_grid(from))
			stop("only regular rasters can be converted to Spatial")
		d = st_dimensions(from)
		xy = attr(d, "raster")$dimensions
		if (!all(match(xy, names(d)) == 1:2))
			stop("x/y dimensions should be at pos 1 and 2")
		offset = c( d[[ 1 ]]$offset, d[[ 2 ]]$offset )
		delta = c( d[[ 1 ]]$delta, d[[ 2 ]]$delta )
		cells.dim = dim(from)
		if (delta[1] < 0)
			stop("negative x cell size not supported")
		if (delta[2] > 0)
			stop("only negative y delta supported")
		cellcentre.offset = setNames(c(offset[1] + 0.5 * delta[1],
			offset[2] + (cells.dim[2] - 0.5) * delta[2]), xy)
		gt = sp::GridTopology(cellcentre.offset, abs(delta), cells.dim)
		sp::SpatialGrid(gt, as(st_crs(d), "CRS"))
	} else {
		if (!has_sfc(from))
			stop("no feature dimension in stars object")
		as(st_dimensions(from)[[1]]$values, "Spatial")
	}
	if (length(from) == 0)
		geom
	else
		sp::addAttrToGeom(geom, 
			as.data.frame(lapply(from, function(y) structure(y, dim = NULL))), match.ID = FALSE)
}
setAs("stars", "Spatial", stars_to_Spatial)

#' @export
st_as_stars.Spatial = function(.x, ...) {
    if (!requireNamespace("sp", quietly = TRUE))
        stop("package sp required, please install it first") # nocov
	if (sp::gridded(.x)) {
		sp::fullgrid(.x) = TRUE
		gt = sp::gridparameters(.x)
		# st_as_stars(raster::stack(.x)) --- we can do better
		# UL corner:
		x = create_dimension(1, gt$cells.dim[1], 
			offset = gt$cellcentre.offset[1] - 0.5 * gt$cellsize[1], delta = gt$cellsize[1],
			refsys = st_crs(.x))
		y = create_dimension(1, gt$cells.dim[2], 
			offset = gt$cellcentre.offset[2] + (gt$cells.dim[2] - 0.5) * gt$cellsize[2],
			delta = -gt$cellsize[2], refsys = st_crs(.x))
		d = create_dimensions(list(x = x, y = y), raster = get_raster(dimensions = c("x", "y")))
		lst = lapply(.x@data, function(x, dims) structure(x, dim = dims), dims = dim(d))
		st_stars(lst, d)
	} else
		st_as_stars(st_as_sf(.x))
}

as.xts.stars <- function(x,...) { 

	if (!requireNamespace("xts", quietly = TRUE))
		stop("xts required: install that first") # nocov
	if (length(x) > 1)
		message("only the first attribute will used; maybe use merge() first?")

	ed = expand_dimensions.stars(x)
	time = which(sapply(ed, inherits, c("Date", "POSIXt")))
	if (length(time) < 1)
		stop("no time dimension present in object")
	if (length(time) > 1) {
		message("more than one time dimension present in object; taking the first")
		time = time[1]
	}
	x = adrop(st_upfront(x[1], time), drop_xy = TRUE)
	if (length(dim(x)) > 2) {
		dims = setNames(c(dim(x)[1], prod(dim(x)[-1])), c(names(ed)[time], "other"))
		x = st_redimension(x, dims)
	} else { # set colnames
		d = st_dimensions(x)
		if (length(dim(x)) == 2 && length(d[[2]]$values) == dim(x)[2] && is.character(d[[2]]$values))
			colnames(x[[1]]) = d[[2]]$values
	}
	xts::xts(x[[1]], ed[[time]])
}

#' @name st_as_stars
#' @param name character; attribute name for array from an \code{xts} object
#' @details for the \code{xts} methods, if \code{dimensions} are provided, time has to be the first dimension.
#' @export
st_as_stars.xts = function(.x, ..., dimensions, name = "attr") {
	if (!requireNamespace("xts", quietly = TRUE))
		stop("xts required: install that first") # nocov
	if (!requireNamespace("zoo", quietly = TRUE))
		stop("zoo required: install that first") # nocov
	time = zoo::index(.x)
	cn = colnames(.x)
	if (!missing(dimensions)) {
		.x = as.matrix(.x)
		dim(.x) = dim(dimensions)
		st_stars(setNames(list(.x), name), dimensions)
	} else {
		.x = st_as_stars(setNames(list(as.matrix(.x)), name))
		.x = st_set_dimensions(.x, 1, time)
		if (!is.null(cn))
			.x = st_set_dimensions(.x, 2, cn)
		st_set_dimensions(.x, names = c("time", "others"))
	}
}

suppressPackageStartupMessages(library(stars))

s5p = system.file("sentinel5p/S5P_NRTI_L2__NO2____20180717T120113_20180717T120613_03932_01_010002_20180717T125231.nc", package = "starsdata")
if (s5p != "") {

lat_ds = paste0("HDF5:\"", s5p, "\"://PRODUCT/latitude")
lon_ds = paste0("HDF5:\"", s5p, "\"://PRODUCT/longitude")
nit_ds = paste0("HDF5:\"", s5p, "\"://PRODUCT/SUPPORT_DATA/DETAILED_RESULTS/nitrogendioxide_summed_total_column")
lat = read_stars(lat_ds)
lon = read_stars(lon_ds)
nit = read_stars(nit_ds)
nit[[1]][nit[[1]] > 9e+36] = NA

ll = setNames(c(lon, lat), c("x", "y"))
nit.c0 = st_as_stars(nit, curvilinear = ll)

# more direct method:
nit.c = read_stars(s5p, sub = "//PRODUCT/SUPPORT_DATA/DETAILED_RESULTS/nitrogendioxide_summed_total_column",
	curvilinear = c("//PRODUCT/longitude", "//PRODUCT/latitude"), driver = NULL)
if (inherits(nit.c[[1]], "units"))
	nit.c[[1]] = units::drop_units(nit.c[[1]])
nit.c[[1]][nit.c[[1]] > 9e+36] = NA
all.equal(nit.c0, nit.c)
st_crs(nit.c) = 4326
print(nit.c)

if (FALSE && capabilities()["png"]) {
 png("nit1.png", 800, 800)
 plot(nit.c, breaks = "equal", reset = FALSE, axes = TRUE, as_points = TRUE, pch = 16)
 maps::map('world', add = TRUE, col = 'red')
 dev.off()

 png("nit2.png", 800, 800)
 plot(nit.c, breaks = "equal", reset = FALSE, axes = TRUE, as_points = FALSE, border = NA)
 maps::map('world', add = TRUE, col = 'red')
 dev.off()

 nit.c = stars:::st_downsample(nit.c, 8)
 print(nit.c)
 
 png("nit3.png", 800, 800)
 plot(nit.c, breaks = "equal", reset = FALSE, axes = TRUE, as_points = TRUE, pch = 16)
 maps::map('world', add = TRUE, col = 'red')
 dev.off()

 png("nit4.png", 800, 800)
 plot(nit.c, breaks = "equal", reset = FALSE, axes = TRUE, as_points = FALSE, border = NA)
 maps::map('world', add = TRUE, col = 'red')
 dev.off()
}

}

first_dimensions_match = function(e1, e2) {
	d1 = st_dimensions(e1)
	d2 = st_dimensions(e2)
	crs1 = st_crs(d1)
	crs2 = st_crs(d2)
	st_crs(d1) = st_crs(NA)
	st_crs(d2) = st_crs(NA)
	n = min(length(d1), length(d2))
	isTRUE(all.equal(d1[1:n], d2[1:n], check.attributes = FALSE)) && crs1 == crs2
}

#' S3 Ops Group Generic Functions for stars objects
#'
#' Ops functions for stars objects, including comparison, product and divide, add, subtract
#'
#' @param e1 object of class \code{stars}
#' @param e2 object of class \code{stars}
#'
#' @return object of class \code{stars}
#' @name ops_stars
#' @examples
#' tif = system.file("tif/L7_ETMs.tif", package = "stars")
#' x = read_stars(tif)
#' x * x
#' x / x
#' x + x
#' x + 10
#' all.equal(x * 10, 10 * x)
#' @export
#' @details if \code{e1} or \code{e2} is is a numeric vector, or \code{e2}
#' has less or smaller dimensions than \code{e1}, then \code{e2} is recycled
#' such that it fits \code{e1}, using usual R array recycling rules. The user
#' needs to make sure this is sensible; it may be needed to use \code{aperm}
#' to permutate dimensions first. 
Ops.stars <- function(e1, e2) {
	if (!missing(e2)) { 
		if (inherits(e1, "stars") && inherits(e2, "stars")) {
			if (!first_dimensions_match(e1, e2))
				stop("(first) dimensions of e1 and e2 do not match")
			dim_final = if (prod(dim(e1)) < prod(dim(e2)))
							st_dimensions(e2)
						else
							st_dimensions(e1)
		} else if (inherits(e1, "stars")) {
			dim_final = st_dimensions(e1)
		} else if (inherits(e2, "stars")) {
			dim_final = st_dimensions(e2)
		}
		if (!inherits(e2, c("stars", "units")))
			e1 = drop_units(e1)
	} else
		dim_final = st_dimensions(e1)
	ret = if (missing(e2))
			lapply(e1, .Generic)
		else if (!inherits(e2, "stars"))
			lapply(e1, .Generic, e2 = e2)
		else { # both e1 and e2 are stars objects:
			# https://github.com/r-spatial/stars/issues/187#issuecomment-834020710 :
			if (!is.null(dim(e1)) &&
					!isTRUE(all.equal(dim(e1), dim(e2), check.attributes = FALSE))) {
				stopifnot(length(e2) == 1)
				lapply(lapply(e1, structure, dim=NULL), .Generic, e2 = structure(e2[[1]], dim = NULL))
			} else
				mapply(.Generic, e1, e2, SIMPLIFY = FALSE)
		}
	if (any(sapply(ret, function(x) is.null(dim(x))))) # happens if e1[[1]] is a factor; #304
		ret = lapply(ret, function(x) { dim(x) = dim(dim_final); x })
	if (! inherits(e1, "stars"))
		st_as_stars(setNames(ret, names(e2)), dimensions = dim_final)
	else
		st_as_stars(ret, dimensions = dim_final)
}

#' Mathematical operations for stars objects
#'
#' @param x object of class stars
#' @param ... parameters passed on to the Math functions
#' 
#' @export
#' @name ops_stars
#' 
#' @examples
#' tif = system.file("tif/L7_ETMs.tif", package = "stars")
#' x = read_stars(tif)
#' a = sqrt(x)
#' b = log(x, base = 10)
#' @export
Math.stars = function(x, ...) {
	ret = lapply(x, .Generic, ...)
	st_as_stars(ret, dimensions = st_dimensions(x))
}

#' @name ops_stars
#' @export
Ops.stars_proxy <- function(e1, e2) {
	if (!inherits(e1, "stars_proxy"))
		stop("first argument in expression needs to be the stars_proxy object") # FIXME: needed?? #nocov
	if (missing(e2))
		collect(e1, match.call(), .Generic, "e1", env = environment())
	else
		collect(e1, match.call(), .Generic, c("e1", "e2"), env = environment())
}

#' @name ops_stars
#' @export
Math.stars_proxy = function(x, ...) {
	collect(x, match.call(), .Generic, env = environment())
}

# https://github.com/r-spatial/stars/issues/390
has_single_arg = function(fun, dots) {
	sum(!(names(as.list(args(fun))) %in% c("", "...", names(dots)))) <= 1
}
can_single_arg = function(fun) {
	!inherits(try(fun(1:10), silent = TRUE), "try-error")
}


#' @export
st_apply = function(X, MARGIN, FUN, ...) UseMethod("st_apply")

#' st_apply apply a function to one or more array dimensions
#' 
#' st_apply apply a function to array dimensions: aggregate over space, time, or something else
#' @name st_apply
#' @param X object of class \code{stars}
#' @param MARGIN see \link[base]{apply}; index number(s) or name(s) of the dimensions over which \code{FUN} will be applied 
#' @param FUN see \link[base]{apply} and see Details.
#' @param ... arguments passed on to \code{FUN}
#' @param CLUSTER cluster to use for parallel apply; see \link[parallel]{makeCluster}
#' @param PROGRESS logical; if \code{TRUE}, use \code{pbapply::pbapply} to show progress bar
#' @param FUTURE logical;if \code{TRUE}, use \code{future.apply::future_apply} 
#' @param rename logical; if \code{TRUE} and \code{X} has only one attribute and 
#' \code{FUN} is a simple function name, rename the attribute of the returned object 
#' to the function name
#' @param .fname function name for the new attribute name (if one or more 
#' dimensions are reduced) or the new dimension (if a new dimension is created); 
#' if missing, the name of \code{FUN} is used
#' @param single_arg logical; if \code{TRUE}, FUN takes a single argument (like \code{fn_ndvi1} below), 
#' if \code{FALSE} FUN takes multiple arguments (like \code{fn_ndvi2} below).
#' @return object of class \code{stars} with accordingly reduced number of dimensions; 
#' in case \code{FUN} returns more than one value, a new dimension is created carrying 
#' the name of the function used; see the examples. Following the logic of 
#' \link[base]{apply}, This new dimension is put before the
#' other dimensions; use \link{aperm} to rearrange this, see last example.
#' @param keep logical; if \code{TRUE}, preserve dimension metadata (e.g. time stamps)
#' @details FUN is a function which either operates on a single object, which will 
#' be the data of each iteration step over dimensions MARGIN, or a function that 
#' has as many arguments as there are elements in such an object. See the NDVI 
#' examples below. The second form can be VERY much faster e.g. when a trivial 
#' function is not being called for every pixel, but only once (example).
#' 
#' The heuristics for the default of \code{single_arg} work often, but not always; try
#' setting this to the right value when \code{st_apply} gives an error.
#' @examples
#' tif = system.file("tif/L7_ETMs.tif", package = "stars")
#' x = read_stars(tif)
#' st_apply(x, 1:2, mean) # mean band value for each pixel
#' st_apply(x, c("x", "y"), mean) # equivalent to the above
#' st_apply(x, 3, mean)   # mean of all pixels for each band
#' \dontrun{
#'  st_apply(x, "band", mean) # equivalent to the above
#'  st_apply(x, 1:2, range) # min and max band value for each pixel
#'  fn_ndvi1 = function(x) (x[4]-x[3])/(x[4]+x[3]) # ONE argument: will be called for each pixel
#'  fn_ndvi2 = function(red,nir) (nir-red)/(nir+red) # n arguments: will be called only once
#'  ndvi1 = st_apply(x, 1:2, fn_ndvi1)
#'    # note that we can select bands 3 and 4 in the first argument:
#'  ndvi2 = st_apply(x[,,,3:4], 1:2, fn_ndvi2) 
#'  all.equal(ndvi1, ndvi2)
#'  # compute the (spatial) variance of each band; https://github.com/r-spatial/stars/issues/430
#'  st_apply(x, 3, function(x) var(as.vector(x))) # as.vector is required!
#'  # to get a progress bar also in non-interactive mode, specify:
#'  if (require(pbapply)) { # install it, if FALSE
#'    pboptions(type = "timer")
#'  }
#'  st_apply(x, 1:2, range) # dimension "range" is first; rearrange by:
#'  st_apply(x, 1:2, range) |> aperm(c(2,3,1))
#' }
#' @export
st_apply.stars = function(X, MARGIN, FUN, ..., CLUSTER = NULL, PROGRESS = FALSE, FUTURE = FALSE, 
		rename = TRUE, .fname, single_arg = has_single_arg(FUN, list(...)) || can_single_arg(FUN),
		keep = FALSE) {
	if (missing(.fname))
		.fname <- paste(deparse(substitute(FUN), 50), collapse = "\n")
	if (is.character(MARGIN))
		MARGIN = match(MARGIN, names(dim(X)))
	dX = dim(X)[MARGIN]

	if (PROGRESS && !requireNamespace("pbapply", quietly = TRUE))
		stop("package pbapply required, please install it first")
	
	if (FUTURE && !requireNamespace("future.apply", quietly = TRUE))
	  stop("package future.apply required, please install it first")

	fn = function(y, ...) {
		ret = if (PROGRESS)
				pbapply::pbapply(X = y, MARGIN = MARGIN, FUN = FUN, ..., cl = CLUSTER)
			else {
				if (is.null(CLUSTER) && !FUTURE)
					apply(X = y, MARGIN = MARGIN, FUN = FUN, ...)
				else if (FUTURE) {
					oopts = options(future.globals.maxSize = +Inf)
					on.exit(options(oopts))
					future.apply::future_apply(y, MARGIN = MARGIN, FUN = FUN, ...)
				} else
					parallel::parApply(CLUSTER, X = y, MARGIN = MARGIN, FUN = FUN, ...)
			}
		if (is.array(ret))
			ret
		else
			array(ret, dX)
	}
	no_margin = setdiff(seq_along(dim(X)), MARGIN)
	ret = if (single_arg) 
			lapply(X, fn, ...) 
		else # call FUN on full chunks:
			lapply(X, function(a) do.call(FUN, setNames(append(asplit(a, no_margin), list(...)), NULL)))
	# fix dimensions:
	dim_ret = dim(ret[[1]])
	ret = if (length(dim_ret) == length(MARGIN)) { # FUN returned a single value
			if (length(ret) == 1 && rename && make.names(.fname) == .fname)
				ret = setNames(ret, .fname)
			st_stars(ret, st_dimensions(X)[MARGIN])
		} else { # FUN returned multiple values: need to set dimension name & values
			dim_no_margin = dim(X)[-MARGIN]
			if (length(no_margin) > 1 && dim(ret[[1]])[1] == prod(dim_no_margin)) {
				r = attr(st_dimensions(X), "raster")
				new_dim = c(dim_no_margin, dim(ret[[1]])[-1])
				for (i in seq_along(ret))
					dim(ret[[i]]) = new_dim
				# set dims:
				dims = st_dimensions(X)[c(no_margin, MARGIN)]
			} else {
				orig = st_dimensions(X)[MARGIN]
				r = attr(orig, "raster")
				dims = if (keep) {
						c(st_dimensions(X)[no_margin], orig)
					} else {
						dim1 = if (!is.null(dimnames(ret[[1]])[[1]])) # FUN returned named vector:
								create_dimension(values = dimnames(ret[[1]])[[1]])
							else
								create_dimension(to = dim_ret[1])
						c(structure(list(dim1), names = .fname), orig)
					}
			}
			st_stars(ret, dimensions = create_dimensions(dims, r))
		}
	for (i in seq_along(ret))
		names(dim(ret[[i]])) = names(st_dimensions(ret))
	ret
}

if (!isGeneric("%in%"))
	setGeneric("%in%", function(x, table) standardGeneric("%in%"))

#' evaluate whether cube values are in a given set
#'
#' evaluate whether cube values are in a given set
#' @docType methods
#' @rdname in-methods
#' @param x data cube value
#' @param table values of the set
#' @exportMethod "%in%"
setMethod("%in%", signature(x = "stars"),
	function(x, table) {
		st_stars(lapply(x, function(y) structure(y %in% table, dim = dim(y))),
			st_dimensions(x))
	}
)

#' Specify parameters to load raster in blocks
#'
#' Helper function for specifying the block parameters (\code{nXOff},
#' \code{nYOff}, \code{nXsize}, and \code{nYSize}) required by \code{RasterIO}
#' argument in \link{read_stars}
#'
#' @param img_rows number of input raster rows (integer)
#' @param img_cols number of input raster columns (integer)
#' @param x_window number of rows in block (integer)
#' @param y_window number of columns in block (integer)
#' @param overlap number of overlapping pixels (integer)
#'
#' @return matrix with specified \code{nXOff}, \code{nYOff}, \code{nXsize},
#' and \code{nYSize} parameters for every block
#'
#' @examples
#' \dontrun{
#' tif = system.file("tif/L7_ETMs.tif", package = "stars")
#' r = read_stars(tif, proxy = TRUE)
#' tiles = st_tile(nrow(r), ncol(r), 256, 256)
#' for (i in seq_len(nrow(tiles))) {
#'   tile = read_stars(tif, proxy = FALSE, RasterIO = tiles[i, ])
#'   # write tiles to separate files
#'   write_stars(tile, dsn = paste0(i, ".tif"))
#' }
#' }
#'
#' @export
st_tile = function(img_rows, img_cols, x_window, y_window, overlap = 0) {

	# make sure input values are integers
	img_rows = as.integer(img_rows)
	img_cols = as.integer(img_cols)
	x_window = as.integer(x_window)
	y_window = as.integer(y_window)
	overlap = as.integer(overlap)

	n = ceiling((img_rows / x_window)) * ceiling((img_cols / y_window))

	x_vec = y_vec = integer(n)
	nXSize_vec = nYSize_vec = integer(n)

	i = 1L
	for (x in seq.int(1L, img_rows, y_window)) {

		if (x + y_window + overlap <= img_rows) {
			nXSize = y_window + overlap
		} else {
			nXSize = img_rows - x + 1L
		}

		for (y in seq.int(1L, img_cols, x_window)) {

			if (y + x_window + overlap <= img_cols) {
				nYSize = x_window + overlap
			} else {
				nYSize = img_cols - y + 1L
			}

			x_vec[i] = x
			y_vec[i] = y
			nXSize_vec[i] = nXSize
			nYSize_vec[i] = nYSize
			i = i + 1L
		}

	}

	mat = matrix(c(x_vec, y_vec, nXSize_vec, nYSize_vec), ncol = 4)
	colnames(mat) = c("nXOff", "nYOff", "nXSize", "nYSize")

	return(mat)

}

mat = matrix(0L, 10, 10)

test1 = st_tile(nrow(mat), ncol(mat), 10, 10)
test_that("one block", {
	expect_equal(nrow(test1), 1)
	expect_setequal(test1, c(1, 1, 10, 10))
	# output matrix should be integer to best performance
	expect_type(test1, "integer")
})

test2 = st_tile(nrow(mat), ncol(mat), 5, 5)
test_that("equal blocks", {
	expect_equal(nrow(test2), 4)
})

test3 = st_tile(nrow(mat), ncol(mat), 9, 5)
test_that("different size blocks", {
	expect_equal(nrow(test3), 4)
})

test4 = st_tile(nrow(mat), ncol(mat), 1, 1)
test_that("many small blocks ", {
	expect_equal(nrow(test4), 100)
	expect_setequal(test4[, "nXOff"], rep(1:10, each = 10))
	expect_setequal(test4[, "nYOff"], rep(1:10, times = 10))
	expect_setequal(test4[, "nXSize"], rep(1, times = 100))
	expect_setequal(test4[, "nYSize"], rep(1, times = 100))
})

test5 = st_tile(nrow(mat), ncol(mat), 5, 5, overlap = 5)
test_that("blocks with overlap", {
	expect_equal(nrow(test5), 4)
	expect_setequal(test5[, "nXSize"], rep(c(10, 5), each = 2))
	expect_setequal(test5[, "nYSize"], rep(c(10, 5), times = 2))
})

split_strings = function(md, split = "=") {
	splt = strsplit(md, split)
	lst = lapply(splt, function(x) if (length(x) <= 1) NA_character_ else x[[2]])
	structure(lst, names = sapply(splt, function(x) x[[1]]), class = "gdal_metadata")
}

.array = function(x, dim) {
	structure(rep_len(x, prod(dim)), dim = dim)
}


#' convert objects into a stars object
#' 
#' convert objects into a stars object
#' @export
#' @param .x object to convert
#' @param ... in case \code{.x} is of class \code{bbox}, arguments passed on to 
#' \link{pretty}. In case \code{.x} is of class \code{nc_proxy}, arguments
#' passed on to \code{\link{read_ncdf}}.
st_as_stars = function(.x, ...) UseMethod("st_as_stars")

#' @name st_as_stars
#' @param dimensions object of class dimensions
#' @export
st_as_stars.list = function(.x, ..., dimensions = NULL) {
	if (length(.x)) {
		if (is.null(names(.x)))
			stop("list should have named elements")
		for (i in seq_along(.x)[-1])
			if (!all(dim(.x[[1]]) == dim(.x[[i]])))
				stop("dim attributes not identical")
		if (!is.null(n <- names(.x)) && (any(n == "") || length(n) != length(unique(n))))
			names(.x) = make.names(names(.x), unique = TRUE)

		# check dimensions, if set:
		if (!is.null(dimensions)) {
			dx = dim(.x[[1]])
			dd = dim(dimensions)
			stopifnot(!is.null(dx), !is.null(dd))
			for (i in seq_along(dimensions)) {
				if (dx[i] < dd[i]) { # create_dimension was called with $values one longer than corresponding array dim,
					v = dimensions[[i]]$values
					if (is.null(v)) # regularly spaced, meaning offset/delta have replaced $values:
						dimensions[[i]]$to = dimensions[[i]]$to - 1
					else if (length(v) == dx[i] + 1) { # convert the one-too-long values into an intervals object:
						dimensions[[i]]$values = head(v, -1)
						dimensions[[i]]$to = dimensions[[i]]$to - 1
					} else
						stop(paste("incorrect length of dimensions values for dimension", i))
					dimensions[[i]]$point = FALSE
				}
			}
		} else
			dimensions = create_dimensions(dim(.x[[1]]))
		if (is.null(names(dim(.x[[1]]))))
			for (i in seq_along(.x))
				names(dim(.x[[i]])) = names(dimensions)
	}
	st_stars(.x, dimensions)
}

st_stars = function(x, dimensions, class = "stars") {
	# sanity checks:
	stopifnot(is.list(x))
	stopifnot(inherits(dimensions, "dimensions"))
	stopifnot(!is.null(attr(dimensions, "raster")))
# 	for (i in seq_along(x))
# 		names(dim(x[[i]])) = names(dimensions)
	structure(x, dimensions = dimensions, class = class)
}


#' @name st_as_stars
#' @export
#' @param raster character; the names of the dimensions that denote raster dimensions
st_as_stars.default = function(.x = NULL, ..., raster = NULL) {
	args = if (is.null(.x))
			list(...)
		else
			append(list(.x), list(...))

	if (length(args) == 0)
		return(st_as_stars(st_bbox()))

	isdim = sapply(args, inherits, what = "dimensions")
	dimensions = if (! any(isdim)) {
			if (is.array(args[[1]]) && !is.null(dimnames(args[[1]])))
				st_dimensions(args[[1]])
			else
				do.call(st_dimensions, lapply(dim(args[[1]]), function(x) seq_len(x) - 1))
		} else {
			d = args[[ which(isdim)[1] ]]
			if (is.null(raster))
				raster = attr(d, "raster")
			d
		}

	if (is.null(raster) && !has_sfc(dimensions)) {
		w = which(sapply(dimensions, function(x) is.null(x$values)))
		raster = get_raster(dimensions = names(dimensions)[w[1:2]])
	}
	dimensions = create_dimensions(dimensions, raster)
	if (any(isdim))
		args = args[-which(isdim)]
	if (is.null(names(args)))
		names(args) = paste0("A", seq_along(args))
	for (i in seq_along(args))
		names(dim(args[[i]])) = names(dimensions)
	st_as_stars.list(args, dimensions = dimensions)
}

#' @param curvilinear only for creating curvilinear grids: named length 2 list holding longitude and latitude matrices or stars arrays, or the names of the corresponding attributes in \code{.x}; the names of this vector should correspond to raster dimensions the matrices are associated with; see Details.
#' @param crs object of class \code{crs} with the coordinate reference system of the values in \code{curvilinear}; see details
#' @details if \code{curvilinear} is a list with \code{stars} objects with longitude and latitude values, its coordinate reference system is typically not that of the latitude and longitude values. If \code{curvilinear} contains the names of two arrays in \code{.x}, then these are removed from the returned object.
#' @export
#' @name st_as_stars
st_as_stars.stars = function(.x, ..., curvilinear = NULL, crs = st_crs('OGC:CRS84')) {
	if (is.null(curvilinear))
		.x
	else {
		stopifnot(names(curvilinear) %in% names(dim(.x)))
		if (all(sapply(curvilinear, is.character)))
			curvilinear = unlist(curvilinear)
		if (is.character(cl <- curvilinear)) {
			if (!cl[1] %in% names(.x) || !cl[2] %in% names(.x))
				stop("curvilinear arrays not present in object")
			curvilinear = setNames(vector("list", 2), names(curvilinear))
			curvilinear[[1]] = .x[ cl[1] ]
			curvilinear[[2]] = .x[ cl[2] ]
			.x[[ cl[1] ]] = NULL
			.x[[ cl[2] ]] = NULL
		}
		stopifnot(is.list(curvilinear))
		if (inherits(curvilinear[[1]], "stars"))
			curvilinear[[1]] = curvilinear[[1]][[1]]
		if (inherits(curvilinear[[2]], "stars"))
			curvilinear[[2]] = curvilinear[[2]][[1]]
		dimensions = st_dimensions(.x)
		xy = names(curvilinear)
		dimensions[[ xy[1] ]]$values = structure(curvilinear[[1]], dim = setNames(dim(curvilinear[[1]])[1:2], xy))
		dimensions[[ xy[2] ]]$values = structure(curvilinear[[2]], dim = setNames(dim(curvilinear[[1]])[1:2], xy))
		# erase regular grid coefficients $offset and $delta:
		dimensions[[ xy[1] ]]$offset = dimensions[[ xy[1] ]]$delta = NA_real_
		dimensions[[ xy[2] ]]$offset = dimensions[[ xy[2] ]]$delta = NA_real_
		raster = get_raster(dimensions = names(curvilinear), curvilinear = TRUE)
		st_set_crs(st_stars(.x, create_dimensions(dimensions, raster)), crs)
	}
}

pretty_cut = function(lim, n, inside = FALSE, ...) {
	stopifnot(n > 2)
	p = pretty(x = lim, n = n, ...)
	d = diff(p[1:2])
	if (! inside)
		lim = c(lim[1] - 0.5 * d, lim[2] + 0.5 * d) # extended limits half a cell
	p[p >= lim[1] & p <= lim[2]]
}

#' @param nx integer; number of cells in x direction; see details
#' @param ny integer; number of cells in y direction; see details
#' @param nz integer; number of cells in z direction; if missing no z-dimension is created.
#' @param dx numeric or object of class units; cell size in x direction; see details
#' @param dy numeric or object of class units; cell size in y direction; see details
#' @param xlim length 2 numeric vector with extent (min, max) in x direction
#' @param ylim length 2 numeric vector with extent (min, max) in y direction
#' @param values value(s) to populate the raster values with
#' @param n the (approximate) target number of grid cells
#' @param pretty logical; should cell coordinates have \link{pretty} values?
#' @param inside logical; should all cells entirely fall inside the bbox, potentially not covering it completely (\code{TRUE}), or always cover the bbox (\code{FALSE}), or find a good approximation (\code{NA}, default)?
#' @param proxy logical; should a \code{stars_proxy} object be created? (requires gdal_create binary when sf < 1.0-6)
#' @details For the \code{bbox} method: if \code{pretty} is \code{TRUE}, raster cells may extend the coordinate range of \code{.x} on all sides. If in addition to \code{nx} and \code{ny}, \code{dx} and \code{dy} are also missing, these are set to a single value computed as \code{sqrt(diff(xlim)*diff(ylim)/n)}. 
#' 
#' If \code{nx} and \code{ny} are missing and \code{values} is a matrix, the number of columns and rows of the matrix are taken.
#'
#' Otherwise, if \code{nx} and \code{ny} are missing, they are computed as the (ceiling, floor, or rounded to integer value) of the ratio of the (x or y) range divided by (dx or dy), depending on the value of \code{inside}. Positive \code{dy} will be made negative. Further named arguments (\code{...}) are passed on to \code{pretty}. If \code{dx} or \code{dy} are \code{units} objects, their value is converted to the units of \code{st_crs(.x)} (only when sf >= 1.0-7). 
#' 
#' @export
#' @name st_as_stars
st_as_stars.bbox = function(.x, ..., nx, ny, dx = dy, dy = dx,
		xlim = .x[c("xmin", "xmax")], ylim = .x[c("ymin", "ymax")], 
		values = 0., n = 64800, pretty = FALSE, inside = FALSE, nz, 
		proxy = FALSE) {

	if (xor(missing(nx), missing(ny)))
		stop("either specify both nx and ny, or none of them")

	adx = abs(diff(xlim))
	ady = abs(diff(ylim))

	if (is.matrix(values) && missing(nx) && missing(ny)) {
		nx = ncol(values)
		ny = nrow(values)
	}

	if (missing(dx) && missing(dy)) {
		if (missing(nx))
			dx = dy = sqrt(adx * ady / n)
		else {
			dx = diff(xlim)/nx
			dy = -diff(ylim)/ny
		}
	} else { 
		u <- st_crs(.x)$ud_unit
		if (inherits(u, "units")) {
			if (inherits(dx, "units"))
				units(dx) = u # might convert value
			if (inherits(dy, "units"))
				units(dy) = u # might convert value
		}
		dx = as.numeric(dx) # drop units if present
		dy = as.numeric(dy) # drop units if present
	}

	consider_inside = function(x, inside) {
		if (is.na(inside))
			round(x)
		else if (inside)
			floor(x)
		else
			ceiling(x)
	}

	if (missing(nx))
		nx = consider_inside(diff(xlim) / dx, inside)

	if (missing(ny)) {
		if (dy > 0)
			dy = -dy
		ny = consider_inside(-diff(ylim) / dy, inside)
	}

	if (pretty) {
		vx = pretty_cut(xlim, nx, inside, ...)
		nx = length(vx)
		x = create_dimension(values = vx, refsys = st_crs(.x))
		vy = pretty_cut(ylim, ny, inside, ...)
		ny = length(vy)
		y = create_dimension(values = vy, refsys = st_crs(.x))
	} else {
		x = create_dimension(from = 1, to = nx, offset = unname(xlim[1]), 
			delta = unname(dx), refsys = st_crs(.x))
		y = create_dimension(from = 1, to = ny, offset = unname(ylim[2]),
			delta = unname(dy), refsys = st_crs(.x))
	}
	if (missing(nz)) { # 2D:
		if (proxy) {
			f = tempfile(fileext = ".tif")
			sf::gdal_create(f, c(nx, ny), values, st_crs(.x), xlim, ylim)
			read_stars(f, proxy = TRUE)
		} else
			st_as_stars(values = .array(values, c(x = nx[[1L]], y = ny[[1L]])),
						dims = create_dimensions(list(x = x, y = y), get_raster()))
	} else {
		stopifnot(proxy == FALSE)
		z = create_dimension(from = 1, to = nz[[1]])
		st_as_stars(values = .array(values, c(x = nx[[1L]], y = ny[[1L]], z = nz[[1]])), 
					dims = create_dimensions(list(x = x, y = y, z = z), get_raster()))
	}
}

## @param x two-column matrix with columns and rows, as understood by GDAL; 0.5 refers to the first cell's centre; 
xy_from_colrow = function(x, geotransform) {
# http://www.gdal.org/classGDALDataset.html , search for geotransform:
# 0-based indices:
# Xp = geotransform[0] + P*geotransform[1] + L*geotransform[2];
# Yp = geotransform[3] + P*geotransform[4] + L*geotransform[5];
	stopifnot(ncol(x) == 2, length(geotransform) == 6, !any(is.na(geotransform)))
	matrix(geotransform[c(1, 4)], nrow(x), 2, byrow = TRUE) + 
		x %*% matrix(geotransform[c(2, 3, 5, 6)], nrow = 2, ncol = 2)
}

colrow_from_xy = function(x, obj, NA_outside = FALSE, flip = FALSE) {
	if (inherits(obj, "stars"))
		obj = st_dimensions(obj)
	xy = attr(obj, "raster")$dimensions
	if (inherits(obj, "dimensions"))
		gt = st_geotransform(obj)

	if (flip && isTRUE(st_is_longlat(st_crs(obj)))) {
		bb = st_bbox(obj)
# see https://github.com/r-spatial/stars/issues/519 where this is problematic;
# not sure whether this introduces new problems.
#		sign = ifelse(x[,1] < bb["xmin"], 1., ifelse(x[,1] > bb["xmax"], -1., 0.))
#		x[,1] = x[,1] + sign * 360.
		## one more try: https://github.com/r-spatial/stars/issues/563
		ix = x[,1] > bb["xmax"] & !is.na(x[,1])
		x[ix,1] = x[ix,1] - 360.
		ix = x[,1] < bb["xmin"] & !is.na(x[,1])
		x[ix,1] = x[ix,1] + 360.
	}
	if (!any(is.na(gt))) { # have geotransform
		inv_gt = gdal_inv_geotransform(gt)
		if (any(is.na(inv_gt)))
			stop("geotransform not invertible")
		ret = floor(xy_from_colrow(x, inv_gt) + 1.) # will return floating point col/row numbers!!
		if (NA_outside)
			ret[ ret[,1] < 1 | ret[,1] > obj[[ xy[1] ]]$to | ret[,2] < 1 | ret[,2] > obj[[ xy[2] ]]$to, ] = NA
		ret
	} else if (is_rectilinear(obj)) {
		ix = obj[[ xy[1] ]]$values 
		if (!inherits(ix, "intervals"))
			ix = as_intervals(ix, add_last = length(ix) == dim(obj)[ xy[1] ])
		cols = find_interval(x[,1], ix)
		iy = obj[[ xy[2] ]]$values 
		if (!inherits(iy, "intervals"))
			iy = as_intervals(iy, add_last = length(iy) == dim(obj)[ xy[2] ])
		rows = find_interval(x[,2], iy) # always NA_outside
		cbind(cols, rows)
	} else if (is_curvilinear(obj)) {
		stop("colrow_from_xy not supported for curvilinear objects")
	} else
		stop("colrow_from_xy not supported for this object")
}

st_cells_from_row_col = function(x, rows, cols) {
	nc = dim(x)[1] # ncols
	(rows - 1) * nc + cols
}

st_cells_from_xy = function(x, xy) {
	x = st_upfront(x)
	cr = colrow_from_xy(xy, x, NA_outside = TRUE)
	st_cells_from_row_col(x, cr[,2], cr[,1])
}


#' return the cell index corresponding to the location of a set of points
#' 
#' If the object has been cropped without normalization, then the indices return
#' are relative to the original uncropped extent.  See \code{\link{st_crop}}
#' @param x object of class \code{stars}
#' @param sf object of class \code{sf} or \code{sfc}
#' @examples
#' set.seed(1345)
#' st_bbox(L7_ETMs) |> 
#'   st_as_sfc() |> 
#'   st_sample(10) -> pts 
#' (x <- st_cells(L7_ETMs, pts))
#' # get the pixel values (first band only):
#' st_as_stars(L7_ETMs)[[1]][x]
#' # get pixel values for all bands:
#' st_as_stars(L7_ETMs) |> split() |> sapply(`[`, x)
#' # compare with st_extract():
#' st_as_stars(L7_ETMs) |> split() |> st_extract(pts)
#' @export
st_cells = function(x, sf) {
	stopifnot(inherits(x, "stars"), inherits(sf, c("sf", "sfc")), st_crs(x) == st_crs(sf))
	st_cells_from_xy(x, st_coordinates(sf))
}

has_rotate_or_shear = function(x) {
	dimensions = st_dimensions(x)
	if (has_raster(x)) {
		r = attr(dimensions, "raster")
		!any(is.na(r$affine)) && any(r$affine != 0.0)
	} else
		FALSE
}

has_raster = function(x) {
	if (inherits(x, "stars"))
		x = st_dimensions(x)
	!is.null(r <- attr(x, "raster")) && all(r$dimensions %in% names(x))
}

is_regular_grid = function(x) {
	has_raster(x) && !(has_rotate_or_shear(x) || is_rectilinear(x) || is_curvilinear(x))
}

is_rectilinear = function(x) {
	d = st_dimensions(x)
	if (has_raster(x) && !is_curvilinear(x)) {
		xy = attr(d, "raster")$dimensions
		dimx = d[[ xy[1] ]]
		dimy = d[[ xy[2] ]]
		(is.na(dimx$delta) || is.na(dimy$delta)) && (!regular_intervals(dimx$values) || !regular_intervals(dimy$values))
	} else
		FALSE 
}

is_curvilinear = function(x) {
	d = st_dimensions(x)
	has_raster(x) && isTRUE(attr(d, "raster")$curvilinear)
}

which_sfc = function(x) {
	if (inherits(x, "stars"))
		x = st_dimensions(x)
	which(sapply(x, function(i) inherits(i$values, "sfc")))
}

which_time = function(x) {
	if (inherits(x, "stars"))
		x = st_dimensions(x)
	which(sapply(x, function(i) 
		inherits(i$values, c("POSIXct", "Date", "CFTime")) ||
		(is.character(i$refsys) && (i$refsys %in% c("POSIXct", "Date", "CFtime")))))
}

#' @export
time.stars = function(x, ..., which = 1) {
	w = which_time(x)
	if (length(w) > 1 && missing(which))
		warning(paste("using the first of", length(w), "time dimensions"))
	if (length(w) == 0)
		stop("object does not have a time dimensions")
	stopifnot(length(which) == 1)
	expand_dimensions(x)[[ w[which] ]]
}

has_sfc = function(x) {
	length(which_sfc(x)) > 0
}

#' retrieve coordinates for raster or vector cube cells
#'
#' retrieve coordinates for raster or vector cube cells
#' @param x object of class \code{stars}
#' @param add_max logical; if \code{TRUE}, dimensions are given with a min (x) and max (x_max) value
#' @param center logical; (only if \code{add_max} is FALSE): should grid cell center coordinates be returned (TRUE) or offset values (FALSE)? \code{center} can be a named logical vector or list to specify values for each dimension.
#' @name st_coordinates
#' @param ... ignored
#' @export
st_coordinates.stars = function(x, ..., add_max = FALSE, center = TRUE) {
	dims = st_dimensions(x)
	xy = attr(dims, "raster")$dimensions
	if (is_curvilinear(x)) {
		x = st_upfront(x) # x and y first...
		cc = setNames(data.frame(as.vector(dims[[ xy[1] ]]$values), as.vector(dims[[ xy[2] ]]$values)), xy)
		dims[ xy ] = NULL # remove
		out = do.call(expand.grid, append(list(ix = seq_len(nrow(cc))), expand_dimensions(dims, center = center))) 
			# cell offsets
		ix = out$ix
		out$ix = NULL
		out = cbind(cc[[1]][ix], cc[[2]][ix], out)
		names(out)[1:2] = xy
		out
	} else if (has_rotate_or_shear(x)) {
		if (add_max)
			stop("add_max will not work for rotated/shared rasters")
		if (isTRUE(!center)) # center = FALSE
			warning("center values are given for spatial coordinates")
		d = dim(x)
		nx = d[ xy[1] ]
		ny = d[ xy[2] ]
		setNames(as.data.frame(xy_from_colrow(as.matrix(expand.grid(seq_len(nx), seq_len(ny))) - 0.5,
			st_geotransform(x))), xy) # gives cell centers
	} else {
		if (add_max) {
			cbind(
				do.call(expand.grid, expand_dimensions(x, center = FALSE)), # cell offsets
				setNames(do.call(expand.grid, expand_dimensions(dims[xy], max = TRUE)),
					paste0(xy, "_max"))
			)
		} else {
			ed = expand_dimensions(x, center = center) # cell centers for x/y if raster
			if (length(ed) > 1)
				do.call(expand.grid, ed)
			else
				data.frame(ed)
		}
	}
}

#' @export
st_coordinates.dimensions = function(x, ...) {
	st_coordinates(st_as_stars(list(), dimensions = x), ...)
}

#' @name st_coordinates
#' @export
#' @param add_coordinates logical; if `TRUE`, columns with dimension values preceed the array values, 
#' otherwise they are omitted
as.data.frame.stars = function(x, ..., add_max = FALSE, center = NA, add_coordinates = TRUE) {
	df = setNames(as.data.frame(lapply(x, function(y) structure(y, dim = NULL))), names(x))
	if (add_coordinates)
		cbind(st_coordinates(x, add_max = add_max, center = center, ...), df)
	else
		df
}

add_units = function(x) {
	f = function(obj) if (inherits(obj, "units")) paste0("[", enc2utf8(as.character(units(obj))), "]") else ""
	paste(names(x), sapply(x, f))
}

#' print stars or dimensions object
#' 
#' print stars or dimensions object
#' @name print_stars
#' @param x object of class stars or of class dimensions
#' @param n when prod(dim(x)) > 10 * n, the first n cells are used for attribute summary statistics
#' @param abbrev number of characters to abbreviate attribute names to
#' @param ... passed on to \code{as.data.frame.dimensions}
#' @export
print.stars = function(x, ..., n = 1e5, abbrev = 30) {
	shorten = function(s) {
		if (nchar(s) > abbrev)
			paste0(substr(s, 1, abbrev - 3), "...")
		else
			s
	}
	cat("stars object with", length(dim(x)), "dimensions and", 
		length(x), if (length(x) != 1) "attributes\n" else "attribute\n")
	if (length(x)) {
		cat("attribute(s)")
		names(x) = sapply(names(x), shorten)
		df = if (prod(dim(x)) > 10 * n) {
			cat(paste0(", summary of first ", n, " cells:\n"))                       # nocov
			as.data.frame(lapply(x, function(y) structure(y, dim = NULL)[1:n]), optional = TRUE) # nocov
		} else {
			cat(":\n")
			as.data.frame(lapply(x, function(y) structure(y, dim = NULL)), optional = TRUE)
		}
		names(df) = add_units(x)
		if (all(sapply(x, is.numeric))) {
			m_summary = function(x) { s = summary(x); if (!"NA's" %in% names(s)) s["NA's"] = 0; s }
			sums = lapply(df, summary)
			if (length(unique(lengths(sums))) > 1)
				sums = lapply(df, m_summary)
			print(do.call(rbind, sums))
		} else
			print(summary(df))
	}
	cat("dimension(s):\n")
	print(st_dimensions(x), ...)
	invisible(x)
}

#' @export
aperm.stars = function(a, perm = NULL, ...) {
	if (is.null(perm))
		perm = rev(seq_along(dim(a)))
	if (all(perm == seq_along(dim(a))) || isTRUE(all(match(perm, names(dim(a))) == seq_along(dim(a)))))
		return(a)
	d = st_dimensions(a)
	if (is.character(perm))
		perm = match(perm, names(d))
	st_stars(lapply(a, aperm, perm = perm, ...), d[perm])
}

#' @export
dim.stars = function(x) {
	d = st_dimensions(x)
	if (length(x) == 0)
		dim(d)
	else {
		stopifnot(length(d) == length(dim(x[[1]])))
		structure(dim(x[[1]]), names = names(d))
	}
}

propagate_units = function(new, old) {
	for (i in seq_along(new))
		if (inherits(old[[i]], "units"))
			units(new[[i]]) <- units(old[[i]])
	new
}

setNamesIfnn = function(x, nms) { # set names if not NULL
	stopifnot(is.character(nms) || is.null(nms))
	if (is.null(nms) || length(nms) != length(x))
		x
	else 
		setNames(x, nms)
}

#' combine multiple stars objects, or combine multiple attributes in a single stars object into a single array
#' 
#' combine multiple stars objects, or combine multiple attributes in a single stars object into a single array
#' @param ... object(s) of class \code{star}: in case of multiple arguments, these are combined into a single stars object, in case of a single argument, its attributes are combined into a single attribute. In case of multiple objects, all objects should have the same dimensionality.
#' @param along integer; see \link{read_stars}
#' @param try_hard logical; if \code{TRUE} and some arrays have different dimensions, combine those that dimensions matching to the first array
#' @param tolerance numeric; values used in \link{all.equal} to compare dimension values
#' combine those that dimensions matching to the first array
#' @param nms character; vector with array names
#' @returns a single \code{stars} object with merged (binded) arrays.
#' @details An error is raised when attempting to combine arrays with different
#' measurement units into a single array. If this was intentded, \code{drop_units} 
#' can be used to remove units of a \code{stars} object before merging.
#' @export
#' @examples
#' tif = system.file("tif/L7_ETMs.tif", package = "stars")
#' x = read_stars(tif)
#' (new = c(x, x))
#' c(new) # collapses two arrays into one with an additional dimension
#' c(x, x, along = 3)
c.stars = function(..., along = NA_integer_, try_hard = FALSE, nms = names(list(...)), tolerance = sqrt(.Machine$double.eps)) {
	dots = list(...)
	if (!all(sapply(dots, inherits, "stars")))
		stop("all arguments to c() should be stars objects")
	if (any(sapply(dots, inherits, "stars_proxy")))
		stop("convert stars_proxy objects to stars first using st_as_stars()")

	if (length(dots) == 1 && length(along) == 1 && missing(along))
		dots[[1]]
	else if (length(along) == 1 && is.na(along)) { 
		# Case 1: merge attributes of several objects by simply putting them together in a single stars object;
		# dim does not change:
		if (identical_dimensions(dots, tolerance = tolerance))
			st_as_stars(setNamesIfnn(do.call(c, lapply(dots, unclass)), nms),
						dimensions = st_dimensions(dots[[1]]))
		else {
			# currently catches only the special case of ... being a broken up time series:
			along = sort_out_along(dots)
			if (!is.na(along))
				do.call(c, c(dots, along = along))
			else if (!try_hard)
				stop("don't know how to merge arrays: please specify parameter along")
			else {
				d = lapply(dots, st_dimensions)
				ident = c(TRUE, sapply(d[-1], identical, d[[1]]))
				if (!all(ident)) {
					warning(paste(
					"ignored subdataset(s) with dimensions different from first subdataset:", 
					paste(which(!ident), collapse = ", "), 
					"\nuse gdal_subdatasets() to find all subdataset names"))
					if (!is.null(nms))
						nms = nms[ident]
				}
				st_as_stars(setNamesIfnn(do.call(c, lapply(dots[ident], unclass)), nms),
							dimensions = st_dimensions(dots[[1]]))
			}
		}
	} else {
		if (is.list(along)) { # custom ordering of ... over dimension(s) with values specified
			if (prod(lengths(along)) != length(dots))
				stop("number of objects does not match the product of lengths of the along argument", call. = FALSE)
			# abind all:
			d = st_dimensions(dots[[1]])
			ret = mapply(abind, ..., along = length(d) + 1, SIMPLIFY = FALSE)
			# make dims:
			newdim = c(dim(dots[[1]]), lengths(along))
			ret = lapply(ret, function(x) { dim(x) = newdim; x })
			ret = propagate_units(ret, dots[[1]])
			# make dimensions:
			for (i in seq_along(along))
				d[[ names(along)[i] ]] = create_dimension(values = along[[i]])
			st_as_stars(ret, dimensions = d)
		} else { # loop over attributes, abind them:
			# along_dim: the number of the dimension along which we merge arrays
			d = st_dimensions(dots[[1]])
			along_dim = if (is.character(along)) {
				along_dim = which(along == names(d))
				if (length(along_dim) == 0)
					length(d) + 1
				else
					along_dim
			} else
				along
			stopifnot_identical_units(dots)
			ret = propagate_units(mapply(abind, ..., along = along_dim, SIMPLIFY = FALSE), dots[[1]])
			dims = combine_dimensions(dots, along_dim)
			if (along_dim == length(d) + 1)
				names(dims)[along_dim] = if (is.character(along)) along else "new_dim"
			st_as_stars(ret, dimensions = dims)
		}
	}
}

stopifnot_identical_units = function(lst) {
	a1 = lst[[1]][[1]]
	for (i in seq_along(lst[-1])) {
		ai = lst[[i+1]][[1]]
		if (inherits(a1, "units") && !identical(units(a1), units(ai)))
			stop("cannot merge subarrays with different units")
	}
	TRUE
}


#' @export
adrop.stars = function(x, drop = which(dim(x) == 1), ..., drop_xy = FALSE) {
	if (missing(drop) && !drop_xy) { # hanlde drop_xy: by default, don't drop x/y
		d = st_dimensions(x)
		xy = attr(d, "raster")$dimensions
		drop = setdiff(drop, match(xy, names(d)))
	}
	if (is.logical(drop))
		drop = which(drop)
	if (any(dim(x) > 1) && length(drop) > 0) {
		l = vector("list", length = length(x))
		f = sapply(x, is.factor)
		l[!f] = lapply(x[!f], adrop, drop = drop, one.d.array = TRUE, ...)
		l[f] = lapply(x[f], function(x) structure(x, dim = dim(x)[-drop]))
		st_as_stars(setNames(l, names(x)), dimensions = st_dimensions(x)[-drop])
	} else 
		x
}

#' @export
st_bbox.default = function(obj, ...) {
	if (!missing(obj))
		stop(paste("no st_bbox method available for object of class", class(obj)))
	obj = st_sfc(st_point(c(-180,-90)), st_point(c(180, 90)), crs = st_crs('OGC:CRS84'))
	st_bbox(obj)
}

#' @export
st_bbox.dimensions = function(obj, ...) {
	if (has_raster(obj)) { # raster
		r = attr(obj, "raster")
		x = obj[[ r$dimensions[1] ]]
		y = obj[[ r$dimensions[2] ]]
		bb = if (is.null(x$values) && is.null(y$values)) {
				gt = st_geotransform(obj)
				if (length(gt) == 6 && !any(is.na(gt))) {
					bb = rbind(c(x$from - 1, y$from - 1), c(x$to, y$from - 1), c(x$to, y$to), c(x$from - 1, y$to))
					xy = xy_from_colrow(bb, gt)
					c(xmin = min(xy[,1]), ymin = min(xy[,2]), xmax = max(xy[,1]), ymax = max(xy[,2]))
				} else
					c(xmin = x$from - 0.5, ymin = y$from - 0.5, xmax = x$to + 0.5, ymax = y$to + 0.5)
			} else {
				if (is_curvilinear(obj))
					c(xmin = min(x$values, na.rm = TRUE),
						ymin = min(y$values, na.rm = TRUE),
						xmax = max(x$values, na.rm = TRUE),
						ymax = max(y$values, na.rm = TRUE))
				else {
					rx = range(x) # dispatches into range.dimension
					ry = range(y)
					c(xmin = rx[1], ymin = ry[1], xmax = rx[2], ymax = ry[2])
				}
			}
		structure(bb, crs = st_crs(x$refsys), class = "bbox")
	} else {
		if (! has_sfc(obj))
			stop("dimensions table does not have x & y, nor an sfc dimension") # nocov
		ix = which_sfc(obj)
		if (length(ix) > 1)
			warning("returning the bounding box of the first geometry dimension")
		st_bbox(obj[[ ix[1] ]]$values)
	}
}

#' @export
st_bbox.stars = function(obj, ...) {
	st_bbox(st_dimensions(obj), ...)
}

#' set bounding box parameters of regular grid
#' @param x object of class dimensions, stars or stars_proxy
#' @param value object of class bbox
#' @param ... ignored
#' @export
st_set_bbox = function(x, value, ...) UseMethod("st_set_bbox")

#' @export
st_set_bbox.dimensions = function(x, value, ...) {
	stopifnot(inherits(value, "bbox"), is_regular_grid(x))
	xy = attr(x, "raster")$dimensions
	if (x[[ xy[1] ]]$from != 1 || x[[ xy[2] ]]$from != 1)
		stop("use st_normalize first so that dimensions start at index 1")
	d = dim(x)
	xsign = sign(x[[ xy[1] ]]$delta)
	ysign = sign(x[[ xy[2] ]]$delta)
	x[[ xy[1] ]]$offset = ifelse(xsign < 0, value[["xmax"]], value[["xmin"]])
	x[[ xy[2] ]]$offset = ifelse(ysign < 0, value[["ymax"]], value[["ymin"]])
	x[[ xy[1] ]]$delta = xsign * (value[["xmax"]] - value[["xmin"]]) / d[[ xy[1] ]]
	x[[ xy[2] ]]$delta = ysign * (value[["ymax"]] - value[["ymin"]]) / d[[ xy[2] ]]
	if (!is.na(st_crs(value)))
		st_crs(x) = st_crs(value)
	x
}

#' @export
st_set_bbox.stars = function(x, value, ...) {
	structure(x, dimensions = st_set_bbox(st_dimensions(x), value))
}

#' @export
st_set_bbox.stars_proxy = function(x, value, ...) {
	structure(x, dimensions = st_set_bbox(st_dimensions(x), value))
}


#' @export
st_crs.stars = function(x, ...) {
	st_crs(st_dimensions(x), ...)
}

#' @export
st_crs.dimensions = function(x, ...) {
	xy = attr(x, "raster")$dimensions
	if (!all(is.na(xy)))
		st_crs(x[[ xy[1] ]]$refsys)
	else if (has_sfc(x)) # search for simple features:
		st_crs(x[[ which_sfc(x)[1] ]]$values)
	else
		NA_crs_
}

#' @export
`st_crs<-.stars` = function(x, value) {
	structure(x, dimensions = st_set_crs(st_dimensions(x), value))
}

#' @export
`st_crs<-.dimensions` = function(x, value) {
	value = if (is.na(value))
			NA_crs_
		else if (is.numeric(value) || is.character(value))
			st_crs(value)
		else if (inherits(value, "crs"))
			value
		else
			stop(paste("crs of class", class(value), "not recognized"))

	drop_if_units = function(x) {
		if (inherits(x, "units"))
			units::drop_units(x)
		else
			x
	}
	# set CRS in dimensions:
	xy = attr(x, "raster")$dimensions
	if (!all(is.na(xy))) { # has x/y spatial dimensions:
		x[[ xy[1] ]]$refsys = value
		x[[ xy[2] ]]$refsys = value
		x[[ xy[1] ]]$offset = drop_if_units(x[[ xy[1] ]]$offset)
		x[[ xy[2] ]]$offset = drop_if_units(x[[ xy[2] ]]$offset)
		x[[ xy[1] ]]$delta  = drop_if_units(x[[ xy[1] ]]$delta)
		x[[ xy[2] ]]$delta  = drop_if_units(x[[ xy[2] ]]$delta)
	}

	if (!all(is.na(xy)) && !is.na(x[[ xy[1] ]]$refsys) && !is.na(value) && st_crs(x) != value)
		warning("replacing CRS does not reproject data: use st_transform, or st_warp to warp to a new CRS")

	# set crs of sfc's, if any:
	for (j in which_sfc(x)) {
		x[[ j ]]$refsys = value
		st_crs(x[[ j ]]$values) = value
	}
	x
}


#' @export
st_geometry.stars = function(obj,...) {
	if (!has_sfc(obj))
		stop("stars object does not have a simple feature dimension")
	d = st_dimensions(obj)
	d[[ which_sfc(obj) ]]$values
}

# make sure asub works for factor too:
#' @export
asub.factor = function(x, idx, dims, drop = NULL, ...) {
	l = levels(x)
	x = unclass(x)
	ret = NextMethod()
	structure(ret, class = class(x), levels = l)
}

#' @name merge
#' @aliases split
#' @param f the name or index of the dimension to split; by default the last dimension
#' @param drop ignored
#' @details split.stars works on the first attribute, and will give an error when more than one attribute is present
#' @export
split.stars = function(x, f = length(dim(x)), drop = TRUE, ...) {
	if (length(x) > 1) {
		l = lapply(seq_along(x), function(i) split(x[i], f, drop = drop, ...))
  		do.call(c, setNames(l, names(x)))
	} else {
		d = st_dimensions(x)
		if (is.character(f))
			f = which(names(d) == f)
		ret = lapply(seq_len(dim(x)[f]), function(y) asub(x[[1]], y, f, drop = TRUE))
		nm = if (!is.null(d[[f]]$values))
				d[[f]]$values
			else
				make.names(seq_along(ret))
		st_as_stars(setNames(ret, nm), dimensions = d[-f])
	}
}

#' merge or split stars object
#' 
#' merge attributes into a dimension, or split a dimension over attributes
#' @param x object of class \code{stars}
#' @param y needs to be missing
#' @param name name for the new dimension
#' @param ... if defined, the first unnamed argument is used for dimension values, if not defined, attribute names are used for dimension values
#' @returns merge merges attributes of a stars object into a new dimension; split splits a dimension over attributes
#' @name merge
#' @export
merge.stars = function(x, y, ..., name = "attributes") {
	dots = list(...)
	if (!missing(y))
		stop("argument y needs to be missing: merging attributes of x")
	old_dim = st_dimensions(x)
	out = st_redimension(x, name = name)
	new_dim = if (length(dots))
			create_dimension(values = dots[[1]])
		else
			create_dimension(values = names(x))
	dims = setNames(c(old_dim, list(new_dim)), make.unique(c(names(old_dim), name)))
	d = create_dimensions(dims, raster = attr(old_dim, "raster"))
	if (!is.null(names(dots)))
		names(d)[length(d)] = names(dots)
	st_stars(out, dimensions = d) # overwrite dimensions of out
}

sort_out_along = function(ret) { 
	# https://github.com/r-spatial/stars/issues/703 :
	# 1. check that time is a dimension name in all objects 
	l = lapply(ret, st_dimensions)
	if (!all(sapply(l, function(x) "time" %in% names(x))))
		return(NA_integer_)
	# 2. check that time values do not overlap
	lv = lapply(l, st_get_dimension_values, "time")
	for (i in seq_along(lv)) {
		if (!inherits(lv[[i]], c("POSIXt", "Date", "CFTime")))
			return(NA_integer_)
		if (i < length(lv) && max(lv[[i]]) >= min(lv[[i+1]])) # no sequence
			return(NA_integer_)
	}
	return("time")
}


#' @export
is.na.stars = function(x, ...) {
	st_as_stars(lapply(x, is.na), dimensions = st_dimensions(x))
}

#' redimension array, or collapse attributes into a new dimension
#' 
#' redimension array, or collapse attributes into a new dimension
#' @name redimension
#' @export
st_redimension = function(x, new_dims, along, ...) UseMethod("st_redimension")

#' @export
#' @name redimension
#' @param x object of class \code{stars}
#' @param new_dims target dimensions: either a `dimensions` object or an integer vector with the dimensions' sizes
#' @param along named list with new dimension name and values
#' @param name character name of the new dimension
#' @param ... ignored
st_redimension.stars = function(x, new_dims = st_dimensions(x), 
		along = setNames(list(names(x)), name), ..., name = "new_dim") {

	d = st_dimensions(x)
	if (inherits(new_dims, "dimensions")) {
		di = dim(new_dims)
	} else {
		di = new_dims
		new_dims = create_dimensions(di)
	}
	if (!isTRUE(all.equal(di, dim(x), check.attributes = FALSE))) {
		if (prod(dim(x)) != prod(di))
			stop("product of dim(new_dim) does not match that of x")
		for (i in seq_len(min(length(di), length(dim(x)))))
			if (di[i] == dim(x)[i]) {
				new_dims[[i]] = d[[i]]
				names(new_dims)[i] = names(d[i])
			}
		x = unclass(x)
		raster = attr(d, "raster")
		if (all(raster$dimensions %in% names(new_dims)))
			attr(new_dims, "raster") = raster
		for (i in seq_along(x))
			dim(x[[i]]) = di
		st_stars(x, dimensions = new_dims)
	} else { # collapse attributes into dimension
		if (length(x) == 1) # only one attribute: do nothing
			x
		else {
			new_dim = create_dimension(values = along[[1]])
			dims = create_dimensions(c(d, new_dim = list(new_dim)), attr(d, "raster"))
			if (length(names(along)) == 1)
				names(dims)[names(dims) == "new_dim"] = names(along)
			ns = names(x)
			x = if (inherits(x[[1]], c("factor", "Date", "POSIXt")))
					structure(do.call(c, x), dim = dim(dims)) # don't do this on large data sets!!
				else
					do.call(abind, append(unclass(x), list(along = length(dims))))
			st_stars(setNames(list(x), paste(ns, collapse = ".")), dimensions = dims)
		}
	}
}

#' @export
"$<-.stars" = function(x, i, value) {
	x[[i]] = value
	x
}

#' @export
"[[<-.stars" = function(x, i, value) {
	if (!is.null(value)) {
		if (prod(dim(x)) %% length(value) != 0) { # error:
			if (is.null(dim(value)))
				stop(paste("replacement has length", length(value), ", data has dim", paste(dim(x), collapse = ", ")))
			else
				stop(paste("replacement has dim", paste(dim(value), collapse = ", "), ", data has dim", paste(dim(x), collapse = ", ")))
		}
		if (inherits(value, "stars")) {
			stopifnot(length(value) == 1)
			value = value[[1]]
		}
		value = if (inherits(value, c("factor", "POSIXct")))
				structure(rep(value, length.out = prod(dim(x))), dim = dim(x), colors = attr(value, "colors"),
					rgba = attr(value, "rgba"), exclude = attr(value, "exclude"))
			else if (!is.array(value) || !isTRUE(all.equal(dim(value), dim(x), check.attributes = FALSE)))
				array(value, dim(x))
			else
				value
	}
	NextMethod()
}

st_upfront = function(x, first = attr(st_dimensions(x), "raster")$dimensions) {
	if (!is.character(first))
		first = names(st_dimensions(x))[first]
	if (!any(is.na(first)))
		aperm(x, c(first, setdiff(names(st_dimensions(x)), first)))
	else
		x
}

#' @export
st_area.stars = function(x, ...) {
	crs = st_crs(x)
	if (is.na(crs))
		message("Missing coordinate reference system: assuming Cartesian coordinates")
	d = st_dimensions(st_upfront(x))[1:2]
	a = if (isTRUE(st_is_longlat(x)) || is_curvilinear(x))
			st_area(st_as_sfc(x, as_points = FALSE)) # has units
		else { 
			a = if (is_regular_grid(x))
					d[[1]]$delta * d[[2]]$delta
				else { # rectilinear:
					x = if (inherits(d[[1]]$values, "intervals"))
							d[[1]]$values
						else
							as_intervals(d[[1]]$values)
					y = if (inherits(d[[2]]$values, "intervals"))
							d[[2]]$values
						else
							as_intervals(d[[2]]$values)
					apply(do.call(cbind, x), 1, diff) %o% apply(do.call(cbind, y), 1, diff)
				}
			if (!is.na(crs))
				units::set_units(abs(a), paste0(crs$units, "^2"), mode = "standard")
			else
				abs(a)
		}
	lst = if (inherits(a, "units"))
			list(area = `units<-`(array(a, dim(d)), units(a)))
		else
			list(area = array(a, dim(d)))
	st_stars(lst, dimensions = d)
}

#' @export
drop_units.stars = function(x) {
	try_drop_units = function(x) {
		if (inherits(x, "units"))
			units::drop_units(x)
		else
			x
	}
	st_stars(lapply(x, try_drop_units), dimensions = st_dimensions(x))
}

#' Predict values, given a model object, for a stars or stars_proxy object
#' @export
#' @name predict.stars
#' @param object object of class `stars`
#' @param model model object of a class that has a predict method; check with `methods(class = class(object))`
#' @param drop_dimensions logical; if `TRUE`, remove dimensions (coordinates etc) from `data.frame` with predictors
#' @param ... arguments passed on to this predict method
#' @details separate predictors in object need to be separate attributes in object; 
#' in case they are e.g. in a band dimension, use `split(object)`
predict.stars = function(object, model, ..., drop_dimensions = FALSE) {
	obj_df = as.data.frame(st_as_stars(object))
	if (drop_dimensions)
		obj_df = obj_df[-seq_along(dim(object))]
	na_ids = which(is.na(obj_df), arr.ind = TRUE) # identify rows with NA's in the predictors
	obj_df[na_ids] = 0  # fill with something valid (e.g. 0)
	pr = try(predict(model, obj_df, ...), silent = TRUE)
	has_method = function(generic, cls) {
		m = row.names(attr(methods(generic), "info"))
		cls %in% substring(m, nchar(generic) + 2, 1e4)
	}
	if (inherits(pr, "try-error") && !has_method("predict", class(model)))
		stop(paste("No predict method found for objects of class", class(model)))
	if (inherits(pr, "try-error")) { # https://github.com/r-spatial/stars/issues/448
		m = paste0("prediction on array(s) `", paste(names(object), collapse = ","), "' failed; will try to split() dimension `", tail(names(dim(object)), 1), "' over attributes")
		message(m)
		predict(split(object), model, ..., drop_dimensions = drop_dimensions) # returns
	} else {
		if (!inherits(pr, "data.frame"))
			pr = if (is.null(colnames(pr)))
					data.frame(prediction = pr)
				else
					as.data.frame(pr)
		pr[unique(data.frame(na_ids)[,1]), ] = NA # Mask with original NA's
		st_stars(lapply(pr, function(y) structure(y, dim = dim(object))), st_dimensions(object))
	}
}


#' create an array with dimension values
#' 
#' create an array with dimension values
#' @param x object of class \code{stars}
#' @param which integer; indices of the dimensions to address (default: all)
#' @return \code{stars} object with dimension values as attributes
#' @export
#' @examples
#' tif = system.file("tif/L7_ETMs.tif", package = "stars")
#' x1 = read_stars(tif)
#' (x = st_dim_to_attr(x1))
#' plot(x)
#' (x = st_dim_to_attr(x1, 2:3))
#' plot(x)
#' (x= st_dim_to_attr(x1, 3))
#' plot(x)
st_dim_to_attr = function(x, which = seq_along(dim(x))) {
	d = dim(x)
	l = vector("list", length = length(which))
	e = expand_dimensions(x)
	for (i in seq_along(which)) {
		l [[i]] = if (is.null(dim(e[[i]]))) {
				dp = c(which[i], setdiff(seq_along(dim(x)), which[i]))
				aperm(array(e[[ which[i] ]], d[dp]), order(dp))
			} else # curvilinear:
				array(e[[i]], d)
	}
	st_stars(setNames(l, names(d)[which]), st_dimensions(x))
}

#' @export
st_interpolate_aw.stars = function(x, to, extensive, ...) {
	ret = sf::st_interpolate_aw(st_as_sf(x), to, extensive, ...)
	geom = attr(ret, "sf_column")
	dx = dim(x)
	if (length(dx) > 2 && length(x) == 1 && length(ret) > 2) {
		ret = merge(st_as_stars(ret))
		nd = names(st_dimensions(x))
		ret = st_set_dimensions(ret, seq_along(dx), 
								names = c(geom, paste0(nd[-(1:2)], collapse = ".")))
		setNames(ret, names(x))
	} else
		ret
}

#' get the raster type (if any) of a stars object
#' @param x object of class \code{stars}
#' @param dimension optional: numbers or names of dimension(s) to get per-dimension type
#' @return if \code{dimension} is not specified, return the spatial raster type: 
#' one of \code{NA} (if the object does not have raster dimensions), 
#' \code{"curvilinear"}, \code{"rectilinear"}, \code{"affine"}, or \code{"regular"}.
#' In case dimension(s) are specified, return one of \code{"regular"}, \code{"rectilinear"}
#' (irregular but numeric), or \code{"discrete"} (anything else).
#' @details categories \code{"curvilinear"} and \code{"affine"} only refer to
#' the relationship between a pair of spatial (raster) dimensions.
#' @examples
#' tif = system.file("tif/L7_ETMs.tif", package = "stars")
#' x = read_stars(tif)
#' st_raster_type(x)
#' st_raster_type(x, 1:3)
#' @export
st_raster_type = function(x, dimension = character(0)) {
	stopifnot(inherits(x, "stars"))
	if (!missing(dimension))
		stopifnot(all(dimension >= 1), all(dimension <= length(dim(x))))
	dimension_type = function(d) {
		if (!any(is.na(c(d$offset, d$delta))))
			"regular"
		else if (!is.null(d$values) && is.numeric(d$values) && is.matrix(d$values))
			"curvilinear"
		else if (!is.null(d$values) && is.numeric(d$values))
			"rectilinear"
		else
			"discrete"
	}
	if (length(dimension))
		sapply(st_dimensions(x)[dimension], dimension_type)
	else if (!has_raster(x))
		NA_character_
	else if (is_curvilinear(x))
		"curvilinear"
	else if (is_rectilinear(x))
		"rectilinear"
	else if (has_rotate_or_shear(x))
		"affine"
	else
		"regular"
}

#' obtain (spatial) resolution of a stars object
#'
#' obtain resolution(s) of a stars object: by default only the (absolute) x/y raster dimensions, optionally all \code{delta} dimension parameters 
#' @param x an object of class \code{stars}
#' @param all logical; if FALSE return a vector with the x/y raster resolution
#' @param absolute logical; only works when \code{all = FALSE}; if TRUE return absolute resolution values, if FALSE return \code{delta} values
#' @returns if \code{all = FALSE} a vector with x/y raster resolutions, otherwise a list with delta values
#' @examples
#' st_res(L7_ETMs)
#' st_res(L7_ETMs, absolute = FALSE)
#' st_res(L7_ETMs, all = TRUE)
#' if (require(starsdata)) {
#'   paste0("netcdf/", c("avhrr-only-v2.19810901.nc", 
#'     "avhrr-only-v2.19810902.nc",
#'     "avhrr-only-v2.19810903.nc",
#'     "avhrr-only-v2.19810904.nc")) |>
#'   system.file(package = "starsdata") |>
#'   read_stars(quiet = TRUE) -> x
#'   st_res(x) |> print()
#'   st_res(x, all = TRUE) |> print()
#' }
#' @export
st_res = function(x, all = FALSE, absolute = !all) {
	stopifnot(inherits(x, "stars"))
	d = st_dimensions(x)
	l = lapply(d, `[[`, "delta")
	if (!all) {
		xy = attr(d, "raster")$dimensions
		l = unlist(l[xy])
		if (absolute)
			l = abs(l)
	}
	l
}

context("st_as_stars tests")

options(rgdal_show_exportToProj4_warnings = "none")

test_that("basic st_as_stars", {
  skip_if_not_installed("ncdfgeom")
  skip_on_os("solaris")
  skip_if_not_installed("ncdfgeom")

  f <- system.file("nc/timeseries.nc", package = "stars")

  test_list <- ncdfgeom::read_timeseries_dsg(f)

  stars_obj <- st_as_stars(test_list)

  expect_true("XYZ" %in% class(st_dimensions(stars_obj)$points$values[[1]]))

  expect_s3_class(stars_obj, "stars")

  dim <- stars::st_dimensions(stars_obj)
  #expect_true(sf::st_crs(dim$points$refsys) == sf::st_crs(4326))
  expect_equal(dim$time$refsys, "POSIXct")

  expect_s3_class(dim$points$values, "sfc_POINT")

  expect_true(dim$points$point)

  geom_point <- sf::st_as_sf(data.frame(id = names(test_list$data_frames[[1]]),
                                        lon = test_list$lons,
                                        lat = test_list$lats, stringsAsFactors = FALSE),
                             coords = c("lon", "lat"), crs = 4326)

  geom_poly <- sf::st_buffer(st_transform(geom_point, 3857), dist = 100)

  stars_obj <- st_as_stars(test_list, sf_geometry = geom_poly)

  dim <- stars::st_dimensions(stars_obj)
  expect_equal(sf::st_crs(dim$geometry$refsys)$proj4string,
               sf::st_crs(geom_poly)$proj4string)

  expect_s3_class(dim$geometry$values, "sfc_POLYGON")
  expect_false(dim$geometry$point)

  stars_obj <- st_as_stars(test_list, sf_geometry = geom_point)
  dim <- stars::st_dimensions(stars_obj)

  expect_equal(sf::st_crs(dim$geometry$refsys)$proj4string,
               sf::st_crs(geom_point)$proj4string)
  expect_true(dim$geometry$point)
  expect_s3_class(dim$geometry$values, "sfc_POINT")

  test_list$alts <- numeric(0)

  stars_obj <- st_as_stars(test_list)

  expect_true("XY" %in% class(st_dimensions(stars_obj)$points$values[[1]]))
})

test_that("st_as_stars.Raster", {
  skip_if_not_installed("raster")
  tif = system.file("tif/L7_ETMs.tif", package = "stars")
  x = raster::raster(tif)
  stars_obj <- st_as_stars(x)
  expect_s3_class(stars_obj, "stars")
  expect_equal(names(stars_obj), names(x))

  # change the name of the attribute
  names(x) <- "newname"
  stars_obj2 <- st_as_stars(x)
  expect_equal(names(stars_obj2), names(x))
})

skip_on_cran()
f <- system.file("nc/reduced.nc", package = "stars")

test_that("proxy", {
	skip_if_not_installed("ncmeta")
	nc <- read_ncdf(f, proxy = TRUE)
	expect_equal(as.character(nc[[1]]), f)
	expect_s3_class(nc, "stars_proxy")

	output <- capture_output(print(nc))

	expect_true(grepl("^netcdf source", output))

	old_opts <- options("stars.n_proxy" = 100)
	nc <- read_ncdf(f)
	expect_equal(as.character(nc[[1]]), f)

	expect_warning(nc <- read_ncdf(f, proxy = FALSE))
	expect_s3_class(nc, "stars")

	options(old_opts)
})

test_that("st_as_stars.nc_proxy", {
    skip_if_not_installed("ncmeta")

	nc <- read_ncdf(f, proxy = TRUE)

	nc2 <- nc[1, 25:35, 25:35, 1, 1]
	
	nc2 <- st_as_stars(nc2)

	expect_s3_class(nc2, 'stars', exact = TRUE)
	expect_equal(names(nc2), "sst")

	dim <- st_dimensions(nc2)

	expect_equal(dim$lon$to, 11)
	expect_equal(dim$lon$offset, 47)

	nc2 <- st_as_stars(nc, var = "sst",
					   proxy = TRUE)

	expect_s3_class(nc2, c("nc_proxy", "stars_proxy", "stars"), exact = TRUE)

	dim <- st_dimensions(nc2)

	expect_equal(dim$lon$to, 180)
	expect_equal(dim$lon$offset, -1)

})

test_that("basics", {
    skip_if_not_installed("ncmeta")
	nc <- read_ncdf(f, proxy = TRUE)
	
	x <- nc["sst"]

	expect_s3_class(x, "nc_proxy")
	
	expect_equal(length(x), 1)
	
	expect_equal(names(x), "sst")
	
	nc[[5]] <- nc[[4]]
	
	expect_equal(length(nc), 5)
	
	expect_equal(names(nc), c("sst", "anom", "err", "ice", ""))
	
	xx <- c(nc["sst"], nc["ice"])
	
	expect_equal(length(xx), 2)
	
	expect_equal(names(xx), c("sst", "ice"))
	
	expect_error(st_mosaic(nc))
	
	expect_error(st_redimension(nc))
})

test_that("subset", {
	skip_if_not_installed("ncmeta")
	nc <- read_ncdf(f, proxy = TRUE)
	
	nc2 <- nc[1, 80:100, 45:55, , ]
	
	nc3 <- nc["sst", 80:100, 45:55, , ]
	
	expect_equal(nc2, nc3)
	
	expect_equal(dim(nc2), 
				 setNames(c(21, 11, 1, 1), 
				 		 c("lon", "lat", "zlev", "time")))
	
	nc3 <- st_as_stars(nc2)
	
	expect_s3_class(nc3, "stars")
	
	expect_equal(as.numeric(dim(nc3)), dim(nc3$sst))
	
	nc <- read_ncdf(system.file("nc/bcsd_obs_1999.nc", package = "stars"),
					proxy = TRUE)
	
	nc2 <- nc[ , , , 5]
	expect_equal(st_dimensions(nc2)$time$from, 5)
	expect_equal(st_dimensions(nc2)$time$values, 
				 structure(928108800, class = c("POSIXct", "POSIXt"), tzone = "UTC"))
	
	nc3 <- st_as_stars(nc2)
	
	expect_equal(dim(nc3$pr), as.numeric(dim(nc2)))
	
	nc_sf <- sf::st_transform(
		read_sf(system.file("gpkg/nc.gpkg", package="sf")),
		sf::st_crs(nc))
	
	nc2 <- st_crop(nc, nc_sf[10, ])
	nc2 <- nc2["pr", , , 1]
	nc2 <- st_as_stars(nc2)

	nc3 <- nc["pr", , , 1]	
	nc3 <- st_crop(nc3, nc_sf[10, ])
	nc3 <- st_as_stars(nc3)
	
	expect_equal(nc2, nc3)
})

test_that("curvilinear", {
	skip_if_not_installed("ncmeta")
	f <- system.file("nc/test_stageiv_xyt.nc", package = "stars")
	
	out <-read_ncdf(f, curvilinear = c(X = "lon", Y = "lat"), 
					proxy = TRUE)
	
	out <- st_as_stars(out)
	
	st_dim <- st_dimensions(out)
	
	expect_true(all(st_dim$x$values < -74 & st_dim$x$values > -81))
	
	expect_true(all(st_dim$y$values < 38 & st_dim$y$values > 32))
	
	expect_equal(dim(st_dim$y$values), setNames(c(87, 118), c("x", "y")))
	
	# Should also find the curvilinear grid.
	suppressWarnings(
		out <- read_ncdf(f, var = "Total_precipitation_surface_1_Hour_Accumulation",
						 proxy = TRUE))
	
	expect_true(attr(st_dimensions(out), "raster")$curvilinear)
	
})

context("rectilinear matrices")

m <- matrix(1:12, nrow = 4, ncol = 3)

x = c(0, 1, 2, 3)
y = c(2, 1, 0)

r_cell_edge = st_as_stars(list(m = m),
                dimensions = st_dimensions(x = x, y = y,
                                           .raster = c("x", "y"),
                                           cell_midpoints = FALSE))
sf::st_crs(r_cell_edge) <- sf::st_crs(5070)
r_ce_dim <- st_dimensions(r_cell_edge)

r_cell_mid = st_as_stars(list(m = m),
                dimensions = st_dimensions(x = x, y = y,
                                           .raster = c("x", "y"),
                                           cell_midpoints = TRUE))
sf::st_crs(r_cell_mid) <- sf::st_crs(5070)
r_cm_dim <- st_dimensions(r_cell_mid)

x = c(0, 1, 2, 3, 4)
y = c(2, 1, 0, -1)

r_cell_edges = st_as_stars(list(m = m),
                dimensions = st_dimensions(x = x, y = y,
                                           .raster = c("x", "y"),
                                           cell_midpoints = FALSE))
sf::st_crs(r_cell_edges) <- sf::st_crs(5070)
r_ces_dim <- st_dimensions(r_cell_edges)

r_cell_mids = st_as_stars(list(m = m),
                dimensions = st_dimensions(x = x, y = y,
                                           .raster = c("x", "y"),
                                           cell_midpoints = TRUE))
sf::st_crs(r_cell_mids) <- sf::st_crs(5070)
r_cms_dim <- st_dimensions(r_cell_edges)

test_that("same number of values as rows/cols", {

  expect_s3_class(r_cell_edge, "stars")

  expect_equal(names(dim(r_cell_edge)), names(r_ce_dim))

  expect_equal(names(r_ce_dim), c("x", "y"))

  #expect_true(st_crs(r_ce_dim) == sf::st_crs(5070))

  expect_equal(r_ce_dim$x$from, 1)
  expect_equal(r_ce_dim$x$to, 4)
  expect_equal(r_ce_dim$x$offset, 0)
  expect_equal(r_ce_dim$x$delta, 1)

  expect_equal(r_ce_dim$y$from, 1)
  expect_equal(r_ce_dim$y$to, 3)
  expect_equal(r_ce_dim$y$offset, 2)
  expect_equal(r_ce_dim$y$delta, -1)

  expect_s3_class(r_cell_mid, "stars")

  expect_equal(names(dim(r_cell_mid)), names(r_cm_dim))

  expect_equal(names(r_cm_dim), c("x", "y"))

  #expect_true(st_crs(r_cm_dim) == sf::st_crs(5070))

  expect_equal(r_cm_dim$x$from, 1)
  expect_equal(r_cm_dim$x$to, 4)
  expect_equal(r_cm_dim$x$offset, -0.5)
  expect_equal(r_cm_dim$x$delta, 1)

  expect_equal(r_cm_dim$y$from, 1)
  expect_equal(r_cm_dim$y$to, 3)
  expect_equal(r_cm_dim$y$offset, 2.5)
  expect_equal(r_cm_dim$y$delta, -1)
})

test_that("one more cell coordinate than matrix cells", {

  expect_s3_class(r_cell_edges, "stars")

  expect_equal(r_ces_dim, r_ce_dim)

  # Make sure cell_midpoint argument is ignored here.
  expect_equal(r_cms_dim, r_ces_dim)
})

test_that("st_as_sf works for these", {
  skip_on_cran() # too costly?
  sf_cell_edge <- sf::st_as_sf(r_cell_edge)

  expect_s3_class(sf_cell_edge, "sf")
  expect_s3_class(sf_cell_edge$geometry, "sfc_POLYGON")

  sf_cell_mid <- sf::st_as_sf(r_cell_mid)

  expect_s3_class(sf_cell_mid, "sf")
  expect_s3_class(sf_cell_mid$geometry, "sfc_POLYGON")

  sf_cell_edges <- sf::st_as_sf(r_cell_edges)

  expect_equal(sf_cell_edge, sf_cell_edges)

  sf_cell_mids <- sf::st_as_sf(r_cell_mids)

  expect_equal(sf_cell_mid, sf_cell_mids)
})

test_that("st_as_stars.OpenStreetMap works", {
  skip_on_cran()
  skip_if_not_installed("OpenStreetMap")
#  m = OpenStreetMap::openmap(c(52,7.3),c(52.1,7.15), type = "osm")
#  s = st_as_stars(m)
#  expect_s3_class(s, "stars")
})

suppressPackageStartupMessages(library(sf))

library(stars)
if (suppressPackageStartupMessages(require(spatstat.geom, quietly = TRUE))) {
tif = system.file("tif/L7_ETMs.tif", package = "stars")
s = adrop(read_stars(tif)[,,,1]) > 70
plot(s)
m = as.owin(s)
plot(m)
print(table(m$m))
}

#' Landsat-7 bands for a selected region around Olinda, BR
#' 
#' Probably containing the six 30 m bands:
#' \itemize{
#'  \item Band 1 Visible (0.45 - 0.52 µm) 30 m
#'  \item Band 2 Visible (0.52 - 0.60 µm) 30 m
#'  \item Band 3 Visible (0.63 - 0.69 µm) 30 m
#'  \item Band 4 Near-Infrared (0.77 - 0.90 µm) 30 m
#'  \item Band 5 Short-wave Infrared (1.55 - 1.75 µm) 30 m
#'  \item Band 7 Mid-Infrared (2.08 - 2.35 µm) 30 m
#' }
#'
"L7_ETMs"

#' Monthly Gridded Meteorological Observations
#'
#' These are the monthly observational data used for BCSD downscaling. See: <https://gdo-dcp.ucllnl.org/downscaled_cmip_projections/dcpInterface.html#About> for more information." ;
#' "Atmospheric Temperature, Air Temperature Atmosphere, Precipitation, Rain, Maximum Daily Temperature, Minimum  Daily Temperature" ;
#'
"bcsd_obs"

#' Sentinel-2 sample tile
#'
#' Sentinel-2 sample tile, downloaded from <https://scihub.copernicus.eu/>
#' reads the four 10-m bands: B2 (490 nm), B3 (560 nm), B4 (665 nm) and B8 (842 nm)
"stars_sentinel2"

#' Extract cell values at point locations
#'
#' Extract cell values at point locations
#' @name st_extract
#' @export
st_extract = function(x, ...) UseMethod("st_extract")

#' @name st_extract
#' @param x object of class \code{stars} or \code{stars_proxy}
#' @param at object of class \code{sf} or \code{sfc} with geometries, or two-column matrix with coordinate points in rows, indicating where to extract values of \code{x}, or a \code{stars} object with geometry and temporal dimensions (vector data cube)
#' @param bilinear logical; use bilinear interpolation rather than nearest neighbour?
#' @param time_column character or integer; name or index of a column with time or date values that will be matched to values of the first temporal dimension (matching classes \code{POSIXct}, \code{POSIXt}, \code{Date}, or \code{CFTime}), in \code{x}, after which this dimension is reduced. This is useful to extract data cube values along a trajectory; see https://github.com/r-spatial/stars/issues/352 .
#' @param interpolate_time logical; should time be interpolated? if FALSE, time instances are matched using the coinciding or the last preceding time in the data cube.
#' @param FUN function used to aggregate pixel values when geometries of \code{at} intersect with more than one pixel
#' @param resampling character; resampling method; for method cubic or cubicspline,
#' `stars_proxy` objects should be used and GDAL should have version >= 3.10.0
#' @param sfc_attribute character; if \code{at} is of class \code{stars} should the aggregation be performed
#' for the attribute geometry rather than the dimension geometry? If \code{NULL} (default),
#' the aggregation is performed at the dimension geometries, else the name of the attribute geometry to perform the aggregation on.
#' If the given attribute geometry does not exist, the aggregation defaults to the dimension geometry.
#' @param ... passed on to \link{aggregate.stars} when geometries are not exclusively POINT geometries
#' @returns if \code{at} is of class \code{matrix}, a matrix with extracted values is returned; 
#' if \code{at} is of class \code{stars} and a temporal dimension was passed to \code{time_column},
#' a \code{stars} object with the original \code{at} dimensions
#' and the extracted values as attributes.
#' otherwise: if \code{x} has more dimensions than only x and y (raster), an 
#' object of class \code{stars} with POINT geometries replacing x and y raster
#' dimensions, if this is not the case, an object of \code{sf} with extracted values.
#' @details points outside the raster are returned as \code{NA} values. For
#' large sets of points for which extraction is needed, passing a matrix as
#' to \code{at} may be much faster than passing an \code{sf} or \code{sfc} object.
#' @export
#' @examples
#' tif = system.file("tif/L7_ETMs.tif", package = "stars")
#' r = read_stars(tif)
#' pnt = st_sample(st_as_sfc(st_bbox(r)), 10)
#' st_extract(r, pnt)
#' st_extract(r, pnt) |> st_as_sf()
#' st_extract(r[,,,1], pnt)
#' st_extract(r, st_coordinates(pnt)) # "at" is a matrix: return a matrix
#' # Extraction on non-POINT geometries
#' poly = st_buffer(pnt, 1000)
#' st_extract(r, poly)
#' 
#' # Extraction with time matching
#' rdate = c(r, r*2, along = "date")
#' dates = c(Sys.Date()-1, Sys.Date())
#' rdate = st_set_dimensions(rdate, "date", values = c(dates))
#' 
#' pntsf = st_sf(date = dates, geometry = pnt)
#' st_extract(split(rdate, "band"), pntsf) # POINT geometries
#' 
#' polysf = st_buffer(pntsf, 1000)
#' st_extract(split(rdate, "band"), polysf, time_column = "date") # POLYGON geometries
#' 
#' vdc = st_sf(rdm = rnorm(20), polygons = st_buffer(st_sample(st_bbox(pnt), 20), 500),
#' 			geometry = rep(pnt, 2), date = rep(dates, each = 10)) |> 
#' 	st_as_stars(dims = c("geometry", "date"))
#' 
#' (vdc_new = st_extract(split(rdate, "band"), vdc)) # stars vector data cube
#' merge(vdc_new, name = "band")
#' 
#' ### Extraction applied to the geometries inside the vector data cube (cell values)
#' (vdc_new2 = st_extract(split(rdate, "band"), vdc,
#' 					   sfc_attribute = "polygons")) # stars vector data cube
#' merge(vdc_new2, name = "band")
st_extract.stars = function(x, at, ..., bilinear = FALSE, time_column = 
		attr(at, "time_column") %||% attr(at, "time_col"),
		interpolate_time = bilinear, FUN = mean,
		resampling = c("nearest", "bilinear", "cubic", "cubicspline"),
		sfc_attribute = NULL) {

	stopifnot(inherits(at, c("sf", "sfc", "stars", "matrix")))
	resampling = match.arg(resampling)
	if (bilinear) {
		stopifnot(resampling %in% c("nearest", "bilinear"))
		resampling = "bilinear"
	}
	at_orig = at
	use_x_agg = FALSE
	if (inherits(at_orig, "stars")) {
		if (is.null(time_column) && length(which_time(at)) > 0)
			time_column = names(which_time(at))[1] # later on matched by name in sf object
		at = st_as_sf(at, long = TRUE) 
		if (!is.null(sfc_attribute)) {
			stopifnot(is.character(sfc_attribute), length(sfc_attribute) == 1)
			sfc_dim = st_geometry(at)
			sfc_dim_name = attr(at, "sf_column")
			if (!is.null(at[[sfc_attribute]])) 
				st_geometry(at) = sfc_attribute
			else
				stop(paste(sfc_attribute, "should be an array in x"))
		}
	}
	if (inherits(at, "matrix"))
		pts = at
	else {
		stopifnot(st_crs(at) == st_crs(x))
		if (! all(st_dimension(at, FALSE) == 0)) { # should check & branch here in case of MULTIPOINT?
			stopifnot(!bilinear) # bilinear interpolation not supported for time matching with lines/polygons
			# from aggregate.stars_proxy:
			by = st_geometry(at)
			# this assumes the result is small, no need to proxy
			l = lapply(seq_along(by), function(i) aggregate(st_normalize(st_as_stars(x[by[i]])), by[i], FUN, ...))
			if(is.null(time_column)) {
				return(do.call(c, c(l, along = list(which_sfc(l[[1]]))))) # RETURNS!
			} else {
				# to pass to time matching		
				x_agg = do.call(c, c(l, along = list(which_sfc(l[[1]]))))
				use_x_agg = TRUE
			}
		}
		sf_column = attr(at, "sf_column") %||% "geometry"

		tm_pts = if (!is.null(time_column))
					at[[time_column]] # else NULL
		at = st_geometry(at)
		pts = st_coordinates(at)
	}

	if (resampling != "nearest" && !inherits(x, "stars_proxy"))
		x = st_as_stars_proxy(x)

	min_dist = NULL
	if (inherits(x, "mdim"))
		x = st_as_stars(x) # realize
	m = if (inherits(x, "stars_proxy")) {
			try_result = try(x0 <- st_as_stars(x, downsample = dim(x)/2), silent = TRUE)
			lapply(x, function(y) do.call(abind, lapply(get_names(y), 
				gdal_extract, pts, resampling)))
		} else if (use_x_agg) {
			# Allow time matching for lines and polygons using aggregate
			st_normalize(st_upfront(x_agg))
		} else {
			x = st_normalize(st_upfront(x))
			if (is_curvilinear(x)) { # https://github.com/r-spatial/stars/issues/632
				d = st_distance(at, st_as_sfc(x, as_points = TRUE))
				ix = apply(d, 1, which.min) # FIXME: handle points outside the raster
				min_dist = d[cbind(seq_along(at), ix)]
			} else {
				cr = colrow_from_xy(pts, x, NA_outside = TRUE)
				ix = (cr[,2] - 1) * dim(x)[1] + cr[,1]
			}
			lapply(x, function(y) 
				array(y, dim = c(prod(dim(x)[1:2]), prod(dim(x)[-(1:2)])))[ix, , drop = FALSE])
		}
	
	# reset factors & units attributes:
	for (i in seq_along(m)) {
		if (inherits(x[[i]], "factor")) {
			if (is.character(m[[i]]))
				m[[i]] = match(as.vector(m[[i]]), levels(x[[i]]))
			m[[i]] = structure(m[[i]], levels = levels(x[[i]]), 
				colors = attr(x[[i]], "colors"), class = class(x[[i]]))
		} else if (inherits(x[[i]], "units"))
			units(m[[i]]) = units(x[[i]])
		else if (inherits(x, "stars_proxy") && !inherits(try_result, "try-error") && inherits(x0[[i]], "units"))
			units(m[[i]]) = units(x0[[i]])
	}
	# match times:
	if (!is.null(time_column)) {
		refsys_time = c("POSIXct", "POSIXt", "Date", "CFtime")
		## If there are more than two temporal dimensions, the first one is taken
		tm = names(which(sapply(
			st_dimensions(x),
			function(i) any(i$refsys %in% refsys_time))))[1]
		if (is.na(tm))
			stop("cannot match times: x does not have a temporal dimension")
		tm_cube = st_dimensions(x)[[tm]]$values %||% st_get_dimension_values(x, tm)
		tm_ix = match_time(tm_pts, tm_cube,
						   intervals = !st_dimensions(x)[[tm]]$point,
						   interpolate_time)
		# need to handle the case that there is more than time, besides space, in x; 
		# https://github.com/r-spatial/stars/issues/748
		if (has_raster(x) && length(dim(x)) > 3 && dim(m[[1]])[2] > length(tm_cube)) {
			# check time is last dimension:
			if (names(dim(x))[length(dim(x))] != tm)
				stop(paste(tm, 'needs to be the last dimensions'))
			if (interpolate_time)
				stop("to use interpolate_time=TRUE, first drop additional dimensions")
			do = dim(x)[-(c(1:2, length(dim(x))))]
			if (length(do) > 1)
				warning("this may not work for higher than 4-dimensional cubes")
			r = rep(1:dim(m[[1]])[1], each = prod(do))
			c = rep(1:prod(do), dim(m[[1]])[1]) + rep((tm_ix - 1) * prod(do), each = prod(do))
			# print(rbind(r,c))
			# print(m[[1]][cbind(r,c)])
			m = lapply(m, function(a) structure(a[cbind(r,c)], dim = NULL))
			df = setNames(as.data.frame(m), names(x))
			df[[sf_column]] = rep(st_geometry(at), each = do)
			# if (inherits(tm_cube, "intervals"))
			#	tm_cube = as.list(tm_cube)
			# df[[tm]] = tm_cube[tm_ix]
			df[[time_column]] = rep(tm_pts, each = do)
			# depth, or whatever is in do: assuming now there is only one dimension in do
			df[[ names(do)[1] ]] = st_get_dimension_values(x, names(do)[1])
			return(st_as_sf(df))
		} else {
			if (!interpolate_time)
				m = lapply(m, function(p) p[cbind(seq_along(at), tm_ix)])
			else {
				interpolate = function(x, ix) { 
					i = floor(ix)
					di = ix - i
					if (is.na(ix) || di == 0)
						x[i]
					else 
						(1 - di) * x[i] + di * x[i+1]
				}
				# each time series is reduced to a _single_ time step, interpolated:
				m = lapply(m, function(n) mapply(interpolate, asplit(n, 1), tm_ix))
			}
		}
	}
	if (NCOL(m[[1]]) > 1) { # multi-band:
		if (inherits(at, "matrix")) {
			if (all(sapply(m, inherits, "units"))) {
				u = lapply(m, units)
				if (!all(sapply(u[-1], ud_are_convertible, u[[1]]))) {
					warning("incompatible units merged in result matrix: dropping all units")
					m = lapply(m, units::drop_units)
				}
			}
			do.call(cbind, m)
		} else { # return stars:
			for (i in seq_along(x))
				dim(m[[i]]) = c(length(at), dim(x)[-(1:2)])
			d = structure(st_dimensions(x),
				raster = get_raster(dimensions = rep(NA_character_,2)))
			d[[1]] = create_dimension(values = at)
			d[[2]] = NULL
			names(d)[1] = sf_column
			st = setNames(st_as_stars(m, dimensions = d), names(x))
			if (!is.null(min_dist))
				st$min_dist = min_dist
			st
		}
	} else {
		df = setNames(as.data.frame(lapply(m, function(i) structure(i, dim = NULL))), names(x))
		if (inherits(at, "matrix")) 
			df
		else { # return sf:
			df[[sf_column]] = st_geometry(at)
			if (!is.null(time_column)) { # add time columns of both cube and at:
				if (inherits(tm_cube, "intervals"))
					tm_cube = as.list(tm_cube)
				df[[tm]] = tm_cube[tm_ix]
				df[[time_column]] = tm_pts
			}
			sf = st_as_sf(df)
			if (!is.null(min_dist))
				sf$min_dist = min_dist
			if (!inherits(at_orig, "stars")) 
				sf
			else {
				if (!is.null(sfc_attribute)) { # add dimension geometry back
					st_geometry(sf) = sfc_dim
					st_geometry(sf) = sfc_dim_name
				}
				st_as_stars(sf, dims = attr(attr(at_orig, "dimensions"), "names"))
			}
		}
	}
}

# match the times in a to those of b, or interpolate:
# if interpolate = FALSE, returns an integer in 1...length(b) or NA if outside
# if interpolate = TRUE, returns a continuous index in 1...length(b) or NA if outside
match_time = function(a, b, intervals = FALSE, interpolate = FALSE) {
	if (inherits(b, "CFTime")) {
		if (interpolate && isFALSE(intervals) && is.null(b$bounds))
			return(b$indexOf(a, method = "linear"))
		else
			return(b$indexOf(a))
	}
	
	if (inherits(a, "POSIXct") && inherits(b, "Date"))
		a = as.Date(a)
	if (inherits(b, "POSIXct") && inherits(a, "Date"))
		b = as.Date(b)
	m = if (inherits(b, "intervals"))
			find_interval(a, b)
		else if (isTRUE(intervals) || interpolate) {
			m = findInterval(a, b, rightmost.closed = TRUE)
			m[ m == 0 | m == length(b) ] = NA
			m
		} else
			match(a, b)
	if (interpolate && !isTRUE(intervals) && !inherits(b, "intervals")) {
		b = as.numeric(b)
		a = as.numeric(a)
		b = c(b, tail(b, 1))
		m + (a - b[m])/(diff(b)[m])
	} else
		m
}

#' create an intervals object
#' 
#' create an intervals object, assuming left-closed and right-open intervals
#' @param start vector with start values, or 2-column matrix with start and end values in column 1 and 2, respectively
#' @param end vector with end values
#' @export
make_intervals = function(start, end) {
	if (missing(end) && is.matrix(start) && ncol(start) == 2) {
		end = start[,2]
		start = start[,1]
	}
	stopifnot(length(start) > 0, length(start) == length(end))
	structure(list(start = start, end = end), class = "intervals")
}

as_intervals = function(x, add_last = FALSE) {
	stopifnot(is.atomic(x))
	if (add_last)
		x = c(x, tail(x, 1) + diff(tail(x, 2)))
	make_intervals(start = head(x, -1), end = tail(x, -1))
}

#' @export
length.intervals = function(x) length(x$start)

#' @export
head.intervals = function(x, n, ...) make_intervals(head(x$start, n, ...), head(x$end, n, ...))

#' @export
tail.intervals = function(x, n, ...) make_intervals(tail(x$start, n, ...), tail(x$end, n, ...))

#' @export
c.intervals = function(...) {
	dots = list(...)
	start = do.call(c, lapply(dots, function(x) x$start))
	end = do.call(c, lapply(dots, function(x) x$end))
	make_intervals(start, end)
}

#' @export
`[.intervals` = function(x, i, ...) {
	make_intervals(x$start[i], x$end[i])
}

#' @export
format.intervals = function(x, ...) {
	mformat = function(x, ..., digits = getOption("digits")) {
		if (inherits(x, "CFTime")) {
			rng = x$range()
			paste0("[", rng[1], ",", rng[2], ")")
		} else
			format(x, digits = digits, ...) 
	}
	if (inherits(x$start, "units") && inherits(x$end, "units")) {
		stopifnot(units(x$start) == units(x$end))
		paste0("[", format(as.numeric(x$start), ...), ",", format(as.numeric(x$end), ...), ") ",
			"[", as.character(units(x$start)), "]")
	} else
		paste0("[", mformat(x$start, ...), ",", mformat(x$end, ...), ")")
}

find_interval = function(x, intervals) {
	if (inherits(intervals$start, "Date") && inherits(x, "POSIXct"))
		x = as.Date(x)
	if (inherits(x, "Date") && inherits(intervals$start, "POSIXct"))
		x = as.POSIXct(x)
	if (all(intervals$start > intervals$end)) { # decreasing intervals
		start = intervals$end
		intervals$end = intervals$start
		intervals$start = start
	}
	if (getRversion() < "4.1.0")
		stop("R >= 4.1.0 required for handling intervals")
	w = apply(outer(x, intervals$start, ">=") & outer(x, intervals$end, "<"), 1, which, simplify = FALSE)
	w[lengths(w) == 0] = NA_integer_
	unlist(w)
}

#' @export
as.list.intervals = function(x, ...) {
	structure(mapply(make_intervals, x$start, x$end, SIMPLIFY = FALSE),
		class = "intervals_list")
}

#' @export
format.intervals_list = function(x, ...) {
	sapply(x, format, ...)
}

#' @export
`[.intervals_list` = function(x, i, ...) {
	ret = NextMethod()
	is_null = sapply(ret, is.null)
	ret[is_null] = list(make_intervals(NA_real_, NA_real_))
	structure(ret, class = "intervals_list")
}

suppressPackageStartupMessages(library(stars))
tif = system.file("tif/L7_ETMs.tif", package = "stars")
x = read_stars(tif)
y = st_set_dimensions(x, "band", point = TRUE)
y[,1:10, 1:10, c(1,4,3,2)]
y = st_set_dimensions(x, "band", point = NA)
y[,1:10, 1:10, c(1,4,3,2)]
y = st_set_dimensions(x, "band", point = FALSE)
y[,1:10, 1:10, c(1,4,3,2)]
y = st_set_dimensions(x, "band", values = letters[1:6])
y[,1:10, 1:10, c(1,4,3,2)]
y = st_set_dimensions(x, "band", values = letters[1:6], point = NA)
y[,1:10, 1:10, c(1,4,3,2)]
y = st_set_dimensions(x, "band", values = letters[1:6], point = TRUE)
y[,1:10, 1:10, c(1,4,3,2)]

# [<- on stars_proxy:
tif = system.file("tif/L7_ETMs.tif", package = "stars")
x = read_stars(tif, proxy = TRUE)
sel = x > 100
x[sel] = 100
st_as_stars(x)

suppressPackageStartupMessages(library(stars))
if (suppressPackageStartupMessages(require(sp, quietly = TRUE)) && require(spacetime, quietly = TRUE)) {
set.seed(1331)
Sys.setenv(TZ="UTC")
example(STFDF)

x = st_as_stars(stfdf)
xx = as(x, "STFDF")

# as.xts.stars:
library(xts)
xts = as.xts(x)
st_as_stars(xts)
st_as_stars(xts, dimensions = st_dimensions(x))

library(sp)
pts = SpatialPoints(rbind(c(0,0), c(0,1), c(1,0), c(1,1)))
gridded(pts) = TRUE

tm = xts(1:10, as.Date("2010-01-01")+0:9)
stfdf = STFDF(pts, tm, data.frame(foo = seq_len(length(pts) * length(tm))))
x = st_as_stars(stfdf)
xx = as(x, "STFDF")

x = st_as_stars(as(stfdf, "STSDF"))
}

#' @name prcomp
#' @param downsample see \link{st_as_stars}
#' @export
prcomp.stars_proxy = function(x, ..., downsample = 0) {
	prcomp(st_as_stars(x, downsample = downsample), ...)
}

#' Principal components of stars object
#'
#' Compute principal components of stars object
#' @param x object of class `stars` or `stars_proxy`
#' @param quiet logical; if `TRUE`, suppress message that PCs will be computed on last dimension; see details
#' @param ... see \link[stats]{prcomp}
#' @details if `x` has only one attribute, principal components will be computed in the space of the last dimension of `x`
#' to predict PC scores into a `stars` object, use \link{predict.stars}; see example below
#' @name prcomp
#' @returns object of class `prcomp`, see \link[stats]{prcomp}
#' @export
#' @examples
#' l7 = split(st_as_stars(L7_ETMs), 3) # use bands as features
#' plot(prcomp(l7))
#' plot(merge(predict(l7, model = prcomp(l7))))
prcomp.stars = function(x, ..., quiet = FALSE) {
	if (length(x) == 1) {
		if (!quiet)
			message(paste0("computing PCs over dimension `", tail(names(dim(x)), 1), "'"))
		x = split(x, length(dim(x)))
	}
	prcomp(as.data.frame(x), ...)
}

# close rings by repeating the first point
close_mat = function(m) {
	if (any(m[1,] != m[nrow(m),]))
		m = rbind(m, m[1,])
	if (nrow(m) < 4)
		stop("polygons require at least 4 points")
	unclass(m)
}

regrp = function(grp, i) {
	# work on single MULTIPOLYGON, group outer rings or outer rings+holes
	if (all(i == 0)) # only outer rings
		lapply(grp, function(x) list(close_mat(x))) # add one list level
	else {
		for (j in seq_along(grp))
			grp[[j]] = close_mat(grp[[j]])
		starts = which(i == 0)
		lst = vector("list", length(starts))
		for (j in seq_along(lst)) {
			end = start = starts[j]
			while (end < length(i) && i[end + 1] == 1)
				end = end + 1 # consume holes from start to end
			lst[j] = list(grp[start:end])
		}
		lst
	}
}

regroup = function(grps, ir) {
	mapply(regrp, grps, ir, SIMPLIFY = FALSE)
}

recreate_geometry = function(l) {
	if (length(geom <- l$geometry)) {
		cc = cbind(geom$x[[1]], geom$y[[1]])
		if (geom$geometry_type == "point") {
			sfc = do.call(st_sfc, lapply(seq_len(nrow(cc)), function(i) st_point(cc[i,])))
			dimension = attr(geom$x[[1]], "d_names")
		} else { 
			nc = geom$node_count[[1]]
			pnc = c(0, cumsum(geom$part_node_count[[1]]))
			parts = vector("list", length(geom$part_node_count[[1]]))
			for (i in seq_along(parts))
				parts[[i]] = cc[(pnc[i]+1):pnc[i+1],]
			i = findInterval(pnc[-1], cumsum(c(0, nc)), left.open = TRUE)
			grps = lapply(split(seq_along(parts), i), function(x) parts[x])
			dimension = attr(nc, "d_names")
			if (geom$geometry_type == "line")
				sfc = do.call(st_sfc, lapply(grps, st_multilinestring))
			else if (geom$geometry_type == "polygon") {
				ir = geom$interior_ring[[1]]
				grps = regroup(grps, split(ir, i))
				sfc = do.call(st_sfc, lapply(grps, st_multipolygon))
					# if all rings are exterior rings: add one list level to each
					# do.call(st_sfc, lapply(lapply(grps, function(x) lapply(x, list)), st_multipolygon))
			} else
				stop(paste("unsupported geometry type:", geom$geometry_type))
		}
		l$dimensions[[dimension]] = sfc
	}
	l
}

get_values_from_bounds = function(x, bnd, center) {
	a = gdal_read_mdim(x, bnd)$array_list[[1]]
	dim(a) = rev(dim(a))
	a = t(a)
	if (length(dim(a)) == 2) {
		if (!dim(a)[2] %in% 1:2)
			warning(paste("bounds variable", bnd, "has", dim(a)[2], "vertices and may be treated incorrectly"))
		if (center) {
			m = apply(a, 1, mean)
			if (regular_intervals(m))
				m
			else
				make_intervals(a[,1], a[,2])
		} else # start of bound:
			a[,1]
	} else
		a
}

mdim_use_bounds = function(dims, x, bnds, center = TRUE) {
	if (isTRUE(bnds)) {
		bnds = character()
		for (d in names(dims))
			if (d != "time" && !is.null(b <- attr(dims[[d]]$values[[1]], "attributes")["bounds"]))
				bnds = c(bnds, setNames(b, d))
	}
	if (length(bnds) && is.null(names(bnds)))
		stop("bounds must be a named vector, names indicating the non-bounds dimension variables")
	for (b in names(bnds))
		if (!is.na(bnds[b])) {
			v <- try(get_values_from_bounds(x, bnds[b], center = TRUE), silent = TRUE)
			if (!inherits(v, "try-error"))
				dims[[b]]$values[[1]] = v
		}
	dims
}

match_raster_dims = function(nms) {
	x = tolower(nms) %in% c("lon", "long", "longitude")
	y = tolower(nms) %in% c("lat", "latitude")
	if (sum(xor(x, y)) == 2)
		c(which(x), which(y))
	else
		1:2
}

#' Read or write data using GDAL's multidimensional array API
#'
#' Read or write data using GDAL's multidimensional array API
#' @name mdim
#' @param filename name of the source or destination file or data source
#' @param variable name of the array to be read; if `"?"`, a list of array names is returned, with group name as list element names.
#' @param options character; driver specific options regarding the opening (read_mdim) or creation (write_mdim) of the dataset
#' @param raster names of the raster variables (default: first two dimensions)
#' @param offset integer; zero-based offset for each dimension (pixels) of sub-array to read, defaults to 0 for each dimension(requires sf >= 1.0-9)
#' @param count integer; size for each dimension (pixels) of sub-array to read (default: read all); a value of NA will read the corresponding dimension entirely; counts are relative to the step size (requires sf >= 1.0-9)
#' @param step integer; step size for each dimension (pixels) of sub-array to read; defaults to 1 for each dimension (requires sf >= 1.0-9)
#' @param proxy logical; return proxy object?
#' @param debug logical; print debug info?
#' @param bounds logical or character: if \code{TRUE} tries to infer from "bounds" attribute; if character, 
#' named vector of the form \code{c(longitude="lon_bnds", latitude="lat_bnds")} with names dimension names
#' @param curvilinear control reading curvilinear (geolocation) coordinate arrays; if \code{NA} try reading the x/y dimension names; if character, defines the arrays to read; if \code{FALSE} do not try; see also \link{read_stars}
#' @param normalize_path logical; if \code{FALSE}, suppress a call to \link{normalizePath} on \code{filename}
#' @details it is assumed that the first two dimensions are easting and northing
#' @param ... ignored
#' @seealso \link[sf]{gdal_utils}, in particular util \code{mdiminfo} to query properties of a file or data source containing arrays
#' @export
read_mdim = function(filename, variable = character(0), ..., options = character(0), 
					 raster = NULL, offset = integer(0), count = integer(0), step = integer(0), proxy = FALSE, 
					 debug = FALSE, bounds = TRUE, curvilinear = NA, normalize_path = TRUE) {

	stopifnot(is.character(filename), is.character(variable), is.character(options))
	cftime_installed = requireNamespace("CFtime", quietly = TRUE) 
	if (! cftime_installed)
		message("for reading time stamps, package CFtime is required: please install it first") # nocov

	if (normalize_path)
		filename = enc2utf8char(maybe_normalizePath(filename, np = normalize_path))
	ret = gdal_read_mdim(filename, variable, options, rev(offset), rev(count), rev(step), proxy, debug)

	if (identical(variable, "?"))
		return(ret) # RETURNS

	if (length(ret$dimensions) == 1 && length(ret$array_list) == 1 && is.data.frame(ret$array_list[[1]]))
		return(ret$array_list[[1]]) ## composite data: RETURNS

	ret = recreate_geometry(ret)
	if (isTRUE(bounds) || is.character(bounds))
		ret$dimensions = mdim_use_bounds(ret$dimensions, filename, bounds)

	create_units = function(x) {
		u = attr(x, "units")
		x = structure(x, units = NULL) # remove attribute
		if (is.null(u) || u %in% c("", "none"))
			x
		else {
			cal = if (!is.null(a <- attr(x, "attributes"))) a["calendar"] else NULL
			time = if (cftime_installed)
					time = try(CFtime::CFtime(u, cal), silent = TRUE)     # cheaply try if we can make CFTime
				else
					FALSE
			if (inherits(time, "CFTime"))
				time + as.numeric(x)                       # if we have CFTime, add the offsets
			else {
				u = try_as_units(u)
				if (inherits(u, "units"))
					units(x) = u
				x
			}
		}
	}
	l = rev(lapply(ret$dimensions, function(x) {
			   if (inherits(x, "sfc")) x else create_units(x$values[[1]])
			}))
	if (length(offset) != 0 || length(step) != 0 || length(count) != 0) {
		if (length(offset) == 0)
			offset = rep(0, length(l))
		if (length(step) == 0)
			step = rep(1, length(l))
		ll = lengths(lapply(l, as.list)) # take care of dimensions of class intervals
		if (length(count) == 0)
			count = floor((ll - offset)/step)
		else if (any(a <- is.na(count)))
			count[a] = floor((ll[a] - offset[a])/step[a])
		for (i in seq_along(l)) {
			l[[i]] = l[[i]][seq(from = offset[i]+1, length.out = count[i], by = step[i])]
		}
	}

	# create dimensions table:
	sf = any(sapply(l, function(x) inherits(x, "sfc")))
	# FIXME: i %in% 1:2 always the case?
	raster_dims = match_raster_dims(names(l))
	d = mapply(function(x, i) create_dimension(values = x, is_raster = !sf && i %in% raster_dims, point = NA),
			   l, seq_along(l), SIMPLIFY = FALSE)
	if (is.null(raster)) {
		raster = if (sf)
					get_raster(dimensions = rep(NA_character_,2))
				else
					get_raster(dimensions = names(d)[raster_dims])
	} else
		raster = get_raster(dimensions = raster)
	dimensions = create_dimensions(d, raster = raster)

	# handle array units:
	for (i in seq_along(ret$array_list))
		if (nchar(u <- attr(ret$array_list[[i]], "units")) && inherits(u <- try_as_units(u), "units"))
			units(ret$array_list[[i]]) = u
	clean_units = function(x) { 
		if (identical(attr(x, "units"), "")) 
			structure(x, units = NULL)
		else
			x
	}

	# create return object:
	if (proxy) {
		lst = lapply(ret$array_list, function(x) filename)
		st = st_stars_proxy(lst, dimensions, NA_value = NA_real_, resolutions = NULL, class = "mdim")
		if (!is.null(ret$srs))
			st_set_crs(st, ret$srs)
		else
			st
	} else { 
		lst = lapply(ret$array_list, function(x) structure(clean_units(x), dim = rev(dim(x))))
		st = st_stars(lst, dimensions)
		if (is.null(ret$srs) || is.na(ret$srs)) {
			if (missing(curvilinear) || is.character(curvilinear)) { # try curvilinear:
				xy = raster$dimensions
				ll = curvilinear
				if (is.character(curvilinear)) {
					if (is.null(names(curvilinear)))
						names(curvilinear) = xy[1:2]
					ret = try(st_as_stars(st, curvilinear = curvilinear), silent = TRUE)
					if (inherits(ret, "stars"))
						st = ret
				}
				if (!is_curvilinear(st) &&
					inherits(x <- try(read_mdim(filename, ll[1], curvilinear = FALSE), silent = TRUE), "stars") &&
					inherits(y <- try(read_mdim(filename, ll[2], curvilinear = FALSE), silent = TRUE), "stars") &&
						identical(dim(x)[xy], dim(st)[xy]) && identical(dim(y)[xy], dim(st)[xy]))
					st = st_as_stars(st, curvilinear = setNames(list(x, y), xy))
			}
			st
		} else
			st_set_crs(st, ret$srs)
	}
}

# WRITE helper functions:

add_attr = function(x, at) { # append at to attribute "attrs"
		structure(x, attrs = c(attr(x, "attrs"), at))
}

add_units_attr = function(l) {
	f = function(x) {
		if (inherits(x, "units"))
			add_attr(x, c(units = as.character(units(x))))
		else if (inherits(x, "POSIXct")) {
			cal = if (!is.null(cal <- attr(x, "cal")))
				c(calendar = paste0(cal, "_day")) # else NULL, intended
			x = as.numeric(x)
			if (all(x %% 86400 == 0))
				add_attr(x/86400, c(units = "days since 1970-01-01", cal))
			else if (all(x %% 3600 == 0))
				add_attr(x / 3600, c(units = "hours since 1970-01-01 00:00:00", cal))
			else
				add_attr(x, c(units = "seconds since 1970-01-01 00:00:00", cal))
		} else if (inherits(x, "CFTime")) {
			add_attr(CFtime::offsets(x), c(units = CFtime::definition(x), CFtime::calendar(x)))
		} else if (inherits(x, "Date"))
			add_attr(as.numeric(x), c(units = "days since 1970-01-01"))
		else
			x
	}
	lapply(l, f)
}

cdl_add_geometry = function(e, i, sfc) {
	stopifnot(inherits(sfc, c("sfc_POINT", "sfc_POLYGON", "sfc_MULTIPOLYGON", "sfc_LINESTRING", "sfc_MULTILINESTRING")))

	if (inherits(sfc, "sfc_POLYGON"))
		sfc = st_cast(sfc, "MULTIPOLYGON")
	if (inherits(sfc, "sfc_LINESTRING"))
		sfc = st_cast(sfc, "MULTILINESTRING")

	cc = st_coordinates(sfc)
	if (inherits(sfc, c("sfc_MULTIPOLYGON"))) {
		e$node = structure(seq_len(nrow(cc)), dim = c(node = nrow(cc)))
		e$x = structure(cc[, 1], dim = c(node = nrow(cc)))
		e$y = structure(cc[, 2], dim = c(node = nrow(cc)))
		e$node_count = structure(rle(cc[,"L3"])$lengths, dim = setNames(length(sfc), names(e)[i]))
		part = rle(cc[,"L3"] * 2 * max(cc[,"L2"]) + 2 * (cc[,"L2"] - 1) + cc[,"L1"])$lengths
		e$part_node_count = structure(part, dim = c(part = length(part)))
		e$interior_ring = structure(as.numeric(cc[cumsum(part), "L1"] > 1), dim = c(part = length(part)))
		attr(e, "dims") = c(node = nrow(cc), part = length(part))
		e$geometry = add_attr(structure(numeric(0), dim = c("somethingNonEx%isting" = 0)),
			c(geometry_type = "polygon", node_count = "node_count", node_coordinates = "x y",
		  	part_node_count = "part_node_count", interior_ring = "interior_ring",
		  	grid_mapping = if (!is.na(st_crs(sfc))) "crs" else NULL))
	} else if (inherits(sfc, "sfc_MULTILINESTRING")) { # LINE:
		e$node = structure(seq_len(nrow(cc)), dim = c(node = nrow(cc)))
		e$x = structure(cc[, 1], dim = c(node = nrow(cc)))
		e$y = structure(cc[, 2], dim = c(node = nrow(cc)))
		e$node_count = structure(rle(cc[,"L2"])$lengths, dim = setNames(length(sfc), names(e)[i]))
		part = rle((cc[,"L2"] - 1) * max(cc[,"L1"]) + cc[,"L1"])$lengths
		e$part_node_count = structure(part, dim = c(part = length(part)))
		attr(e, "dims") = c(node = nrow(cc), part = length(part))
		e$geometry = add_attr(structure(numeric(0), dim = c("somethingNonEx%isting" = 0)),
			c(geometry_type = "line", node_count = "node_count", node_coordinates = "x y",
		  	part_node_count = "part_node_count",
		  	grid_mapping = if (!is.na(st_crs(sfc))) "crs" else NULL))
	} else { # POINT:
		if (ll <- isTRUE(st_is_longlat(sfc))) {
			e$lon = add_attr(structure(cc[, 1], dim = setNames(nrow(cc), names(e)[i])), 
						 	c(units = "degrees_north", standard_name = "longitude"))
			e$lat = add_attr(structure(cc[, 2], dim = setNames(nrow(cc), names(e)[i])), 
						 	c(units = "degrees_east", standard_name = "latitude"))
		} else {
			e$x = structure(cc[, 1], dim = setNames(nrow(cc), names(e)[i]))
			e$y = structure(cc[, 2], dim = setNames(nrow(cc), names(e)[i]))
		}
		e$geometry = add_attr(structure(numeric(0), dim = c("somethingNonEx%isting" = 0)),
				c(geometry_type = "point", node_coordinates = ifelse(ll, "lon lat", "x y"),
				grid_mapping = if (!is.na(st_crs(sfc))) "crs" else NULL))
	}
	e
}

# convert stars object into a list of CDL-like variables with named dimensions
st_as_cdl = function(x) {
	e = add_units_attr(expand_dimensions(x)) # TODO: handle sfc dimensions
	d = st_dimensions(x)
	dimx = dim(x)
	xy = attr(st_dimensions(x), "raster")$dimensions
	co = NULL
	for (i in seq_along(e)) {
		if (is.null(dim(e[[i]])) && !is.list(e[[i]]))
			e[[i]] = structure(e[[i]], dim = setNames(length(e[[i]]), names(e)[i]))
		else if (is_curvilinear(x)) { # curvilinear coordinate matrices:
			if (names(e)[i] == xy[1]) names(e)[i] = "lon"
			if (names(e)[i] == xy[2]) names(e)[i] = "lat"
		} else if (inherits(e[[i]], "sfc")) { # vector data cube:
			sfc = e[[i]]
			e[[i]] = structure(seq_along(sfc), dim = setNames(length(sfc), names(e)[i]))
			e = cdl_add_geometry(e, i, sfc)
			if (inherits(sfc, "sfc_POINT")) {
				if (isTRUE(st_is_longlat(x))) {
					co = c(coordinates = "lat lon")
					xy = c("lon", "lat") # hack
				} else {
					co = c(coordinates = "x y")
					xy = c("x", "y") # hack
				}
			} else {
				dimx = c(dimx, attr(e, "dims"))
				co = c(coordinates = "x y")
				xy = c("x", "y") # hack
			} 
			co = c(co, geometry = "geometry")
		}
	}

	if (is_curvilinear(x)) {
		e[["lat"]] = add_attr(e[["lat"]], c(units = "degrees_north", "_CoordinateAxisType" = "Lat", axis = "Y"))
		e[["lon"]] = add_attr(e[["lon"]], c(units = "degrees_east",  "_CoordinateAxisType" = "Lon", axis = "X"))
		co = paste(rev(setdiff(names(d), c("x", "y", "lat", "lon"))), "lat lon")
	} else {
		if (!any(is.na(xy))) {
			e[[ xy[1] ]] = add_attr(e[[ xy[1] ]], c(axis = "X"))
			e[[ xy[2] ]] = add_attr(e[[ xy[2] ]], c(axis = "Y"))
		}
	}
	for (i in seq_along(x))
		x[[i]] = add_attr(x[[i]], co)

	x = add_units_attr(x) # unclasses x
	for (i in seq_along(x))
		if (is.null(names(dim(x[[i]])))) # FIXME: read_ncdf() doesn't name array dim
			names(dim(x[[i]])) = names(d)

	for (i in names(e)) # copy over dimension values
		x[[i]] = e[[i]]

	which_dims = function(a, dimx) {
		m = match(names(dim(a)), names(dimx), nomatch = numeric(0)) - 1
		if (all(is.na(m)))
			integer(0)
		else
			m
	}
	ret = lapply(x, function(a) structure(a, which_dims = which_dims(a, dimx)))
	structure(ret, which_crs = !(names(ret) %in% names(e)), 
			  is_numeric = sapply(ret, is.numeric),
			  dims = attr(e, "dims"))
}


#' @name mdim
#' @param x stars object 
#' @param driver character; driver name
#' @param root_group_options character; driver specific options regarding the creation of the root group
#' @param as_float logical; if \code{TRUE} write 4-byte floating point numbers, if \code{FALSE} write 8-byte doubles
#' @param ... ignored
#' @export
#' @examples
#' set.seed(135)
#' m = matrix(runif(10), 2, 5)
#' names(dim(m)) = c("stations", "time")
#' times = as.Date("2022-05-01") + 1:5
#' pts = st_as_sfc(c("POINT(0 1)", "POINT(3 5)"))
#' s = st_as_stars(list(Precipitation = m)) |>
#'  st_set_dimensions(1, values = pts) |>
#'  st_set_dimensions(2, values = times)
#' nc = tempfile(fileext=".nc")
#' if (compareVersion(sf_extSoftVersion()["GDAL"], "3.4.0") > -1) {
#'   write_mdim(s, nc)
#'   # try ncdump on the generated file
#'   print(read_mdim(nc))
#' }
write_mdim = function(x, filename, driver = detect.driver(filename), ..., 
					  root_group_options = character(0), options = character(0),
					  as_float = TRUE, normalize_path = TRUE) {

	if (normalize_path)
		filename = enc2utf8(maybe_normalizePath(filename, TRUE))
	if (inherits(x, "stars_proxy"))
		x = st_as_stars(x)
	cdl = st_as_cdl(x)
	wkt = if (is.na(st_crs(x)))
			character(0)
		else
			st_crs(x)$wkt
	xy = attr(st_dimensions(x), "raster")$dimensions
	gdal_write_mdim(filename, driver, c(dim(x), attr(cdl, "dims")), cdl, wkt, xy, 
					root_group_options = root_group_options, options = options,
					as_float = as_float)
	invisible(x)
}

#' @export
st_as_stars.mdim = function(.x, ..., downsample = 0, debug = FALSE,
		envir = parent.frame()) {
	d = st_dimensions(.x)
	if (length(downsample) == 1) { # implies only spatial coordinates
		ds = dim(.x) * 0
		ds[1:2] = downsample
		downsample = ds
	}
	step = rep_len(round(downsample), length(dim(.x))) + 1
	offset = sapply(d, function(x) x$from) - 1
	count = (sapply(d, function(x) x$to) - offset) %/% step
	if (debug)
		print(data.frame(offset = offset, step = step, count = count))
	# read:
	l = vector("list", length(.x))
	for (i in seq_along(l))
		l[[i]] = read_mdim(.x[[i]], names(.x)[i], offset = offset, count = count, step = step, proxy = FALSE, ...)
	ret = do.call(c, l)
	if (!is.na(st_crs(.x)))
		st_crs(ret) = st_crs(.x)
	# post-process:
	cl = attr(.x, "call_list")
	if (!all(downsample == 0))
		lapply(cl, check_xy_warn, dimensions = st_dimensions(.x))
	process_call_list(ret, cl, envir = envir, downsample = downsample)
}

#' @export
`[.mdim` = function(x, i, j, ...) {
	structure(NextMethod(), class = class(x))
}

#' @export
#' @name st_crop
st_crop.mdim = function(x, y, ...) {
	structure(NextMethod(), class = class(x))
}

Sys.setenv(TZ="UTC")
suppressPackageStartupMessages(library(sf))
suppressPackageStartupMessages(library(stars))
tif = system.file("tif/L7_ETMs.tif", package = "stars")
(r = read_stars(tif, proxy = TRUE))
if (interactive()) {
	plot(r)
}
dim(r)
r[,,,2:3]
r+r
st_as_stars(sin(r), downsample = 2)
aperm(r, c(3,2,1))
if (interactive()) {
	plot(r)
}
(xx = c(r,r))
st_redimension(xx)
st_as_stars(r)
(y = st_apply(r, 1:2, max))
(yy = st_as_stars(y, downsample = 1))
(y = adrop(st_apply(r, c("x", "y"), max)))
(yy = st_as_stars(y, downsample = 1))
plot(yy)
(y = adrop(st_apply(r, c("x", "band"), max)))
(yy = st_as_stars(y, downsample = 1))
rm(yy)

(xx = c(r,r))
names(xx) = c("a", "b")
xx["a"]
pt = st_point(c(x = 290462.103109179, y = 9114202.32594085))
buf = st_buffer(st_sfc(pt, crs = st_crs(r)), 1500)
buf = st_sfc(st_polygon(list(st_buffer(pt, 1500)[[1]], st_buffer(pt, 1000)[[1]])),
  crs = st_crs(r))
r = r[buf] # crops
r = r[buf, epsilon = 1e-5] # crops with shrinked bounding box

# c:
f = system.file("netcdf/avhrr-only-v2.19810902.nc", package = "starsdata")
if (FALSE && f != "") {
  files = c("avhrr-only-v2.19810901.nc",
  "avhrr-only-v2.19810902.nc",
  "avhrr-only-v2.19810903.nc",
  "avhrr-only-v2.19810904.nc",
  "avhrr-only-v2.19810905.nc",
  "avhrr-only-v2.19810906.nc",
  "avhrr-only-v2.19810907.nc",
  "avhrr-only-v2.19810908.nc",
  "avhrr-only-v2.19810909.nc")
  l = list()
  for (f in files) {
	from = system.file(paste0("netcdf/", f), package = "starsdata")
  	l[[f]] = read_stars(from, sub = c("sst", "anom"), proxy = TRUE)
  }
  ret = do.call(c, l)
  print(ret)
  all = system.file(paste0("netcdf/", files), package = "starsdata")
  ret = read_stars(all, sub = c("sst", "anom"))
  print(ret)
  print(st_redimension(ret)) # collapse the two attributes into new dimension

  try(ret <- c(l[[1]], l[[2]], l[[3]], along = list(times = as.Date("1981-09-01") + 0:2)))
  #print(ret)
  #ret = adrop(adrop(c(l[[1]], l[[2]], l[[3]], along = "times")))
  #print(ret)
  ret <- st_redimension(l[[1]], along = list(times = as.Date("1981-09-01") + 0:1))
  print(ret)
}

# demonstrate that environments work:
x = read_stars(tif, proxy = TRUE)
f = function(x) {
	g = function(xx) {
		ndvi = function(z1, z2, z3, z4, z5, z6) (z2-z1)/(z2+z1)
		st_apply(xx, 1:2, ndvi)
	}
	g(x)
}
(x = f(x))
st_as_stars(x)

suppressPackageStartupMessages(library(stars))
tif = system.file("tif/L7_ETMs.tif", package = "stars")
x = read_stars(tif)
(a = st_area(x[,1:5,1:5]))

m = matrix(1:20, nrow = 5, ncol = 4)
x = c(0, 0.5, 1, 2, 4, 5)  # 6 numbers: boundaries!
y = c(0.3, 0.5, 1, 2, 2.2) # 5 numbers: boundaries!
(r = st_as_stars(list(m = m), dimensions = st_dimensions(x = x, y = y)))
# image(r, axes = TRUE, col = grey((1:20)/20))
(a = st_area(r))
#plot(a, axes = TRUE)

s5p = system.file("sentinel5p/S5P_NRTI_L2__NO2____20180717T120113_20180717T120613_03932_01_010002_20180717T125231.nc", package = "starsdata")
if (s5p != "") {
 nit.c = read_stars(s5p, sub = "//PRODUCT/SUPPORT_DATA/DETAILED_RESULTS/nitrogendioxide_summed_total_column",
	curvilinear = c("//PRODUCT/longitude", "//PRODUCT/latitude"), driver = NULL)
 if (inherits(nit.c[[1]], "units"))
	nit.c[[1]] = units::drop_units(nit.c[[1]])
 nit.c[[1]][nit.c[[1]] > 9e+36] = NA
 st_crs(nit.c) = 4326
 print((a <- st_area(nit.c[,1:5,1:5])))
 # plot(a, axes = TRUE, border = NA)
}

suppressPackageStartupMessages(library(stars))

if (require(dplyr, quietly = TRUE)) {
tif = system.file("tif/L7_ETMs.tif", package = "stars")
x <- read_stars(tif)
(y <- x |> select(L7_ETMs.tif))
(y <- x |> filter(band > 2))
(z <- x |> mutate(foo = 2* L7_ETMs.tif))
x |> pull(L7_ETMs.tif) |> class()
x |> filter(x > 289900)
x |> slice("x", 50:100)

suppressPackageStartupMessages(library(ggplot2))
ggplot() + geom_stars(data = x) + coord_equal() + facet_wrap(~band)
ggplot() + geom_stars(data = x[,c(1:50,52:100),c(1:45, 50:60)]) +
	coord_equal()
ggplot() + geom_stars(data = x[,1:10,1:10,1:2], sf = TRUE) + facet_wrap(~band)


# see https://github.com/r-spatial/stars/issues/539 :
x <- read_stars(system.file("tif/L7_ETMs.tif", package = "stars"))
y <- transmute(x, L7_ETMs.tif = units::set_units(L7_ETMs.tif, degree_C)) # same as x but with units

x - slice(x, 'band', 1) # runs as expected, recycling the second object
try(y - slice(x, 'band', 1)) # fails as expected because x is unitless
try(x - slice(y, 'band', 1)) # fails as expected 
y - slice(y, 'band', 1) # runs as expected
}

check_spatstat <- function(pkg, X = NULL){
  if(!requireNamespace(pkg, quietly = TRUE)){
    stop("package ", pkg, " required, please install it (or the full spatstat package) first")
  } else{
    spst_ver <- try(packageVersion("spatstat"), silent = TRUE)
    if (!inherits(spst_ver, "try-error") && spst_ver < "2.0-0") {
      stop("You have an old version of spatstat installed which is incompatible with ", pkg, 
           ". Please update spatstat (or uninstall it).")
    }
  }
  if (!is.null(X) && isTRUE(st_is_longlat(X)))
    stop("Only projected coordinates may be converted to spatstat class objects", call. = FALSE)
}

as.owin.stars = function(W, ..., fatal) {
	check_spatstat("spatstat.geom", W)
	if (length(dim(W)) != 2)
		stop("as.owin.stars requires a 2-dimensional object (single raster layer)")
	m = t(W[[1]])
	bb = st_bbox(W)
	spatstat.geom::owin(bb[c("xmin", "xmax")], bb[c("ymin", "ymax")], mask = m[nrow(m):1,])
}

#' @export
st_as_stars.im = function(.x, ...) {
	# see https://github.com/r-spatial/stars/issues/648
	#setNames(st_as_stars(as.data.frame(.x, ...)), "v")
	d = dim(.x)
	nd = create_dimensions(
		list(x = create_dimension(1, d[2], .x$xrange[1], diff(.x$xrange)/d[2]),
			 y = create_dimension(1, d[1], .x$yrange[1], diff(.x$yrange)/d[1])),
		get_raster(affine = c(0.0, 0.0), dimensions = c("x", "y"), curvilinear = FALSE))
	st_stars(list(v = t(.x$v)), nd)
}

as.im.stars = function(X, ...) {
	check_spatstat("spatstat.geom", X)
	if (!has_raster(X) || !is_regular_grid(X))
		stop("as.im.stars only works for stars objects with regular raster data")
	if (length(dim(X)) > 2)
		stop("as.im.stars only works for two-dimensional rasters")
	X = as.data.frame(X)
	spatstat.geom::as.im.data.frame(X, ...)
}

options(rgdal_show_exportToProj4_warnings = "none")
if (suppressPackageStartupMessages(require(sp, quietly = TRUE))) {
demo(meuse)
# remove +init=xxx
meuse@proj4string = CRS(
"+proj=sterea +lat_0=52.1561605555556 +lon_0=5.38763888888889 +k=0.9999079 +x_0=155000 +y_0=463000 +ellps=bessel +units=m +no_defs"
)
meuse.grid@proj4string = CRS(
"+proj=sterea +lat_0=52.1561605555556 +lon_0=5.38763888888889 +k=0.9999079 +x_0=155000 +y_0=463000 +ellps=bessel +units=m +no_defs"
)

suppressPackageStartupMessages(library(stars))
(m = st_as_stars(meuse))
(gr = st_as_stars(meuse.grid))

m2 = as(m, "Spatial")
summary(m2)
gr2 = as(gr, "Spatial")
summary(gr2)

all.equal(meuse, m2, check.attributes = FALSE)
fullgrid(meuse.grid) = TRUE
all.equal(meuse.grid, gr2, check.attributes = FALSE)

#all.equal(meuse, m2)
#all.equal(meuse.grid, gr2)
}

# Create 'stars' object
set.seed(1331)
suppressPackageStartupMessages(library(stars))
volcano = rbind(volcano, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA) # add NA rows
d = st_dimensions(x = 1:ncol(volcano), y = 1:nrow(volcano))
(r = st_as_stars(t(volcano)))
r = st_set_dimensions(r, 1, offset = 0, delta = 1)
r = st_set_dimensions(r, 2, offset = nrow(volcano), delta = -1)

# Create points
pnt = st_sample(st_as_sfc(st_bbox(r)), 100)
pnt = st_as_sf(pnt)

# Extract - 'st_join'
x = st_join(pnt, st_as_sf(r))

# Extract - 'st_extract'
y = st_extract(r, pnt)

# check there are NAs:
any(is.na(x))
# Compare
all.equal(x$A1, y[[1]])

################################################
if (FALSE) {

## tic: segfaults
# check equal results with stars_proxy:
#x = st_extract(stars:::st_as_stars_proxy(r), pnt)
#all.equal(x$A1, y[[1]])
#all.equal(x, y)

r = c(r, 2*r, 10*r)
x = st_join(pnt, st_as_sf(r))
y = st_as_sf(st_extract(r, pnt))
all.equal(x, y)

## tic: segfaults
#x = st_extract(stars:::st_as_stars_proxy(merge(r)), pnt)
#all.equal(st_as_sf(x), y)

tif = system.file("tif/L7_ETMs.tif", package = "stars")
xp = read_stars(tif, proxy = TRUE)
xm = read_stars(tif, proxy = FALSE)
pts = st_sample(st_as_sfc(st_bbox(xp)), 10)
pts = c(pts, st_as_sfc("POINT(0 0)"), pts)
em = st_extract(xm, pts)
if (utils::packageVersion("sf") >= "0.9-7") {
	ep = st_extract(xp, pts)
	print(all.equal(ep, em, check.attributes = TRUE))
}

# two-attribute objects:
library(stars)
tif = system.file("tif/L7_ETMs.tif", package = "stars")
xp = read_stars(c(tif, tif), proxy = TRUE)
xm = read_stars(c(tif, tif), proxy = FALSE)
pts = st_sample(st_as_sfc(st_bbox(xp)), 10)
pts = c(pts, st_as_sfc("POINT(0 0)"), pts)
em = st_extract(xm, pts)
if (utils::packageVersion("sf") >= "0.9-7") {
	ep = st_extract(xp, pts)
	print(all.equal(ep, em, check.attributes = TRUE))
}

# single-attribute, single raster objects:
tif1 = paste0(tempfile(), ".tif")
write_stars(xm[1,,,1], "x.tif")
xp = read_stars("x.tif", proxy = TRUE)
xm = read_stars("x.tif", proxy = FALSE)
em = st_extract(xm, pts)
if (utils::packageVersion("sf") >= "0.9-7") {
	ep = st_extract(xp, pts)
	print(all.equal(ep, em, check.attributes = TRUE))
}

# multiple-file attributes:
x = c(
  "avhrr-only-v2.19810901.nc",
  "avhrr-only-v2.19810902.nc",
  "avhrr-only-v2.19810903.nc",
  "avhrr-only-v2.19810904.nc",
  "avhrr-only-v2.19810905.nc",
  "avhrr-only-v2.19810906.nc",
  "avhrr-only-v2.19810907.nc",
  "avhrr-only-v2.19810908.nc",
  "avhrr-only-v2.19810909.nc"
)
file_list = system.file(paste0("netcdf/", x), package = "starsdata")
if (!identical(file_list, "")) {
  y = read_stars(file_list, quiet = TRUE)
  print(y)
  st_crs(y) = "OGC:CRS84"
  pts = st_sample(st_as_sfc(st_bbox(y)), 10)
  em = st_extract(y, pts)

  (y = read_stars(file_list, quiet = TRUE, proxy = TRUE))
  print(y)
  st_crs(y) = "OGC:CRS84"
  if (utils::packageVersion("sf") >= "0.9-7") {
	  ep = st_extract(y, pts)
	  print(all.equal(em, ep))
  }
}

# nearest & bilinear comparison:
if (utils::packageVersion("sf") >= "0.9-7") {
  set.seed(12331)
  s = st_as_stars(matrix(rnorm(16), 4))
  pts = st_sample(st_as_sfc(st_bbox(s)), 10000, type = "regular")
  s1 = st_extract(s, pts, bilinear = FALSE)
  s2 = st_extract(s, pts, bilinear = TRUE)
  s1$s2 = s2[[1]]
  names(s1)[c(1,3)] = c("nearest", "bilinear")
  print(s1[sample(10000, 5),])
}
}

suppressPackageStartupMessages(library(stars))
m = matrix(1:121, 11, 11)
m
s = st_as_stars(m)
st_downsample(s, 1)
st_downsample(s, 1)[[1]]
st_downsample(s, 1, offset = 1)
st_downsample(s, 1, offset = 1)[[1]]
st_downsample(s, 1, FUN = mean)
st_downsample(s, 1, FUN = mean)[[1]]
st_downsample(s, 1, offset = 1, FUN = mean)
st_downsample(s, 1, offset = c(0,1), FUN = mean)[[1]]
st_downsample(s, 1, offset = c(0,1))
st_downsample(s, 1, offset = c(0,1))[[1]]

.update_dims <- function(dims, proxy_dimensions, coords) {
	between <- function(x, ymin, ymax, z = x) {
		z[x >= ymin & x <= ymax, drop = FALSE]
	}
	
	uc <- coords
	for(coord in names(coords)) {
		ndx <- which(dims$name == coord)
		
		if (!is.na(proxy_dimensions[[coord]]$refsys) &&
			all(proxy_dimensions[[coord]]$refsys %in% c("POSIXct", "CFtime"))) {
			dims$start[ndx] <- proxy_dimensions[[coord]]$from
			dims$count[ndx] <- proxy_dimensions[[coord]]$to - proxy_dimensions[[coord]]$from + 1
		} else {
			if(attr(proxy_dimensions, "raster")$curvilinear) {
				uc[[coord]] <- seq(proxy_dimensions[[coord]]$from, proxy_dimensions[[coord]]$to)
			} else {
				pv <- st_get_dimension_values(proxy_dimensions, coord)
				uc[[coord]] <- between(uc[[coord]], min(pv), max(pv))
			}
			dims$start[dims$name == coord] <- which(coords[[coord]] == uc[[coord]][1])[1]
			dims$count[dims$name == coord] <- dims$length[dims$name == coord] <- length(uc[[coord]])
		}
	}
	dims
}

#' @export
print.nc_proxy = function(x, ..., n = 1e5, nfiles = 10, simplify = TRUE) {
	cat("netcdf source stars proxy object from:\n")

	if(simplify) print(shorten_names(x[[1]], nfiles))

	cat("\nAvailable nc variables:\n")
	cat(paste(names(x), collapse = "\n"))
	cat("\n\ndimension(s):\n")
	print(st_dimensions(x), ...)
	cat("\n")
	if (!is.null(attr(x, "call_list"))) {
		cat("call_list:\n")
		print(unlist(attr(x, "call_list")))
		cat("This object has pending lazy operations: dimensions as printed may not reflect this.\n")
	}
}

# as.data.frame.stars_proxy works for nc_proxy.

#' @export
st_as_stars.nc_proxy <- function(.x, ..., downsample = 0, envir = parent.frame()) {
	process_call_list(read_ncdf(.x, downsample = downsample, ...), 
					  attr(.x, "call_list"), envir = envir, downsample = downsample)
}

#' @name plot
#' @param max_times integer; maximum number of time steps to attempt to plot. 
#' @export
#' @details when plotting a subsetted \code{stars_proxy} object, the default value for argument \code{downsample} will not be computed correctly, and has to be set manually.
plot.nc_proxy = function(x, y, ..., downsample = get_downsample(dim(x)), max_times = 16) {
	
	if(length(x) > 1) {
		message("Plotting first variable only.")
		x <- x[1]
	}
			
	tdim <- which(sapply(st_dimensions(x), function(x) any(grepl("^POSIX|^CFtime", x$refsys))))
	
	if(length(tdim)) {
		if(length(st_get_dimension_values(x, tdim)) > max_times) {
			stop("Time dimension of nc_proxy is longer than max_times in plot.nc_proxy.")
		}
	}
	
	plot(st_as_stars(x, downsample = downsample, ...), ..., downsample = 0)
}

#' @export
"[.nc_proxy" <- function(x, i = TRUE, ...) {
	# stars_proxy works but need to match class.
	structure(NextMethod(), class = class(x))
}

#' @export
st_crop.nc_proxy <- function(x, y, ..., 
							 crop = TRUE, 
							 epsilon = sqrt(.Machine$double.eps), 
							 collect = TRUE) {
	# will this work? Couild be slick if it does.
	structure(NextMethod(), class = class(x))
}

# "[[<-.nc_proxy" not needed.
	
#' @export
c.nc_proxy = function(..., along = NA_integer_, along_crs = FALSE, try_hard = FALSE, 
						 nms = names(list(...)), tolerance = sqrt(.Machine$double.eps)) {
	
	dots = list(...)
	if (!all(sapply(dots, function(x) inherits(x, "nc_proxy"))))
		stop("all arguments to c() should be nc_proxy objects")
	
	x <- dots[[1]]
	
	structure(NextMethod(), class = class(x))
}


#' @export
st_redimension.nc_proxy <- function(x, new_dims, along, ...) stop("st_redimension not support for nc_proxy")

#' @export
st_mosaic.nc_proxy = function(.x, ...) stop("st_mosaic not supported for nc_proxy")

#' @param .x stars object to add curvilinear coordinates too.
#' @param curvilinear only for creating curvilinear grids: named length 
#' 2 list holding longitude and latitude matrices; the names of this 
#' list should correspond to raster dimensions referred to
#' @param crs object of class \code{crs} with the coordinate reference 
#' system of the values in \code{curvilinear}; see details
#' @details if \code{curvilinear} is a \code{stars} object with longitude 
#' and latitude values, its coordinate reference system is typically not 
#' that of the latitude and longitude values.
#' @noRd
add_curvilinear <- function(.x, 
							curvilinear = NULL, 
							crs = st_crs('OGC:CRS84')) {
	# so we can just call add_curvilinear regardless
	if (is.null(curvilinear))
		.x
	else {
		stopifnot(is.list(curvilinear), 
				  names(curvilinear) %in% names(dim(.x)))
		
		if (inherits(curvilinear[[1]], "stars"))
			curvilinear[[1]] = curvilinear[[1]][[1]]
		
		if (inherits(curvilinear[[2]], "stars"))
			curvilinear[[2]] = curvilinear[[2]][[1]]
		
		dimensions = st_dimensions(.x)
		xy = names(curvilinear)
		
		dimensions[[ xy[1] ]]$values = structure(curvilinear[[1]], dim = setNames(dim(curvilinear[[1]]), xy))
		dimensions[[ xy[2] ]]$values = structure(curvilinear[[2]], dim = setNames(dim(curvilinear[[1]]), xy))
		
		# erase regular grid coefficients $offset and $delta:
		dimensions[[ xy[1] ]]$offset = dimensions[[ xy[1] ]]$delta = NA_real_
		dimensions[[ xy[2] ]]$offset = dimensions[[ xy[2] ]]$delta = NA_real_
		
		raster = get_raster(dimensions = names(curvilinear), 
							curvilinear = TRUE)
		
		st_set_crs(st_stars(.x, 
							create_dimensions(dimensions, raster),
							class = class(.x)), 
				   crs)
	}
}

#' spatially or temporally aggregate stars object
#' 
#' spatially or temporally aggregate stars object, returning a data cube with lower spatial or temporal resolution
#' 
#' @param x object of class \code{stars} with information to be aggregated
#' @param by object of class \code{sf} or \code{sfc} for spatial aggregation, for temporal aggregation a vector with time values (\code{Date}, \code{POSIXct}, or \code{CFTime}) that is interpreted as a sequence of left-closed, right-open time intervals or a string like "months", "5 days" or the like (see \link{cut.POSIXt}), or a function that cuts time into intervals; if by is an object of class \code{stars}, it is converted to sfc by \code{st_as_sfc(by, as_points = FALSE)} thus ignoring its time component. Note: each pixel is assigned to only a single group (in the order the groups occur) so non-overlapping spatial features and temporal windows are recommended.
#' @param FUN aggregation function, such as \code{mean}
#' @param ... arguments passed on to \code{FUN}, such as \code{na.rm=TRUE}
#' @param drop logical; ignored
#' @param join function; function used to find matches of \code{x} to \code{by}
#' @param rightmost.closed see \link{findInterval}
#' @param left.open logical; used for time intervals, see \link{findInterval} and \link{cut.POSIXt}
#' @param as_points see \link[stars]{st_as_sf}: shall raster pixels be taken as points, or small square polygons?
#' @param exact logical; if \code{TRUE}, use \link[exactextractr]{coverage_fraction} to compute exact overlap fractions of polygons with raster cells
#' @param weights single-layer \code{SpatRaster} or \code{stars} object on the same grid as \code{x} with secondary per-cell weights, e.g. population or cropland area. Only used when \code{exact = TRUE}, and cells with \code{NA} weight are treated as zero weight. Not supported for \code{stars_proxy} objects
#' @param transform function or one-sided formula, evaluated by \code{rlang::as_function}, applied to cell values before aggregation; e.g. \code{~ .x^2} or \code{~ pmax(0, .x - 10) - pmax(0, .x - 30)}. Units on \code{x} are dropped when set. Note: any transform that derives parameters from the values it is given is applied separately to each attribute (and to each chunk for \code{stars_proxy} objects), so its results may not be comparable across attributes or chunks, e.g. \code{splines::bs(.x)} places knots at quantiles of its input, so fix such parameters explicitly, as in \code{splines::bs(.x, knots = ..., Boundary.knots = ...)}
#' @seealso \link[sf]{aggregate}, \link[sf]{st_interpolate_aw}, \link{st_extract}, https://github.com/r-spatial/stars/issues/317
#' @export
#' @aliases aggregate
#' @examples
#' # aggregate time dimension in format Date
#' tif = system.file("tif/L7_ETMs.tif", package = "stars")
#' t1 = as.Date("2018-07-31")
#' x = read_stars(c(tif, tif, tif, tif), along = list(time = c(t1, t1+1, t1+2, t1+3)))[,1:30,1:30]
#' st_get_dimension_values(x, "time")
#' x_agg_time = aggregate(x, by = t1 + c(0, 2, 4), FUN = max) 
#'
#' # aggregate time dimension in format Date - interval
#' by_t = "2 days"
#' x_agg_time2 = aggregate(x, by = by_t, FUN = max) 
#' st_get_dimension_values(x_agg_time2, "time")
#' #TBD:
#' #x_agg_time - x_agg_time2
#'
#' # aggregate time dimension in format POSIXct
#' x = st_set_dimensions(x, 4, values = as.POSIXct(c("2018-07-31", 
#'                                                   "2018-08-01", 
#'                                                   "2018-08-02", 
#'                                                   "2018-08-03")), 
#'                       names = "time")
#' by_t = as.POSIXct(c("2018-07-31", "2018-08-02"))
#' x_agg_posix = aggregate(x, by = by_t, FUN = max)
#' st_get_dimension_values(x_agg_posix, "time")
#' #TBD:
#' # x_agg_time - x_agg_posix
#' aggregate(x, "2 days", mean)
#' if (require(ncmeta, quietly = TRUE)) {
#'  # Spatial aggregation, see https://github.com/r-spatial/stars/issues/299
#'  prec_file = system.file("nc/test_stageiv_xyt.nc", package = "stars")
#'  prec = read_ncdf(prec_file, curvilinear = c("lon", "lat"))
#'  prec_slice = dplyr::slice(prec, index = 17, along = "time")
#'  nc = sf::read_sf(system.file("gpkg/nc.gpkg", package = "sf"), "nc.gpkg")
#'  nc = st_transform(nc, st_crs(prec_slice))
#'  agg = aggregate(prec_slice, st_geometry(nc), mean)
#'  plot(agg)
#' }
#'
#' # example of using a function for "by": aggregate by month-of-year
#' d = c(10, 10, 150)
#' a = array(rnorm(prod(d)), d) # pure noise
#' times = Sys.Date() + seq(1, 2000, length.out = d[3])
#' m = as.numeric(format(times, "%m"))
#' signal = rep(sin(m / 12 * pi), each = prod(d[1:2])) # yearly period
#' s = (st_as_stars(a) + signal) |>
#'       st_set_dimensions(3, values = times)
#' f = function(x, format = "%B") {
#' 	  months = format(as.Date(paste0("01-", 1:12, "-1970")), format)
#' 	  factor(format(x, format), levels = months)
#' }
#' agg = aggregate(s, f, mean)
#' plot(agg)
#'
#' \donttest{
#' # exact = TRUE with secondary weights: population-weighted mean
#' # population density per municipality
#' if (requireNamespace("exactextractr", quietly = TRUE) &&
#'     requireNamespace("terra", quietly = TRUE)) {
#'   dens = read_stars(system.file("sao_miguel/gpw_v411_2020_density_2020.tif", 
#'      package = "exactextractr"))
#'   pop = read_stars(system.file("sao_miguel/gpw_v411_2020_count_2020.tif", 
#'      package = "exactextractr"))
#'   conc = sf::read_sf(system.file("sao_miguel/concelhos.gpkg", 
#'      package = "exactextractr"))
#'   aggregate(dens, conc, mean, exact = TRUE, weights = pop, na.rm = TRUE)
#' }
#' } # donttest
aggregate.stars = function(x, by, FUN, ..., drop = FALSE, join = st_intersects, 
		as_points = any(st_dimension(by) == 2, na.rm = TRUE), rightmost.closed = FALSE,
		left.open = FALSE, exact = FALSE, weights = NULL, transform = NULL) {

	fn_name = substr(deparse1(substitute(FUN)), 1, 20)
	classes = c("sf", "sfc", "POSIXct", "Date", "character", "function", "stars")
	if (!is.function(by) && !inherits(by, classes))
		stop(paste("currently, only `by' arguments of class", 
			paste(classes, collapse= ", "), "supported"))
	if (inherits(by, "stars"))
		by = st_as_sfc(by, as_points = FALSE) # and if not, then use st_normalize(by)

	if (inherits(by, "sf")) {
		geom = attr(by, "sf_column")
		by = st_geometry(by)
	} else
		geom = "geometry"
	stopifnot(!missing(FUN), is.function(FUN))
	
	if (!exact && !is.null(weights))
		warning("for exact=FALSE, weights is ignored")
	if (!is.null(transform)) {
		tf = rlang::as_function(transform)
		# tf may return a vector or a matrix, the matrix branch appends a `term`
		# dimension to x, the vector branch keeps dim(y). for example splines::bs(...)
		# or ~ cbind(.x, .x^2, .x^3) returns a matrix while x^2 returns a vector
		tr = lapply(x, function(y) tf(as.vector(y)))
		d = st_dimensions(x)
		if (is.matrix(tr[[1]])) {
			k = ncol(tr[[1]])
			if (!all(sapply(tr, function(v) is.matrix(v) && ncol(v) == k)))
				stop("transform must return a matrix with the same number of columns for every attribute")
			if (!all(mapply(function(v, y) nrow(v) == length(y), tr, x)))
				stop("transform must return one row per cell value")
			cn = colnames(tr[[1]])
			# any unnamed column makes all names default to t1..tk:
			term_names = if (is.null(cn) || any(is.na(cn) | !nzchar(cn))) paste0("t", seq_len(k)) else cn
			d = create_dimensions(append(d, list(term = create_dimension(values = term_names))), attr(d, "raster"))
			x = st_as_stars(mapply(function(v, y) array(v, dim = c(dim(y), k)), tr, x, SIMPLIFY = FALSE),
				dimensions = d)
		} else {
			if (!all(mapply(function(v, y) length(v) == length(y), tr, x)))
				stop("transform must return one value per cell value")
			x = st_as_stars(mapply(function(v, y) array(v, dim = dim(y)), tr, x, SIMPLIFY = FALSE),
				dimensions = d)
		}
	}

	if (exact && inherits(by, c("sfc_POLYGON", "sfc_MULTIPOLYGON")) && has_raster(x)) {
    	if (!requireNamespace("terra", quietly = TRUE))
        	stop("package terra required, please install it first") # nocov
    	if (!requireNamespace("exactextractr", quietly = TRUE))
        	stop("package exactextractr required, please install it first") # nocov
		x = st_upfront(x)
		d = st_dimensions(x)[1:2]
		r = st_as_stars(list(a = array(1, dim = dim(d))), dimensions = d)
		template = as(r, "SpatRaster")
		e = exactextractr::coverage_fraction(template, by)
		m = terra::values(do.call(c, e))
		if (!is.null(weights)) {
			if (inherits(weights, "stars"))
				weights = as(weights, "SpatRaster")
			if (!inherits(weights, "SpatRaster"))
				stop("weights must be a SpatRaster or single-attribute stars object")
			if (terra::nlyr(weights) > 1)
				stop("weights must be a single-layer raster")
			terra::compareGeom(weights, template) # errors if weights does not align with x
			w = terra::values(weights)[, 1]
			w[is.na(w)] = 0
			if (all(w == 0))
				stop("weights are all zero")
			m = m * w
		}
		is_mean = !identical(FUN, sum) # see https://github.com/r-spatial/stars/issues/289
		if (is_mean && !isTRUE(as.character(as.list(FUN)[[3]])[2] == "mean"))
			stop("for exact=TRUE, FUN should either be mean or sum")
		na.rm = isTRUE(list(...)$na.rm)
		new_dim = c(prod(dim(x)[1:2]), prod(dim(x)[-(1:2)]))
		out_dim = c(ncol(m), dim(x)[-(1:2)])
		agg = lapply(x, function(a) {
			v = array(a, dim = new_dim)
			nas = is.na(v)
			v[nas] = 0
			num = crossprod(m, v)
			if (is_mean) # na.rm also drops NA cells from the denominator:
				num = num / (if (na.rm) crossprod(m, !nas) else colSums(m))
			if (!na.rm && any(nas)) # a group is NA when it covers an NA cell:
				num[crossprod(m, nas) > 0] = NA
			array(num, dim = out_dim)
		})
		# crossprod dropped units, so to propagate units, if present we need to copy (mean/sum):
		d = create_dimensions(append(setNames(list(create_dimension(values = by)), geom),
			st_dimensions(x)[-(1:2)]))
		for (i in seq_along(x)) {
			if (inherits(x[[i]], "units")) 
				agg[[i]] = units::set_units(agg[[i]], units(x[[i]]), mode = "standard")
			names(dim(agg[[i]])) = names(d)
		}
		return(st_as_stars(agg, dimensions = d))
	}
	if (exact && !is.null(weights))
		warning("weights is ignored: exact aggregation requires a polygonal `by' and a raster x")

	values = NULL
	drop_y = FALSE
	grps = if (inherits(by, c("sf", "sfc"))) {
			x = if (has_raster(x)) {
					ndims = 2
					drop_y = TRUE
					st_upfront(x)
				} else if (has_sfc(x)) {
					ndims = 1
					st_upfront(x, which_sfc(x))
				}
	
			# find groups:
			# don't use unlist(join(x_geoms, by)) as this would miss the empty groups, 
			#      and may have multiple if geometries in by overlap, hence:
			if (identical(join, st_intersects) && has_raster(x))
				sapply(join(x, by, as_points = as_points),
					function(x) if (length(x)) x[1] else NA)
			else {
				x_geoms = if (has_raster(x))
						st_as_sfc(x, as_points = as_points)
					else
						st_dimensions(x)[[ which_sfc(x) ]]$values
				sapply(join(x_geoms, by), function(x) if (length(x)) x[1] else NA)
			}
		} else { # time: by is POSIXct/Date or character
			ndims = 1
			x = st_upfront(x, which_time(x))
			values = expand_dimensions(x)[[1]]
			if (is.function(by)) {
				i = by(values)
				if (!is.factor(i))
					i = as.factor(i)
				by = levels(i)
			} else if (inherits(by, "character")) {
				if (inherits(values, "CFTime")) {
					i  = values$cut(by)
					by = levels(i)
					new_time = attr(i, "CFTime")
				} else {
					i = cut(values, by, right = left.open)
					by = if (inherits(values, "Date"))
							as.Date(levels(i))
						else
							as.POSIXct(levels(i))
				}
			} else {
				if (!inherits(values, class(by)))
					warning(paste0('argument "by" is of a different class (', class(by)[1], 
						') than the time values (', class(values)[1], ')'))
				i = findInterval(values, by, left.open = left.open, rightmost.closed = rightmost.closed)
				i[ i == 0 | i == length(by) ] = NA
			}
			as.integer(i)
		}

	d = st_dimensions(x)
	dims = dim(d)

	agr_grps = function(x, grps, uq, FUN, bind, ...) { 
		do.call(bind, lapply(uq, function(i) {
				sel <- which(grps == i)
				if (!isTRUE(any(sel)))
					NA_real_
				else
					apply(x[sel, , drop = FALSE], 2, FUN, ...)
			}
		))
	}

	bind = if (length(FUN(1:10, ...)) > 1)
			cbind
		else
			rbind
	# rearrange:
	x = structure(x, dimensions = NULL, class = NULL) # unclass
	newdims = c(prod(dims[1:ndims]), prod(dims[-(1:ndims)]))
	for (i in seq_along(x)) {
		a = array(x[[i]], newdims)
		u = if (inherits(x[[i]], "units") && dim(a)[2] > 0) {
				a = units::set_units(a, units(x[[i]]), mode = "standard")
				try(out <- FUN(a[,1], ...))
				if (inherits(out, "units"))
					units(out)
				else
					NULL
			} else
				NULL
		x[[i]] = agr_grps(a, grps, seq_along(by), FUN, bind, ...)
		if (is.numeric(x[[i]]) && !is.null(u))
			x[[i]] = units::set_units(x[[i]], u, mode = "standard")
	}

	# reconstruct dimensions table:
	if (!is.null(values) && inherits(values, "CFTime")) {
		d[[1]] = create_dimension(refsys = "CFTime", values = new_time)
		names(d)[1] <- "time"
	} else {
		d[[1]] = create_dimension(values = by)
		names(d)[1] = if (is.function(by) || inherits(by, c("POSIXct", "Date", "function")))
				"time"
			else
				geom
	}
	if (drop_y)
		d = d[-2] # y
	
	# suppose FUN resulted in more than one value?
	if ((r <- prod(dim(x[[1]])) / prod(dim(d))) > 1) {
		if (r %% 1 != 0)
			stop("unexpected array size: does FUN return a consistent number of values?")
		a = attributes(d)
		values = if (is.null(rn <- rownames(x[[1]])))
				seq_len(r)
			else
				rn[1:r]
		d = append(d, list(create_dimension(values = values)))
		n = length(d)
		names(d)[n] = fn_name
		d = d[c(n, 1:(n-1))]
		attr(d, "raster") = a$raster
		attr(d, "class") = a$class
		newdim = setNames(c(r, length(by), dims[-(1:ndims)]), names(d))
	} else
		newdim = setNames(c(length(by), dims[-(1:ndims)]), names(d))

	st_stars(lapply(x, structure, dim = newdim), dimensions = d)
}

	# aggregate is done over one or more dimensions
	# say we have dimensions 1,...,k and we want to aggregate over i,...,j
	# with 1 <= i <= j <= k; 
	# let |n| = j-1+1 be the number of dimensions to aggregate over, n
	# let |m| = k - n be the number of remaining dimensions, m
	# permute the cube such that the n dimensions are followed by the m
	# rearrange the cube to a 2D matrix with |i| x ... x |j| rows, and remaining cols
	# find the grouping of the rows
	# (rearrange such that groups are together)
	# for each sub matrix, belonging to a group, do
	#   apply FUN to every column
	#   assing the resulting row to the group
	# now we have |g| rows, with |g| the number of groups
	# assign each group to the target group of "by"
	# redimension the matrix such that unaffected dimensions match again

#' @export
aggregate.stars_proxy = function(x, by, FUN, ...) {
	if (!inherits(by, c("sf", "sfc", "sfg", "stars")))
		collect(x, match.call(), "aggregate", c("x", "by", "FUN"), env = environment(), ...)
	else {
		if (inherits(by, "stars"))
			by = st_as_sfc(by, as_points = FALSE)
		by = st_geometry(by)
		if (isTRUE(list(...)$exact) && !is.null(list(...)$weights))
			stop("weights is not supported for stars_proxy objects")

		# this assumes each result of a [ selection is small enough to hold in memory
		l = lapply(seq_along(by), 
		   	function(i) {
			   	sel_i = st_normalize(st_as_stars(x[by[i]]))
			   	aggregate(sel_i, by[i], FUN, ...)
		   	}
		)
		do.call(c, c(l, along = list(which_sfc(l[[1]]))))
	}
}

toRad = function(x) x * pi / 180

toDeg = function(x) x * 180 / pi

rotate = function(lonlat, lon0, lat0, north = TRUE) {
	# https://gis.stackexchange.com/questions/10808/manually-transforming-rotated-lat-lon-to-regular-lat-lon/14445 , by Miha!
	lon0 = toRad(lon0)
	lat0 = toRad(lat0)
	if (north) {
		lat0 = -lat0
		lon0 = pi + lon0
	}
	lonlat = toRad(lonlat)
	lon = lonlat[,1]
	lat = lonlat[,2]
	vartheta = -(pi/2 + lat0)
	varphi = -lon0
	toDeg(cbind(
		atan2(sin(lon), tan(lat) * sin(vartheta) + 
				cos(lon) * cos(vartheta)) - varphi,
		asin(cos(vartheta) * sin(lat) - cos(lon) * sin(vartheta) * cos(lat))
	))
}

#' @export
st_rotate = function(.x, lon0, lat0, north, ...) UseMethod("st_rotate")


#' Transform rotated pole long/lat regular grid to unrotated curvilinear grid
#'
#' Transform rotated long/lat regular grid to unrotated curvilinear grid
#' @param .x object of class \code{stars}
#' @param lon0 longitude of the rotated pole in degrees
#' @param lat0 latitude of the rotated pole in degrees
#' @param north logical; if \code{TRUE} the pole refers to the North pole, otherwise the South pole
#' @param ... ignored
#' @returns curvilinear stars object with coordinates in regular long/lat (North pole at lat=90)
#' @export
#' @name st_rotate
#' @examples
#' if (require("starsdata") && require("maps")) {
#'   # data downloaded from https://esgf-data.dkrz.de/search/cosmo-rea/
#'   nc = "netcdf/ts_EUR-6km_ECMWF-ERAINT_REA6_r1i1p1f1_COSMO_v1_mon_201801-201812.nc"
#'   f = system.file(nc, package = "starsdata")
#'   m = read_mdim(f, "ts")
#'   print(m)
#'   # NOTE this function is obsolete when reading m as
#'   # m = read_mdim(f, "ts", curvilinear = c("longitude", "latitude"))
#'   if (require(RNetCDF)) {
#'      x = open.nc(f)
#'      lon = att.get.nc(x, "rotated_latitude_longitude", "grid_north_pole_longitude")
#'      lat = att.get.nc(x, "rotated_latitude_longitude", "grid_north_pole_latitude")
#'      close.nc(x)
#'      print(c(lon = lon, lat = lat))
#'   } else {
#'      lon = -162
#'      lat = 39.25
#'   } 
#'   m1 = st_rotate(m, lon, lat)
#'   print(m1)
#'   h = function() maps::map(add = TRUE)
#'   plot(m1, downsample = c(10, 10, 5), axes = TRUE, hook = h, mfrow = c(1, 2)) 
#'     # curvilinear grid: downsample for plotting speed
#'   m2 = st_warp(m1, crs = st_crs("OGC:CRS84"), threshold = .1)
#'   plot(m2, hook = h, mfrow = c(3, 4)) # regular grid: plots fast
#' }
st_rotate.stars = function(.x, lon0, lat0, north = TRUE, ...) {
	stopifnot(is.na(st_crs(.x)) || st_is_longlat(.x), 
			  !is_curvilinear(.x), 
			  is.logical(north), 
			  !is.na(north),
			  lat0 <= 90,
			  lat0 >= -90,
			  length(list(...)) == 0)
	if (has_sfc(.x))
		st_set_dimensions(.x, which_sfc(.x), values = rotate(st_geometry(.x), lon0, lat0, north))
	else {
		.x = st_upfront(.x)
		d = dim(.x)
		ccr = rotate(st_coordinates(st_dimensions(.x)[1:2]), lon0, lat0, north)
		st_as_stars(.x, curvilinear = setNames(list(matrix(ccr[,1], d[1], d[2]),
					  matrix(ccr[,2], d[1], d[2])), names(d)[1:2]))
	}
}

#' @name st_rotate
#' @export
st_rotate.sfc = function(.x, lon0, lat0, north = TRUE, ...) {
	r = rapply(.x, rotate, how = "replace", lon0 = lon0, lat0 = lat0, north = north)
	st_set_crs(st_sfc(r, recompute_bbox = TRUE), NA_crs_)
}

#' @name st_rotate
#' @export
st_rotate.sf = function(.x, lon0, lat0, north = TRUE, ...) {
	st_set_geometry(.x, st_rotate(st_geometry(.x), lon0, lat0, north))
}

# examples with the sf::nc polygon dataset NOT ABOUT NETCDF's, see e.g. stars.R netcdf.R
suppressPackageStartupMessages(library(sf))
nc = read_sf(system.file("gpkg/nc.gpkg", package="sf"))
m = st_set_geometry(nc, NULL)
n = as.matrix(m[c("BIR74", "SID74", "NWBIR74", "BIR79", "SID79", "NWBIR79")])
dim(n) = c(county = 100, var = 3, year = 2)
dimnames(n) = list(county = nc$NAME, var = c("BIR", "SID", "NWBIR"), year = c(1974, 1979))
suppressPackageStartupMessages(library(stars))
(st = st_as_stars(pop = n))
foo <- st |> st_set_dimensions(1, st_geometry(nc)) # |> st_set_dimensions(3, c(1974, 1979))
st |> st_set_dimensions(1, st_geometry(nc)) |> st_set_dimensions(names = c("geometries", "var", "year"))
foo
st_bbox(foo)
(x = st_as_sf(foo))
frac = function(x) x[2] / x[1]
frac2 = function(x) c(sidsr = x[2] / x[1], nwbr = x[3] / x[1])
frac2an = function(x) c(x[2] / x[1], x[3] / x[1])
st_apply(foo, c(1,3), frac)
st_apply(foo, c(1,3), frac2)
st_apply(foo, c(1,3), frac2an)
library(abind)
(x = aperm(st_apply(foo, c(1,3), frac2), c(2,3,1)))
y = aperm(st_apply(foo, c(1,3), frac2), c("county","year","frac2"))
all.equal(st_dimensions(x), st_dimensions(y))

split(foo, 2)
split(foo, 3)
# subset vector cube:
foo[nc[1]]

to_curvilinear = function(x) {
	if (! is_curvilinear(x)) {
		xy = attr(st_dimensions(x), "raster")$dimensions
		nx = dim(x)[ xy[1] ]
		ny = dim(x)[ xy[2] ]
		cl = if (has_rotate_or_shear(x)) {
			pts = xy_from_colrow(as.matrix(expand.grid(seq_len(nx), seq_len(ny))) - 0.5,
				st_geotransform(x))
			list(matrix(pts[,1], nx, ny), matrix(pts[,2], nx, ny))
		} else {
			ed = expand_dimensions(x)
			list(matrix(ed[[ xy[1] ]], nx, ny), matrix(ed[[ xy[2] ]], nx, ny, byrow = TRUE))
		}
		st_as_stars(x, curvilinear = setNames(cl, xy), crs = st_crs(x))
	} else
		x
}

transform_curvilinear = function(x, crs, ...) {
	if (is.numeric(crs))
		crs = st_crs(crs)  	 # nocov

	to = crs
	from = if (sf_extSoftVersion()["proj.4"] < "5.0.0") {
			if (inherits(crs, "crs"))
				to = crs$proj4string
			else
				stopifnot(is.character(crs))
			st_crs(x)$proj4string
		} else
			st_crs(x)

	get_yx = function(x) isTRUE(sf::st_crs(sf::st_sfc(sf::st_point(), crs = x), parameters=TRUE)$yx)

	d = st_dimensions(x)
	xy = attr(d, "raster")$dimensions
	cc = cbind(as.vector(d[[ xy[1] ]]$values), as.vector(d[[ xy[2] ]]$values))
	pts = sf::sf_project(from, to, cc, ...)
	d[[ xy[1] ]]$refsys = d[[ xy[2] ]]$refsys = st_crs(crs)
	d[[ xy[1] ]]$values = matrix(pts[,1], dim(x)[xy])
	d[[ xy[2] ]]$values = matrix(pts[,2], dim(x)[xy])
	if (sf::st_axis_order() && get_yx(st_crs(from)) != get_yx(st_crs(crs))) {
		message("swapping [x] and [y] roles")
		attr(d, "raster")$dimensions = rev(attr(d, "raster")$dimensions)
		d[[ xy[1] ]]$values = t(d[[ xy[1] ]]$values)
		d[[ xy[2] ]]$values = t(d[[ xy[2] ]]$values)
	}
	st_stars(x, d)
}


#' transform geometries in stars objects to a new coordinate reference system, without warping
#'
#' @name st_transform
#' @param x object of class \code{stars}, with either raster or simple feature geometries
#' @param crs object of class \code{crs} with target crs
#' @param ... ignored
#' @examples
#' geomatrix = system.file("tif/geomatrix.tif", package = "stars")
#' (x = read_stars(geomatrix))
#' new = st_crs('OGC:CRS84')
#' y = st_transform(x, new)
#' plot(st_transform(st_as_sfc(st_bbox(x)), new), col = NA, border = 'red')
#' plot(st_as_sfc(y, as_points=FALSE), col = NA, border = 'green', axes = TRUE, add = TRUE)
#' image(y, col = heat.colors(12), add = TRUE)
#' plot(st_as_sfc(y, as_points=TRUE), pch=3, cex=.5, col = 'blue', add = TRUE)
#' plot(st_transform(st_as_sfc(x, as_points=FALSE), new), add = TRUE)
#' @seealso \link{st_warp}
#' @details For simple feature dimensions, \link[sf]{st_transform} is called, leading to lossless transformation. For gridded spatial data, a curvilinear grid with transformed grid cell (centers) is returned, which is also lossless. To convert this to a regular grid in the new \code{CRS}, use \link{st_warp} (which is in general lossy).
#' 
#' If array values contain geometries and an array as a whole is of class `sfc` and 
#' has a non-missing CRS, array geometries are also transformed. 
#' @export
st_transform.stars =  function(x, crs, ...) {

	stopifnot(!is.na(crs), !is.na(st_crs(x)))

	for (i in seq_along(x)) {
		if (inherits(x[[i]], "sfc") && !is.na(st_crs(x[[i]]))) # array cells are geometries
			x[[i]] = structure(st_transform(x[[i]], crs), dim = dim(x[[i]]))
	}
	if (has_sfc(x)) {
		if (!inherits(crs, "crs") && !inherits(crs, "stars"))
			crs = st_crs(crs) # needed for GDAL's transform of features
		d = st_dimensions(x)
		ix = which_sfc(x)
		for (j in ix) {
			d[[j]]$values = st_transform(d[[j]]$values, crs, ...)
			d[[j]]$refsys = crs
		}
		structure(x, dimensions = d)
	} else {
		if (! has_raster(x)) {
			warning("no spatial coordinates present: st_transform does nothing")
			x
		} else
			transform_curvilinear(to_curvilinear(x), crs, ...)
	}
}

#' @name st_transform
st_transform_proj.stars =  function(x, crs, ...) {

	stopifnot(!is.na(crs), !is.na(st_crs(x)))

	if (has_sfc(x)) {
		try_proj = function(x, crs) {
			ret = try(st_transform(x, crs), silent = TRUE)
			if (inherits(ret, "try-error")) {
				if (!requireNamespace("lwgeom", quietly = TRUE))
					stop("package lwgeom required, please install it first") # nocov
				if (inherits(crs, "crs"))
					crs = crs$proj4string
				ret = lwgeom::st_transform_proj(x, crs)
			}
			ret
		}
		d = st_dimensions(x)
		ix = which_sfc(x)
		for (j in ix) {
			d[[j]]$values = try_proj(d[[j]]$values, crs)
			d[[j]]$refsys = crs
		}
		structure(x, dimensions = d)
	} else {
		if (! has_raster(x)) {
			warning("no spatial coordinates present: st_transform_proj does nothing")
			x
		} else
			transform_curvilinear(to_curvilinear(x), crs)
	}
}

suppressPackageStartupMessages(library(stars))
tif = system.file("tif/L7_ETMs.tif", package = "stars")
x = read_stars(tif)
(xx = x[,c(1:347,349),c(1:350,352),]) # rectilinear: one-but-last row & col missing
image(xx) # chooses useRaster itself
  # see also tidy.R for a ggplot example, leaving out the row.

m = matrix(1:16, 4, 4)
x = 1:4
y = 2:5
(st = st_as_stars(list(m = m), dimensions = st_dimensions(x = x, y = y)))
st = st_set_dimensions(st, 1, values = c(1,2,3,4,6))
st = st_set_dimensions(st, 2, values = c(2,3,4,5,8))
try(st_set_dimensions(st, 2, values = c(2,3,4,5,8,20))) # one too many

# $<-.stars:
st$foo = st$m * 2
st

# st_as_sf:

if (utils::packageVersion("sf") > "0.7.2") {
	plot(st_as_sf(st, as_points = FALSE), axes = TRUE)
}

image(st, text_values = TRUE, axes = TRUE)

## https://mastodon.social/@EvenRouault/111738407524055989

has_global_longitude = function(x) {
	st_is_longlat(x) && isTRUE(all.equal(as.numeric(st_bbox(x)$xlim), c(-180.,180.)))
}

cut_latitude_to_3857 = function(bb) { # Pseudomercator does not have global coverage:
	bb["ymin"] = max(bb["ymin"], -85.06)
	bb["ymax"] = min(bb["ymax"],  85.06)
	bb
}

# create a target grid from x and crs, determining cellsize if unknown
# return a dimensions object
default_target_grid = function(x, crs, cellsize = NA_real_, segments = NA) {
	bb_x = st_bbox(x)
	old_crs = st_crs(x)
	if (st_is_longlat(x) && crs == st_crs(3857)) # common case:
		bb_x = cut_latitude_to_3857(bb_x)
	envelope = st_as_sfc(st_set_crs(bb_x, NA_crs_))
	# global adjustment: needed to have st_segmentize span global extent
	# https://github.com/r-spatial/mapview/issues/256
				# https://github.com/r-tmap/tmap/issues/526 :
	envelope = st_set_crs(st_segmentize(envelope, st_length(st_cast(envelope, "LINESTRING"))/segments), old_crs)
	envelope_new = st_transform(envelope, crs)
	bb = st_bbox(envelope_new) # in new crs
	if (any(is.na(cellsize))) {
		area = if (st_is_longlat(crs)) # we need a cell size in degree lon lat
				diff(bb[c("xmin", "xmax")]) * diff(bb[c("ymin", "ymax")])
			else
				st_area(envelope_new)
		ratio = if (has_rotate_or_shear(x)) {
				d = st_dimensions(x)
				xy = xy_from_colrow(rbind(c(0,0), c(1,1)), st_geotransform(d))
				cellarea = sum((xy[1,] - xy[2,])^2) / 2 # lenght is cell diagonal
				xy_dims = attr(d, "raster")$dimensions
				unclass(st_area(envelope)) / (prod(dim(x)[xy_dims]) * cellarea) # > 1
			} else
				1.0
		dxy = attr(st_dimensions(x), "raster")$dimensions
		cellsize = sqrt(unclass(area)/prod(dim(x)[dxy])/ratio)
		# TODO: divide by st_area(evelope_new)/st_area(envelope) ?
	}
	cellsize = rep(abs(cellsize), length.out = 2)
	nx = ceiling(diff(bb[c("xmin", "xmax")])/cellsize[1])
	ny = ceiling(diff(bb[c("ymin", "ymax")])/cellsize[2])
	if (has_global_longitude(x)) { # if global coverage, don't cross the boundaries:
		cellsize[1] = diff(bb[c("xmin", "xmax")])/nx
		cellsize[2] = diff(bb[c("ymin", "ymax")])/ny
	}
	x = create_dimension(from = 1, to = nx, offset = bb["xmin"], delta =  cellsize[1], refsys = crs)
	y = create_dimension(from = 1, to = ny, offset = bb["ymax"], delta = -cellsize[2], refsys = crs)
	create_dimensions(list(x = x, y = y), get_raster())
}

rename_xy_dimensions = function(x, dims) {
	#stopifnot(inherits(x, "stars"), inherits(dims, "dimensions"))
	dx = st_dimensions(x)
	xxy = attr(dx, "raster")$dimensions
	dimsxy = attr(dims, "raster")$dimensions
	if (all(xxy == dimsxy))
		x
	else {
		na = names(dx)
		names(dx)[match(xxy, na)] = dimsxy
		attr(dx, "raster")$dimensions = dimsxy
		structure(x, dimensions = dx)
	}
}

# transform grid x to dimensions target
# x is a stars object, target is a dimensions object
transform_grid_grid = function(x, target, threshold) {
	stopifnot(inherits(x, "stars"), inherits(target, "dimensions"))
	x = rename_xy_dimensions(x, target) # so we can match by name
	xy_names = attr(target, "raster")$dimensions
	new_pts = st_coordinates(target[xy_names])
	dxy = attr(target, "raster")$dimensions

	from = st_crs(target)
	pts = sf::sf_project(from = from, to = st_crs(x), pts = new_pts)

	# at xy (target) locations, get values from x, or put NA
	# to array:
	d = st_dimensions(x)
	# get col/row from x/y:
	xy = if (is_curvilinear(x)) {
    		if (!requireNamespace("FNN", quietly = TRUE))
        		stop("package FNN required, please install it first") #nocov
			if (st_is_longlat(x))
				warning("using Euclidean distance measures on geodetic coordinates")
			fnn = FNN::get.knnx(cc_x <- st_coordinates(d[xy_names]), pts, 1)
			if (is.na(threshold)) {
				p12 = st_as_sf(as.data.frame(cc_x[1:2,]), coords = 1:2)
				threshold = signif(st_distance(p12)[1,2])
				message(paste("threshold set to", format(threshold), 
					 ": set a larger value if you see missing values where there shouldn't be"))
			}
			i = fnn$nn.index - 1
			i[fnn$nn.dist > threshold] = NA
			ny = dim(x)[1]
			cbind(i %% ny, i %/% ny) + 1
		} else
			colrow_from_xy(pts, x, NA_outside = TRUE, flip = TRUE)
	dims = dim(x)
	index = matrix(seq_len(prod(dims[dxy])), dims[ dxy[1] ], dims[ dxy[2] ])[xy]
	x = unclass(x) # avoid using [[<-.stars:
	if (length(dims) > 2) {
		remaining_dims = dims[setdiff(names(dims), dxy)]
		newdim = c(prod(dims[dxy]), prod(remaining_dims))
		for (i in seq_along(x)) {
			dim(x[[i]]) = newdim
			x[[i]] = x[[i]][index,]
			dim(x[[i]]) = c(dim(target)[dxy], remaining_dims)
		}
	} else {
		for (i in seq_along(x)) {
			x[[i]] = x[[i]][index]
			dim(x[[i]]) = dim(target)
		}
	}
	d[dxy] = target[1:2]
	st_stars(x, dimensions = create_dimensions(d, attr(target, "raster")))
}


#' Warp (resample) grids in stars objects to a new grid, possibly in an new coordinate reference system
#'
#' @param src object of class \code{stars} with source raster
#' @param dest object of class \code{stars} with target raster geometry
#' @param crs coordinate reference system for destination grid, only used when \code{dest} is missing
#' @param cellsize length 1 or 2 numeric; cellsize in target coordinate reference system units
#' @param segments (total) number of segments for segmentizing the bounding box before transforming to the new crs
#' @param use_gdal logical; if \code{TRUE}, use gdal's warp or warper, through \link[sf]{gdal_utils}
#' @param options character vector with options, passed on to gdalwarp
#' @param no_data_value value used by gdalwarp for no_data (NA) when writing to temporary file; 
#'  not setting this when \code{use_gdal} is \code{TRUE} leads to a warning
#' @param debug logical; if \code{TRUE}, do not remove the temporary gdalwarp destination file, and print its name
#' @param method character; see details for options; methods other than \code{near} only work when \code{use_gdal=TRUE}
#' @param threshold numeric; distance threshold for warping curvilinear grids: new cells at distances larger than threshold are assigned NA values.
#' @param ... ignored
#' @details \code{method} should be one of \code{near}, \code{bilinear}, \code{cubic}, \code{cubicspline}, \code{lanczos}, \code{average}, \code{mode}, \code{max}, \code{min}, \code{med}, \code{q1} or \code{q3}; see https://github.com/r-spatial/stars/issues/109
#' @examples
#' geomatrix = system.file("tif/geomatrix.tif", package = "stars")
#' (x = read_stars(geomatrix))
#' new_crs = st_crs('OGC:CRS84')
#' y = st_warp(x, crs = new_crs)
#' plot(st_transform(st_as_sfc(st_bbox(x)), new_crs), col = NA, border = 'red')
#' plot(st_as_sfc(y, as_points=FALSE), col = NA, border = 'green', axes = TRUE, add = TRUE)
#' image(y, add = TRUE, nbreaks = 6)
#' plot(st_as_sfc(y, as_points=TRUE), pch=3, cex=.5, col = 'blue', add = TRUE)
#' plot(st_transform(st_as_sfc(x, as_points=FALSE), new_crs), add = TRUE)
#' # warp 0-360 raster to -180-180 raster:
#' r = read_stars(system.file("nc/reduced.nc", package = "stars"))
#' r |> st_set_crs('OGC:CRS84') |> st_warp(st_as_stars(st_bbox(), dx = 2)) -> s
#' plot(r, axes = TRUE) # no CRS set, so no degree symbols in labels
#' plot(s, axes = TRUE)
#' # downsample raster (90 to 270 m)
#' r = read_stars(system.file("tif/olinda_dem_utm25s.tif", package = "stars"))
#' r270 = st_as_stars(st_bbox(r), dx = 270)
#' r270 = st_warp(r, r270)
#' @details For gridded spatial data (dimensions \code{x} and \code{y}), see figure; the existing grid is transformed into a regular grid defined by \code{dest}, possibly in a new coordinate reference system. If \code{dest} is not specified, but \code{crs} is, the procedure used to choose a target grid is similar to that of \link[raster]{projectRaster}. This entails: (i) the envelope (bounding box polygon) is transformed into the new crs, possibly after segmentation (red box); (ii) a grid is formed in this new crs, touching the transformed envelope on its East and North side, with (if cellsize is not given) a cellsize similar to the cell size of \code{src}, with an extent that at least covers \code{x}; (iii) for each cell center of this new grid, the matching grid cell of \code{x} is used; if there is no match, an \code{NA} value is used.
#' @export
st_warp = function(src, dest, ..., crs = NA_crs_, cellsize = NA_real_, segments = 100,
		use_gdal = FALSE, options = character(0), no_data_value = NA_real_, debug = FALSE,
		method = "near", threshold = NA_real_) {

	if (!inherits(src, "stars_proxy"))
		src = st_normalize(src)

	if (!is.na(crs))
		crs = st_crs(crs)

	if (!missing(dest)) {
		if (inherits(dest, "crs"))
			stop("target crs should be specified with crs = ..., not as argument dest")
	} else if (is.na(crs) && !use_gdal)
		stop("when use_gdal is FALSE, either dest or crs need to be specified")

	ret = if (use_gdal) {
		if (is_curvilinear(src))
			stop("for warping curvilinear grids using GDAL, use gdal_utils(\"warp\",...) directly on the source dataset of this object")
		if (!is.na(no_data_value))
			options = c(options, "-dstnodata", no_data_value)
		else 
			warning("no_data_value not set: missing values will appear as zero values")
		options = c(options, "-r", method)
		if (all(!is.na(cellsize))) {
			cellsize = rep(abs(cellsize), length.out = 2)
			options = c(options, "-tr", cellsize[1], cellsize[2])
		}
		if (! inherits(src, "stars_proxy")) {
			src = st_as_stars_proxy(src, NA_value = no_data_value)
			if (debug)
				cat("Writing input to: ", src[[1]], "\n")
			else
				on.exit(unlink(src[[1]])) # temp file
		}
		# collapse a multi-file proxy objects into single-file/multi-band:
		if (length(src[[1]]) > 1 || length(src) > 1) {
			tmp = tempfile(fileext = ".vrt") # multi-band destination
			sf::gdal_utils("buildvrt", unlist(src), tmp, options = "-separate")
			src[[1]] = tmp
		}
		if (missing(dest) && is.na(crs)) {
			delete = TRUE
			dest = tempfile(fileext = ".tif")
			sf::gdal_utils("warp", src[[1]], dest, options = options)
		} else {  # dest exists, and should be used: should use warper rather than warp
			# https://github.com/r-spatial/stars/issues/407
			if (missing(dest)) {
				dest = default_target_grid(src, crs = crs, cellsize = cellsize, segments = segments) # dimensions
				dest = st_stars(list(values = array(NA_real_, dim(dest))), dest)
			}
			if (length(dim(src)) == 3 && length(dim(dest)) == 2)
				dest = merge(do.call(c, lapply(seq_len(dim(src)[3]), function(x) dest)))
			dest = if (! inherits(dest, "stars_proxy")) {
					dest[[1]] = NA_real_ * dest[[1]] # blank out values
					delete = !debug
					st_as_stars_proxy(dest, NA_value = no_data_value)[[1]]
				} else {
					delete = FALSE
					dest[[1]] # the file name of the stars_proxy, not to be deleted
				}
			sf::gdal_utils("warper", src[[1]], dest, options = method) # not "warp"!
		}
		if (debug)
			cat("Writing result to: ", dest, "\n")
		else if (delete && !inherits(src, "stars_proxy"))
			on.exit(unlink(dest)) # a temp file
		read_stars(dest)
	} else {
		if (method != "near")
			stop("methods other than \"near\" are only supported if use_gdal=TRUE")
		if (missing(dest))
			dest = default_target_grid(src, crs = crs, cellsize = cellsize, segments = segments)
		if (!inherits(dest, "stars") && !inherits(dest, "dimensions"))
			stop("dest should be a stars object, or a dimensions object")
		transform_grid_grid(st_as_stars(src), st_dimensions(dest), threshold)
	}
	# restore attributes?
	if (method %in% c("near", "mode")) {
		a = lapply(src, attributes)
		for (i in seq_along(ret))
			ret[[i]] = structure(ret[[i]],
				levels = attr(ret[[i]], "levels") %||% a[[i]]$levels,
				colors = attr(ret[[i]], "colors") %||% a[[i]]$colors,
				class =  attr(ret[[i]], "class")  %||% a[[i]]$class)
	}
	ret
}

get_geotransform = function(x) {
	r = attr(x, "raster")
	if (is.null(r))
		rep(NA_real_, 6)
	else {
		xd = x[[ r$dimensions[1] ]]
		yd = x[[ r$dimensions[2] ]]
		as.numeric(c(xd$offset, xd$delta, r$affine[1], yd$offset, r$affine[2], yd$delta))
	}
}

#' get or set the geotransform, or rotation matrix
#' @param x object of class stars or dimensions
#' @param ... ignored
#' @export
st_geotransform = function(x, ...) UseMethod("st_geotransform")

#' @export
st_geotransform.default = function(x, ...) {
	stop(paste("no st_geotransform method available for objects of class", class(x)[1]))
}

#' @export
st_geotransform.stars = function(x, ...) st_geotransform(st_dimensions(x))

#' @export
st_geotransform.dimensions = function(x, ...) get_geotransform(x)


#' @export
#' @name st_geotransform
#' @param value length 6 numeric vector, or 2 x 2 (scaled) rotation matrix
`st_geotransform<-` = function(x, value) UseMethod("st_geotransform<-")

#' @export
#' @name st_geotransform
#' @examples
#' # using the "classical" rotation matrix, see https://en.wikipedia.org/wiki/Rotation_matrix :
#' rot = function(theta, dxdy = c(1., -1.)) {
#'    th = theta / 180 * pi
#'    matrix(c(cos(th), sin(th), -sin(th), cos(th)), 2, 2) %*% 
#' 		   matrix(c(dxdy[2], 0, 0, dxdy[1]), 2, 2)
#' }
#' l = st_downsample(st_as_stars(L7_ETMs), 9) # save time in plotting
#' st_geotransform(l) = rot(20, c(28.5, 28.5)) # clockwise, 20 degrees, scale by cell size
#' plot(l[,,,1])
#' m = rot(20, c(1, 2))
#' g = expand.grid(x = 0:4, y = 0:4)
#' plot(g[1:2], asp = 1)
#' text(g[,1], g[,2], labels = seq_along(g[,1]), pos = 4)
#' g = t(m %*% t(as.matrix(g)))
#' points(g, col = 'red')
#' text(g[,1], g[,2], labels = seq_along(g[,1]), pos = 4, col = 'red')
#'
#' m = matrix(1:20, 4)
#' s0 = st_as_stars(m)
#' s = s0
#' # dy > 0, clockwise rotation:
#' st_geotransform(s) = rot(10, c(1,1))
#' plot(s0, reset = FALSE)
#' plot(s, add = TRUE)
#' # dy < 0, counter clockwise rotation, + expansion in x-direction:
#' layout(1)
#' s0 = st_as_stars(st_bbox(s0), dx = 1)
#' s0$values = 1:20
#' s0
#' plot(s0, reset = FALSE)
#' s = s0
#' st_geotransform(s) = rot(10, c(2,1))
#' plot(s, add = TRUE)
`st_geotransform<-.stars` = function(x, value) {
	d = st_dimensions(x)
	r = attr(d, "raster")
	# https://gdal.org/tutorials/geotransforms_tut.html
	if (is.matrix(value)) { # gdal has col-row (pixel-line) order, where linear algebra uses row-col
		stopifnot(all(dim(value) == c(2, 2)))
		r$affine = c(value[2,1] * sign(d[[ r$dimensions[1] ]]$delta), 
					 value[1,2] * sign(d[[ r$dimensions[2] ]]$delta))
		d[[ r$dimensions[1] ]]$delta = value[2,2] * sign(d[[ r$dimensions[1] ]]$delta)
		d[[ r$dimensions[2] ]]$delta = value[1,1] * sign(d[[ r$dimensions[2] ]]$delta)
	} else {
		stopifnot(is.numeric(value), length(value) == 6)
		d[[ r$dimensions[1] ]]$offset = value[1]
		d[[ r$dimensions[1] ]]$delta  = value[2]
		r$affine[1]                   = value[3]
		d[[ r$dimensions[2] ]]$offset = value[4]
		r$affine[2]                   = value[5]
		d[[ r$dimensions[2] ]]$delta  = value[6]
	}
	attr(d, "raster") = r
	st_stars(x, d)
}

suppressPackageStartupMessages(library(stars))
geomatrix = system.file("tif/geomatrix.tif", package = "stars")
x = read_stars(geomatrix)
# can stars reproduce what gdal does, by default?
x2 = st_warp(x, use_gdal = TRUE, no_data_value = -9999)
y = st_warp(x, x2)
if (interactive()) { plot(x2, breaks = "equal", axes=TRUE) }
if (interactive()) { plot(y, breaks = "equal", axes=TRUE) }
names(x2) = names(y)
all.equal(x2, y) # yes?

# does gdal reproduce with stars template object?
(x2 = setNames(st_warp(x, y, use_gdal = TRUE, no_data_value=-1), "file.tif"))

# does gdal reproduce what stars does, default cell size?
(x2 = st_warp(x, crs = st_crs(x), use_gdal = FALSE))
(y = setNames(st_warp(x, x2, use_gdal = TRUE, debug = FALSE, no_data_value=-1), "file.tif"))

# try with multiple bands:
tif = system.file("tif/L7_ETMs.tif", package = "stars")
(x1 = read_stars(tif))
tifcp = tempfile(fileext = ".tif")
file.copy(tif, tifcp)
x1p = read_stars(tifcp, proxy = TRUE)
st_dimensions(x1p)
(x1a = st_warp(x1, crs = st_crs(4326)))
(x1b = setNames(st_warp(x1, x1p, use_gdal = TRUE, no_data_value=-1), "file.tif"))

# does gdal reproduce what stars does? Smaller grid:
x2 = st_warp(x, crs = st_crs(x), use_gdal = FALSE, cellsize = 3)
# x2 = x2[,2:43,2:43]
if (interactive()) { plot(x2, breaks = "equal", axes=TRUE, reset = FALSE) }
if (interactive()) { plot(st_as_sfc(st_bbox(x2)), add = TRUE, col = NA, border = 'red') }
### doesn't work: FIXME: check with more recent GDAL:
(y = setNames(st_warp(x, x2, use_gdal = TRUE, debug = FALSE, no_data_value=-1), "file.tif"))
if (interactive()) { plot(y, breaks = "equal") }
names(x2) = names(y)
# isTRUE(all.equal(x2, y, check.attributes=FALSE))
m = mean(as.vector(x2[[1]]==y[[1]]), na.rm = TRUE)
u = unique(as.vector(x2[[1]]-y[[1]]))

g = st_as_stars()
attr(g, "dimensions")$x$offset = 0
g$values = as.vector(t(matrix(seq_len(prod(dim(g))), 180, 360)))
if (interactive()) { plot(g, axes=TRUE) }
a = st_warp(g, st_as_stars(), use_gdal=FALSE)
b = st_warp(g, st_as_stars(), use_gdal=TRUE, no_data_value = -1)
all.equal(a, b, check.attributes = FALSE)

#' Read NetCDF into stars object
#'
#' Read data from a file (or source) using the NetCDF library directly.
#'
#' The following logic is applied to coordinates. If any coordinate axes have
#' regularly spaced coordinate variables they are reduced to the
#' offset/delta form with 'affine = c(0, 0)', otherwise the values of the coordinates
#' are stored and used to define a rectilinear grid.
#'
#' If the data has two or more dimensions and the first two are regular
#' they are nominated as the 'raster' for plotting.
#'
#' If the \code{curvilinear} argument is used it specifies the 2D arrays
#' containing coordinate values for the first two dimensions of the data read. It is currently
#' assumed that the coordinates are 2D and that they relate to the first two dimensions in
#' that order.
#' 
#' Any time dimension in the source is captured in a 'CFTime' object, unless the
#' \code{make_time} argument is false. In that case, the numeric offsets 
#' representing time are returned instead.
#' @examples
#' f <- system.file("nc/reduced.nc", package = "stars")
#' if (require(ncmeta, quietly = TRUE)) {
#'  read_ncdf(f)
#'  read_ncdf(f, var = c("anom"))
#'  read_ncdf(f, ncsub = cbind(start = c(1, 1, 1, 1), count = c(10, 12, 1, 1)))
#' }
#'
#' @param .x NetCDF file or source as a character vector or an nc_proxy object.
#' @param ... ignored
#' @param var variable name or names (they must be on matching grids)
#' @param ncsub matrix of start, count columns (see Details)
#' @param curvilinear length two character named vector with names of variables holding
#' longitude and latitude values for all raster cells. `stars` attempts to figure out appropriate
#' curvilinear coordinates if they are not supplied.
#' @param eps numeric; dimension value increases are considered identical when they differ less than \code{eps}
#' @param ignore_bounds logical; should bounds values for dimensions, if present, be ignored?
#' @param make_time if \code{TRUE} (the default), an attempt is made to provide
#'   a CFTime class for any "time" dimension.
#' @param make_units if \code{TRUE} (the default), an attempt is made to set the units property of each variable
#' @param proxy logical; if \code{TRUE}, an object of class \code{stars_proxy} is read which contains array
#' metadata only; if \code{FALSE} the full array data is read in memory. If not set, defaults to \code{TRUE}
#' when the number of cells to be read is larger than \code{options(stars.n_proxy)}, or to 1e8 if that option was not set.
#' @param downsample integer; number of cells to omit between samples along each dimension. 
#' e.g. \code{c(1,1,2)} would return every other cell in x and y and every third cell
#' in the third dimension (z or t). If 0, no downsampling is applied. Note that this transformation
#' is applied AFTER NetCDF data are read using st_downsample. As such, if proxy=TRUE, this 
#' option is ignored.
#' @details
#' If \code{var} is not set the first set of variables on a shared grid is used.
#'
#' \code{start} and \code{count} columns of ncsub must correspond to the variable dimension (nrows)
#' and be valid index using \code{\link[RNetCDF]{var.get.nc}} convention (start is 1-based). If the count value
#' is \code{NA} then all steps are included. Axis order must match that of the variable/s being read.
#' @export
#' @examples
#' \donttest{
#' if (require(ncmeta, quietly = TRUE)) {
#'  #' precipitation data in a curvilinear NetCDF
#'  prec_file = system.file("nc/test_stageiv_xyt.nc", package = "stars")
#'  prec = read_ncdf(prec_file, curvilinear = c("lon", "lat"), ignore_bounds = TRUE)
#' }
#'
#' ##plot(prec) ## gives error about unique breaks
#' ## remove NAs, zeros, and give a large number
#' ## of breaks (used for validating in detail)
#' qu_0_omit = function(x, ..., n = 22) {
#'   x = units::drop_units(na.omit(x))
#'   c(0, quantile(x[x > 0], seq(0, 1, length.out = n)))
#' }
#' if (require(dplyr, quietly = TRUE)) {
#'   prec_slice = slice(prec, index = 17, along = "time")
#'   plot(prec_slice, border = NA, breaks = qu_0_omit(prec_slice[[1]]), reset = FALSE)
#'   nc = sf::read_sf(system.file("gpkg/nc.gpkg", package = "sf"), "nc.gpkg")
#'   plot(st_geometry(nc), add = TRUE, reset = FALSE, col = NA)
#' }
#' } # donttest
read_ncdf = function(.x, ..., var = NULL, ncsub = NULL, curvilinear = character(0),
    eps = sqrt(.Machine$double.eps), ignore_bounds = FALSE, make_time = TRUE, make_units = TRUE,
    proxy = NULL, downsample = 0) {

  if (!requireNamespace("ncmeta", quietly = TRUE)) # This will pull in package CFtime as well
    stop("package ncmeta required, please install it first") # nocov
  if (!requireNamespace("RNetCDF", quietly = TRUE))
    stop("package RNetCDF required, please install it first") # nocov

  if(inherits(.x, "nc_proxy")) {
  	# grab the source file and all variables needed.
    x <- as.character(.x[[1]]) # FIXME: only supports one data source.
    var <- names(.x)
    proxy_dimensions <- st_dimensions(.x)
    nc_prj <- sf::st_crs(.x)
    
    # proxy already has CFTime in dimension[["time"]]$values
    make_time <- FALSE
    
    if(!is.null(ncsub)) {
    	warning("ncsub ignored when .x is class nc_proxy")
    	ncsub <- NULL
    }
    
  } else {
    x <- .x
    proxy_dimensions <- NULL
  }
	
  # Get all the nc metadata
  meta <- .fix_meta(ncmeta::nc_meta(x))

  if(.is_netcdf_cf_dsg(meta)) {
    if (!requireNamespace("ncdfgeom", quietly = TRUE))
      stop("package ncdfgeom required, please install it first") # nocov

    if(!is.null(proxy) && proxy) warning("proxy behavior not supported for timeseries netcdf")

    geom <- .get_geom_name(meta)

    return(st_as_stars(ncdfgeom::read_timeseries_dsg(x), sf_geometry = geom))
  }

  # Get relevant variables
  var <- .get_vars(var, meta)
  rep_var <- var[1L]

  # Get coordinate variable info
  all_coord_var <- ncmeta::nc_coord_var(x)

  if(ncol(all_coord_var) == 0) all_coord_var <- data.frame(variable = NA, X = NA, Y = NA,
                                                           Z = NA, T = NA, bounds = NA)

  coord_var <- .clean_coord_var(all_coord_var, rep_var, meta, curvilinear)

  canon_order <- c("X", "Y", "Z", "T")[!is.na(coord_var[c("X", "Y", "Z", "T")])]
  
  curvilinear <- if(coord_var$curvilinear) {
    c(X = coord_var$X, Y = coord_var$Y) } else character(0)

  # Get dimensions for representative var in correct axis order.
  dims <- .get_dims(meta, rep_var, coord_var, meta$axis, canon_order)

  # Validate that ncsub matches dims
  dims <- .add_ncsub(dims, ncsub)

  nc <- RNetCDF::open.nc(x)
  on.exit(RNetCDF::close.nc(nc), add = TRUE)

  # Get coordinates from netcdf or create them
  coords <- .get_coords(nc, dims)
  coords <- .clean_coords(coords, coord_var, meta$attribute, eps)

  # Figure out if we have a raster or not
  raster <- .get_nc_raster(coords)

  # Get matcher for netcdf requests
  dimid_matcher <- .get_dimid_matcher(nc, coord_var, var)
  
  if(!is.null(proxy_dimensions)) {
  	dims <- .update_dims(dims, proxy_dimensions, coords)
  	coords <- .get_coords(nc, dims)
  	coords <- .clean_coords(coords, coord_var, meta$attributes, eps)
  }
  
  pull <- .should_pull(proxy, 
  					   array_size = prod(dims[, "count", drop = TRUE]),
  					   num_vars = length(var))
  
  out_data <- if(pull) {
  	# Get all the data from the nc file
  	.set_nc_units(.get_data(nc, var, dims, dimid_matcher, pull = pull),
  				  meta$attribute, make_units)
  } else {
  	# Just return the source file for each variable.
  	setNames(as.list(rep(x, length(var))), var)
  }
  
  # Create stars dimensions object

  if(is.null(nc_dim <- dim(out_data[[1]]))) nc_dim <- dims$length
  
  dimensions <- create_dimensions(setNames(nc_dim, dims$name),
  								  raster)
  	
  dimensions <- .get_nc_dimensions(dimensions,
                                   coord_var = all_coord_var,
                                   coords = coords,
                                   nc = nc,
                                   dims = dims,
                                   var_names = meta$variable$name,
                                   curvilinear,
                                   eps = eps,
                                   ignore_bounds = ignore_bounds,
                                   atts = meta$attribute)
  if (make_time)
  	dimensions <- .get_cf_time(dimensions, rep_var, meta)
  
  if (is.character(out_data[[1]])) {
    
    # this is a proxy
    out_data <- st_stars_proxy(out_data, dimensions, NA_value = NA_real_, resolutions = NULL)

    class(out_data) <- c("nc_proxy", "stars_proxy", "stars")

  } else {
    # Make initial response data
    out_data <- st_stars(out_data, dimensions)
  }

  if(is.null(proxy_dimensions)) {
  	if(length(curvilinear) == 2) {
  		cv <- curvilinear[1]
  	} else if (length(attr(coords, "cv")) > 0) {
  		cv <- attr(coords, "cv")[1]
  	} else {
  		cv <- all_coord_var[all_coord_var$variable == rep_var, ]$X
  	}
  	
  	nc_prj <- .get_nc_projection(meta$attribute, rep_var, cv)
  }
  
  st_crs(out_data) <- nc_prj

  # Add curvilinear and return
  if (length(curvilinear) == 2) {
    curvi_coords = .get_curvilinear_coords(curvilinear, dimensions, nc, dims)
    out_data <- add_curvilinear(out_data, curvilinear = curvi_coords, nc_prj)
  }
  
  if(!all(downsample == 0)) {
    out_data <- st_downsample(out_data, downsample)
  }
  
  out_data
  
}

.fix_meta <- function(meta) {
  nas <- is.na(meta$axis$dimension)
  if (any(nas)) meta$axis$dimension[nas] <- -1 # nocov

  if(!is.null(names(meta$attribute))) {
    names(meta$attribute)[names(meta$attribute) == "attribute"] <- "name" # future proofing.
  }

  meta
}

.is_regular <- function(coords_list, eps) {
  sapply(coords_list, function(x) regular_intervals(x, epsilon = eps))
}

.unique_fuzz <- function(x, eps) {
  u = unique(x)
  if (all(diff(sort(u)) < eps))
    mean(x) # rather than mean(u)
  else
    u
}

.as_rectilinear = function(d) {
    ed = expand_dimensions(d, center = FALSE)
    for (i in attr(d, "raster")$dimensions) {
        if (!is.na(d[[ i ]]$offset)) {
            d[[i]]$values = as_intervals(ed[[i]], add_last = TRUE)
            d[[i]]$offset = d[[i]]$delta = NA
        }
    }
    d
}

.should_pull <- function(proxy, array_size, num_vars,
                         n_proxy = options("stars.n_proxy")[[1]] %||% 1.e8) {
  if(is.null(proxy)) {
    if(array_size > n_proxy) {
      pull <- FALSE
      message("Large netcdf source found, returning proxy object.")
    } else {
      pull <- TRUE
      message(paste("Will return stars object with", array_size, "cells."))
    }
  } else {
    pull <- !proxy
  }

  if(pull & array_size > n_proxy)
    warning("Large netcdf source will be requested. Consider using stars proxy.")

  pull
}

.get_vars <- function(var, meta) {

  if (is.null(var)) {
    ix <- 1
    if (meta$grid$grid[ix] == "S") {
      ix <- which(!meta$grid$grid == "S")[1L] # nocov

      if (length(ix) < 1)  stop("only scalar variables found, not yet supported") # nocov
    }
    grd = meta$grid$grid[which.max(nchar(meta$grid$grid))]
    var = meta$grid$variables[[match(grd, meta$grid$grid)]]$variable

    message(sprintf("no 'var' specified, using %s", paste(var, collapse = ", ")))
    other_vars <- setdiff(meta$variable$name, var)
    if (length(other_vars) > 0)
      message(sprintf("other available variables:\n %s", paste(other_vars, collapse = ", ")))
  }
  return(var)
}

.get_grid_mapping <- function(atts, rep_var) {
	if(!is.null(atts)) {
		suppressWarnings(ncmeta::nc_grid_mapping_atts(atts, rep_var))
	} else {
		list()
	}
}

.get_nc_projection <- function(atts, rep_var, rep_coord_var) {

  nc_grid_mapping <- .get_grid_mapping(atts, rep_var)
  
  if(length(nc_grid_mapping) == 0) {
    
    if(.is_degrees(atts, rep_coord_var)) {
      message(paste("No projection information found in nc file. \n",
                    "Coordinate variable units found to be degrees, \n",
                    "assuming WGS84 Lat/Lon."))
      st_crs('OGC:CRS84')
    } else {
      warning("No projection information found in nc file.")
      st_crs(NULL)
    }
  } else {
  	tryCatch({
  		cv_units <- .get_gm_units(atts, rep_coord_var)
  	
  		base_gm <- ncmeta::nc_gm_to_prj(nc_grid_mapping)
  	
  		base_gm <- paste0(gsub("\\+units=m ", "", base_gm), 
  						  " +units=", cv_units)
  	
    	st_crs(base_gm)
  	}, error = function(e) {
  		warning(paste0("failed to create crs based on grid mapping\n",
  					  "and coordinate variable units. Will return NULL crs.\n",
  					  "Original error: \n", e))
  		st_crs(NULL)
  	})
  }
}

.is_degrees <- function(atts, var) {
  units <- .get_attributes(atts, "units", var)$value[[1]]
  !is.null(units) && grepl("degrees", units, ignore.case = TRUE)
}

.get_gm_units <- function(atts, var) {
	if(.is_degrees(atts, var)) {
		"degrees"
	} else {
		.get_attributes(atts, "units", var)$value[[1]]	
	}
}

.clean_coord_var <- function(c_v, var, meta, curvilinear) {
  c_v <- c_v[c_v$variable == var, ]

  if(nrow(c_v) == 0) c_v[1, ] <- NA

  check_curvi <- .check_curvilinear(c_v, var, meta$variable, curvilinear)

  if(length(curvilinear)) {
  	c_v$X <- rep(check_curvi[1], nrow(c_v))
  	c_v$Y <- rep(check_curvi[2], nrow(c_v))
  }

  not_na <- apply(c_v, 1, function(x) sum(!is.na(x)))
  
  c_v <- c_v[which(not_na == max(not_na)), ][1, ]
  
  has_prj <- length(.get_grid_mapping(meta$attribute, var)) != 0
  
  if(length(curvilinear) > 0) {
    if(length(check_curvi) == 0) stop("Curvilinear coordinate variables provided by not found in file.")
    c_v <- c_v[c_v$variable == var & c_v$X == check_curvi[1], ]
    c_v$curvilinear <- TRUE
  } else if(nrow(c_v) > 1 | length(check_curvi) == 2) {
    # Should use convenience function?
    sn_atts <- meta$attribute[meta$attribute$name == "standard_name", ]
    auxiliary_x <- sn_atts[grepl("projection_x_coordinate", sn_atts$value), ]$variable
    auxiliary_y <- sn_atts[grepl("projection_y_coordinate", sn_atts$value), ]$variable

    if(length(auxiliary_x) > 0 && length(auxiliary_y) > 0 &&
       auxiliary_x %in% c_v$X && auxiliary_y %in% c_v$Y &
       has_prj) {
      c_v <- c_v[c_v$variable == var & c_v$X == auxiliary_x, ]
      c_v$curvilinear <- FALSE
    } else if (length(check_curvi) == 2) {
      c_v <- c_v[c_v$variable == var & c_v$X == check_curvi[1], ]
      c_v$curvilinear <- TRUE
    } else{
      c_v <- c_v[1, ]
      warning(paste("Found two coordinate variable pairs. Chosing:",
                    paste(as.character(c_v)[2:5], collapse = " "),
                    "for", as.character(c_v)[1])) #nocov
    }
  } else {
    c_v$curvilinear <- FALSE
  }
  return(c_v)
}

.check_curvilinear <- function(coord_var, var, variables, curvilinear) {
  if(length(curvilinear) > 0) {
    if(!all(curvilinear %in% variables$name)) stop("Curvilinear variables not found in file.")
    if(!length(curvilinear) == 2 | all(!is.character(curvilinear))) stop("Curvilinear input must be a length two character vector.")
    if(all(variables$ndims[match(curvilinear, variables$name)] == 2)) {
      xy_coords <- coord_var[c("X", "Y")]
      if(!curvilinear[1] %in% c(xy_coords$X, xy_coords$Y) | !curvilinear[2] %in% c(xy_coords$X, xy_coords$Y)) {
        stop("Specified curvilinear coordinate variables not found as X/Y coordinate variables.")
      } else {
        if(!curvilinear[1] %in% xy_coords$X) {
          curvilinear <- curvilinear[2:1]
        }
        return(stats::setNames(curvilinear, c("X", "Y")))
      }
    } else {
      stop("Specified curvilinear coordinates are not 2-dimensional.")
    }
  }
  # If we have X and Y coordvars to look at.
  if (all(c("X", "Y") %in% names(coord_var))) {

    # Actually look at this one.
    XY_curvi = unlist(coord_var[coord_var$variable == var, ][c("X", "Y")])

    if (all(!is.na(XY_curvi)) && # If all coordinate variables have 2 dims.
        all(variables$ndims[match(XY_curvi, variables$name)] == 2)) {
      return(XY_curvi)
    } else {
      return(character(0))
    }
  } else {
    return(character(0))
  }
}

#' Gets dimension info ensuring they are in an order suited to coordinate
#' variables that need to be subset appropriately.
#' @noRd
.get_dims <- function(meta, var, c_v, axis, canon_order) {
  # This matches the axes of a given variable to the correct dimension
  dims <-meta$dimension[match(meta$axis$dimension[meta$axis$variable == var],
                              meta$dimension$id), ]

  dims$coord_var <- ""
  dims$axis <- ""

  rep_var_coordinates <- .get_attributes(meta$attribute, "coordinates", var)
  if(!is.null(rep_var_coordinates)) {
    rep_var_coordinates <- strsplit(rep_var_coordinates$value[[1]], " ")[[1]]
  }

  x_cv <- axis[axis$variable == c_v$X, ]
  y_cv <- axis[axis$variable == c_v$Y, ]

  if(nrow(x_cv) == 1 & nrow(y_cv) == 1) {
    dims[dims$id == x_cv$dimension, ][c("coord_var", "axis")] <- list(x_cv$variable, "X")
    dims[dims$id == y_cv$dimension, ][c("coord_var", "axis")] <- list(y_cv$variable, "Y")
  }

  dim_matcher <- axis[axis$variable == c_v$X, ]$dimension

  if(length(dim_matcher) < 2) {
    dim_matcher <- c(dim_matcher,
                     axis[axis$variable == c_v$Y, ]$dimension)
  }

  z_axis <- integer(0)
  if(!is.na(c_v$Z)) {
    z_cv <- axis[axis$variable == c_v$Z, ]
    
    z_axis <- z_cv$dimension
    
    dims[dims$id == z_cv$dimension, ][c("coord_var", "axis")] <- list(z_cv$variable, "Z")
    
    dim_matcher <- c(dim_matcher, z_axis)
    
  } else if(nrow(dims) == 4 & !c_v$curvilinear) {
    z_axis <- unique(axis[!axis$variable %in% c(c_v$X, c_v$Y, c_v$T) &
                            !axis$dimension %in% dim_matcher, ]$dimension)
  }

  if(!is.na(c_v$T)) {
    t_cv <- axis[axis$variable == c_v$T, ]
    
    t_axis <- t_cv$dimension
    
    dims[dims$id == t_cv$dimension, ][c("coord_var", "axis")] <- list(t_cv$variable, "T")
    
    if(length(z_axis) > 1) {
      z_axis <- z_axis[z_axis != t_axis]
    }
    dim_matcher <- c(dim_matcher, z_axis, t_axis)
  }

  if(all(!is.na(dim_matcher))) {
    if(length(dim_matcher) != length(dims$id)) {
      dim_matcher <- c(dim_matcher,
                       dims$id[!dims$id %in% dim_matcher])
      dim_matcher <- unique(dim_matcher)
    }
    dims <- dims[match(dim_matcher, dims$id, nomatch = 0L), ]
  }
  return(dims)
}

#' Gets a list of dimension id matching indexes for use in
#' requesting data from a given NetCDF file.
#' @noRd
.get_dimid_matcher <- function(nc, coord_var, var) {
  setNames(lapply(var, FUN = function(.v, c_v, nc) {

    matcher <- NULL

    if(all(!is.na(c_v[2:3]))) {
      coordvar_dimids <- RNetCDF::var.inq.nc(nc, c_v$X)$dimids

      if(length(coordvar_dimids) == 1) {
        coordvar_dimids <- c(coordvar_dimids,
                             RNetCDF::var.inq.nc(nc, c_v$Y)$dimids)
      }

      if(!is.na(c_v$Z)) {
        coordvar_dimids <- c(coordvar_dimids,
                             RNetCDF::var.inq.nc(nc, c_v$Z)$dimids)
      }

      if(!is.na(c_v$T)) {
        coordvar_dimids <- c(coordvar_dimids,
                             RNetCDF::var.inq.nc(nc, c_v$T)$dimids)
      }

      var_dimids <- RNetCDF::var.inq.nc(nc, .v)$dimids

      matcher <- match(coordvar_dimids, var_dimids)
      
      if(!all(diff(matcher[1:2]) == 1)) {
        warning("Non-canonical axis order found, attempting to correct.")
      }
    }
    matcher
  }, c_v = coord_var, nc = nc),
  var)
}

.get_nc_var <- function(nc, .v, start, count) {
  RNetCDF::var.get.nc(nc,
                      variable = .v,
                      start = start,
                      count = count,
                      collapse = FALSE, ## keep 1-dims
                      unpack = TRUE,     ## offset and scale applied internally
                      rawchar = TRUE)  ## needed for NC_CHAR, as per
}

.get_data <- function(nc, var, dims, dimid_matcher, pull = pull) {
  out_data <- lapply(var, pull = pull, FUN = function(.v, pull) {

    dm <- match(RNetCDF::var.inq.nc(nc, .v)$dimids,
                dims[, "id", drop = TRUE])
    if(is.null(dm)) dm <- c(1:nrow(dims))

    request <- list(start = dims[, "start", drop = TRUE][dm],
                    count = dims[, "count", drop = TRUE][dm])

    request$size <- prod(request$count)

    request$dimid_match <- dm
    
    request$axis <- dims$axis[dm]

    if(pull) {

      ret <- .get_nc_var(nc, .v, request$start, request$count)

      if(length(dm) > 1 && !all(diff(dm[1:2]) == 1)) {
        ret <- aperm(ret, dm)
      }

    }
    
    return(ret)

  })
  
  ## "../rasterwise/extdata/R13352.nc"
  ## https://github.com/hypertidy/tidync/issues/75
  ## check for NC_CHAR case
  setNames(lapply(out_data, function(.v) {
  	if (mode(.v) == "raw") {
  		array(unlist(lapply(.v, rawToChar)), dims$length)
  	} else .v
  }), var)
}

.get_attributes <- function(attribute, att, var = NULL) {
  if(!is.null(var)) attribute <- attribute[attribute$variable == var, ]
  if(att %in% attribute$name)
    attribute[attribute$name == att, ]
  else {
    NULL
  }
}

.add_ncsub <- function(dims, ncsub, proxy_dimensions) {

  if (is.null(ncsub)) {
    dims$start <- 1
    dims$count <- dims$length
  } else {
    if (nrow(dims) != nrow(ncsub))
      stop("input ncsub doesn't match available dims")
    ix <- is.na(ncsub[, "count"])
    if (any(ix))  ncsub[ix, "count"] <- dims$length[ix] - ncsub[ix, "start"] + 1
    if (any(ncsub[, "start"] < 1) ||
        any((ncsub[, "count"] - ncsub[, "start"] + 1) > dims$length))
      stop("start or count out of bounds")
    dims$start <- ncsub[, "start"]
    dims$count <- ncsub[, "count"]
  }
  return(dims)
}

.set_nc_units <- function(data_list, nc_atts, make_units) {
  # Get all the units
  nc_units <- .get_attributes(nc_atts, "units")

  if (!is.null(nc_units) && nrow(nc_units) > 0 && make_units) {
    for (i in names(data_list)) {
      if (i %in% nc_units$variable) {
        uval <- unlist(nc_units$value[nc_units$variable == i])
        if(is.numeric(data_list[[i]])) {
          units(data_list[[i]]) <- try_as_units(uval[1L])
        } else {
          data_list[[i]]$units <- try_as_units(uval[1L])
        }
      }
    }
  }
  return(data_list)
}

.get_coords <- function(nc, dims) {
  ## cannot assume we have COORDS style coordinate varibales
  ## - so create them as 1:length if needed
  coords = setNames(vector("list", length(dims$name)), dims$name)
  
  attr(coords, "cv") <- c()
  
  for (ic in seq_along(coords)) {
    subidx <- seq(dims$start[ic], length = dims$count[ic])

    ## create_dimvar means we can var_get it
    ## test checks if there's actually a variable of the dim name
    if (dims$name[ic] %in% ncmeta::nc_vars(nc)$name) {
      coords[[ic]] <- RNetCDF::var.get.nc(nc, variable = dims$name[ic])[subidx]
		
      attr(coords, "cv") <- c(attr(coords, "cv"), dims$name[ic])
      
    } else {
      coords[[ic]] <- subidx
    }
  }
  return(coords)
}

.get_nc_raster <- function(coords) {
  ## can we create a raster?
  raster = NULL
  ## which coords are regular
  if (length(coords) > 1) {
    raster = get_raster(affine = c(0, 0),
                        dimensions = names(coords)[1:2], curvilinear = FALSE)
  }
  return(raster)
}

.clean_coords <- function(coords, coord_var, atts, eps) {
	## hack longitudes for #277
	if (any(lon_coord <- grepl("lon", coord_var$X, ignore.case = TRUE))) {
		lon_coord <- coord_var$X[lon_coord][1]
		lons <- coords[[lon_coord]]
		if (!is.null(lons) && !regular_intervals(lons, epsilon = eps) &&
			.is_degrees(atts, lon_coord) &&
			lons[1L] > 180 && min(lons) > 0) {
			coords[[lon_coord]] <- ((coords[[lon_coord]] + 180) %% 360) - 180
		} else if(!is.null(lons) && .is_degrees(atts, lon_coord) &&
				  max(lons) > 180 && min(lons) > 180) {
			coords[[lon_coord]] <- ((coords[[lon_coord]] + 180) %% 360) - 180
		} else if(!is.null(lons) && .is_degrees(atts, lon_coord) &&
				  max(lons) > 180) {
			message(paste("0-360 longitude crossing the international date", 
						   "line encountered.\nLongitude coordinates will be",
						   "0-360 in output."))

		}
	}
	coords
}

.get_nc_dimensions <- function(dimensions, coord_var, coords, nc, dims,
                               var_names, curvilinear, eps, ignore_bounds,
                               atts) {

  to_rectilinear = FALSE

  regular <- .is_regular(coords, eps)

  for (i in seq_along(coords)) {
  	
  	from_to <- dimensions[[i]]$from:dimensions[[i]]$to
  	
  	try_bounds <- names(coords)[i] %in% var_names && 
  		!ignore_bounds &&
  		length(bounds <- coord_var[coord_var$variable == names(coords)[i], ]$bounds) > 0 &&
  		bounds %in% var_names
  	
  	if (try_bounds) {
  		
  		bounds = RNetCDF::var.get.nc(nc, bounds)
  		if (!is.matrix(bounds)) # single instance, returns a vector
  			bounds = matrix(bounds, nrow = 2)
  		
  		if(!bounds[1, 1] < coords[[i]][1] | !bounds[2, 1] > coords[[i]][1]) {
  			warning(paste("bounds not enveloping", names(coords)[i], "coordinates. Ignoring."))
  			try_bounds <- FALSE
  		} else {
  			
  			is_reg = ncol(bounds) > 1 && length(u <- .unique_fuzz(apply(bounds, 2, diff), eps)) == 1 &&
  				length(v <- .unique_fuzz(diff(bounds[1,]), eps)) == 1
  			
  			if (is_reg && abs(u + v) < eps) {
  				warning(paste("bounds for", names(coords)[i], "seem to be reversed; reverting them"))
  				bounds = apply(bounds, 2, sort) # should not be needed according to CF, but see #133
  				u = v
  			}
  			
  			if (is_reg && abs(v - u) < eps) {
  				dimensions[[i]]$offset = bounds[1,1]
  				dimensions[[i]]$delta = v
  			} else {
  				dimensions[[i]]$values = make_intervals(bounds[1, from_to], bounds[2,from_to])
  				dimensions[[i]]$point = FALSE
  				if (i %in% 1:2 && length(curvilinear) < 1) # FIXME: ? hard-coding here that lon lat are in the first two dimensions:
  					to_rectilinear = TRUE
  			}
  		}
  	}
  	
  	if (regular[i] & !try_bounds) {
      mdc = mean(diff(coords[[i]]))
      coord_name = names(coords)[i]
      t_dim = coord_var[coord_var$variable == coord_name,]$T
      if (is.null(t_dim) || length(t_dim) == 0)
        t_dim = NA_character_
      # https://github.com/r-spatial/stars/issues/378
      if (!is.na(t_dim) && names(coords)[i] == t_dim)
        dimensions[[i]]$offset[1L] = coords[[i]][1]
      else
        dimensions[[i]]$offset[1L] = coords[[i]][1] - mdc/2
      ## NaN for singleton dims, but that seems ok unless we have explicit interval?
      dimensions[[i]]$delta[1L]  = mdc
    } else {
      dimensions[[i]]$values = coords[[i]][from_to]
      ## offset/delta for fall-back index (and for NA test )
      ## https://github.com/r-spatial/stars/blob/master/R/dimensions.R#L294-L303
      dimensions[[i]]$offset[1L] = NA_real_
      dimensions[[i]]$delta[1L] = NA_real_
    }
  }
  if (to_rectilinear)
    dimensions = .as_rectilinear(dimensions)
  return(dimensions)
}

#' Get the CFTime instances for the variable
#' 
#' @param dimensions All dimensions of the data source.
#' @param var The variable to search for. This is a single var.
#' @param meta The metadata of the data source.
#' @noRd
.get_cf_time <- function(dimensions, var, meta) {
	var_dims  <- meta$axis$dimension[which(meta$axis$variable == var)]
	time_dims <- meta$extended$dimension[which(!is.na(meta$extended$time))]
	time_idx  <- which(meta$extended$dimension == time_dims[which(time_dims %in% var_dims)])
	for (t in seq_along(time_idx)) {
		idx <- time_idx[t]
		time <- meta$extended$time[[idx]]
		if (time$calendar$name %in% CF_calendar_regular)
			dimensions[[ meta$extended$name[[idx]] ]] <-
			create_dimension(from = 1, to = length(time$offsets),
							 refsys = "POSIXct", point = TRUE,
							 values = time$as_timestamp(asPOSIX = TRUE))
		else
			dimensions[[ meta$extended$name[[idx]] ]] <- 
			create_dimension(from = 1, to = length(time), values = time, 
							 refsys = "CFtime", point = TRUE)
	}
	dimensions
}

.get_curvilinear_coords <- function(curvilinear, dimensions, nc, dims) {
  curvi_coords <- lapply(curvilinear, function(.v) {

    cv_matcher <- RNetCDF::var.inq.nc(nc, .v)$dimids
    cv_matcher <- match(cv_matcher, dims$id[1:2])

    RNetCDF::var.get.nc(nc,
                        variable = .v,
                        start = dims[1:2, "start", drop = TRUE][cv_matcher],
                        count = dims[1:2, "count", drop = TRUE][cv_matcher],
                        collapse = FALSE,
                        unpack = TRUE)
  })
  names(curvi_coords)[1:2] <- names(dimensions)[1:2]

  expected_shape <- c(dimensions[[1]]$to, dimensions[[2]]$to)

  if(!all(dim(curvi_coords[[1]]) == expected_shape)) {
    curvi_coords[1] <- t(curvi_coords[1])
  }

  if(!all(dim(curvi_coords[2]) == expected_shape)) {
    curvi_coords[2] <- t(curvi_coords[2])
  }

  return(curvi_coords)
}

.is_netcdf_cf_dsg <- function(meta) {
  featuretype <- .get_attributes(meta$attribute, "featureType", "NC_GLOBAL")

  if(!is.null(featuretype)) {
    if(grepl( "timeseries", featuretype$value, ignore.case = TRUE)) {
      return(TRUE)
    }
  }
  return(FALSE)
}

.get_geom_name <- function(meta) {
  geometry <- unlist(unique(.get_attributes(meta$attribute, "geometry")$value))
  if(length(geometry) > 1) {
    warning(paste("Only a single geometry is supported. Using", geometry[1])) #nocov
    geometry <- geometry[1]
  }

  if(!is.null(geometry)) {
    ncdfgeom::read_geometry(meta$source$source)
  } else {
    NA
  }
}

test_that("slice", {
  skip_if_not_installed("dplyr", "1.0.0")
  library(dplyr)

  f <- system.file("nc/reduced.nc", package = "stars")
  nc <- read_ncdf(f, proxy = TRUE)
  expect_s3_class(nc, "stars_proxy")

  # Slicing after other lazy operations should be lazy
  nc2 <- adrop(nc)
  nc3 <- slice(nc2, "lon", 90)
  expect_equal(length(attr(nc3, "call_list")), 2)

  # Slicing dimension that are part of the file should be lazy
  nc4 <- slice(nc, "lon", 90)
  expect_equal(length(attr(nc4, "call_list")), 1)

  # Trailing singleton dimensions in files
  nc5 <- slice(nc, "time", 1)
  expect_null(attr(nc5, "call_list"))
  expect_equal(st_dimensions(nc), st_dimensions(nc5))

  # concatenate to simulate an extra dimension
  nc6 <- c(nc, nc, nc, nc, along = "dummy")
  expect_equal(dim(nc6)["dummy"], c(dummy = 4))
  expect_s3_class(nc6, "stars_proxy")

  nc7 <- slice(nc6, "dummy", 1)
  expect_s3_class(nc7, "stars_proxy")
  expect_equal(dim(nc7)["dummy"], c(dummy = 1))
  expect_null(attr(nc7, "call_list"))
  expect_equal(names(nc7), names(nc6))

  nc8 <- slice(nc6, "dummy", 2:3)
  expect_s3_class(nc8, "stars_proxy")
  expect_equal(dim(nc8)["dummy"], c(dummy = 2))
  expect_null(attr(nc8, "call_list"))
  expect_equal(names(nc8), names(nc6))
})

#' @name st_as_stars
#' @param as_col logical; return rgb numbers (FALSE) or (character) color values (TRUE)?
#' @export
st_as_stars.OpenStreetMap = function(.x, ..., as_col = FALSE) {

	ncols = .x$tiles[[1]]$yres # not sure why OpenStreetMap has swapped y and x...
	nrows = .x$tiles[[1]]$xres
  
	a = if (as_col)
			array(.x$tiles[[1]]$colorData, dim = c(x = ncols, y = nrows))
		else {
			rgb_values = grDevices::col2rgb(.x$tiles[[1]]$colorData)
			array(t(rgb_values), dim = c(x = ncols, y = nrows, band = 3))
		}
	
	bbx <- unname(unlist(.x$bbox))[c(1,3,4,2)]
	names(bbx) <- c("xmin", "xmax", "ymin", "ymax")
  
	crs = st_crs(.x$tiles[[1]]$projection)
  
	dimensions = list(
	x = create_dimension(from = 1, to = ncols, offset = unname(bbx["xmin"]),
		delta = unname((bbx["xmax"] - bbx["xmin"]) / ncols), refsys = crs),
	y = create_dimension(from = 1, to = nrows, offset = unname(bbx["ymax"]),
		delta = unname((bbx["ymin"] - bbx["ymax"]) / nrows), refsys = crs),
	band = create_dimension(values = c("red", "green", "blue")))
    
	if (as_col)
		st_as_stars(a, dimensions = create_dimensions(dimensions[1:2], get_raster()))
	else
		st_as_stars(a, dimensions = create_dimensions(dimensions, get_raster()))
}

make_label = function(x, i = 1) {
	dims = dim(x)
	if ((length(dims) <= 2 || dims[3] == 1) && inherits(x[[i]], "units"))
		make_unit_label(paste0("'", names(x)[i], "'"), units(x[[i]]))
	else
		names(x)[i]
}

#kw_dflt = function(x, key.pos) {
#	if (is.null(key.pos) || key.pos <= 0)
#		lcm(0)
#	else if (key.pos %in% c(2,4) && is.factor(x[[1]]))
#		lcm(max(strwidth(levels(x[[1]]), "inches")) * 2.54 * 1.1 + par("ps")/12) # cm
#	else
#		lcm(1.8 * par("ps")/12)
#}
# copy from sf:
kw_dflt = function(x, key.pos) {
	if (is.null(key.pos) || key.pos[1] == 0) # no key:
		return(lcm(0))

	font_scale = par("ps") / 12
	if (key.pos[1] == -1)
		lcm(1.8 * font_scale)
	else if (key.pos[1] %in% c(2, 4) && (is.character(x[[1]]) || is.factor(x[[1]]))) {
		strings = if (is.factor(x[[1]]))
				levels(x[[1]])
			else
				x[[1]]
		lcm(cm(max(strwidth(strings, "inches"))) * 1.3 + font_scale) # cm
		#lcm(cm(max(strwidth(strings, "inches"))) * 1.3) # cm
	} else
		lcm(1.8 * font_scale)
}


#' plot stars object, with subplots for each level of first non-spatial dimension
#'
#' plot stars object, with subplots for each level of first non-spatial dimension, and customization of legend key
#'
#' @name plot
#' @param x object of class \code{stars}
#' @param y ignored
#' @param join_zlim logical; if \code{TRUE}, compute a single, joint zlim (color scale) for all subplots from \code{x}
#' @param main character; subplot title prefix; use \code{""} to get only time, use \code{NULL} to suppress subplot titles
#' @param axes logical; should axes and box be added to the plot?
#' @param downsample logical or numeric; if \code{TRUE} will try to plot not many more pixels than actually are visible, if \code{FALSE}, no downsampling takes place, if numeric, the number of pixels/lines/bands etc that will be skipped; see Details.
#' @param nbreaks number of color breaks; should be one more than number of colors. If missing and \code{col} is specified, it is derived from that.
#' @param breaks numeric vector with actual color breaks, or a \code{style} name used in \link[classInt]{classIntervals}.
#' @param col colors to use for grid cells, or color palette function
#' @param ... further arguments: for \code{plot}, passed on to \code{image.stars}; for \code{image}, passed on to \code{image.default} or \code{rasterImage}.
#' @param key.pos numeric; side to plot a color key: 1 bottom, 2 left, 3 top, 4 right; set to \code{NULL} to omit key. Ignored if multiple columns are plotted in a single function call. Default depends on plot size, map aspect, and, if set, parameter \code{asp}. If it has length 2, the second value, ranging from 0 to 1, determines where the key is placed in the available space (default: 0.5, center).
#' @param key.width amount of space reserved for width of the key (labels); relative or absolute (using lcm)
#' @param key.length amount of space reserved for length of the key (labels); relative or absolute (using lcm)
#' @param key.lab character; label for color key in case of multiple subplots, use \code{""} to suppress
#' @param reset logical; if \code{FALSE}, keep the plot in a mode that allows adding further map elements; if \code{TRUE} restore original mode after plotting
#' @param box_col color for box around sub-plots; use \code{NA} to suppress plotting of boxes around sub-plots.
#' @param center_time logical; if \code{TRUE}, sub-plot titles will show the center of time intervals, otherwise their start
#' @param hook NULL or function; hook function that will be called on every sub-plot; see examples.
#' @param mfrow length-2 integer vector with nrows, ncolumns of a composite plot, to override the default layout
#' @param compact logical; place facets compactly (TRUE), or spread over the plotting device area?
#' @details
#' Downsampling: a value for \code{downsample} of 0: no downsampling, 1: after every dimension value (pixel/line/band), one value is skipped (half of the original resolution), 2: after every dimension value, 2 values are skipped (one third of the original resolution), etc. If \code{downsample} is \code{TRUE} or a length 1 numeric vector, downsampling is only applied to the raster [x] and [y] dimensions.
#'
#' To remove unused classes in a categorical raster, use the \link[base]{droplevels} function.
#'
#' When bitmaps show visual artefacts (Moiré effects), make sure that device \link{png} is used rather than \code{ragg::agg_png} as the latter uses antialiasing for filled polygons which causes this; see also https://github.com/r-spatial/stars/issues/573 .
#' @export
#' @examples
#' st_bbox(L7_ETMs) |> st_as_sfc() |> st_centroid() |> st_coordinates() -> pt
#' hook1 = function() {
#'     text(pt[,"X"], pt[,"Y"], "foo", col = 'orange', cex = 2)
#' }
#' plot(L7_ETMs, hook = hook1)
#' x = st_set_dimensions(L7_ETMs, 3, paste0("B_", 1:6))
#' hook2 = function(..., row, col, nr, nrow, ncol, value, bbox) {
#'    str = paste0("row ", row, "/", nrow, ", col ", col, "/", ncol, "\nnr: ", nr, " value: ", value)
#'    bbox |> st_as_sfc() |> st_centroid() |> st_coordinates() -> pt
#'    text(pt[,"X"], pt[,"Y"], str, col = 'red', cex = 2)
#' }
#' plot(x, hook = hook2, col = grey(c(.2,.25,.3,.35)))
#' if (isTRUE(dev.capabilities()$rasterImage == "yes")) {
#'   lc = read_stars(system.file("tif/lc.tif", package = "stars"))
#'   levels(lc[[1]]) = abbreviate(levels(lc[[1]]), 6) # so it's not only legend
#'   plot(lc, key.pos=4)
#' }
plot.stars = function(x, y, ..., join_zlim = TRUE, main = make_label(x, 1), axes = FALSE,
		downsample = TRUE, nbreaks = 11, breaks = "quantile", col = grey(1:(nbreaks-1)/nbreaks),
		key.pos = get_key_pos(x, ...), key.width = kw_dflt(x, key.pos), key.length = 0.618, 
		key.lab = ifelse(length(main) == 1, main, ""), reset = TRUE, box_col = NA, center_time = FALSE, hook = NULL, 
		mfrow = NULL, compact = TRUE) {

	if (!missing(y))
		stop("y argument should be missing")
	flatten = function(x, i) { # collapse all non-x/y dims into one, and select "layer" i
		d = st_dimensions(x)
		dxy = attr(d, "raster")$dimensions
		dims = dim(x)
		nms = names(x)
		x = x[[1]]
		aux = setdiff(names(dims), dxy)
		newdims = c(dims[dxy], prod(dims[aux]))
		dim(x) = newdims
		st_as_stars(setNames(list(x[,,i]), nms[1]), dimensions = d[dxy])
	}
	x = st_normalize(x)
	if (is.character(x[[1]])) # rgb values
		key.pos = NULL
	if (!missing(col) && is.function(col))
		col = col(nbreaks - 1)
	if (is.factor(x[[1]])) {
		if (missing(col))
			col = attr(x[[1]], "colors") %||% sf.colors(length(levels(x[[1]])), categorical = TRUE)
		else
			attr(x[[1]], "colors") = col
	}
	
	key.pos.missing = missing(key.pos)
	breaks.missing = missing(breaks)
	if (missing(nbreaks) && !missing(col))
		nbreaks = length(col) + 1
	opar = par(no.readonly = TRUE)
	dots = list(...)
	need_to_reset = TRUE

	#if (any(dim(x) == 1))
	#	x = adrop(x)
	if (is.factor(x[[1]]) && any(is.na(levels(x[[1]]))))
		x = droplevels(x) # https://github.com/r-spatial/stars/issues/339

	zlim = if (join_zlim && is.null(dots$rgb))
			range(unclass(x[[1]]), na.rm = TRUE)
		else
			rep(NA_real_, 2)

	# downsample if raster:
	if (has_raster(x) && !identical(downsample, FALSE)) {
		dxy = attr(st_dimensions(x), "raster")$dimensions
		loop = setdiff(names(dim(x)), dxy) # dimension (name) over which we loop, if any
		x = aperm(x, c(dxy, loop))
		dims = dim(x)
		if (isTRUE(downsample)) {
			n = dims * 0 # keep names
			n[dxy] = get_downsample(dims, rgb = is.numeric(dots$rgb))
			x = st_downsample(x, n)
		} else if (is.numeric(downsample)) {
			x = st_downsample(x, downsample)
		} 
	}

	# find breaks:
	if (join_zlim && !is.character(x[[1]]) && is.null(dots$rgb)) {
		breaks = if (is.factor(x[[1]]))
					seq(.5, length.out = length(levels(x[[1]])) + 1)
				else
					get_breaks(x, breaks, nbreaks, dots$logz)
		if (!inherits(breaks, c("POSIXt", "Date")))
			breaks = as.numeric(breaks)
		# https://github.com/r-spatial/stars/issues/728:
		if (packageVersion("sf") < "1.0.20" && length(breaks) > 2)
			breaks = unique(breaks)
		nbreaks = length(breaks) # might be shorter than originally intended!
		if (breaks.missing && length(unique(breaks)) <= 2) # unlucky default!
			warning('breaks="quantile" leads to a single class; maybe try breaks="equal" instead?')
	}

	if (isTRUE(dots$logz) && !((has_raster(x) && (is_curvilinear(x) || has_rotate_or_shear(x))) || has_sfc(x)))
		x = log10(x) # otherwise, defer log-transforming to sf::plot.sf

	if (has_raster(x)) {
		dims = dim(x)
		if (length(dims) == 2 || dims[3] == 1 || (!is.null(dots$rgb) && is.numeric(dots$rgb))) { ## ONE IMAGE:
			# set up key region
			values = structure(x[[1]], dim = NULL) # array -> vector
			if (! isTRUE(dots$add) && ! is.null(key.pos) && !all(is.na(values)) && is.null(dots$rgb) &&
					(is.factor(values) || length(unique(na.omit(values))) > 1) &&
					length(col) > 1 && !is_curvilinear(x)) { # plot key?
				switch(key.pos[1],
					layout(matrix(c(2,1), nrow = 2, ncol = 1), widths = 1, heights = c(1, key.width)),  # 1 bottom
					layout(matrix(c(1,2), nrow = 1, ncol = 2), widths = c(key.width, 1), heights = 1),  # 2 left
					layout(matrix(c(1,2), nrow = 2, ncol = 1), widths = 1, heights = c(key.width, 1)),  # 3 top
					layout(matrix(c(2,1), nrow = 1, ncol = 2), widths = c(1, key.width), heights = 1)   # 4 right
				)
				if (missing(key.lab))
					key.lab = ""
				if (is.factor(values))
					.image_scale_factor(levels(values), col, key.pos = key.pos,
						key.width = key.width, key.length = key.length, axes = axes, ...)
				else
					.image_scale(values, col, breaks = breaks, key.pos = key.pos,
						key.width = key.width, key.length = key.length, axes = axes, ..., lab = key.lab)
			} else
				need_to_reset = FALSE

			# map panel:
			mar = c(axes * 2.1, axes * 2.1, 1.2 * !is.null(main), 0)
			if (!is.null(key.pos) && key.pos[1] %in% 1:4)
				mar[key.pos[1]] = mar[key.pos[1]] + .5
			par(mar = mar)

			# plot the map:
			image(x, ..., axes = axes, breaks = breaks, col = col, key.pos = key.pos,
				key.width = key.width, key.length = key.length, main = NULL)
			if (!is.null(main))
				title(main)

		} else { ## MULTIPLE IMAGES -- now loop over dimensions 3:
			draw.key = !is.null(key.pos) && join_zlim
			if (! draw.key)
				key.pos = NULL
			lt = sf::.get_layout(st_bbox(x), dims[3], par("din"),
						if (join_zlim && key.pos.missing) -1 else key.pos[1], key.width, mfrow = mfrow, main = main)
			if (key.pos.missing)
				key.pos = lt$key.pos
			title_size = if (is.null(main))
					0
				else
					1.2
			layout(lt$m, widths = lt$widths, heights = lt$heights, respect = compact)
			par(mar = c(axes * 2.1, axes * 2.1, title_size, 0))
			labels = st_get_dimension_values(x, 3, center = center_time)
			if (inherits(labels, "CFTime")) {
				labels <- labels$as_timestamp()
			}
			for (i in seq_len(dims[3])) {
				im = flatten(x, i)
				if (! join_zlim) {
					zlim = range(im[[1]], na.rm = TRUE)
					br = get_breaks(im, breaks, nbreaks, dots$logz)
				} else
					br = breaks
				image(im, xlab = "", ylab = "", axes = axes, zlim = zlim, breaks = br, col = col, key.pos = NULL, main = NULL, ...)
				if (!is.null(main)) {
					if (length(main) == dims[3])
						title(main[i])
					else if (identical(main, names(x)[1])) # default value: omit on multi
						title(paste(format(labels[i])))
					else # user-defined
						title(paste(main, format(labels[i])))
				}
				if (!is.null(hook) && is.function(hook)) {
					if (is.null(formals(hook)))
						hook()
					else {
						nc = lt$mfrow[[2]] # nr of columns in the plot
						hook(row = ((i - 1) %/% nc) + 1,
							 col = ((i - 1)  %% nc) + 1, 
							 nrow = lt$mfrow[[1]],
							 ncol = nc,
							 nr = i, value = labels[i],
							 bbox = st_bbox(x))
					}
				}
				if (!is.na(box_col))
					box(col = box_col)
			}
			for (i in seq_len(prod(lt$mfrow) - dims[3])) # empty panels:
				plot.new()

			if (draw.key) {
				if (missing(key.lab) && inherits(x[[1]], "units"))
					key.lab = units::make_unit_label(names(x)[1], x[[1]])
				values = structure(x[[1]], dim = NULL)
				if (is.factor(values))
					.image_scale_factor(levels(values), col, key.pos = key.pos,
						key.width = key.width, key.length = key.length, axes = axes,...)
				else
					.image_scale(values, col, breaks = breaks, key.pos = key.pos,
						key.width = key.width, key.length = key.length, axes = axes,..., lab = key.lab)
			}
		}
	} else if (has_sfc(x)) {
#		if (key.pos.missing)
#			key.pos = -1
		if (! join_zlim)
			key.pos = NULL # omit key
		plot(st_as_sf(x[1]), ..., breaks = breaks, key.pos = key.pos, key.length = key.length,
			key.width = key.width, reset = reset, axes = axes, main = main)
	} else
		stop("no raster, no features geometries: no default plot method set up yet!")
	if (reset && need_to_reset) {
		layout(matrix(1)) # reset
		par(opar)
	}
	invisible(NULL)
}

get_breaks = function(x, breaks, nbreaks, logz = NULL) {
	if (is.character(breaks)) { # compute breaks from values in x:
		pdx = prod(dim(x[[1]]))
		## take a regular sample from x[[1]]:
		# values = structure(x[[1]], dim = NULL)[seq(1, pdx, length.out = min(pdx, 10000))]
		values = structure(x[[1]], dim = NULL)
		if (isTRUE(logz))
			values = log10(values)
		if (is.factor(values) || is.logical(values))
			values = as.numeric(values)
		values = na.omit(values)
		n.unq = length(unique(values))
		if (n.unq > 2) {
			n = min(nbreaks - 1, n.unq)
			if (breaks == "pretty")
				pretty(values, n)
			else if (breaks == "quantile")
				setNames(quantile(values, seq(0., 1., 1./n)), NULL)
			else if (breaks == "equal")
				seq(min(values), max(values), length.out = n + 1)
			else {
				if (!requireNamespace("classInt", quietly = TRUE))
					stop("package classInt required, please install it first")
				classInt::classIntervals(na.omit(values), n, breaks,
					warnSmallN = FALSE)$brks
			}
		} else if (n.unq == 2) {
			r = range(values)
			d = diff(r)
			c(r[1] - 0.5 * d, mean(r), r[2] + 0.5 * d)
		} else
			range(values, na.rm = TRUE) # lowest and highest!
	} else # breaks was given:
		breaks
}

#' @name plot
#' @param band integer; which band (dimension) to plot
#' @param attr integer; which attribute to plot
#' @param asp numeric; aspect ratio of image
#' @param rgb integer; specify three bands to form an rgb composite. Experimental: rgb color table; see Details.
#' @param maxColorValue numeric; passed on to \link{rgb}
#' @param xlab character; x axis label
#' @param ylab character; y axis label
#' @param xlim x axis limits
#' @param ylim y axis limits
#' @param text_values logical; print values as text on image?
#' @param text_color character; color for printed text values
#' @param interpolate logical; when using \link{rasterImage} (rgb), should pixels be interpolated?
#' @param as_points logical; for curvilinear or sheared grids: parameter passed on to \link{st_as_sf}, determining whether raster cells will be plotted as symbols (fast, approximate) or small polygons (slow, exact)
#' @param logz logical; if \code{TRUE}, use log10-scale for the attribute variable. In that case, \code{breaks} and \code{at} need to be given as log10-values; see examples.
#' @param add.geom object of class \code{sfc}, or list with arguments to \code{plot}, that will be added to an image or sub-image
#' @param border color used for cell borders (only in case \code{x} is a curvilinear or rotated/sheared grid)
#' @param useRaster logical; use the rasterImage capabilities of the graphics device?
#' @param extent object which has a \code{st_bbox} method; sets the plotting extent
#' @details use of an rgb color table is experimental; see https://github.com/r-spatial/mapview/issues/208
#' @export
#' @examples
#' tif = system.file("tif/L7_ETMs.tif", package = "stars")
#' x = read_stars(tif)
#' image(x, col = grey((3:9)/10))
#' if (isTRUE(dev.capabilities()$rasterImage == "yes")) {
#'   image(x, rgb = c(1,3,5)) # false color composite
#' }
image.stars = function(x, ..., band = 1, attr = 1, asp = NULL, rgb = NULL,
		maxColorValue = ifelse(inherits(rgb, "data.frame"), 255, max(x[[attr]], na.rm = TRUE)),
		xlab = if (!axes) "" else names(d)[1], ylab = if (!axes) "" else names(d)[2],
		xlim = st_bbox(extent)$xlim, ylim = st_bbox(extent)$ylim, text_values = FALSE,
		text_color = 'black', axes = FALSE,
		interpolate = FALSE, as_points = FALSE, key.pos = NULL, logz = FALSE,
		key.width = kw_dflt(x, key.pos), key.length = 0.618, add.geom = NULL, border = NA,
		useRaster = isTRUE(dev.capabilities()$rasterImage == "yes"), extent = x) {

	dots = list(...)

	#stopifnot(!has_rotate_or_shear(x)) # FIXME: use rasterImage() with rotate, if only rotate & no shear

	if (any(dim(x) == 1))
		x = adrop(x, drop_xy = TRUE)

	d = st_dimensions(x)
	if (has_raster(x) || has_sfc(x)) {
		if (missing(xlim) || missing(ylim)) {
			force(xlim)
			force(ylim)
		}
		dimxy = attr(d, "raster")$dimensions
		dimx =  dimxy[1]
		dimy =  dimxy[2]
		dimxn = match(dimx, names(d))
		dimyn = match(dimy, names(d))
		is_rectilinear = is_rectilinear(x)
	} else {
		e = expand_dimensions(x)
		if (missing(xlim))
			xlim = range(e[[1]])
		if (missing(ylim))
			ylim = range(e[[2]])
		dimx = names(dim(x))[1]
		dimy = names(dim(x))[2]
		dimxn = 1
		dimyn = 2
		is_rectilinear = (is.na(d[[1]]$delta) || is.na(d[[2]]$delta)) && (!regular_intervals(d[[1]]$values) || !regular_intervals(d[[2]]$values))
		if (missing(axes))
			axes = TRUE
	}

	if (has_sfc(x) && length(dim(x)) == 2) { # time series of features:
		ed = lapply(expand_dimensions(x), function(x) if (inherits(x, "sfc")) seq_along(x) else x)
		names(ed) = c("x", "y")
		ed$z = x[[attr]]
		axes = TRUE
		return(image(ed, ..., xlab = xlab, ylab = ylab))
	}

	if (! is_curvilinear(x)) {
		dims = expand_dimensions(x, center = FALSE, max = FALSE)
		d_max = expand_dimensions(x, center = FALSE, max = TRUE)
		if (tail(dims[[dimx]], 1) != tail(d_max[[dimx]], 1))
			dims[[ dimx ]] = c(dims[[dimx]], tail(d_max[[dimx]], 1))
		if (tail(dims[[dimy]], 1) != tail(d_max[[dimy]], 1))
			dims[[ dimy ]] = c(dims[[dimy]], tail(d_max[[dimy]], 1))

		y_is_neg = all(diff(dims[[ dimy ]]) < 0)
		if (y_is_neg)
			dims[[ dimy ]] = rev(dims[[ dimy ]])
	} 

	if (is.null(asp))
		asp = if (isTRUE(st_is_longlat(x)))
				1 / cos((mean(st_bbox(x)[c(2,4)]) * pi)/180)
			else if (!has_raster(x) && !has_sfc(x))
				NA_real_
			else
				1.0

	ar = unclass(x[[ attr ]]) # raw data matrix/array

	# handle color table, if present:
	co = attr(ar, "colors")
	if (! is.null(co)) {
		if (is.null(rgb))
			rgb = TRUE
		interpretation = attr(co, "interpretation")
		if (!is.null(interpretation) && interpretation != 1)
			warning("color interpretation is not RGB, but rgb is used nevertheless: colors will probably be wrong")
	}

	# rearrange ar:
	others = setdiff(seq_along(dim(ar)), c(dimxn, dimyn))
	ar = aperm(ar, c(dimxn, dimyn, others))
	if (text_values)
		ar_text = ar # keep original order for cell text labels

	if (is.null(rgb) && is.character(ar))
		rgb = TRUE

	if (! is.null(rgb)) {
		if (is_curvilinear(x)) {
			x.sf = st_as_sf(x, as_points = as_points)
			plot_sf = function(x, col, ...) plot(x, ...) # absorb col
			if (!is.null(co)) # #456:
				plot_sf(x.sf, pal = co, border = FALSE, ...)
			else
				plot(x.sf, border = FALSE, ...)
			return()
		}
		xy = dim(ar)[1:2]
		if (! y_is_neg) { # need to flip y?
			ar = if (length(dim(ar)) == 3)
					ar[ , rev(seq_len(dim(ar)[2])), ]
				else
					ar[ , rev(seq_len(dim(ar)[2]))]
		}
		if (!useRaster)
			stop("rgb plotting not supported on this device")
		if (! isTRUE(dots$add)) {
			plot.new()
			plot.window(xlim = xlim, ylim = ylim, asp = asp)
		}
		if (is.numeric(rgb) && length(rgb) == 3) {
			ar = structure(ar[ , , rgb], dim = c(prod(xy), 3)) # flattens x/y
			#nas = apply(ar, 1, function(x) any(is.na(x))) #503
			nas = !complete.cases(ar)
			ar = grDevices::rgb(ar[!nas,], maxColorValue = maxColorValue)
			mat = rep(NA_character_, prod(xy))
			mat[!nas] = ar
			dim(mat) = xy
		} else {
			stopifnot(isTRUE(rgb) || inherits(rgb, "data.frame"))
			# rgb has col 1: index, col 2: label, col 3-5: R, G, B
			# ar = as.vector(ar[ , , 1]) # flattens x/y to 1-D index vector
			mat = if (isTRUE(rgb)) {
					if (!is.null(co))
						structure(co[as.vector(ar)], dim = dim(ar))
					else
						ar
				} else {
					ar = as.vector(ar) # flattens x/y to 1-D index vector
					rgb = grDevices::rgb(rgb[match(ar, rgb[[1]]), 3:5], maxColorValue = maxColorValue)
					structure(rgb, dim = xy)
				}
		}
		myRasterImage = function(x, xmin, ymin, xmax, ymax, interpolate, ..., breaks, add, zlim) # absorbs breaks, add & zlim
			rasterImage(x, xmin, ymin, xmax, ymax, interpolate = interpolate, ...)
		myRasterImage(t(mat), xlim[1], ylim[1], xlim[2], ylim[2], interpolate = interpolate, ...)
	} else if (is_curvilinear(x) || has_rotate_or_shear(x)) {
		x = st_as_sf(x[1], as_points = as_points)
		mplot = function(x, col, ..., zlim) {
			if (missing(col))
				plot(x, ...)
			else
				plot(x, pal = col, ...) # need to swap arg names: FIXME:?
		}
		mplot(x, xlab = xlab, ylab = ylab, xlim = xlim, ylim = ylim, axes = axes, reset = FALSE,
			key.pos = key.pos, key.width = key.width, key.length = key.length, logz = logz,
			border = border, ...)
	} else { # regular & rectilinear grid, no RGB:
		if (y_is_neg) { # need to flip y?
			ar = if (length(dim(ar)) == 2)
					ar[ , rev(seq_len(dim(ar)[2])), drop = FALSE]
				else
					ar[ , rev(seq_len(dim(ar)[2])), band] # FIXME: breaks if more than 3?
		}
		image.default(dims[[ dimx ]], dims[[ dimy ]], ar, asp = asp, xlab = xlab, ylab = ylab,
			xlim = xlim, ylim = ylim, axes = FALSE, useRaster = useRaster && !is_rectilinear, ...)
	}
	if (text_values) {
		dims = expand_dimensions.stars(x, center = TRUE)
		text(do.call(expand.grid, dims[1:2]), labels = as.character(as.vector(ar_text)),
			col = text_color)
	}
	if (axes) { # FIXME: see sf::plot.sf for refinements to be ported here?
		if (!is.null(dots$cex.axis))
			par(cex.axis = dots$cex.axis)
        if (isTRUE(st_is_longlat(x))) {
			if (isTRUE(all.equal(st_bbox(x), st_bbox(), tolerance = .1, 
								 check.attributes = FALSE))) {
				.degAxis(1, at = seq(-180, 180, 60))
				.degAxis(2, at = seq(-90, 90, 30))
			} else {
				.degAxis(1)
				.degAxis(2)
			}
        } else {
            axis(1)
            axis(2)
		}
	}
	if (!is.null(add.geom)) {
		if (inherits(add.geom, c("sf", "sfc")))
			add.geom = list(add.geom)
		do.call(plot, c(add.geom, add = TRUE))
	}
}

# compute the degree of downsampling allowed to still have more than
# one cell per screen/device pixel:
get_downsample = function(dims, px = dev.size("px"), rgb = FALSE) {
	if (rgb)
		dims = dims[1:2]
	if (n <- max(0L, floor(sqrt(prod(dims) / prod(px))) - 1.0))
		message(paste0("downsample set to ", n))
	n
}


#' plot contours of a stars object
#'
#' plot contours of a stars object
#' @param x object of class \code{stars}
#' @param ... other parameters passed on to \link[graphics]{contour}
#' @details this uses the R internal contour algorithm, which (by default) plots contours; \link{st_contour} uses the GDAL contour algorithm that returns contours as simple features.
#' @export
#' @examples
#' d = st_dimensions(x = 1:ncol(volcano), y = 1:nrow(volcano))
#' r = st_as_stars(t(volcano))
#' r = st_set_dimensions(r, 1, offset = 0, delta = 1)
#' r = st_set_dimensions(r, 2, offset = 0, delta = -1)
#' plot(r, reset = FALSE)
#' contour(r, add = TRUE)
contour.stars = function(x, ...) {
	if (!(is_regular_grid(x) || is_rectilinear(x)))
		stop("contour only works for regular or rectilinear grids")
	if (inherits(x[[1]], "units"))
		x[[1]] = units::drop_units(x[[1]])
	x = st_upfront(adrop(x)) # drop singular dimensions, put x/y first
	dx = dim(x)
	if (length(dx) != 2)
		stop("contour only supported for 2-D arrays") # nocov
	e = expand_dimensions(x)
	contour(z = x[[1]][,rev(seq_len(dx[2]))], x = e[[1]], y = rev(e[[2]]), ...)
}

#' reduce dimension to rgb (alpha) hex values
#'
#' @export
#' @param x object of class \code{stars}
#' @param dimension dimension name or number to reduce
#' @param use_alpha logical; if TRUE, the fourth band will be used as alpha values
#' @param maxColorValue integer; maximum value for colors
#' @param stretch logical or character; if \code{TRUE} or \code{"percent"},
#' each band is stretched to 0 ... maxColorValue by "percent clip" method using
#' probs values. If \code{"histogram"}, a "histogram equalization" is performed
#' (\code{probs} values are ignored). If stretch is \code{NULL} or \code{FALSE}, no stretching
#' is performed. Other character values are interpreted as "percent" and a message
#' will be printed.
#' @param probs probability values for quantiles used for stretching by "percent".
#' @seealso \link{st_apply}, \link[grDevices]{rgb}
#' @details the dimension's bands are mapped to red, green, blue, alpha; if a different
#' ordering is wanted, use \link{[.stars} to reorder a dimension, see examples.
#' Alternatively, you can use \link{plot.stars} with the \code{rgb} argument to create a three-band composition.
#' @examples
#' tif = system.file("tif/L7_ETMs.tif", package = "stars")
#' x = read_stars(tif)
#' st_rgb(x[,,,3:1])
#' r = st_rgb(x[,,,c(6,5,4,3)], 3, use_alpha=TRUE) # now R=6,G=5,B=4,alpha=3
#' if (require(ggplot2)) {
#'  ggplot() + geom_stars(data = r) + scale_fill_identity()
#' }
#' r = st_rgb(x[,,,3:1],
#' 		   probs = c(0.01, 0.99),
#' 		   stretch = "percent")
#' plot(r)
#' r = st_rgb(x[,,,3:1],
#' 		   probs = c(0.01, 0.99),
#' 		   stretch = "histogram")
#' plot(r)
st_rgb = function (x,
					dimension = 3,
					use_alpha = dim(x)[dimension] == 4,
					maxColorValue = 255L,
					probs = c(0, 1),
					stretch = NULL) {

	if (inherits(x, "stars_proxy"))
		return(collect(x, match.call(), "st_rgb", 
					   c("x", "dimension", "use_alpha", "maxColorValue", "probs", "stretch"), env = environment())) # RETURNS!!

	# if not stars_proxy:
	if (is.character(dimension))
		dimension = match(dimension, names(dim(x)))
	stopifnot(is.numeric(dimension), length(dimension) == 1)
	if (!dim(x)[dimension] %in% c(3, 4))
		stop(paste("number of bands along dimension", dimension,
				   "should be 3 or 4"))
	dims = setdiff(seq_along(dim(x)), dimension)
	cutoff = function(x, probs, stretch.method = "percent") {
		if (stretch.method == "percent"){
			qs = if (all(probs == c(0, 1)))
				range(x)
			else quantile(x, probs, na.rm = TRUE)
			x = (x - qs[1])/(qs[2] - qs[1])
			x[x > 1] = 1
			x[x < 0] = 0
			x * maxColorValue
		} else if(stretch.method == "histogram"){
			x = stats::ecdf(x)(x)
			x * maxColorValue
		} else {
			qs = range(x)
			(x - qs[1])/(qs[2] - qs[1]) * maxColorValue
		}
	}

	if (is.null(stretch)) {
		stretch.method = "none"
		stretch = FALSE
	}

	if (is.logical(stretch)) {
		if(stretch)
			stretch.method = "percent"
		else
			maxColorValue = max(maxColorValue, max(x[[1]], na.rm = TRUE))
	}

	if (is.character(stretch)) {
		if (!stretch %in% c("percent", "histogram"))
			stretch.method = "percent"
		else
			stretch.method = stretch
		stretch = TRUE
	}

	if (stretch)
		x = st_apply(x, dimension, cutoff, probs = probs, stretch.method = stretch.method)

	if (anyNA(x[[1]])) {
		rgb4 = function(r, g, b, a) {
			r = cbind(as.vector(r), as.vector(g), as.vector(b),
					  as.vector(a))
			sel = !apply(r, 1, anyNA)
			ret = rep(NA_character_, nrow(r))
			ret[sel] = rgb(r[sel, 1:3], alpha = a[sel], maxColorValue = maxColorValue)
			structure(ret, dim = dim(g))
		}
		rgb3 = function(r, g, b) {
			r = cbind(as.vector(r), as.vector(g), as.vector(b))
			sel = !apply(r, 1, anyNA)
			ret = rep(NA_character_, nrow(r))
			ret[sel] = rgb(r[sel, 1:3], maxColorValue = maxColorValue)
			structure(ret, dim = dim(g))
		}
	} else {
		rgb4 = function(r, g, b, a)
			structure(rgb(r, g, b, a,maxColorValue = maxColorValue), dim = dim(r))
		rgb3 = function(r, g, b)
			structure(rgb(r, g, b, maxColorValue = maxColorValue), dim = dim(r))
	}
	st_apply(x, dims, if (use_alpha) rgb4 else rgb3)
}

#' @export
hist.stars = function(x, ..., main = names(x)[1]) {
	hist(x[[1]], ..., main = main)
}

#' @export
hist.stars_proxy = function(x, ..., downsample = 0) {
	x = st_as_stars(x, downsample = downsample)
	NextMethod()
}

## IGNORE_RDIFF_BEGIN
library(stars)
## IGNORE_RDIFF_END

if (suppressPackageStartupMessages(require(sp, quietly = TRUE)) && require(spacetime, quietly = TRUE)) {
data(air) # this loads several datasets in .GlobalEnv
dim(air)
d = st_dimensions(station = st_set_crs(st_as_sfc(stations), 4326), time = dates)

blocks = st_make_grid(st_as_sfc("POLYGON ((5.871619 47.26986, 15.03811 47.26986, 15.03811 55.05653, 5.871619 55.05653, 5.871619 47.26986))", crs = 4326),
	n = c(3,3))

print(aq <- st_as_stars(list(PM10 = air), dimensions = d))
print(a <- aggregate(aq, blocks, mean, na.rm = TRUE))
}

# adapted from ?read_stars:
m = array(1:720, dim = c(x = 10, y = 12, t = 6)) # named dim
st = st_as_stars(m)
attr(st, "dimensions")$y$delta = -1
attr(st, "dimensions")$y$offset = 12
tm = as.Date("2019-02-19") + 1:6
st = st_set_crs(st_set_dimensions(st, 3, values = tm), 4326)

tmp = tempfile(fileext = ".tif")
write_stars(st, tmp)

(red <- setNames(read_stars(tmp, RasterIO = list(nXOff = 1, nYOff = 1, nXsize = 10, nYSize = 12,
   nBufXSize = 2, nBufYSize = 2)), "foo"))

sfc = st_set_crs(st_as_sfc(red, as_points = FALSE), st_crs(st))
(a = aggregate(st, st_sf(a = 1, geom = sfc), mean))
(a = aggregate(st, sfc, mean))
if (requireNamespace("terra", quietly = TRUE) && requireNamespace("exactextractr", quietly = TRUE)) {
 print(a <- aggregate(st, sfc, mean, exact = TRUE))
 print(a[[1]])
 print(sum(a[[1]])*30 == sum(1:720))
 # weights:
 w = st_as_stars(list(w = array(rep(1:5, 24), dim = c(x = 10, y = 12))), dimensions = st_dimensions(st)[1:2])
 print(all.equal(aggregate(st, sfc, mean, exact = TRUE, weights = w)[[1]][1,1], weighted.mean(st[[1]][1:5,1:6,1], rep(1:5, 6))))
 print(sum(aggregate(st, sfc, sum, exact = TRUE, weights = w)[[1]][,1]) == sum(st[[1]][,,1] * rep(1:5, 24)))
 # na.rm: only groups covering an NA cell become NA
 na_st = st
 na_st[[1]][1,1,1] = NA
 a = aggregate(na_st, sfc, mean, exact = TRUE)[[1]]
 print(is.na(a[1,1]) && !anyNA(a[-1,]) && !anyNA(a[,-1]))
 print(all.equal(aggregate(na_st, sfc, mean, exact = TRUE, na.rm = TRUE)[[1]][1,1], mean(na_st[[1]][1:5,1:6,1], na.rm = TRUE)))
}

# transform:
print(all.equal(aggregate(st, sfc, mean, transform = ~ .x^2)[[1]], aggregate(st^2, sfc, mean)[[1]]))
a = aggregate(st, sfc, mean, transform = ~ cbind(lin = .x, sq = .x^2))
print(st_get_dimension_values(a, "term"))
print(all.equal(unname(a[[1]][,,2]), unname(aggregate(st^2, sfc, mean)[[1]][,])))

tm0 = as.Date("2019-02-19") + -1:8
(a = aggregate(st, tm0, mean, na.rm = TRUE))
(a = aggregate(st, "days", mean, na.rm = TRUE))
yd = function(x) as.POSIXlt(x)$yday
(a = aggregate(st, yd, mean, na.rm = TRUE))

# with "by" geometry not overlapping x
pt = st_point(c(-10,-10))
(sfc = c(sfc, st_sfc(pt, crs = st_crs(sfc))))
(a = aggregate(st, sfc, mean))
(a = aggregate(st, st, mean))

# align bb with the first two dimensions of d
st_align = function(bb, d) {
	up1 = function(x) ifelse(sign(x) < 0, -ceiling(abs(x)), floor(x))
	up2 = function(x) ifelse(sign(x) > 0, ceiling(abs(x)), -floor(abs(x)))
	dx = d[[1]]$delta
	dy = d[[2]]$delta
	ox = d[[1]]$offset
	oy = d[[2]]$offset
	dbbx = diff(bb[c("xmin", "xmax")])
	dbby = diff(bb[c("ymin", "ymax")])
	bb["xmin"] = ox - up2((ox - bb["xmin"])/dx) * dx
	bb["ymin"] = oy - up1((oy - bb["ymin"])/dy) * dy
	bb["xmax"] = ox - up1((ox - bb["xmax"])/dx) * dx
	bb["ymax"] = oy - up2((oy - bb["ymax"])/dy) * dy
	bb
}

#' rasterize simple feature geometries
#' 
#' rasterize simple feature geometries
#' @export
#' @param sf object of class \code{sf}
#' @param template optional; stars object with desired target geometry, or target geometry alignment if \code{align=TRUE}; see details
#' @param file temporary file name
#' @param driver driver for temporary file
#' @param options character; options vector for \code{GDALRasterize}
#' @param align logical; if \code{TRUE}, \code{template} is only used for the geometry _alignment_, 
#' informing target resolution and offset 
#' @param proxy logical; should a proxy object be returned?
#' @param ... arguments passed on to \link{st_as_stars}
#' @details if `template` is a `stars` object, non-NA cells that are not covered by `sf` receive the value in `template`; see also argument `align`.
#' @examples
#' demo(nc, echo = FALSE, ask = FALSE)
#' (x = st_rasterize(nc)) # default grid:
#' plot(x, axes = TRUE)
#' # a bit more customized grid:
#' (x = st_rasterize(nc, st_as_stars(st_bbox(nc), nx = 100, ny = 50, values = NA_real_)))
#' plot(x, axes = TRUE)
#' (ls = st_sf(a = 1:2, st_sfc(st_linestring(rbind(c(0.1, 0), c(1.1, 1))),
#'    st_linestring(rbind(c(0, 0.05), c(1, 0.05))))))
#' (grd = st_as_stars(st_bbox(ls), nx = 10, ny = 10, xlim = c(0, 1.0), ylim = c(0, 1),
#'    values = NA_real_))
#' # Only the left-top corner is part of the grid cell:
#' sf_extSoftVersion()["GDAL"]
#' plot(st_rasterize(ls, grd), axes = TRUE, reset = FALSE) # ALL_TOUCHED=FALSE; 
#' plot(ls, add = TRUE, col = "red")
#' plot(st_rasterize(ls, grd, options = "ALL_TOUCHED=TRUE"), axes = TRUE, reset = FALSE)
#' plot(ls, add = TRUE, col = "red")
#' # add lines to existing 0 values, summing values in case of multiple lines:
#' (grd = st_as_stars(st_bbox(ls), nx = 10, ny = 10, xlim = c(0, 1.0), ylim = c(0, 1), values = 0))
#' r = st_rasterize(ls, grd, options = c("MERGE_ALG=ADD", "ALL_TOUCHED=TRUE"))
#' plot(r, axes = TRUE, reset = FALSE)
#' plot(ls, add = TRUE, col = "red")
st_rasterize = function(sf, template = guess_raster(sf, ...) %||% 
			st_as_stars(st_bbox(sf), values = NA_real_, ...), 
		file = tempfile(), driver = "GTiff", options = character(0),
		align = FALSE, proxy = FALSE, ...) {

	if (align) {
		if (missing(template))
			stop("align=TRUE requires template to be specified")
		stopifnot(is_regular_grid(template))
		bb = st_align(st_bbox(sf), d <- st_dimensions(st_normalize(template)))
		dx = d[[1]]$delta
		dy = d[[2]]$delta
		template = st_as_stars(bb, values = NA, dx = dx, dy = dy,
			nx = round(diff(bb[c("xmin", "xmax")])/abs(dx)), 
			ny = round(diff(bb[c("ymin", "ymax")])/abs(dy)),
			proxy = proxy)
	} else
		template = st_normalize(template)
	is_numeric = function(x) is.numeric(x) || is.factor(x)
	isn = sapply(sf, is_numeric)
	sf = if (!any(isn)) {
			sf$ID = seq_len(nrow(sf))
			sf["ID"]
		} else
			sf[isn]
	n_attr = length(which(isn))
	geoms = which(sapply(sf, inherits, "sfc"))
	attrs = as.data.frame(sf)[-geoms]
	if (length(dim(template)) == 2 && n_attr > 1)
		template = st_set_dimensions(merge(do.call(c, lapply(seq_len(n_attr), function(x) template))),
									 3, names(attrs))
	for (i in which(sapply(sf, inherits, "factor"))) # factors:
		sf[[i]] = as.numeric(sf[[i]])
	sf::gdal_rasterize(sf, template, st_geotransform(template), file, driver, options)
	ret = read_stars(file, driver = driver, proxy = proxy)
	if (! proxy) {
		if (length(dim(ret)) > 2)
			ret = split(ret)
		for (i in seq_along(ret)) {
			ret[[i]][is.nan(ret[[i]])] = NA_real_
			if (inherits(sf[[i]], "units"))
				units(ret[[i]]) = units(sf[[i]])
			if (is.factor(attrs[[i]]))
				ret[[i]] = structure(ret[[i]], class = class(attrs[[i]]), levels = levels(attrs[[i]]))
		}
	}
	if (n_attr <= 1) # no names in bands
		setNames(ret, names(attrs))
	else
		ret
}

guess_raster = function(x, ...) {
	if (length(list(...))) # ... hints at other arguments meant at not guessing the raster
		return(NULL)
	get_cell_size = function(x) {
		du = unique(diff(x))
		if (length(du) > 1) {
			if (var(du)/mean(du) < 1e-8) # some fuzz:
				du = mean(du)
			else if (all(du %% min(du) == 0)) # missing cols/rows:
				du = min(du)
			else
				numeric(0) # no regular step size
		} else
			du
	}
	if (all(st_dimension(x) == 0)) { # POINT
		cc = st_coordinates(x)
		ux = sort(unique(cc[,1]))
		if (length(dux <- get_cell_size(ux)) == 0)
			return(NULL);
		uy = sort(unique(cc[,2]))
		if (length(duy <- get_cell_size(uy)) == 0)
			return(NULL);
		bb = st_bbox(x)
		bb = st_bbox(setNames(c(bb["xmin"] - 0.5 * dux, 
			bb["ymin"] - 0.5 * duy, 
			bb["xmax"] + 0.5 * dux, 
			bb["ymax"] + 0.5 * duy), c("xmin", "ymin", "xmax", "ymax")), crs = st_crs(x))
		return(st_as_stars(bb, dx = dux, dy = duy))
	}
	NULL
}


#' @export
#' @param dims the column names or indices that form the cube dimensions
#' @param coords same as dims, for symmetry with \link[sf]{st_as_sf}
#' @param xy the x and y raster dimension names or indices; only takes effect after \code{dims} has been specified, see details
#' @param y_decreasing logical; if TRUE, (numeric) y values get a negative delta (decrease with increasing index)
#' @name st_as_stars
#' @details
#' If \code{xy} is not specified and the first two dimensions in \code{dims} are both numeric,
#' then it is set to these two dimensions.
#' @examples
#' if (require(plm, quietly = TRUE)) {
#'  data(Produc, package = "plm")
#'  st_as_stars(Produc)
#' }
#' if (require(dplyr, quietly = TRUE)) {
#'   # https://stackoverflow.com/questions/77368957/
#' spatial_dim <- st_sf(
#'   ID = 1:3,
#'   geometry = list(
#'     st_polygon(list(
#'       cbind(c(0, 1, 1, 0, 0), c(0, 0, 1, 1, 0))
#'     )),
#'     st_polygon(list(
#'       cbind(c(1, 2, 2, 1, 1), c(0, 0, 1, 1, 0))
#'     )),
#'     st_polygon(list(
#'       cbind(c(2, 3, 3, 2, 2), c(0, 0, 1, 1, 0))
#'     ))
#'   )
#' )
#' weekdays_dim <- data.frame(weekdays = c("Monday", "Tuesday", "Wednesday", 
#'     "Thursday", "Friday", "Saturday", "Sunday"))
#' hours_dim <- data.frame(hours = c("8am", "11am", "4pm", "11pm"))
#' sf_dta <- spatial_dim |>
#'   cross_join(weekdays_dim)|>
#'   cross_join(hours_dim) |>
#'   mutate(population = rnorm(n(), mean = 1000, sd = 200)) |>
#'   select(everything(), geometry)
#' 
#' st_as_stars(sf_dta, dims = c("weekdays", "hours", "geometry"))
#' }
#' demo(nc, echo=FALSE,ask=FALSE)
#' st_as_stars(nc)
#' st_as_stars(st_drop_geometry(nc), dims = "NAME")
#' data.frame(expand.grid(x=1:5, y = 1:5), z = rnorm(25)) |> st_as_stars()
st_as_stars.data.frame = function(.x, ..., dims = coords, xy, y_decreasing = TRUE, coords = 1:2) {

	if (missing(dims) && !missing(xy))
		stop("parameter xy only takes effect when the cube dimensions are set with dims")
	if (is.character(dims))
		dims = match(dims, names(.x))
	if (missing(xy)) {
		if (length(dims) > 1 && all(sapply(.x[dims[1:2]], is.numeric)))
			xy = dims[1:2]
		else 
			xy = c(-1, -1) # cancels y_decreasing
	}
	if (is.character(xy)) {
		xy = match(xy, names(.x))
		if (any(is.na(xy)))
			stop("xy coordinates not found in data")
	}
	stopifnot(length(dims) >= 1, all(dims >= 1), !any(is.na(dims)), is.numeric(xy))
	
	index = NULL
	dimensions = list()
	this_dim = 1
	for (i in dims) {
		v = .x[[i]]
		ix = if (inherits(v, "sfc")) {
    			if (!requireNamespace("digest", quietly = TRUE))
       				stop("package digest required, please install it first") # nocov
				dig = sapply(st_equals(v), digest::digest)
				uv = unique(dig) # no need to sort
				dimensions[[this_dim]] = create_dimension(values = v[match(uv, dig)])
				match(dig, uv) # but look up "hash collision"
			} else {
				suv = if (is.factor(v))
						levels(v)
					else if (is.character(v))
						unique(v)
					else # numeric:
						sort(unique(v), decreasing = y_decreasing && i == xy[2])
				dimensions[[this_dim]] = create_dimension(values = suv, is_raster = i %in% xy)
				match(v, suv)
			}
		index = cbind(index, ix)
		this_dim = this_dim + 1
	}
	names(dimensions) = names(.x)[dims]
	
	raster_xy = if (length(xy) == 2 && all(xy > 0)) 
			names(.x)[xy] 
		else 
			c(NA_character_, NA_character_)
	d = create_dimensions(dimensions, raster = get_raster(dimensions = raster_xy))
	l = lapply(.x[-dims], function(x) {
			m = if (is.factor(x))
					structure(factor(rep(NA_character_, prod(dim(d))), levels = levels(x)),
							  dim = dim(d))
				else 
					array(NA, dim = dim(d))
			m[index] = x # match order
			if (inherits(x, "sfc")) 
				array(st_sfc(m, crs = st_crs(x), precision = st_precision(x)), dim = dim(d))
			else 
				m 
		}
	)
	st_stars(l, d)
}

## $values interface

# if the offset and delta fields are not set, dimension values need
# to be in the $values field of a dimension.
# this can be in two forms:
#   1. curvilinear grid (raster, 2d space only): here, a matrix keeps the values for all x (y) 
#      coordinates for the full set of coordinates
#   2. rectilinear grid (space, but also time or sth else): here, a data.frame with values
#      that specify the start, and that specify the end of the dimensions values

values_get_start = function(x) {
	stopifnot(inherits(x, "dimension"))
	v = x$values
	stopifnot(inherits(v, "intervals") && !is.null(v$start))
	v$start
}

values_get_end = function(x) {
	stopifnot(inherits(x, "dimension"))
	v = x$values
	stopifnot(inherits(v, "intervals") && !is.null(v$end))
	v$end
}

values_get_where = function(x, where) {
	if (where == 0.0)
		values_get_start(x)
	else if (where == 1.0)
		values_get_end(x)
	else {
		stopifnot(inherits(x, "dimension"))
		v = x$values
		stopifnot(inherits(v, "intervals"))
		v$start + where * (v$end - v$start)
	}
}

# x is an object of class dimension
# where specifies, in case of linear mapping, where in [0,1] the value is to be returned: 
#    0 begin, 1 end, 0.5 center
# geotransform: the geotransform (for rotated or sheared raster)
# what: "x" or "y" if geotransform is given
get_dimension_values = function(x, where = 0.0, geotransform, what = NA_character_) {
	stopifnot(inherits(x, "dimension"))
	stopifnot(where >= 0.0 && where <= 1.0)
	if (!is.null(x$values)) {
		if (is.atomic(x$values) || inherits(x$values, "sfc"))
			x$values
		else
			values_get_where(x, where)
	} else {
		if (! any(is.na(geotransform)))
			switch(what,
				x = xy_from_colrow(cbind(seq(x$from, x$to) - 1 + where, 0), geotransform)[,1],
				y = xy_from_colrow(cbind(0, seq(x$from, x$to) - 1 + where), geotransform)[,2],
				stop("argument what needs to be x or y")
			)
		else if (!any(c(is.na(x$offset), is.na(x$delta)))) {
			seq(from = x$offset + (x$from - 1 + where) * x$delta, by = x$delta, length.out = x$to - x$from + 1)
		} else if (!is.na(x$offset) && x$from == 1 && x$to == 1)
			x$offset
		else
			seq(x$from, x$to)
	}
}

#values_range = function(x) {
#	stopifnot(inherits(x, "dimension"))
#	range(as.vector(as.matrix(x$values)))
#}

# sets the values (start, end) field of a dimension, from centers, from start, or from start & end.
# if end is not given, the last two intervals are assumed to be equal sized
# if centers is given, the first two, and the last two intervals are assumed to be equal sized.
set_dimension_values = function(..., start = NULL, end = NULL, centers = NULL) {
	if (!is.null(centers) && (!is.null(start) || !is.null(end)))
		stop("if centers is given, start and end should not be given")
	if (!is.null(centers)) {
		l = length(centers)
		if (l <= 1)
			stop("cannot derive cell boundaries from a single center: specify start and end")
		d = diff(centers)
		start = c(centers - 0.5 * c(d[1], d))
		end = c(start[-1], centers[l] + 0.5 * d[l-1])
	} else if (missing(end)) {
		l = length(start)
		end = if (l == 1)
				NA_real_ #stop("cannot infer end values from a single start value")
			else
				c(start[-1], start[l] + diff(c(start[l-1], start[l])))
	}
	make_intervals(start = start, end = end)
}

#' @export
range.dimension = function(..., na.rm = FALSE) {
	dots = list(...)
	stopifnot(length(dots) == 1)
	x = dots[[1]]
	range(get_dimension_values(x, 0.0, NA, NA_character_), 
		get_dimension_values(x, 1.0, NA, NA_character_))
}

options(rgdal_show_exportToProj4_warnings = "none")
suppressPackageStartupMessages(library(stars))

if (suppressPackageStartupMessages(require(sp, quietly = TRUE)) && require("raster", quietly = TRUE) && 
	suppressPackageStartupMessages(require("terra", quietly = TRUE))) {

tif = system.file("tif/L7_ETMs.tif", package = "stars")
(x = read_stars(tif))
(r = as(x, "Raster"))
(y = st_as_stars(r))

# single band:
x = adrop(x[,,,1])
r = as(x, "Raster")
(y = st_as_stars(r))

# proxy:
(x = read_stars(tif, proxy = TRUE))
(r = as(x, "Raster"))
(y = st_as_stars(r))
(y2 = st_as_stars(r, proxy = TRUE))
}

if (require(terra, quietly = TRUE)) {
## terra -------------
stopifnot(require(terra))
(x = read_stars(tif))
(r = as(x, "SpatRaster"))
(y = st_as_stars(r))

# single band:
x = adrop(x[,,,1])
r = as(x, "SpatRaster")
(y = st_as_stars(r))

# proxy:
(x = read_stars(tif, proxy = TRUE))
(r = as(x, "SpatRaster"))
(y = st_as_stars(r))
(y2 = st_as_stars(r, proxy = TRUE))

f = system.file("tif/lc.tif", package = "stars")
lc = read_stars(system.file("tif/lc.tif", package = "stars"))
levels(lc[[1]])
r = rast(lc)
e = attr(lc[[1]], "exclude")

if (packageVersion("terra") >= "1.5.48") {

print(all.equal(terra::levels(r)[[1]][,2], terra::levels(rast(f))[[1]][!e,2]))
print(all.equal(terra::coltab(r)[[1]][!e,], terra::coltab(rast(f))[[1]][!e,], check.attributes = FALSE))

r<- rast(nrow=10, ncol=10)
set.seed(5)
values(r)<- sample(1:3, size = ncell(r), replace = TRUE)
levels(r) <- data.frame(ID=1:3, category = c("c1", "c2", "c3"))

r2<- rast(nrow=10, ncol=10)
set.seed(6)
values(r2)<- sample(c(0,2,3), size = ncell(r2), replace = TRUE)
levels(r2) <- data.frame(ID=c(0,2,3), category = c("x1", "x2", "x3"))

r3<- c(r,r2)
names(r3)<- c("lyr.1", "lyr.2")

r3_stars<- st_as_stars(r3)

st_redimension(r3_stars) |> print()
rast(r3_stars) |> levels() |> print()

m = matrix(1:121, 11, 11)
bb = st_bbox(c(xmin=0, xmax=11, ymin=0, ymax=11))
s = st_as_stars(bb, values = m)
r = rast(s)
aggregate(r, 2) |> st_as_stars() -> r.s
stopifnot(isTRUE(all.equal(r.s[,1:5,1:5], st_downsample(s, 1, FUN = mean))))
}
}

suppressPackageStartupMessages(library(stars))
if (suppressPackageStartupMessages(require(sp, quietly = TRUE)) && require(spacetime, quietly = TRUE)) {
 data(air) # this loads several datasets in .GlobalEnv
 dim(air)
 d = st_dimensions(station = st_as_sfc(stations), time = dates)
 aq = st_as_stars(list(PM10 = air), dimensions = d)
 image(aperm(log(aq), 2:1), main = "NA pattern (white) in PM10 station time series")
 st_geometry(aq)
}

csv = system.file("tif/ESACCI-LC-Legend.csv", package = "starsdata")
if (csv != "") {
	rgb = read.csv(csv, header = TRUE, sep = ";")
	luc = read_stars(system.file("tif/LUC3.tif", package = "starsdata"))
	plot(luc, rgb = rgb, axes = TRUE, key.pos = NULL)
}

# https://github.com/r-spatial/stars/issues/233 :
nc = read_sf(system.file("shape/nc.shp", package = "sf"))
output = st_as_stars(st_bbox(nc), nx = 1, ny = 1, values = 1)
plot(output)

suppressPackageStartupMessages(library(stars))
nc = read_sf(system.file("gpkg/nc.gpkg", package="sf"))
(x = st_rasterize(nc)) # default grid:
plot(x, axes = TRUE)
# a bit more customized grid:
(x = st_rasterize(nc, st_as_stars(st_bbox(nc), nx = 100, ny = 50, values = NA_real_)))
plot(x, axes = TRUE)
(ls = st_sf(a = 1:2, st_sfc(st_linestring(rbind(c(0.1, 0), c(1.1, 1))), st_linestring(rbind(c(0, 0.05), c(1, 0.05))))))
(grd = st_as_stars(st_bbox(ls), nx = 10, ny = 10, xlim = c(0, 1.0), ylim = c(0, 1), values = NA_real_))
# the following two plots confirm that (only) the
# upper-left corner is part of the grid cell (when dy is negative), leading 
# to a seemingly half-gridcell-shift problem:
## sf_extSoftVersion()["GDAL"]
plot(st_rasterize(ls, grd), axes = TRUE, reset = FALSE) # ALL_TOUCHED=FALSE;
plot(ls, add = TRUE, col = "red")
plot(st_rasterize(ls, grd, options = "ALL_TOUCHED=TRUE"), axes = TRUE, reset = FALSE)
plot(ls, add = TRUE, col = "red")
# add lines to existing 0 values, summing values in case of multiple lines:
(grd = st_as_stars(st_bbox(ls), nx = 10, ny = 10, xlim = c(0, 1.0), ylim = c(0, 1), values = 0))
r = st_rasterize(ls, grd, options = c("MERGE_ALG=ADD", "ALL_TOUCHED=TRUE"))
plot(r, axes = TRUE, reset = FALSE)
plot(ls, add = TRUE, col = "red")

#' subset stars objects
#' 
#' subset stars objects
#' @name stars_subset
#' @param x object of class \code{stars}
#' @param i first selector: integer, logical or character vector indicating attributes to select, or object of class \code{sf}, \code{sfc}, \code{bbox}, or \code{stars} used as spatial selector; see details
#' @param ... further (logical or integer vector) selectors, matched by order, to select on individual dimensions
#' @param drop logical or character; if \code{TRUE}, degenerate dimensions (with only one value) are dropped; if \code{character} name(s) of the degenerate dimension(s) to drop
#' @param crop logical; if \code{TRUE} and parameter \code{i} is a spatial geometry (\code{sf} or \code{sfc}) object, the extent (bounding box) of the result is cropped to match the extent of \code{i} using \link{st_crop}. Cropping curvilinear grids is not supported.
#' @details If \code{i} is an object of class \code{sf}, \code{sfc} or \code{bbox}, the spatial subset covering this geometry is selected, possibly followed by cropping the extent. Array values for which the cell centre is not inside the geometry are assigned \code{NA}. If \code{i} is of class \code{stars}, and attributes of \code{i} are \code{logical}, cells in \code{x} corresponding to \code{NA} or \code{FALSE} cells in \code{i} are assigned an \code{NA}. Dimension ranges containing negative values or \code{NA} may be partially supported. Character selectors are matched against the names of a dimension if it has names, otherwise to the dimension values.
#' @export
#' @examples
#' tif = system.file("tif/L7_ETMs.tif", package = "stars")
#' x = read_stars(tif)
#' x[,,,1:3] # select bands
#' x[,1:100,100:200,] # select x and y by range
#' x["L7_ETMs.tif"] # select attribute
#' xy = structure(list(x = c(293253.999046018, 296400.196497684), y = c(9113801.64775462,
#' 9111328.49619133)), names = c("x", "y"))
#' pts = st_as_sf(data.frame(do.call(cbind, xy)), coords = c("x", "y"), crs = st_crs(x))
#' image(x, axes = TRUE)
#' plot(st_as_sfc(st_bbox(pts)), col = NA, add = TRUE)
#' bb = st_bbox(pts)
#' (xx = x[bb])
#' image(xx)
#' plot(st_as_sfc(bb), add = TRUE, col = NA)
#' image(x)
#' pt = st_point(c(x = 290462.103109179, y = 9114202.32594085))
#' buf = st_buffer(st_sfc(pt, crs = st_crs(x)), 1500)
#' plot(buf, add = TRUE)
#' 
#' buf = st_sfc(st_polygon(list(st_buffer(pt, 1500)[[1]], st_buffer(pt, 1000)[[1]])),
#'    crs = st_crs(x))
#' image(x[buf])
#' plot(buf, add = TRUE, col = NA)
#' image(x[buf, crop=FALSE])
#' plot(buf, add = TRUE, col = NA)
#' # with i of class stars:
#' x[x > 75] # generates lots of NA's; pattern for each band
#' x[x[,,,1] > 75] # recycles a single band template for all bands
"[.stars" = function(x, i = TRUE, ..., drop = FALSE, crop = !is_curvilinear(x)) {
	missing.i = missing(i)

	# special case: i is sf/sfc/bbox
	if (! missing.i && inherits(i, c("sf", "sfc", "bbox"))) {
		x = if (has_raster(x))
			st_crop(x, i, crop = crop, ...)
		else {
			x = st_upfront(x, which_sfc(x))
			sel = which(lengths(st_intersects(st_geometry(x), i)) > 0)
			x[, sel]
		}
		return(x)
	}
	# special case: i is stars
	if (! missing.i && inherits(i, "stars")) {
		stopifnot(all(sapply(i, is.logical)))
		fun = function(x, y) { x[is.na(y) | !y] = NA; x }
		ret = mapply(fun, x, i, SIMPLIFY = FALSE)
		return(st_as_stars(ret, dimensions = st_dimensions(x)))
	}

	d = st_dimensions(x)
	args = rep(list(rlang::missing_arg()), length(dim(x)))

	x = unclass(x)
	if (length(x))
		x = x[i]
	mc = match.call(expand.dots = TRUE)
	# remove [, x, i from mc:
	mc = if (missing.i)
			mc[-(1:2)]
		else
			mc[-(1:3)]
	if ("drop" %in% names(mc))
		mc = mc[-match("drop", names(mc))]

	do_select = FALSE
	for (i in seq_along(mc)) { 
		if ((is.call(mc[[i]]) || is.name(mc[[i]])) && !identical(as.character(mc[[i]]), "")) # try to "get" it:
			mc[[i]] = eval(mc[[i]], parent.frame())
		if (is.numeric(mc[[i]]) || is.call(mc[[i]]) || is.name(mc[[i]]) || is.character(mc[[i]])) { # FIXME: or something else?
			args[[i]] = if (is.character(mc[[i]])) {
						m = if (!is.null(names(d[[i]]$values)))
								match(mc[[i]], names(d[[i]]$values))
							else
								match(mc[[i]], d[[i]]$values)
						if (length(m) == 0 || any(is.na(m)))
							stop("selecting using invalid value label(s)?")
						m
					} else
						mc[[i]]
			do_select = TRUE
		}
	}

	if (is_curvilinear(d)) { # fill the row/col selectors for $values matrices:
		args_xy = rep(list(rlang::missing_arg()), 2)
		xy = attr(d, "raster")$dimensions
		for (i in seq_along(mc)) {
			if (is.numeric(mc[[i]]) || is.call(mc[[i]]) || is.name(mc[[i]])) { # FIXME: or something else?
				if (names(d)[i] == xy[1])
					args_xy[[1]] = mc[[i]]
				if (names(d)[i] == xy[2])
					args_xy[[2]] = mc[[i]]
			}
		}
	}

	# subset arrays:
	args[["drop"]] = FALSE
	for (i in names(x))
		x[[i]] = structure(eval(rlang::expr(x[[i]][ !!!args ])), levels = attr(x[[i]], "levels"),
			colors = attr(x[[i]], "colors"), exclude = attr(x[[i]], "exclude"))

	# now do dimensions:
	if (do_select) {
		ed = expand_dimensions(d)
		xy = attr(d, "raster")$dimensions
		if (is_curvilinear(d)) { # subset curvilinear lat/lon matrices/rasters: can't do one-at-a-time!
			d[[ xy[1] ]]$values = eval(rlang::expr(d[[ xy[1] ]]$values[!!!args_xy]))
			d[[ xy[2] ]]$values = eval(rlang::expr(d[[ xy[2] ]]$values[!!!args_xy]))
		}
		
		# dimensions:
		#mc0 = mc[1:3] # "[", x, first dim
		for (i in seq_along(d)) { # one-at-a-time:
			if (is_CFTime(d[[i]]$refsys)) {
				time = d[[i]]$values
				bnds = CFtime::bounds(time)
				time = CFtime::CFtime(CFtime::definition(time), 
									  CFtime::calendar(time), 
									  CFtime::offsets(time)[ args[[i]] ])
				if (!is.null(bnds))
					CFtime::bounds(time) <- bnds[, args[[i]] ]
				d[[i]]$values = time
				d[[i]]$to = length(args[[i]])
			} else {
				name_i = names(d)[i]
				argi = args[i]
				if (!(is_curvilinear(d) && name_i %in% xy) &&  # as that case was handled above
						!(is.name(argi[[1]]) && all(argi[[1]] == rlang::missing_arg())) &&  # empty arg
						is.numeric(e <- eval(argi[[1]])) && !any(is.na(e)) && !all(diff(e) == 1)) # sequence with gaps
					d[[i]]$values = if (isTRUE(d[[i]]$point) || !is.numeric(unclass(ed[[i]][1])))
							ed[[i]]
						else
							as_intervals(ed[[i]], add_last = TRUE)
				d[[i]] = eval(rlang::expr(d[[i]] [!!!argi]))
			}
		}
	}
	x = st_as_stars(x, dimensions = d)
	if (is.character(drop))
		drop = match(drop, names(dim(x)))
	if (isTRUE(drop))
		adrop(x)
	else if (is.integer(drop) || length(drop) > 1)
		adrop(x, drop = drop)
	else
		x
}

#' @name stars_subset
#' @param downsample downsampling rate used in case \code{i} is a \code{stars_proxy} object
#' @param value array of dimensions equal to those in \code{x}, or a vector or value that will be recycled to such an array
#' @export
#' @details in an assignment (or replacement form, \code{[<-}), argument \code{i} needs to be either (i) a \code{stars} object with logical attribute(s) that has dimensions matching (possibly after recycling) those of \code{x}, in which case the \code{TRUE} cells will be replaced and \code{i} and/or \code{value} will be recycled to the dimensions of the arrays in \code{x}, or (ii) a length-one integer or character vector indicating which array to replace, in which case \code{value} may be stars object or a vector or array (that will be recycled).
#' @examples
#' x = read_stars(tif)
#' # replace, using a logical stars selector: cuts all values above 90 to 90
#' x[x > 90] = 90
#' # replace a single attribute when there are more than one:
#' s = split(x)
#' names(s) = paste0("band", 1:6)
#' # rescale only band 1:
#' s[1] = s[1] * 0.75 
#' # rescale only attribute named "band2":
#' s["band2"] = s["band2"] * 0.85 
#' # create a new attribute from a numeric vector:
#' s["rnorm"] = rnorm(prod(dim(s))) 
#' s
"[<-.stars" = function(x, i, value) {
	if (inherits(i, "stars")) {
		fun = function(x, y, value) { x[y] = value; x }
		st_as_stars(mapply(fun, x, i, value = value, SIMPLIFY = FALSE), dimensions = st_dimensions(x))
	} else if (inherits(i, c("numeric", "character")) && length(i) == 1) {
		if (inherits(value, "stars")) {
			stopifnot(length(value) == 1, first_dimensions_match(x, value))
			value = value[[1]]
		}
		if (is.numeric(i))
			stopifnot(i > 0, i <= length(x))
		y = unclass(x)
		y[[i]] = array(value, dim(x))
		st_as_stars(y, dimensions = st_dimensions(x))
	} else
		stop("selector i should be a stars object or a lenght-one integer or character vector")
}

#' @export
st_intersects.bbox = function(x, y, ...) { # FIXME: segmentize first if geographic coords? see sf::st_sample
	if (!inherits(y, "sfc"))
		y = st_as_sfc(y)
	st_intersects(st_as_sfc(x), y, ...)
}

#' crop a stars object
#' 
#' crop a stars object
#' @name st_crop
#' @export
#' @param x object of class \code{stars}
#' @param y object of class \code{sf}, \code{sfc} or \code{bbox}; see Details below.
#' @param epsilon numeric; factor to shrink the bounding box of \code{y} towards its center before cropping.
#' @param as_points logical; only relevant if \code{y} is of class \code{sf} or \code{sfc}: if \code{FALSE}, treat \code{x} as a set of points, else as a set of small polygons. Default: \code{TRUE} if \code{y} is two-dimensional, else \code{FALSE}; see Details
#' @param ... ignored
#' @param crop logical; if \code{TRUE}, the spatial extent of the returned object is cropped to still cover \code{obj}, if \code{FALSE}, the extent remains the same but cells outside \code{y} are given \code{NA} values.
#' @param normalize logical; if \code{TRUE} then pass the cropped object to \code{\link[sf]{st_normalize}} before returning. This typically changes the `offset` field and resets the `from` field to 1, and changes the bounding box of the returned object accordingly.
#' @details for raster \code{x}, \code{st_crop} selects cells that intersect with \code{y}. 
#' For intersection, are raster cells interpreted as points or as small polygons? 
#' If \code{y} is of class \code{stars}, \code{x} raster cells are interpreted as points; if \code{y} is of class \code{bbox}, \code{x} cells are interpreted as cells (small polygons). Otherwise, if \code{as_points} is not given, cells are interpreted as points if \code{y} has a two-dimensional geometry.
#' 
#' @examples
#' l7 = read_stars(system.file("tif/L7_ETMs.tif", package = "stars"))
#' d = st_dimensions(l7)
#' 
#' # area around cells 3:10 (x) and 4:11 (y):
#' offset = c(d[["x"]]$offset, d[["y"]]$offset)
#' res = c(d[["x"]]$delta, d[["y"]]$delta)
#' bb = st_bbox(c(xmin = offset[1] + 2 * res[1],
#' 	ymin = offset[2] + 11 * res[2],
#' 	xmax = offset[1] + 10 * res[1],
#' 	ymax = offset[2] +  3 * res[2]), crs = st_crs(l7))
#' l7[bb]
#' # equivalent:
#' st_crop(l7, bb)
#' 
#' plot(l7[,1:13,1:13,1], reset = FALSE)
#' image(l7[bb,,,1], add = TRUE, col = sf.colors())
#' plot(st_as_sfc(bb), add = TRUE, border = 'green', lwd = 2)
#' 
#' # slightly smaller bbox:
#' bb = st_bbox(c(xmin = offset[1] + 2.1 * res[1],
#' 	ymin = offset[2] + 10.9 * res[2],
#' 	xmax = offset[1] +  9.9 * res[1],
#' 	ymax = offset[2] +  3.1 * res[2]), crs = st_crs(l7))
#' l7[bb]
#' 
#' plot(l7[,1:13,1:13,1], reset = FALSE)
#' image(l7[bb,,,1], add = TRUE, col = sf.colors())
#' plot(st_as_sfc(bb), add = TRUE, border = 'green', lwd = 2)
#' 
#' # slightly larger bbox:
#' bb = st_bbox(c(xmin = offset[1] + 1.9 * res[1],
#' 	ymin = offset[2] + 11.1 * res[2],
#' 	xmax = offset[1] + 10.1 * res[1],
#' 	ymax = offset[2] +  2.9 * res[2]), crs = st_crs(l7))
#' l7[bb]
#' 
#' plot(l7[,1:13,1:13,1], reset = FALSE)
#' image(l7[bb,,,1], add = TRUE, col = sf.colors())
#' plot(st_as_sfc(bb), add = TRUE, border = 'green', lwd = 2)
#' 
#' # half a cell size larger bbox:
#' bb = st_bbox(c(xmin = offset[1] + 1.49 * res[1],
#' 	ymin = offset[2] + 11.51 * res[2],
#' 	xmax = offset[1] + 10.51 * res[1],
#' 	ymax = offset[2] +  2.49 * res[2]), crs = st_crs(l7))
#' l7[bb]
#' 
#' plot(l7[,1:13,1:13,1], reset = FALSE)
#' image(l7[bb,,,1], add = TRUE, col = sf.colors())
#' plot(st_as_sfc(bb), add = TRUE, border = 'green', lwd = 2)
st_crop.stars = function(x, y, ..., crop = TRUE, epsilon = sqrt(.Machine$double.eps), 
		as_points = all(st_dimension(y) == 2, na.rm = TRUE), normalize = FALSE) {
	x = st_upfront(x) # put spatial dimensions up front; https://github.com/r-spatial/stars/issues/457
	d = dim(x)
	dm = st_dimensions(x)
	args = rep(list(rlang::missing_arg()), length(d)+1)
	if (inherits(y, c("stars", "sf", "sfc", "bbox")) && st_crs(x) != st_crs(y))
		stop("for cropping, the CRS of both objects have to be identical")
	if (inherits(y, "stars")) {
		as_points = TRUE
		y = st_as_sfc(st_bbox(y))
	}
	if (!as.matrix(st_intersects(st_bbox(x), st_bbox(y))))
		warning("st_crop: bounding boxes of x and y do not overlap")
	if (crop && (is_regular_grid(x) || has_rotate_or_shear(x))) {
		rastxy = attr(dm, "raster")$dimensions
		xd = rastxy[1]
		yd = rastxy[2]
		bb = st_bbox(y)
		if (any(is.na(as.numeric(bb)))) # as.numeric() can go after sf 0.7-5
			stop("NA values in bounding box of y")
		if (epsilon != 0)
			bb = bb_shrink(bb, epsilon)
		if (!all(is.na(colrow_from_xy(matrix(bb, 2, byrow = TRUE), dm, NA_outside = TRUE)))) { 
			cr = colrow_from_xy(matrix(bb, 2, byrow = TRUE), dm, NA_outside = FALSE) # FALSE: https://github.com/r-spatial/stars/issues/455
			cr[,1] = cr[,1] - dm[[xd]]$from + 1
			cr[,2] = cr[,2] - dm[[yd]]$from + 1
			for (i in seq_along(d)) {
				if (names(d[i]) == xd)
					args[[i+1]] = seq(max(1, cr[1, 1], na.rm = TRUE), min(d[xd], cr[2, 1], na.rm = TRUE))
				if (names(d[i]) == yd) {
					if (dm[[ yd ]]$delta < 0)
						cr[1:2, 2] = cr[2:1, 2]
					args[[i+1]] = seq(max(1, cr[1, 2], na.rm = TRUE), min(d[yd], cr[2, 2], na.rm = TRUE))
				}
			}
			x = eval(rlang::expr(x[!!!args]))
		}
	} else if (crop)
		warning("crop only crops regular grids: maybe use st_warp() first?")

	if (!inherits(y, "bbox")) { # post-process: burn in geometry mask
		dxy = attr(dm, "raster")$dimensions
		xy_grd = if (is_curvilinear(x) || !as_points) # FIXME: for curvilinear as_points should work too!
				st_as_sfc(st_dimensions(x)[dxy], as_points = as_points, geotransform = st_geotransform(x))
			else
				st_as_sf(do.call(expand.grid, expand_dimensions.stars(x)[dxy]), coords = dxy, crs = st_crs(x))
		inside = st_intersects(st_union(y), xy_grd)[[1]]
		d = dim(x) # cropped x
		mask = rep(TRUE, prod(d[dxy]))
		mask[inside] = FALSE
		mask = array(mask, d) # recycles over dims >= 3
		for (i in seq_along(x))
			x[[i]][mask] = NA
	}
	if (normalize[1]) 
		st_normalize(x)
	else 
		x
}

#' Move raster offset such that grid start at row/col (1,1)
#' 
#' Move raster offset such that grid start at row/col (1,1)
#' @name st_normalize
#' @details the `stars` method "moves" the `offset` of the raster dimensions, such that the `from`
#' field becomes `(1,1)` for the raster dimensions; as a side effect, it moves the raster
#' dimensions `x` and `y` to first and second position.  If the `from` field is already `(1,1)` or the
#' object does not have raster dimensions it does nothing. 
#' @export
st_normalize.stars = function(x, domain = c(0, 0, 1, 1), ...) {
	stopifnot(all(domain == c(0,0,1,1)))
	if (has_raster(x)) {
		x = st_upfront(x)
		d = st_dimensions(x)
		if (d[[1]]$from != 1) {
			d[[1]]$offset = d[[1]]$offset + (d[[1]]$from - 1) * d[[1]]$delta
			d[[1]]$to = d[[1]]$to - d[[1]]$from + 1
			d[[1]]$from = 1
		}
		if (d[[2]]$from != 1) {
			d[[2]]$offset = d[[2]]$offset + (d[[2]]$from - 1) * d[[2]]$delta
			d[[2]]$to = d[[2]]$to - d[[2]]$from + 1
			d[[2]]$from = 1
		}
		st_stars(x, dimensions = d)
	} else
		x
}

#' @name stars_subset
#' @param which character or integer; dimension(s) to be flipped
#' @export
#' @return \code{st_flip} flips (reverts) the array values along the chosen dimension 
#' without(s) changing the dimension properties
#' @examples
#' lc = read_stars(system.file("tif/lc.tif", package = "stars"))
#' x = c(orig = lc, 
#'       flip_x = st_flip(lc, "x"), 
#'       flip_y = st_flip(lc, "y"), 
#'       flip_xy = st_flip(lc, c("x", "y")), 
#'       along = 3)
#' plot(x)
st_flip = function(x, which = 1) {
	if (is.character(which))
		which = match(which, names(dim(x)))
	stopifnot(all(which %in% seq_along(dim(x))))
	dims = lapply(dim(x), seq_len)
	for (i in which)
	  dims[[ i ]] = rev(dims[[ i ]])
	for (i in seq_along(x))
		x[[i]] = do.call(`[`, c(list(x[[i]]), dims))
	x
}

#' get dimensions from stars object
#' @name st_dimensions
#' @export
#' @param .x object to retrieve dimensions information from 
#' @param ... further arguments
#' @param value new object of class \code{dimensions}, with matching dimensions
#' @return the \code{dimensions} attribute of \code{x}, of class \code{dimensions}
st_dimensions = function(.x, ...) UseMethod("st_dimensions")

#' @export
#' @name st_dimensions
st_dimensions.stars = function(.x, ...) attr(.x, "dimensions")

#' @export
st_dimensions.dimensions = function(.x, ...) .x

#' @export
#' @name st_dimensions
`st_dimensions<-` = function(x, value) UseMethod("st_dimensions<-")

#' @export
#' @name st_dimensions
#' @param x object of class \code{dimensions}
`st_dimensions<-.stars` = function(x, value) {
	stopifnot(inherits(value, "dimensions"), length(x) && all(dim(x[[1]]) == dim(value)))
	st_stars(x, value)
}

#' @export
#' @name st_dimensions
`st_dimensions<-.stars_proxy` = function(x, value) {
	if (!is.null(attr(x, "call_list")))
		stop("st_dimensions<- on a stars_proxy object only works if there is no call list")
	structure(NextMethod(), class = class(x))
}


#' @export
#' @name st_dimensions
`st_dimensions<-.list` = function(x, value) {
	st_as_stars(x, dimensions = value)
}


#' @export
#' @name st_dimensions
st_dimensions.array = function(.x, ...) {
	if (length(list(...)) > 0)
		stop("only one argument expected")
	dn = dimnames(.x)

	# raster?
	r = if (all(c("x", "y") %in% names(dn)))
			get_raster(affine = c(0.0, 0.0), dimensions = c("x", "y"), curvilinear = FALSE)
		else
			get_raster(affine = c(0.0, 0.0), dimensions = rep(NA_character_, 2), curvilinear = FALSE)

	ret = if (is.null(dn))
			st_dimensions(list(.x)) # default
		else # try to get dimension names and default values from dimnames(.x):
			create_dimensions(lapply(dn, function(y) create_dimension(values = y)), r)

	if (is.null(names(ret)) || any(names(ret) == ""))
		names(ret) = make.names(seq_along(ret))

	ret
}

#' @export
st_dimensions.matrix = st_dimensions.array


#' @export
#' @name st_dimensions
#' @param .raster length 2 character array with names (if any) of the raster dimensions
#' @param affine numeric; specify parameters of the affine transformation
#' @param cell_midpoints logical; if \code{TRUE} AND the dimension values are strictly regular, the values are interpreted as the cell midpoint values rather than the cell offset values when calculating offset (i.e., the half-cell-size correction is applied); can have a value for each dimension, or else is recycled
#' @param point logical; does the pixel value (measure) refer to a point (location) value or to an pixel (area) summary value?
#' @details dimensions can be specified in two ways. The simplest is to pass a vector with numeric values for a numeric dimension, or character values for a categorical dimension. Parameter \code{cell_midpoints} is used to specify whether numeric values refer to the offset (start) of a dimension interval (default), or to the center; the center case is only available for regular dimensions. For rectilinear numeric dimensions, one can specify either a vector with cell borders (start values), or a data.frame with two columns named "start" and "end", with the respective interval start and end values. In the first case, the end values are computed from the start values by assuming the last two intervals have equal width.
#' 
st_dimensions.default = function(.x, ..., .raster, affine = c(0, 0), 
		cell_midpoints = FALSE, point = FALSE) {
	d = list(...)
	if (! missing(.x))
		d = append(list(.x), d)
	if (missing(.raster)) { 
		.raster = if (all(c("x", "y") %in% names(d)) && is.numeric(d$x) && is.numeric(d$y))
				c("x", "y") 
			else 
				rep(NA_character_, 2)
	}
	create_dimensions(mapply(create_dimension, values = d, is_raster = cell_midpoints, point = point, SIMPLIFY = FALSE),
		raster = get_raster(dimensions = .raster, affine = affine))
}

#' @name st_dimensions
#' @param which integer or character; index or name of the dimension to be changed
#' @param values values for this dimension (e.g. \code{sfc} list-column), or length-1 \code{dimensions} object; setting special value \code{NULL} removes dimension values, for instance to remove curvilinear raster coordinates
#' @param names character; vector with new names for all dimensions, or with the single new name for the dimension indicated by \code{which}
#' @param xy length-2 character vector; (new) names for the \code{x} and \code{y} raster dimensions
#' @export
#' @examples
#' x = read_stars(system.file("tif/L7_ETMs.tif", package = "stars"))
#' # Landsat 7 ETM+ band semantics: https://landsat.gsfc.nasa.gov/the-enhanced-thematic-mapper-plus/
#' # set bands to values 1,2,3,4,5,7:
#' (x1 = st_set_dimensions(x, "band", values = c(1,2,3,4,5,7), names = "band_number", point = TRUE))
#' # set band values as bandwidth
#' rbind(c(0.45,0.515), c(0.525,0.605), c(0.63,0.69), c(0.775,0.90), c(1.55,1.75), c(2.08,2.35)) |>
#'   units::set_units("um") -> bw # or: units::set_units(µm) -> bw
#' # set bandwidth midpoint:
#' (x2 = st_set_dimensions(x, "band", values = 0.5 * (bw[,1]+bw[,2]), 
#'    names = "bandwidth_midpoint", point = TRUE))
#' # set bandwidth intervals:
#' (x3 = st_set_dimensions(x, "band", values = make_intervals(bw), names = "bandwidth"))
st_set_dimensions = function(.x, which, values = NULL, point = NULL, names = NULL, xy, ...) {
	dots = list(...)
	if (inherits(.x, "mdim"))
		stop("for mdim objects, use st_set_dimensions() after st_as_stars()")
	d = st_dimensions(.x)
	if (!missing(which) && is.character(which))
		which = match(which, base::names(d))
	if (! missing(values)) {
		if (is.na(which))
			stop("which should be a name or index of an existing dimensions")
		if (inherits(values, "units") && names(d)[which] %in% attr(d, "raster")$xy 
				&& !is.na(st_crs(.x)))
			stop("units in x/y dimension values only allowed if object has no CRS")
		if (inherits(values, "dimensions")) {
			if (is.null(names))
				names <- names(values)
			values <- values[[1]]$values
		}
		if (!is.null(values) && !inherits(values, "intervals") && dim(.x)[which] != length(values)) {
			if (dim(.x)[which] == length(values) - 1) # create intervals:
				values = as_intervals(values)
			else
				stop(paste("length of values (", length(values), 
					") does not match length of dimension", which, "(", dim(.x)[which], ")"))
		}
		if (is.null(values))
			d[[which]]["values"] = list(NULL) # avoid removing element values
		else {
			d[[which]] = create_dimension(values = values, point = point %||% d[[which]]$point, ...)
			dots = list()
		}
		r = attr(d, "raster")
		if (isTRUE(r$curvilinear)) {
			# FIXME: there's much more that should be checked for curvilinear grids...
			# https://github.com/r-spatial/stars/issues/460
			if (base::names(d)[which] %in% r$dimensions && !any(sapply(d, function(v) is.matrix(v$values)))) {
				attr(d, "raster")$curvilinear = FALSE 
				st_crs(d) = NA
			}
		}
#		else if (inherits(values, "sfc"))
#			base::names(d)[which] = "sfc"
	}
	if (!is.null(point) && is.logical(point)) {
		d[[which]]$point = point
	}
	if (! missing(xy)) {
		if (is.null(xy) || (length(xy) == 1 && is.na(xy)))
			xy = c(NA, NA)
		stopifnot(length(xy) == 2)
		r = attr(d, "raster")
		r$dimensions = as.character(xy)
		attr(d, "raster") = r
	}
	if (!missing(names)) {
		if (length(d) != length(names) && length(names) != 1)
			stop("length of names should match number of dimensions")
		r = attr(d, "raster")
		if (length(d) == length(names)) {
			# handle names in raster attribute, #46
			# problematic again in #379
			if (any(!is.na(r$dimensions)))
				r$dimensions = names[match(r$dimensions, names(d))]
			new_names = names
		} else { # replace the name of dimension `which`, #354
			if (names(d)[which] %in% r$dimensions)
				r$dimensions[match(names(d)[which], r$dimensions)] = names
			new_names = names(d)
			new_names[which] = names
		}
		attr(d, "raster") = r
		base::names(d) = new_names
	}
	if (length(dots))
		d[[which]] = create_dimension(from = 1, to = dim(.x)[which], ...)

	if (inherits(.x, "stars_proxy"))
		structure(.x, dimensions = d)
	else
		st_as_stars(unclass(.x), dimensions = d)
}


#' @name st_dimensions
#' @param where character, one of 'start', 'center' or 'end'. Set to NA (default) to ignore and use \code{max} and \code{center} explictly.  This argument provides a convenient alternative to setting \code{max} and \code{center}.
#' @param max logical; if \code{TRUE} return the end, rather than the beginning of an interval
#' @param center logical; if \code{TRUE} return the center of an interval; if \code{NA} return the center for raster dimensions, and the start of intervals in other cases
#' @export
#' @examples
#' m = matrix(1:20, nrow = 5, ncol = 4)
#' dim(m) = c(x = 5, y = 4) # named dim
#' (s = st_as_stars(m))
#' st_get_dimension_values(s, 'x', where = "start")
#' st_get_dimension_values(s, 'x', center = FALSE)
#' st_get_dimension_values(s, 'x', where = "center")
#' st_get_dimension_values(s, 'x', center = TRUE)
#' st_get_dimension_values(s, 'x', where = "end")
#' st_get_dimension_values(s, 'x', max = TRUE)
st_get_dimension_values = function(.x, which, ..., where = NA, max = FALSE, center = NA) {
	if ((!is.numeric(which) && !is.character(which)) || length(which) != 1)
		stop("argument which should be a length 1 dimension index or name") # nocov
  if (!is.na(where)){
    w = tolower(where[1])
    if (w == 'center'){
      max = FALSE
      center = TRUE
    } else if (w == 'start'){
      max = FALSE
      center = FALSE
    } else if (w == 'end'){
      max = TRUE
      center = NA
    } else {
      stop("where, if not NA, must be 'start', 'center' or 'end': ", where)
    }
  }
  expand_dimensions(.x, ..., max = max, center = center)[[which]]
}


#' @export
"[.dimensions" = function(x, i, j,..., drop = FALSE) {
	raster = attr(x, "raster")
	ret = unclass(x)[i]
	if (isTRUE(all(raster$dimensions %in% names(ret))))
		create_dimensions(ret, raster)
	else
		create_dimensions(ret) # drop raster
}

regular_intervals = function(x, epsilon = options("stars.regular")[[1]] %||% 1.0e-4) {
	if (length(x) <= 1)
		FALSE
	else {
		ud = if (is.atomic(x) && (is.numeric(x) || inherits(x, c("POSIXt", "Date"))))
				unique(diff(x))
			else {
				if (inherits(x, "intervals") && identical(tail(x$end, -1), head(x$start, -1)))
					x$end - x$start
				else
					return(FALSE)
			}
		ud = as.numeric(ud)
		isTRUE((abs(diff(range(ud)) / mean(ud))) < epsilon)
	}
}

create_dimension = function(from = 1, to, offset = NA_real_, delta = NA_real_, 
		refsys = NA_character_, point = NA, values = NULL, is_raster = FALSE)  {

	example = NA

	if (! is.null(values)) { # figure out from values whether we have sth regular:
		from = 1
		to = length(values)
		if (inherits(values, "CFTime") && values$calendar$name %in% CF_calendar_regular) {
			as_date = FALSE
			if (length(values) == 1)
				as_date = values$unit == "days"
			values = values$as_timestamp(asPOSIX = TRUE)
			if (length(values) > 1) {
				dd = diff(values)
				as_date = attr(dd, "units") == "days" && all((as.numeric(dd) %% 1) == 0)
			}
			if (as_date)
				values = as.Date(values)
		}
		if (!(is.character(values) || is.factor(values)) && is.atomic(values)) { 
			if (! all(is.finite(values)))
				stop("dimension value(s) non-finite")
			if (length(values) == 1 || regular_intervals(values)) {
				offset = values[1]
				if (length(values) > 1) {
					#delta = diff(values[1:2])
					delta = mean(diff(values))
				# shift half grid cell size if x or y raster dim cell midpoints:
					if (is_raster)
						offset = offset - 0.5 * delta
				}
				values = NULL
				example = offset
			} else
				example = values
		} else if (inherits(values, "intervals")) {
			example = values$start
		} else if (inherits(values, "CFTime"))
			example = values

		# refsys:
		if (inherits(example, "POSIXct"))
			refsys = "POSIXct"
		else if (inherits(example, "Date"))
			refsys = "Date"
		else if (inherits(example, "CFTime"))
			refsys = paste("CFTime", example$cal$name, sep = "_")
		else if (inherits(example, "units"))
			refsys = "udunits"

		if (inherits(values, "sfc")) {
			point = inherits(values, c("sfc_POINT", "sfc_MULTIPOINT"))
			if (!is.na(st_crs(values)) && is.na(refsys)) # inherit:
				refsys = st_crs(values)
		}
		if (is.numeric(values) && (is.na(point) || !point)) {
			values = if (is_raster)
					set_dimension_values(centers = values)
				else
					set_dimension_values(start = values)
		}
	}
	structure(list(from = unname(from), to = unname(to), offset = unname(offset), 
		delta = unname(delta), refsys = refsys, point = unname(point), values = values), 
		class = "dimension")
}

create_dimensions = function(lst, raster = NULL) {
	if (is.numeric(lst)) # when called with a dim(array) argument:
		lst = setNames(lapply(seq_along(lst), function(i) create_dimension(from = 1, to = lst[i])), 
					names(lst))
	if (is.null(names(lst)))
		names(lst) = make.names(seq_along(lst))
	if (any(names(lst) == "")) {
		sel = which(names(lst) == "")
		names(lst)[sel] = make.names(seq_along(sel))
	}

	if (is.null(raster))
		structure(lst, raster = get_raster(dimensions = c(NA_character_, NA_character_)), class = "dimensions")
	else { 
		d = structure(lst, raster = raster, class = "dimensions")
		rd = raster$dimensions
		if (identical(d[[rd[1]]]$refsys, "udunits") && identical(d[[rd[2]]]$refsys, "udunits")) {
			deg = units(as_units("degree"))
			ux = units(d[[rd[1]]])
			uy = units(d[[rd[2]]])
			u1 = units(units::as_units("1"))
			if (inherits(ux, "symbolic_units") && inherits(uy, "symbolic_units") &&
					(!identical(ux, u1) && !identical(uy, u1)) &&
					units::ud_are_convertible(ux, deg) && units::ud_are_convertible(uy, deg)) {
				d[[rd[1]]] = drop_units(d[[rd[1]]])
				d[[rd[2]]] = drop_units(d[[rd[2]]])
				d[[rd[1]]]$refsys = d[[rd[2]]]$refsys = st_crs('OGC:CRS84') # specifies units
			}
		}
		d
	}
}

get_crs = function(pr) {
	if (!is.null(pr$crs))
		pr$crs
	else if (!is.null(pr$wkt)) # newer sf, but GDAL < 3.0.0
		st_crs(pr$wkt)
	else if (!is.null(pr$proj_wkt))
		st_crs(pr$proj_wkt)
	else 
		st_crs(NA)
}

create_dimensions_from_gdal_meta = function(dims, pr) {
	#if (all(is.na(pr$geotransform)))
	#	pr$geotransform = c(0.0,  1.0,  0.0,  0.0,  0.0, -1.0) # some GTiffs...
	lst = vector("list", length(dims))
	names(lst) = names(dims)
	for (i in names(lst)) {
		lst[[i]] = switch(i,
			x = create_dimension(from = pr$cols[1], to = pr$cols[2], 
				offset = pr$geotransform[1], 
				delta = pr$geotransform[2],
				point = pr$point,
				refsys = get_crs(pr)),
			y = create_dimension(from = pr$rows[1], to = pr$rows[2], 
				offset = pr$geotransform[4],
				delta = pr$geotransform[6],
				point = pr$point,
				refsys = get_crs(pr)),
			# default:
			create_dimension(from = 1, to = dims[i]) # time? depth+units? To be filled in later...
		)
	}
	if (! is.null(pr$dim_extra)) { # netcdf...
		for (d in names(pr$dim_extra)) {
			de = pr$dim_extra[[d]]
			lst[[d]] = create_dimension(from = 1, to = length(de), values = de, point = NA)
		}
		lst[["band"]] = NULL
	}
	# handle band descriptions, if present:
	if (!is.null(lst$band) && !is.null(pr$descriptions) && all(pr$descriptions != ""))
		lst$band$values = pr$descriptions
	else if (!is.null(pr$band_meta)) {
		bm = unlist(pr$band_meta)
		if (any(a <- grepl("DESCRIPTION=", bm)) && length(which(a)) > 1)
			lst$band$values = substring(bm[a], 13)
	}
	# set up raster:
	raster = get_raster(affine = pr$geotransform[c(3,5)],
						dimensions = c("x", "y"),
						curvilinear = FALSE, 
						blocksizes = pr$blocksizes)
	create_dimensions(lst, raster)
}

get_raster = function(affine = rep(0, 2), dimensions = c("x", "y"),
					  curvilinear = FALSE, blocksizes = NULL) {
	if (any(is.na(affine)))
		affine = c(0, 0)
	if (!is.null(blocksizes))
		colnames(blocksizes) = dimensions # columns, rows!
	structure(list(affine = affine,
				   dimensions = dimensions,
				   curvilinear = curvilinear,
				   blocksizes = blocksizes), 
		  class = "stars_raster")
}


#' @export
print.stars_raster = function(x, ...) {
	if (any(is.na(x$affine)))
		cat(paste("affine parameters:", x$affine[1], x$affine[2], "\n"))
	else if (any(x$affine != 0.0))
		cat(paste("sheared raster with parameters:", format(x$affine[1], ...), format(x$affine[2], ...), "\n"))
	if (x$curvilinear)
		cat("curvilinear grid\n")
	invisible(x)
}

get_val = function(pattern, meta) {
	i = grep(pattern, meta)
	if (length(i))
		strsplit(meta[i], "=")[[1]][2]
	else
		NA_character_
}

parse_netcdf_meta = function(pr, name) {
	meta = pr$meta
	spl = strsplit(name, ":")[[1]] # "C:\..." results in length 2:
	name = if (length(spl) < 3) # name is not the variable, but the file name; FIXME: how to make it the variable?
			"zzzzz40163a99980" # bogus string, to avoid match
		else
			tail(spl, 1) # last one contains 
	# longname:
	pr$long_name = get_val(paste0(get_val("NC_GLOBAL#variable_id", meta), "#long_name"), meta)
	# unit:
	pr$units = get_val(paste0(name, "#units"), meta)
	# extra dims: NETCDF_DIM_EXTRA={time,zlev}
	val = get_val("NETCDF_DIM_EXTRA", meta)
	if (! is.na(val)) {
		val = substr(val, 2, nchar(val)-1) # e.g. "{time,depth}" removes { }
		val = strsplit(val, ",")[[1]]
		if (length(val)) {
			pr$dim_extra = vector("list", length(val))
			names(pr$dim_extra) = val
			for (v in val) {
				rhs = get_val(paste0("NETCDF_DIM_", v, "_VALUES"), meta)
				if (!is.na(rhs)) {
					rhs = strsplit(gsub("[\\{\\}]" , "", rhs), ",")
					pr$dim_extra[[v]] = as.numeric(rhs[[1]])
				} else {
					rhs = get_val(paste0("NETCDF_DIM_", v), meta) # nocov # FIXME: find example?
					pr$dim_extra[[v]] = as.numeric(rhs)           # nocov
				}
				u = get_val(paste0(v, "#units"), meta)
				if (! is.na(u)) {
					cal = get_val(paste0(v, "#calendar"), meta)
					if (is.null(cal) || is.na(cal))
						cal = "standard"
					time = try(CFtime::CFtime(u, cal), silent = TRUE)
					if (inherits(time, "CFTime"))
						pr$dim_extra[[v]] = time + pr$dim_extra[[v]] # add the time offsets
					else if (inherits(u <- try_as_units(u), "units"))
						units(pr$dim_extra[[v]]) = u
				}
			}
			pr$dim_extra = rev(pr$dim_extra)
		}
	}
	pr
}

try_as_units = function(u) {
	un = try(suppressWarnings(as_units(u)), silent = TRUE)
	if (inherits(un, "try-error")) # try without ^:
		un = try(suppressWarnings(as_units(gsub("^", "", u, fixed = TRUE))), silent = TRUE)
	if (inherits(un, "try-error")) # try without **
		un = try(suppressWarnings(as_units(gsub("**", "", u, fixed = TRUE))), silent = TRUE)
	if (inherits(un, "try-error") && u != "none") {
		warning(paste("ignoring unrecognized unit:", u), call. = FALSE)
		NULL
	} else
		un
}

parse_gdal_meta = function(properties) {
	point = get_val("AREA_OR_POINT", properties$meta)
	properties$point = switch(point,
	  Area=FALSE,
	  Point=TRUE,
	  NA)
	properties
}

#' expand the dimension values into a list
#'
#' expand the dimension values into a list
#' @param x object of class `stars` or `dimensions`
#' @param ... ignored
#' @export
expand_dimensions = function(x, ...) UseMethod("expand_dimensions")

#' @export
expand_dimensions.stars = function(x, ...) {
	expand_dimensions(st_dimensions(x), ...)
}

#' @export
#' @name expand_dimensions
#' @param max logical; if `TRUE` return the max (end) values of the dimensions intervals
#' @param center logical; if `TRUE` return the center values of intervals, otherwise return offset (start) of intervals; if `NA` (default) return centers for x/y dimensions, offsets for all others
expand_dimensions.dimensions = function(x, ..., max = FALSE, center = NA) {
# returns, in case of numeric dimensions:
# 	center = TRUE: return center values for x and y coordinates, interval start values otherwise
# 	center = FALSE: return start values
#   center = NA: return centers for x/y raster, otherwise start values
#   add_max = TRUE: add in addition to x and y start values an x_max and y_max end values


	if (length(center) == 1)
		center = setNames(rep(center, length(x)), names(x))
	if (length(max) == 1)
		max = setNames(rep(max, length(x)), names(x))

	if (is.list(max))
		max = unlist(max)
	if (is.list(center))
		center = unlist(center)

	if (any(max & center, na.rm = TRUE))
		stop("only one of max and center can be TRUE, not both")

	where = setNames(rep(0.0, length(x)), names(x)) # offset
	for (i in seq_along(x)) {
		if (max[i])
			where[i] = 1.0
		if (isTRUE(center[i]))
			where[i] = 0.5
	}

	dimensions = x
	xy = attr(x, "raster")$dimensions
	gt = st_geotransform(x)
	lst = vector("list", length(dimensions))
	names = names(dimensions)
	names(lst) = names
	if (! is.null(xy) && all(!is.na(xy))) { # we have raster: where defaulting to 0.5
		where[xy] = ifelse(!max[xy] & (is.na(center[xy]) | center[xy]), 0.5, where[xy])
		if (xy[1] %in% names) # x
			lst[[ xy[1] ]] = get_dimension_values(dimensions[[ xy[1] ]], where[[ xy[1] ]], gt, "x")
		if (xy[2] %in% names)  # y
			lst[[ xy[2] ]] = get_dimension_values(dimensions[[ xy[2] ]], where[[ xy[2] ]], gt, "y")
	}

	if ("crs" %in% names)
		lst[[ "crs" ]] = dimensions[[ "crs" ]]$values
	
	is_time = sapply(dimensions, function(d) inherits(d$values, "CFTime"))
	if (all(is.logical(is_time)) && length(time_dim <- which(is_time))) {
		lst[[ time_dim ]] = dimensions[[ time_dim ]]$values$offsets
		time_name = names[time_dim]
	} else
		time_name = ""

	for (nm in setdiff(names, c(xy, "crs", time_name))) # non-xy, non-crs, non-time dimensions
		lst[[ nm ]] = get_dimension_values(dimensions[[ nm ]], where[[nm]], NA, NA)

	lst
}


#' @export
dim.dimensions = function(x) {
	if (is_curvilinear(x))
		setNames(sapply(x, function(x) { x$to - x$from + 1 } ), names(x))
	else
		lengths(expand_dimensions(x)) # FIXME: optimise?
}

#' @name print_stars
#' @param digits number of digits to print numbers
#' @param usetz logical; used to format \code{POSIXct} values
#' @param stars_crs maximum width of string for CRS objects
#' @param all logical; if \code{TRUE} print also fields entirely filled with \code{NA} or \code{NULL}
#' @export
as.data.frame.dimensions = function(x, ..., digits = max(3, getOption("digits")-3), usetz = TRUE, stars_crs = getOption("stars.crs") %||% 28, all = FALSE) {
	mformat = function(x, ..., digits) {
		if (inherits(x, "POSIXct"))
			format(x, ..., usetz = usetz)
		else
			format(x, digits = digits, ...) 
	}
	abbrev_dim = function(y) {
		if (inherits(y$values, "CFTime")) {
			rng = range(y$values, ...)
			y$values = paste0(rng[1], ",...,", rng[2])
		} else if (length(y$values) > 3 || (inherits(y$values, "sfc") && length(y$values) > 2)) {
			y$values = if (is.array(y$values))
					paste0("[", paste(dim(y$values), collapse = "x"), "] ", 
						mformat(min(y$values), digits = digits), ",...,", 
						mformat(max(y$values), digits = digits))
				else if (inherits(y$values[[1]], "crs"))
					paste0(format(y$values[[1]]), ",...,", format(y$values[[length(y$values)]]))
				else if (inherits(y$values, "sfc"))
					paste0(format(y$values[[1]], width = stars_crs), ",...,", 
						   format(y$values[[length(y$values)]], width = stars_crs))
				else
					paste0(format(head(y$values, 1)), ",...,", 
						format(tail(y$values, 1)))
		}
		if (is.na(y$refsys))
			y$refsys = NA_character_
		else if (nchar(tail(format(y$refsys), 1)) > stars_crs)
			y$refsys = paste0(substr(tail(format(y$refsys), 1), 1L, stars_crs - 3), "...")
		y
	}
	lst = lapply(x, abbrev_dim)
	lst = lapply(lst, function(x) sapply(x, mformat, digits = digits))
	ret = data.frame(do.call(rbind, lst), stringsAsFactors = FALSE)
	if (! all) { # remove fields entirely NA or NULL:
		if (all(ret$offset == "NA"))
			ret$offset = NULL
		if (all(ret$delta == "NA"))
			ret$delta = NULL
		if (all(ret$refsys == "NA"))
			ret$refsys = NULL
		if (all(ret$point == "NA"))
			ret$point = NULL
		if (all(ret$values == "NULL"))
			ret$values = NULL
	}
	r = attr(x, "raster")
	if (! any(is.na(r$dimensions))) {
		ret$raster = rep("", nrow(ret))
		ret[r$dimensions[1], "raster"] = "[x]"
		ret[r$dimensions[2], "raster"] = "[y]"
		names(ret)[ncol(ret)] = "x/y"
	}
	ret
}

#' @name print_stars
#' @export
print.dimensions = function(x, ...) {
	ret = as.data.frame(x, ...)
	print(ret)
	print(attr(x, "raster"))
	invisible(x)
}

identical_dimensions = function(lst, ignore_resolution = FALSE, tolerance = 0) {
	if (length(lst) > 1) {
		d1 = attr(lst[[1]], "dimensions")
		crs1 = st_crs(d1)
		st_crs(d1) = NA_crs_
		for (i in 2:length(lst)) {
			di = attr(lst[[i]], "dimensions")
			if (st_crs(di) != crs1) # check semantical equivalence; https://github.com/r-spatial/stars/issues/703
				return(FALSE)
			st_crs(di) = NA_crs_
			if (ignore_resolution) {
				for (j in seq_along(d1))
					d1[[j]]$delta = d1[[j]]$to = NA_real_
				for (j in seq_along(di))
					di[[j]]$delta = di[[j]]$to = NA_real_
				attr(d1, "raster")$blocksizes = NULL
				attr(di, "raster")$blocksizes = NULL
			}
			if (! isTRUE(all.equal(d1, di, tolerance = tolerance, check.attributes = FALSE)))
				return(FALSE)
		}
	}
	TRUE
}

#' @method all.equal dimensions
#' @export
all.equal.dimensions = function(target, current, ..., ignore_blocksizes = TRUE) {
	if (ignore_blocksizes) {
		attr(target, "raster")$blocksizes = NULL
		attr(current, "raster")$blocksizes = NULL
	}
	NextMethod()
}

combine_dimensions = function(dots, along, check_dims_identical = TRUE) {
	dims = st_dimensions(dots[[1]])
	if (along > length(dims)) { # along new dimension:
		if (length(dots) > 1 && check_dims_identical) {
			for (i in 2:length(dots))
				if (! isTRUE(report <- all.equal(dims, st_dimensions(dots[[i]])))) {
					print(report)
					stop(paste("dimensions of element", 1, "and", i, "are not identical"))
				}
		}
		dims[[along]] = create_dimension(from = 1, to = length(dots), values = names(dots))
	} else {
		no_values = all(sapply(dots, function(x) is.null(st_dimensions(x)[[along]]$values)))
		dims[[along]] = if (no_values)
				create_dimension(from = 1, to = sum(sapply(dots, function(x) dim(x)[along])))
			else {
				values = do.call(c, lapply(dots, function(y) st_get_dimension_values(y, along)))
				create_dimension(values = values, point = dims[[along]]$point)
			}
	}
	dims
}

#' @export
seq.dimension = function(from, ..., center = FALSE) { # does what expand_dimensions also does, for single dimension
	get_dimension_values(from, where = ifelse(center, 0.5, 0.0), NA, what = NA_character_)
}

#' @export
`[.dimension` = function(x, i, ...) {
	if (!missing(i)) {
		if (is_CFTime(x$refsys)) {
			ind = x$values$indexOf(i)
			newcf = attr(ind, "CFTime")
			x$to = length(newcf)
			x$values = newcf
			return(x)
		}
		
		if (!is.null(x$values) && !is.matrix(x$values))
			x$values = x$values[i]
		if (!is.na(x$from)) {
			rang = x$from:x$to # valid range
			if (!all(is.na(i)) && max(i, na.rm = TRUE) > -1 && min(i, na.rm = TRUE) < 0)
				stop("cannot mix positive and negative indexes")
			if (is.logical(i))
				i = which(i)
			else if (all(i < 0))
				i = setdiff(rang, abs(i)) # subtract
			if (!any(is.na(i)) && all(diff(i) == 1)) {
				if (max(i) > length(rang))
					stop("invalid range selected")
				sel = rang[i]
				x$from = min(sel)
				x$to = max(sel)
			} else { # invalidate offset & delta
				x$delta = x$offset = NA
				x$from = 1
				x$to = length(i)
			}
		}
	}
	x
}

#' @export
`[<-.dimensions` = function(x, i, value) {
	# stopifnot(length(i) == length(value))
	if (!is.null(value))
		create_dimensions(NextMethod(), raster = attr(x, "raster"))
	else
		NextMethod()
}

#' @export
dimnames.stars = function(x) {
	names(st_dimensions(x))
}

#' @export
`dimnames<-.stars` = function(x, value) {
	for (i in seq_along(x))
		names(dim(x[[i]])) = value
	names(attr(x, "dimensions")) = value
	x
}

#' @export
as.POSIXct.stars = function(x, ...) {
	d = st_dimensions(x)
	e = expand_dimensions(d)
	for (i in seq_along(d)) {
		if (inherits(e[[i]], c("Date", "POSIXt", "udunits"))) {
			p = try(as.POSIXct(e[[i]]), silent = TRUE)
			if (!inherits(p, "try-error"))
				d[[i]] = create_dimension(values = p)
		}
		if (inherits(d[[i]]$values, "CFTime")) {
			warning("converting between incompatible calendars: results are approximate")
			tm = d[[i]]$values
			p = as.POSIXct(tm$format())
			d[[i]] = create_dimension(values = p)
		}
	}
	structure(x, dimensions = d)
}

#' @export
as.Date.stars = function(x, ...) {
	d = st_dimensions(x)
	e = expand_dimensions(d)
	for (i in seq_along(d)) {
		if (inherits(e[[i]], c("POSIXt", "udunits"))) {
			p = try(as.Date(e[[i]]), silent = TRUE)
			if (!inherits(p, "try-error"))
				d[[i]] = create_dimension(values = p)
		}
		if (inherits(d[[i]]$values, "CFTime")) {
			warning("converting between incompatible calendars: results are approximate")
			tm = d[[i]]$values
			p = as.Date(tm$format())
			d[[i]] = create_dimension(values = p)
		}
	}
	structure(x, dimensions = d)
}

#' @export
drop_units.dimension = function(x) {
	du = function(y) {
		if (inherits(y, "units"))
			units::drop_units(y)
		else if (inherits(y, "intervals")) {
			y$start = du(y$start)
			y$end = du(y$end)
			y
		} else
			y
	}
	if (!is.null(x$offset))
		x$offset = du(x$offset)
	if (!is.null(x$delta))
		x$delta  = du(x$delta)
	if (!is.null(x$values))
		x$values = du(x$values)
	x
}

#' @export
units.dimension = function(x) {
	if (inherits(x$offset, "units"))
		units(x$offset)
	else if (inherits(x$delta, "units"))
		units(x$delta)
	else if (inherits(x$values, "units"))
		units(x$values)
	else if (inherits(x$values$start, "units"))
		units(x$values$start)
	else
		NULL
}

context("aggregate.stars with weights and transform")

make_test_data = function() {
	skip_if_not_installed("exactextractr")
	skip_if_not_installed("terra")

	tif = system.file("tif/L7_ETMs.tif", package = "stars")
	x = read_stars(tif)[, 1:30, 1:30, 1]

	bb = sf::st_bbox(x)
	midx = mean(c(bb["xmin"], bb["xmax"]))
	midy = mean(c(bb["ymin"], bb["ymax"]))
	p1 = sf::st_polygon(list(rbind(
		c(bb["xmin"], bb["ymin"]), c(midx, bb["ymin"]),
		c(midx, midy), c(bb["xmin"], midy),
		c(bb["xmin"], bb["ymin"]))))
	p2 = sf::st_polygon(list(rbind(
		c(midx, midy), c(bb["xmax"], midy),
		c(bb["xmax"], bb["ymax"]), c(midx, bb["ymax"]),
		c(midx, midy))))
	polys = sf::st_sfc(p1, p2, crs = sf::st_crs(x))

	list(x = x, polys = polys)
}

test_that("exact = TRUE without weights/transform is backwards compatible", {
	d = make_test_data()

	a_mean = aggregate(d$x, d$polys, mean, exact = TRUE)
	a_sum = aggregate(d$x, d$polys, sum, exact = TRUE)

	expect_s3_class(a_mean, "stars")
	expect_s3_class(a_sum, "stars")
	expect_true(all(is.finite(c(a_mean[[1]]))))
	expect_true(all(is.finite(c(a_sum[[1]]))))
})

test_that("transform one-to-one captures Jensen's inequality", {
	d = make_test_data()

	a_linear = aggregate(d$x, d$polys, mean, exact = TRUE)
	a_squared = aggregate(d$x, d$polys, mean, exact = TRUE, transform = ~ .x^2)

	expect_true(all(c(a_squared[[1]]) >= c(a_linear[[1]])^2 - 1e-6))
	expect_true(any(c(a_squared[[1]]) > c(a_linear[[1]])^2 + 1))
})

test_that("transform accepts function and formula equivalently", {
	d = make_test_data()

	a_form = aggregate(d$x, d$polys, mean, exact = TRUE, transform = ~ .x^2)
	a_func = aggregate(d$x, d$polys, mean, exact = TRUE, transform = function(z) z^2)

	expect_equal(c(a_form[[1]]), c(a_func[[1]]))
})

test_that("one-to-many transform appends a term dimension with given names", {
	d = make_test_data()

	a = aggregate(d$x, d$polys, mean, exact = TRUE,
		transform = ~ cbind(lin = .x, sq = .x^2, cu = .x^3))

	expect_s3_class(a, "stars")
	expect_true("term" %in% names(st_dimensions(a)))
	expect_equal(dim(a)[["term"]], 3L)
	expect_equal(st_get_dimension_values(a, "term"), c("lin", "sq", "cu"))
})

test_that("one-to-many transform falls back to t1, t2, ... when colnames are NULL", {
	d = make_test_data()

	a = aggregate(d$x, d$polys, mean, exact = TRUE,
		transform = ~ cbind(.x, .x^2, .x^3))

	expect_equal(st_get_dimension_values(a, "term"), c("t1", "t2", "t3"))
})

test_that("one-to-many transform with partial colnames defaults all terms to t1, t2, ...", {
	d = make_test_data()

	# a single unnamed column makes every term fall back to t1..tk, not just the unnamed one
	a = aggregate(d$x, d$polys, mean, exact = TRUE,
		transform = ~ cbind(a = .x, .x^2))

	expect_equal(st_get_dimension_values(a, "term"), c("t1", "t2"))
})

test_that("weights produces a different result from unweighted", {
	d = make_test_data()

	w_raster = methods::as(d$x, "SpatRaster")
	terra::values(w_raster) = seq_len(terra::ncell(w_raster))

	a_unweighted = aggregate(d$x, d$polys, mean, exact = TRUE)
	a_weighted = aggregate(d$x, d$polys, mean, exact = TRUE, weights = w_raster)

	expect_false(isTRUE(all.equal(c(a_unweighted[[1]]), c(a_weighted[[1]]))))
})

test_that("weighted mean matches hand calculation", {
	skip_if_not_installed("exactextractr")
	skip_if_not_installed("terra")

	tif = system.file("tif/L7_ETMs.tif", package = "stars")
	x = read_stars(tif)[, 1:4, 1:4, 1]
	x[[1]][] = as.numeric(1:16)

	bb = sf::st_bbox(x)
	pad = (bb["xmax"] - bb["xmin"]) / 100
	p = sf::st_polygon(list(rbind(
		c(bb["xmin"] - pad, bb["ymin"] - pad),
		c(bb["xmax"] + pad, bb["ymin"] - pad),
		c(bb["xmax"] + pad, bb["ymax"] + pad),
		c(bb["xmin"] - pad, bb["ymax"] + pad),
		c(bb["xmin"] - pad, bb["ymin"] - pad))))
	polys = sf::st_sfc(p, crs = sf::st_crs(x))

	w_raster = methods::as(x, "SpatRaster")

	a = aggregate(x, polys, mean, exact = TRUE, weights = w_raster)

	# polygon strictly encloses every cell, so coverage_fraction = 1 throughout;
	# weights = data per cell, so weighted_mean = sum(data^2) / sum(data) = 1496 / 136 = 11
	expect_equal(as.numeric(a[[1]]), sum((1:16)^2) / sum(1:16), tolerance = 1e-9)
})

test_that("all-zero weights raise an error rather than silently returning NaN", {
	skip_if_not_installed("exactextractr")
	skip_if_not_installed("terra")

	d = make_test_data()
	w0 = methods::as(d$x, "SpatRaster")
	terra::values(w0) = 0

	expect_error(
		aggregate(d$x, d$polys, mean, exact = TRUE, weights = w0),
		"all zero")
})

test_that("weights with stars_proxy input errors cleanly", {
	d = make_test_data()

	tif = system.file("tif/L7_ETMs.tif", package = "stars")
	p = read_stars(tif, proxy = TRUE)
	w = methods::as(d$x, "SpatRaster")

	expect_error(
		aggregate(p, d$polys, mean, exact = TRUE, weights = w),
		"not supported for stars_proxy")
})

test_that("weights warns when exact = FALSE", {
	d = make_test_data()
	w_raster = methods::as(d$x, "SpatRaster")

	expect_warning(
		aggregate(d$x, d$polys, mean, weights = w_raster),
		"weights is ignored")
})

test_that("transform applies in the non-exact path as well", {
	d = make_test_data()

	a = aggregate(d$x, d$polys, mean, exact = FALSE, transform = ~ .x^2)

	expect_s3_class(a, "stars")
})

test_that("na.rm drops NA cells from the mean denominator, not just numerator", {
	skip_if_not_installed("exactextractr")
	skip_if_not_installed("terra")

	bb = sf::st_bbox(c(xmin = 0, ymin = 0, xmax = 2, ymax = 2), crs = 4326)
	x = st_as_stars(bb, nx = 2, ny = 2, values = NA_real_)
	x[[1]][] = c(10, 20, 30, NA)
	poly = sf::st_as_sfc(bb)

	# polygon covers all four cells (coverage 1); the mean must average the three
	# non-NA cells, not divide their sum by all four
	a = aggregate(x, poly, mean, exact = TRUE, na.rm = TRUE)
	expect_equal(as.numeric(a[[1]]), mean(c(10, 20, 30)), tolerance = 1e-9)

	# a heavily weighted NA cell must drop from both sides
	x[[1]][] = c(20, 20, 20, NA)
	w = x
	w[[1]][] = c(1, 1, 1, 5)
	aw = aggregate(x, poly, mean, exact = TRUE,
		weights = methods::as(w, "SpatRaster"), na.rm = TRUE)
	expect_equal(as.numeric(aw[[1]]), 20, tolerance = 1e-9)
})

skip_on_cran()
context("read ncdf")
f <- system.file("nc/reduced.nc", package = "stars")

test_that("basic reduced comes back as expected", {
  skip_if_not_installed("ncmeta")
  nc <- read_ncdf(f)
  expect_equal(names(nc), c("sst", "anom", "err", "ice"))
  st_dim <- st_dimensions(nc)
  expect_equal(names(st_dim), c("lon", "lat", "zlev", "time"))

  expect_equal(st_dim$lon$delta, 2)
#  expect_equal(length(st_dim$zlev$values), 1)
#  expect_equal(length(st_dim$time$values), 1)
})

test_that("variable subsetting", {
  skip_if_not_installed("ncmeta")
  nc <- read_ncdf(f, var = c("anom"))
  expect_equal(names(nc), "anom")
})

test_that("domain subsetting", {
  skip_if_not_installed("ncmeta")
  nc <- read_ncdf(f, ncsub = cbind(start = c(20, 1, 1, 1),
                                   count = c(10, 12, 1, 1)))
  st_dim <- st_dimensions(nc)
  expect_equal(st_dim$lon$to - st_dim$lon$from, 9)
  expect_equal(st_dim$lat$to - st_dim$lat$from, 11)
  expect_equal(st_dim$lon$offset, 37)

  expect_error(nc <- read_ncdf(f, ncsub = cbind(start = c(1, 1, 1, 1),
                                   count = c(200, 12, 1, 1))),
               "start or count out of bounds")

  expect_error(nc <- read_ncdf(f, ncsub = cbind(start = c(1, 1, 1),
                                                count = c(200, 12, 1))),
               "input ncsub doesn't match available dims")




  # Leaving this here -- NA or -1 counts should return all but this causes other errors.
  # nc <- read_ncdf(f, ncsub = cbind(start = c(1, 1, 1, 1),
  #                                  count = c(NA, NA, 1, 1)))
  # st_dim <- st_dimensions(nc)
  # expect_equal(st_dim$lon$to - st_dim$lon$from, c("lon" = 179))
  # expect_equal(st_dim$lat$to - st_dim$lat$from, c("lat" = 89))
})

test_that("normal bcsd", {
  skip_if_not_installed("ncmeta")
  f <- system.file("nc/bcsd_obs_1999.nc", package = "stars")
  nc <- read_ncdf(f, package = "stars")
  expect_equal(names(nc), c("pr", "tas"))
  st_dim <- st_dimensions(nc)
  expect_equal(names(st_dim), c("longitude", "latitude", "time"))
  expect_equal(st_dim$longitude$delta, 0.125)
  
  file.copy(f, (tf <- tempfile()))
  
  nc_tf <- RNetCDF::open.nc(tf, write = TRUE)
  
  RNetCDF::var.put.nc(nc_tf, "longitude", 
  					(360 + stars::st_get_dimension_values(nc, "longitude")))
  
  RNetCDF::close.nc(nc_tf)
  
  nc <- read_ncdf(tf, package = "stars")
  
  expect_equal(stars::st_get_dimension_values(nc, "longitude")[1], -84.9375)
})

test_that("non canonical axis order is handled right", {
  skip_if_not_installed("ncmeta")
  expect_warning(nc <- read_ncdf(system.file("nc/3B42_Daily.19991231.7.test.nc",
                                             package = "stars")),
                 "Non-canonical axis order found, attempting to correct.")
  expect_equal(names(st_dimensions(nc)), c("lon", "lat"))
  expect_equal(st_dimensions(nc)[[1]]$to, 4L)
  expect_equal(st_dimensions(nc)[[2]]$to, 5L)
})


test_that("euro cordex extra dimvars", {
  skip_if_not_installed("ncmeta")
  f <- unzip(zipfile = system.file("nc/EURO-CORDEX_81_DOMAIN000_mask_int.nc.zip", package = "stars"), exdir = tempdir())

  expect_warning(out <- read_ncdf(f, var = "mask"),
                 "Didn't find a longitude of prime meridian for datum, assuming 0.")

  expect_equal(names(out), c("mask"))

  #expect_true(sf::st_crs(out) ==
  #sf::st_crs("+proj=lcc +lat_1=30 +lat_2=65 +lat_0=48 +lon_0=9.75 +x_0=-6000 +y_0=-6000 +a=6371229 +b=6371229 +units=m +no_defs"))

  expect_error(out <- suppressWarnings(read_ncdf(f, var = "mask", curvilinear = c("X", "Y"))),
                 "Curvilinear variables not found in file.")

  out1 <- suppressWarnings(read_ncdf(f, var = "mask", curvilinear = c("xlat", "xlon")))
  out2 <- suppressWarnings(read_ncdf(f, var = "mask", curvilinear = c("xlon", "xlat")))

  expect_equal(sf::st_crs(out1), sf::st_crs(NULL))
  
  expect_true(attr(st_dimensions(out1), "raster")$curvilinear)
  expect_true(attr(st_dimensions(out2), "raster")$curvilinear)

  expect_equal(st_dimensions(out1), st_dimensions(out2))
})

test_that("curvilinear", {
  skip_if_not_installed("ncmeta")
  f <- system.file("nc/test_stageiv_xyt.nc", package = "stars")

  out <-read_ncdf(f, curvilinear = c(X = "lon", Y = "lat"))

  # expect_match(warn[1], "bounds for time seem to be reversed; reverting them")

  st_dim <- st_dimensions(out)

  expect_true(all(st_dim$x$values < -74 & st_dim$x$values > -81))

  expect_true(all(st_dim$y$values < 38 & st_dim$y$values > 32))

  expect_equal(dim(st_dim$y$values), setNames(c(87, 118), c("x", "y")))
  
  #nc <- RNetCDF::open.nc(f)
  # Below test will always fail because of different timestamp formats: "2018-09-13 19:00:00 UTC" vs "2018-09-13T19:00:00" etc
  #expect_equal(st_get_dimension_values(st_dim, "time"), 
  #			 RNetCDF::utcal.nc(RNetCDF::att.get.nc(nc, "time", "units"),
  #			 				  RNetCDF::var.get.nc(nc, "time"), type = "c"))
  #RNetCDF::close.nc(nc)
  
  # Should also find the curvilinear grid.
  suppressWarnings(out <- read_ncdf(f, var = "Total_precipitation_surface_1_Hour_Accumulation"))

  expect_true(attr(st_dimensions(out), "raster")$curvilinear)

})

test_that("curvilinear broked", {
  skip_if_not_installed("ncmeta")
  f <- system.file("nc/test_stageiv_xyt_borked.nc", package = "stars")

  expect_error(suppressMessages(read_ncdf(f, curvilinear = c("lon", "lat", "time"))),
               "Curvilinear input must be a length two character vector.")

  expect_error(read_ncdf(f, curvilinear = c("time", "time_bounds")),
               "Specified curvilinear coordinates are not 2-dimensional.")

  expect_error(suppressMessages(read_ncdf(f, curvilinear = c("lon", "time_bounds"))),
               "Specified curvilinear coordinate variables not found as X/Y coordinate variables.")

  warn <- capture_warnings(out <-read_ncdf(f, curvilinear = c(X = "lon", Y = "lat")))

  expect_match(warn[1], "Non-canonical axis order found, attempting to correct.")

  st_dim <- st_dimensions(out)

  expect_true(all(st_dim$y$values < -74 & st_dim$y$values > -81))

  expect_true(all(st_dim$x$values < 38 & st_dim$x$values > 32))

  expect_equal(dim(st_dim$y$values), setNames(c(118, 87), c("y", "x")))
  
  # expect_equal(st_dim$time$values, as.POSIXct("2018-09-14 05:00:00", tz = "UTC"))
  
})

test_that("high-dim from rasterwise", {
  skip_if_not_installed("ncmeta")
  f <- system.file("nc/test-1.nc", package = "stars")
  out <- suppressWarnings(read_ncdf(f, var = "a"))

  expect_equal(names(st_dimensions(out)), c("x", "y", "c3", "c4", "c5"))
})

test_that("timeseries.nc", {
  skip_if_not_installed("ncmeta")
  skip_if_not_installed("ncdfgeom")
  skip_on_os("solaris")

  f <- system.file("nc/timeseries.nc", package = "stars")
  nc <- read_ncdf(f)
  dims <- st_dimensions(nc)
  expect_s3_class(nc, "stars")
  expect_equal(names(nc), "pr")
  expect_equal(names(dims), c("time", "points"))

  poly <- sf::st_buffer(st_transform(st_sf(id = seq_along(dims$points$values),
                              geom = dims$points$values, crs = 4326), 3857),
                        dist = 100, nQuadSegs = 1)

  temp_f <- tempfile()
  file.copy(f, temp_f)

  suppressWarnings(temp_f <- ncdfgeom::write_geometry(temp_f, poly, "station", "pr"))
  nc <- read_ncdf(temp_f)

  dims <- st_dimensions(nc)
  expect_s3_class(nc, "stars")
  expect_equal(names(nc), "pr")
  expect_equal(names(dims), c("time", "points", "geometry"))
})

test_that("curvilinear 2", {
  skip_if_not_installed("ncmeta")
  f <- system.file("nc/c201923412.out1_4.nc", package = "stars")
  nc <- read_ncdf(f)

  expect_equal(names(st_dimensions(nc)), c("nx", "ny", "time"))

  expect_true(attributes(st_dimensions(nc))$raster$curvilinear)
  expect_equal(attributes(st_dimensions(nc))$raster$dimensions, c("nx", "ny"))

})

test_that("lon cross 360", {
  skip_if_not_installed("ncmeta")
  f <- system.file("nc/test_adaptor.cams_regional_fc.nc", package = "stars")

  suppressWarnings(nc <- read_ncdf(f))

  expect_true(head(st_dimensions(nc)$longitude$values, 1) < 0 &
                tail(st_dimensions(nc)$longitude$values, 1) > 0)
})

test_that("4d not 4d", {
  skip_if_not_installed("ncmeta")
  f <- system.file("nc/sub.nc", package = "stars")

  nc <- read_ncdf(f)

  dim <- st_dimensions(nc)

  expect_equal(dim$time$to, 10)

  expect_equal(dim$level$to, 2)

})

test_that("units are right with lcc km", {
	skip_if_not_installed("ncmeta")
	f <- system.file("nc/lcc_km.nc", package = "stars")
	
	nc <- expect_warning(read_ncdf(f), "prime meridian")
	
	expect_equal(units(sf::st_crs(nc)$ud_unit)$numerator, "km")
})

test_that("axis attribute order -- see #680", {
	
	# Example with two netCDFs
	tas <- array(
		data = rowSums(
			expand.grid(seq_len(9), 10 * (seq_len(2) - 1), 100 * (seq_len(3) - 1))
		),
		dim = c(9, 2, 3)
	)
	
	# File1 has no "axis" attributes
	file1 <- tempfile("tas_example_", fileext = ".nc")
	nc <- RNetCDF::create.nc(file1)
	
	id_lat <- RNetCDF::dim.def.nc(nc, "lat", 3)
	iv_lat <- RNetCDF::var.def.nc(nc, "lat", "NC_FLOAT", id_lat)
	RNetCDF::var.put.nc(nc, "lat", c(40, 45, 50))
	
	id_lon <- RNetCDF::dim.def.nc(nc, "lon", 2)
	iv_lon <- RNetCDF::var.def.nc(nc, "lon", "NC_FLOAT", id_lon)
	RNetCDF::var.put.nc(nc, "lon", c(-100, -95))
	
	id_bnds <- RNetCDF::dim.def.nc(nc, "bnds", 2)
	
	id_time <- RNetCDF::dim.def.nc(nc, "time", 9)
	iv_time <- RNetCDF::var.def.nc(nc, "time", "NC_INT", id_time)
	RNetCDF::var.put.nc(nc, "time", 1:9)
	
	iv_tas <- RNetCDF::var.def.nc(nc, "temperature", "NC_FLOAT", c(id_time, id_lon, id_lat))
	RNetCDF::var.put.nc(nc, "temperature", tas)
	
	RNetCDF::close.nc(nc)
	
	# File2 has X, Y, T axis attributes
	file2 <- tempfile("tas_example_", fileext = ".nc")
	file.copy(from = file1, to = file2)
	#> [1] TRUE
	nc <- RNetCDF::open.nc(file2, write = TRUE)
	
	RNetCDF::att.put.nc(nc, "lon", "axis", "NC_CHAR", "X")
	RNetCDF::att.put.nc(nc, "lat", "axis", "NC_CHAR", "Y")
	RNetCDF::att.put.nc(nc, "time", "axis", "NC_CHAR", "T")
	
	RNetCDF::close.nc(nc)
	
	file3 <- tempfile("tas_example_", fileext = ".nc")
	file.copy(from = file1, to = file3)
	#> [1] TRUE
	nc <- RNetCDF::open.nc(file3, write = TRUE)
	
	RNetCDF::att.put.nc(nc, "lon", "standard_name", "NC_CHAR", "longitude")
	RNetCDF::att.put.nc(nc, "lat", "standard_name", "NC_CHAR", "latitude")
	RNetCDF::att.put.nc(nc, "time", "standard_name", "NC_CHAR", "time")
	RNetCDF::att.put.nc(nc, "lon", "units", "NC_CHAR", "degrees")
	RNetCDF::att.put.nc(nc, "lat", "units", "NC_CHAR", "degrees")
	RNetCDF::att.put.nc(nc, "time", "units", "NC_CHAR", "days since 1900-01-01")
	
	RNetCDF::close.nc(nc)

	s1 <- suppressWarnings(stars::read_ncdf(file1))
		
	expect_equal(names(stars::st_dimensions(s1)), c("time", "lon", "lat"))
	
	s2 <- suppressWarnings(stars::read_ncdf(file2))

	expect_equal(names(stars::st_dimensions(s2)), c("lon", "lat", "time"))

	s3 <- suppressWarnings(stars::read_ncdf(file3))
	
	expect_equal(names(stars::st_dimensions(s3)), c("lon", "lat", "time"))
})

#' @export
st_as_stars.STSDF = function(.x, ...) st_as_stars(as(.x, "STFDF"), ...)

#' @export
st_as_stars.STS = function(.x, ...) st_as_stars(as(.x, "STF"), ...)

#' @export
st_as_stars.STFDF = function(.x, ...) {
    if (!requireNamespace("sp", quietly = TRUE))
        stop("package sp required, please install it first") #nocov
    if (!requireNamespace("zoo", quietly = TRUE))
        stop("package zoo required, please install it first") #nocov
	#ix = 1:prod(dim(.x)[c("space", "time")])
	d = if (sp::gridded(.x@sp)) {
			gp = sp::gridparameters(.x@sp)
			nx = gp[1,3]
			ny = gp[2,3]
			if (! sp::fullgrid(.x@sp)) { # SpatialPixels
				lst = vector("list", ncol(.x@data))
				for (j in seq_along(.x@data)) { # attributes
					px2vec = function(pix) {
						sp::fullgrid(pix) = TRUE
						as.vector(as.matrix(pix))
					}
					a = array(NA_real_, c(nx, ny, dim(.x)[2]))
					for (i in seq_len(dim(.x)[2])) # time
						a[,,i] = px2vec(.x[,i,j])
					lst[[j]] = as.vector(a)
				}
				.x@data = as.data.frame(setNames(lst, names(.x@data)))
			}
			# offs_x dx 0 offs_y 0 dy
			gt = c(gp[1,1] - gp[1,2]/2, gp[1,2], 0.0, gp[2,1] + (gp[2,3] - 0.5) * gp[2,2], 0.0, -gp[2,2])
			vals = lapply(seq_len(nrow(gp)), function(i) seq(gp[i,1], by = gp[i,2], length.out = gp[i,3]))
			create_dimensions(list(
				x = create_dimension(values = vals[[1]] - gp[1,2]/2),
				y = create_dimension(values = rev(vals[[2]]) + gp[2,2]/2),
				time = create_dimension(values = zoo::index(.x@time))),
				raster = get_raster())
		} else
			create_dimensions(list(
				sfc = create_dimension(values = st_as_sfc(.x@sp)), # FIXME: doesn't do SpatialPixels -> x/y
				time = create_dimension(values = zoo::index(.x@time))))
	vals = lapply(.x@data, function(y) { dim(y) = dim(d); y })
	st_set_crs(st_as_stars(vals, dimensions = d), st_crs(.x@sp))
}

#' @export
st_as_stars.STF = function(.x, ...) {
    if (!requireNamespace("sp", quietly = TRUE))
        stop("package sp required, please install it first") #nocov
    if (!requireNamespace("zoo", quietly = TRUE))
        stop("package zoo required, please install it first") #nocov
	#ix = 1:prod(dim(.x)[c("space", "time")])
	d = if (sp::gridded(.x@sp)) {
			gp = sp::gridparameters(.x@sp)
			nx = gp[1,3]
			ny = gp[2,3]
			# offs_x dx 0 offs_y 0 dy
			gt = c(gp[1,1] - gp[1,2]/2, gp[1,2], 0.0, gp[2,1] + (gp[2,3] - 0.5) * gp[2,2], 0.0, -gp[2,2])
			vals = lapply(seq_len(nrow(gp)), function(i) seq(gp[i,1], by = gp[i,2], length.out = gp[i,3]))
			create_dimensions(list(
				x = create_dimension(values = vals[[1]] - gp[1,2]/2),
				y = create_dimension(values = rev(vals[[2]]) + gp[2,2]/2),
				time = create_dimension(values = zoo::index(.x@time))),
				raster = get_raster())
		} else
			create_dimensions(list(
				sfc = create_dimension(values = st_as_sfc(.x@sp)), # FIXME: doesn't do SpatialPixels -> x/y
				time = create_dimension(values = zoo::index(.x@time))))
	st_set_crs(st_as_stars(list(), dimensions = d), st_crs(.x@sp))
}

st_as_STFDF = function(x) {

    if (!requireNamespace("sp", quietly = TRUE))
        stop("package sp required, please install it first") #nocov
    if (!requireNamespace("spacetime", quietly = TRUE))
        stop("package spacetime required, please install it first") #nocov
	rst = has_raster(x)
	d = st_dimensions(x)
	geom = if (rst) {
			x = st_upfront(x)
			time_dim = 3
			sp::geometry(as(adrop(x[,,,1], drop = 3), "Spatial"))
		} else {
			x = st_upfront(x, which_sfc(x))
			time_dim = 2
			sp::geometry(as(adrop(x[,,1], drop = 2), "Spatial"))
		}
	tm = st_get_dimension_values(x, time_dim, center = FALSE)

	st = suppressWarnings(spacetime::STF(geom, tm)) # would warn for SpatialGrid -> SpatialPixels

	sp::addAttrToGeom(st, as.data.frame(lapply(x, function(y) structure(y, dim = NULL))), match.ID = FALSE)
}

setAs("stars", "STFDF", function(from) { 
    if (!requireNamespace("sp", quietly = TRUE))
        stop("package sp required, please install it first") #nocov
    if (!requireNamespace("zoo", quietly = TRUE))
        stop("package zoo required, please install it first") #nocov
    if (!requireNamespace("xts", quietly = TRUE))
        stop("package xts required, please install it first") #nocov
    if (!requireNamespace("spacetime", quietly = TRUE))
        stop("package spacetime required, please install it first") #nocov
	st_as_STFDF(from)
})

suppressPackageStartupMessages(library(stars))
tif = system.file("tif/L7_ETMs.tif", package = "stars")
r = read_stars(list(a = c(tif,tif), b = c(tif, tif)), proxy = TRUE)
(xx = st_redimension(r, along = list(foo = 1:4)))
(rr = c(xx, xx))
# st_as_stars(rr)
(rrr = st_redimension(rr, along = list(bar = as.Date(c("2001-01-01", "2002-01-01")))))
(rr = c(a = xx, b = xx))
# st_as_stars(rr)
(rrr = st_redimension(rr, along = list(bar = as.Date(c("2001-01-01", "2002-01-01")))))

#' @importFrom graphics image.default image par plot title box text axis plot.new plot.window rasterImage layout lcm contour hist strwidth
#' @importFrom grDevices dev.capabilities dev.size grey rgb col2rgb cm
#' @importFrom utils head tail setTxtProgressBar txtProgressBar packageVersion methods modifyList
#' @importFrom stats na.omit runif aggregate setNames predict quantile var complete.cases na.pass time prcomp
#' @importFrom tools file_ext
#' @importFrom methods as slotNames new slot
#' @importFrom abind abind adrop asub
#' @importFrom parallel parApply
#' @importFrom rlang %||%
#' @import sf
#' @import units
NULL

setOldClass("stars")
setOldClass("stars_proxy")

.onLoad = function(libname, pkgname) {
	register_all_s3_methods() # dynamically registers non-imported pkgs (tidyverse) # nocov
}

.onAttach = function(libname, pkgname) {
	if (nrow(sf::st_drivers("raster", "netCDF")) != 1)
		packageStartupMessage("Note: sf is linked to a GDAL version without netCDF driver, some tests or examples may fail")
}

Sys.setenv(TZ="")
suppressPackageStartupMessages(library(stars))
matrix(1:4,2)
m = matrix(1:4,2)
s = st_as_stars(m)

s = st_set_dimensions(s, st_dimensions(s), xy = NULL) 
sfc = st_sfc(st_point(0:1), st_point(3:2))
s = st_set_dimensions(s, 1, sfc) |> st_set_crs(4326)
tm = as.POSIXct("2023-03-04 12:35") + c(0, 3600)
s = st_set_dimensions(s, 2, tm)
f = tempfile(fileext = ".nc")
write_mdim(s, f, as_float = FALSE)
s2 = read_mdim(f)
#print(s2)
print(all.equal(s, s2, check.attributes = FALSE))
print(all.equal(s, s2))

st_crs(sfc) = st_crs(s2)
st_crs(s) = st_crs(s2)
tm = as.Date("2023-02-25") + 0:1
sd = st_set_dimensions(s, 2, tm)
print(sd)
sd = st_set_dimensions(sd, 1, sfc)
print(sd)
f = tempfile(fileext = ".nc")
write_mdim(sd, f)
sd2 = read_mdim(f)
print(sd2)
print(all.equal(sd, sd2, check.attributes = FALSE))
print(all.equal(sd, sd2))
# if (.Platform$OS.type == "unix") {
#   print(read_mdim(paste0("/vsizip/", (system.file("zarr/sic_daily_sample.zarr.zip", package = "stars"))))
# }

#' spatial intersect predicate for stars and sfc object
#'
#' spatial intersect predicate for stars and sfc object
#' @param x object of class stars
#' @param y object that has an `st_geometry` method: of class `sf` or `sfc`, or `stars` object with an `sfc` dimension
#' @param sparse logical; if TRUE, return the a sparse logical matrix (object of class `sgbp`), if FALSE, return a logical matrix
#' @param as_points logical, should grid cells be considered as points (TRUE) or polygons (FALSE)? Default: FALSE and warning emitted
#' @param ... ignored, or passed on to `st_intersects.sf` for curvilinear grids
#' @param transpose logical; should the transpose of the `sgbp` object be returned?
#' @return `sgbp` object if sparse = TRUE, logical matrix otherwise
#' @details curvilinear grids are always converted to polygons, so points on grid boundaries may intersect with two cells touched; for other grids each cell boundary or corner belongs only to one cell.
#' 
#' @export
st_intersects.stars = function(x, y, sparse = TRUE, ..., as_points = NA, transpose = FALSE) { 

	y = st_geometry(y)
	d = st_dimensions(x)
	xy = attr(d, "raster")$dimensions
	if (is.na(as_points)) {
		p = d[[ xy[1] ]]$point || d[[ xy[2] ]]$point
		if (is.na(p)) {
			warning("as_points is NA: assuming here that raster cells are small polygons, not points")
			as_points = FALSE
		} else
			as_points = p
	}
	ret = if (has_raster(x) && !is_curvilinear(x) &&
			inherits(y, "sfc_POINT") && 
			!as_points) {

		cr = colrow_from_xy(st_coordinates(y), x, NA_outside = TRUE)
		ncells = prod(dim(d[xy]))
		ncols = dim(x)[ xy[1] ]
		ix = (cr[,2] - 1) * ncols + cr[,1] # for all points, the 1-based index in (as.vector) x
		# now create the TRANSPOSE of the sgbp:
   		ret = structure(lapply(as.list(ix), function(x) if (is.na(x)) integer(0) else x),
       		predicate = "intersects",
       		region.id = seq_along(y),
       		ncol = ncells,
       		class = "sgbp")
		if (! sparse)
			ret = as.matrix(ret)
		if (! transpose)
			t(ret)
		else
			ret
	} else { # curvilinear or !as_points:
		ret = st_intersects(st_as_sf(x, as_points = as_points, na.rm = FALSE), y, 
			sparse = sparse, ...)
		if (! sparse)
			ret = as.matrix(ret)
		if (transpose) 
			t(ret)
		else
			ret
	}
}

# FIXME: of_y should be left/right/full?

#' Spatially join a stars and an `sf` object
#'
#' Spatially join a stars and an `sf` object
#' @param x object of class stars
#' @param y object of class sf, or one that can be coerced into that by \link{st_as_sf}
#' @param join the join function, which should return an sgbp object; see details
#' @param ... arguments that will be passed on to the join function
#' @param as_points logical; controls whether grid cells in \code{x} will be treated as points, or as cell areas; the \link{st_intersects.stars} method by default will derive this from \code{x}'s metadata, or else assume areas.
#' @param what "left1", "right" or "inner"; see details
#' @param warn logical; if TRUE, warn on 1-to-many matches when \code{what} is \code{"left1"}
#' @return If what is "left1", an object of class stars with the (first) value of y at spatial instances of x
#' @details When there is more than one match to a single x value, the first matching record from y is taken (and if \code{warn} is TRUE a warning is raised). If what is "inner", an object of class \code{sf} with all matching records of x and y.
#' @export
st_join.stars = function(x, y, join = st_intersects, ..., what = "left1", as_points = NA, warn = TRUE) {
	if (!requireNamespace("dplyr", quietly = TRUE)) 
		stop("dplyr needed: install first?")

	if (!inherits(y, "sf")) {
		try(y <- st_as_sf(y))
		if (inherits(y, "try-error"))
			stop("argument y should be of class sf, or sfc")
	}

	x = st_upfront(x)

	if (what == "left1") {
		ix = join(x, y, ..., as_points = as_points)
		li = lengths(ix)
		if (warn && any(li > 1)) {
			warning(paste("st_join found", sum(li > 1),"1-to-n matches, taking the first match for each of those"))
			ix[li > 1] = lapply(ix[li > 1], head, 1)
		}
		sel = li > 0
		ix = unlist(ix)
		# white-out existing cells not matched; No,see https://github.com/r-spatial/stars/issues/643:
		# for (i in seq_along(x))
		#	x[[i]][!sel] = NA

		# add matching fields from y:
		for (i in setdiff(names(y), attr(y, "sf_column"))) {
			na_mode = NA
			storage.mode(na_mode) = storage.mode(y[[i]])
			a = array(na_mode, dim(x))
			attr(a, "units") = attr(y[[i]], "units")
			attr(a, "levels") = attr(y[[i]], "levels")
			a[sel] = y[[i]][ix]
			x[[ i ]] = a
		}
		x
	} else if (what == "inner") {
		ix = if (missing(join))
				st_intersects(x, y, sparse = TRUE, ..., transpose = TRUE, as_points = as_points)
			else
				join(y, st_as_sf(x, as_points = as_points), ...)
		li = lengths(ix)
		i = unlist(ix) # sequence of x indexes
		j = rep(seq_along(ix), li) # possibly repeated y indexes
		stopifnot(length(i) == length(j))
		if (length(dim(x)) > 2) {
			recycle = prod(dim(x)[-(1:2)]) # 1 in case of no dimensions beyond x and y
			i = rep(i, recycle)
			j = rep(j, recycle)
			# FIXME: create a stars object with an sfc dimension
			warning("a stars object should probably have been created here; please file an issue")
		}
		st_sf(dplyr::bind_cols(as.data.frame(x)[i,], y[j,]))
	} else
		stop('value for "what" not supported')
}

#' @name st_as_stars
#' @param att see \link[raster:factor]{factorValues}; column in the RasterLayer's attribute table
#' @param ignore_file logical; if \code{TRUE}, ignore the Raster object file name
#' @export
st_as_stars.Raster = function(.x, ..., att = 1, ignore_file = FALSE) {
	if (!requireNamespace("raster", quietly = TRUE))
		stop("package raster required, please install it first") # nocov

	#0 360 -90  90
	e = as.vector(raster::extent(.x)) # xmin xmax ymin ymax

	RasterIO = if ("data" %in% slotNames(.x) && inherits(.x@data, ".SingleLayerData"))
			list(bands = .x@data@band)
		else
			list()
	if (!ignore_file) {
		file = if ("file" %in% slotNames(.x))
				.x@file@name
			else if ("filename" %in% slotNames(.x))
				.x@filename
			else
				""
		if (file != "") {
			r = try(read_stars(file, RasterIO = RasterIO, ...), silent = TRUE)
			if (!inherits(r, "try-error")) {

				if (is.na(st_crs(r)))
					r = st_set_crs(r, st_crs(raster::crs(.x)))

				r = fix_dims(r, e)

				if (inherits(.x, "RasterLayer"))
					names(r) <- names(.x)

				return(r)
			}
		}
	}

	if (!requireNamespace("sp", quietly = TRUE))
		stop("package sp required, please install it first") # nocov
	v = raster::values(.x)
	dim(v) = dim(.x)[c(2,1,3)]
	if (all(raster::is.factor(.x))) {
		l = raster::levels(.x)[[1]]$levels
		if (length(l) == 0) # get the layer's RAT, column att:
			l = raster::factorValues(.x, seq_len(max(v, na.rm = TRUE)), att = att)[[1]]
		colors = try(.x@legend@colortable, silent = TRUE)
		if (inherits(colors, "try-error") || length(colors) == 0)
			colors = NULL
		else
			colors = colors[-1]
		v = structure(v, class = "factor", levels = as.character(l), colors = colors)
		# FIXME: should we handle levels for all layers here, or break on multiple different ones?
	}
	dimensions = list(
		x = create_dimension(from = 1, to = dim(v)[1], offset = e[1],
			delta = (e[2]-e[1])/dim(v)[1], refsys = st_crs(raster::crs(.x))),
		y = create_dimension(from = 1, to = dim(v)[2], offset = e[4],
			delta = (e[3]-e[4])/dim(v)[2], refsys = st_crs(raster::crs(.x))))
	z = raster::getZ(.x)
	dimensions$band = if (is.null(z))
			create_dimension(values = names(.x))
		else
			create_dimension(values = z)
	l = if (length(names) > 1)
			setNames(list(v), deparse(substitute(.x), 50))
		else
			setNames(list(v), names(.x)[1])
	ret = st_as_stars(l, dimensions = create_dimensions(dimensions, get_raster()))
	if (dim(ret)[3] == 1)
		adrop(ret, 3)
	else
		ret
}
#' Coerce stars object into a RasterLayer, RasterBrick, or SpatRaster
#'
#' Coerce stars object into a RasterLayer, RasterBrick, or SpatRaster
#' @name as
#' @rdname coerce-methods
#' @aliases coerce,stars,Raster-method
#' @aliases coerce,stars_proxy,Raster-method
#' @returns RasterLayer or RasterBrick
#' @details If the stars object has more than three dimensions, all dimensions higher than the third will be collapsed into the third dimensions. If the stars object has only an x/y raster but multiple attributes, these are merged first, then put in a raster brick or SpatRaster.
setAs("stars", "Raster", function(from) {
	if (!requireNamespace("sp", quietly = TRUE))
		stop("package sp required, please install it first") # nocov
	if (!requireNamespace("raster", quietly = TRUE))
		stop("package raster required, please install it first") # nocov
	if (!is_regular_grid(from))
		stop("only regular rasters can be converted to Raster* objects")
	st_as_raster(from, class = "Raster")
})

setAs("stars_proxy", "Raster", function(from) {
	if (!requireNamespace("raster", quietly = TRUE))
		stop("package raster required, please install it first") # nocov
	if (!is_regular_grid(from))
		stop("only regular rasters can be converted to Raster* objects")
	if (length(attr(from, "call_list"))) {
		fname = paste0(tempfile(), ".tif")
		write_stars(from, fname)
		from = fname
	}
	raster::brick(unlist(from))
})

get_terra_levels = function(x, colors) {
	# create factor levels, as used by stars, from SpatRaster levels in a data.frame
	# see https://github.com/r-spatial/stars/pull/484
	x = x[order(x[[1]]), ] # sort table on level
	missing_labels = x[[2]] == ""
	levels = x[[1]]
	if (any(levels < 0))
		stop("negative IDs in SpatRaster levels not supported")
	ex = setdiff(0:max(levels), levels)
	exclude = rep(FALSE, max(levels) + 1)
	exclude[ex + 1] = TRUE # 0-based vector
	list(levels = levels[!missing_labels],
		 labels = x[[2]][!missing_labels],
		 exclude = exclude,
		 colors = colors[!missing_labels])
}

#' @name st_as_stars
#' @param ignore_file logical; if \code{TRUE}, ignore the SpatRaster object file name
#' @param as_attributes logical; if \code{TRUE} and \code{.x} has more than one layer, load these as separate attributes rather than as a band or time dimension (only implemented for the case where \code{ignore_file} is \code{TRUE})
#' @export
st_as_stars.SpatRaster = function(.x, ..., ignore_file = FALSE,
			  as_attributes = all(terra::is.factor(.x))) {
	if (!requireNamespace("terra", quietly = TRUE))
		stop("package terra required, please install it first") # nocov

	#0 360 -90  90
	e = as.vector(terra::ext(.x)) # xmin xmax ymin ymax

	src = terra::sources(.x, bands=TRUE)

	attr_name = basename(src$source[1])

	if (!ignore_file && all(src$source != "")) {
	# there can be multiple files, but only the first one is used here.
	# perhaps a warning should be given; better would be to iterate over "sid"
	# but you might have a situation where some sources are filenames and others are not
		if (attr_name == "")
			attr_name = "values"
		lst = vector("list", length(unique(src$sid)))
		for (i in unique(src$sid)) {
			file = unique(src$source[src$sid == i])
			if (length(file) > 1)
				stop("more than one file per sid: giving up; try ignore_file=FALSE")
			# 	RasterIO = if (dim(.x)[3] == 1)
			# > 1 would be more sensible? 
			# But this can only be ignored if the _file_ has 1 band
			RasterIO = list(bands = src$bands[src$sid == i])
			r = try(read_stars(file, RasterIO = RasterIO, ...), silent = TRUE)
			if (! inherits(r, "try-error")) {
				if (is.na(st_crs(r)) && terra::crs(.x) != "")
					r = st_set_crs(r, st_crs(terra::crs(.x)))
				r = fix_dims(r, e)
				if (length(unique(src$sid)) > 1 && length(dim(r)) > 2)
					r = split(r)
				#transfer the layer/band names as well?
				# ... = names(.x)[1:(dim(r)[3])]
				# perhaps check whether they represent a time dimension (all(!is.na(time(.x))))
			} else
				stop(paste("error reading", file, "bands", paste0(RasterIO$bands, collapse = " ")))
			lst[[i]] = r
		}
		ret = if (length(lst) > 1)
				merge(setNames(do.call(c, lst), names(.x)))
			else
				r
		if (!all(is.na(terra::time(.x))))
			ret = st_set_dimensions(ret, 3, values = terra::time(.x), names = "time")

	} else { # ignore_file TRUE:
		if (terra::nlyr(.x) > 1 && as_attributes) {
			ret = do.call(c, lapply(seq_len(terra::nlyr(.x)), function(i) st_as_stars(.x[[i]], ignore_file = TRUE)))
			if (!is.null(names(.x)))
				names(ret) = names(.x)
			return(ret) # RETURNS
		}
		if (attr_name == "") {
			if (all(names(.x) == ""))
				attr_name = "values"
			else
				attr_name = paste(names(.x)[1], collapse = ".")
		}
		v = terra::values(.x, mat = FALSE)
		dimv = dim(v) = setNames(dim(.x)[c(2,1,3)], c("x", "y", "band"))
		if (all(terra::is.factor(.x))) {
			if (length(terra::levels(.x)) > 1)
				warning("ignoring categories/levels for all but first layer")
			colors = try(rgb(terra::coltab(.x)[[1]][-1], maxColorValue = 255), silent = TRUE)
			if (inherits(colors, "try-error") || length(colors) == 0)
				colors = NULL
			if (inherits(l <- terra::levels(.x)[[1]], "data.frame"))
				l = get_terra_levels(l, colors)
			else
				stop("terra levels should return a list of data.frame's; pls update terra")
			v = structure(factor(as.vector(v), levels = l$levels, labels = l$labels),
					dim = dimv, colors = l$colors, exclude = l$exclude)
		}
		crs = if (terra::crs(.x) == "")
				NA_crs_
			else
				st_crs(terra::crs(.x))
		dimensions = list(
				x = create_dimension(from = 1, to = dim(v)[1], offset = e[1],
							 	delta = (e[2]-e[1])/dim(v)[1], refsys = crs),
				y = create_dimension(from = 1, to = dim(v)[2], offset = e[4],
							 	delta = (e[3]-e[4])/dim(v)[2], refsys = crs))
		dimensions$band = create_dimension(values = names(.x))
		ret = st_as_stars(setNames(list(v), attr_name), dimensions = create_dimensions(dimensions, get_raster()))
		if (dim(ret)[3] == 1)
			ret = adrop(ret, 3)
		else if (!all(is.na(terra::time(.x))))
			ret = st_set_dimensions(ret, 3, values = terra::time(.x), names = "time")
	}
	ret
}

#' @name as
#' @rdname coerce-methods
#' @aliases coerce,stars,Terra-method
#' @aliases coerce,stars_proxy,Terra-method
#' @returns SpatRaster
setAs("stars", "SpatRaster", function(from) {
	if (!requireNamespace("terra", quietly = TRUE))
		stop("package terra required, please install it first") # nocov
	if (!is_regular_grid(from))
		stop("only regular rasters can be converted to SpatRaster objects")
	st_as_raster(from, class = "SpatRaster")
})

setAs("stars_proxy", "SpatRaster", function(from) {
	if (!requireNamespace("terra", quietly = TRUE))
		stop("package terra required, please install it first") # nocov
	if (!is_regular_grid(from))
		stop("only regular rasters can be converted to SpatRaster objects")
	if (length(attr(from, "call_list"))) {
		fname = paste0(tempfile(), ".tif")
		write_stars(from, fname)
		from = fname
	}
	terra::rast(unlist(from))
})

fix_dims = function(r, e){
	e = unname(e)
	rdims = attr(r, "dimensions")
	dxy = attr(rdims, "raster")$dimensions
	dimx = rdims[[dxy[1]]]
	dimy = rdims[[dxy[2]]]

	xrev = dimx$delta < 0
	yrev = dimy$delta < 0

	if (xrev) {
		dimx$offset = e[2]
		dimx$delta = (e[1] - e[2]) / (dimx$to - dimx$from + 1)
	} else {
		dimx$offset = e[1]
		dimx$delta = (e[2] - e[1]) / (dimx$to - dimx$from + 1)
	}

	if (yrev) {
		dimy$offset = e[4]
		dimy$delta = (e[3] - e[4]) / (dimy$to - dimy$from + 1)
	} else {
		dimy$offset = e[3]
		dimy$delta = (e[4] - e[3]) / (dimy$to - dimy$from + 1)
	}

	attr(r, "dimensions")[[dxy[1]]] = dimx
	attr(r, "dimensions")[[dxy[2]]] = dimy

	r
}

st_as_raster = function(x, class, ...) {
	stopifnot(inherits(x, "stars"))
	if (is.null(names(x)))
		names(x) = make.names(seq_along(x)) # FIXME: why? sdsr_exercises
	x_crs = st_crs(x)
	x = st_upfront(x) # x/y dimensions first
	if (length(dim(x)) > 3) {
		warning("folding all higher dimensions into the third dimension") # nocov
		x = st_apply(x, 1:2, as.vector) # fortunes::fortune("side effect") # nocov
	}
	if (length(dim(x)) == 2 && length(x) > 1)
		x = merge(x)
	d = st_dimensions(x)
	if (d[[2]]$delta > 0) { # swap:
		ny = dim(x)[2]
		d[[2]]$offset = d[[2]]$offset + ny * d[[2]]$delta # top
		d[[2]]$delta = -d[[2]]$delta # going down
		x[[1]] = if (length(dim(x)) == 2)
			x[[1]][,ny:1]
		else
			x[[1]][,ny:1,]
	}
	dxy = attr(d, "raster")$dimensions
	stopifnot(all(dxy %in% names(d)))
	bb = st_bbox(x)
	levels = NULL
	coltab = vector("list", length(x))
	values = if (all(sapply(x, is.factor))) {
			ex = attr(x[[1]], "exclude")
			if (is.null(ex) || class != "SpatRaster")
				structure(merge(x)[[1]], dim = NULL) # return the factor
			else {
				v = vector("list", length(x))
				levels = vector("list", length(x))
				for (i in seq_along(v)) {
					ex = attr(x[[i]], "exclude")
					ix = which(!ex) - 1 # 0-based index
					n = as.numeric(structure(x[[i]], dim = NULL)) # factor -> numeric
					v[[i]] = ix[n]
					levels[[i]] = data.frame(IDs = ix, categories = levels(x[[i]]))
					if (!is.null(ct <- attr(x[[i]], "colors"))) {
						coltab[[i]] = t(col2rgb(rep("#000000", length(ex)), alpha = TRUE))
						coltab[[i]][which(!ex),] = t(col2rgb(ct, alpha = TRUE))
					}
				}
				do.call(c, v)
			}
		} else  {
			if (any(sapply(x, is.factor)))
				warning("mix of factor and non-factor attributes: all factor levels are ignored")
			as.vector(merge(x)[[1]])
		}
	if (class == "SpatRaster") {
		third = setdiff(names(d), dxy)
		b = terra::rast(nrows = dim(x)[ dxy[2] ], ncols=dim(x)[ dxy[1] ],
						xmin = bb[1], xmax = bb[3], ymin = bb[2], ymax = bb[4],
						nlyrs = ifelse(length(dim(x)) == 2, 1, dim(x)[third]),
						crs = x_crs$wkt)
		terra::values(b) = values
		if (!is.null(levels)) {
			levels(b) = levels
			if (!all(sapply(coltab, is.null)))
				for (i in seq_len(terra::nlyr(b))) {
					m = coltab[[i]]
					terra::coltab(b, layer = i) = cbind(value = seq_len(nrow(m)) - 1, m)
				}
		}
		if (length(dim(x)) != 2) {
			z = seq(d[[third]])
			if (!any(is.na(z))) {
				if (is.character(z)) {
					names(b) = z
				} else if (inherits(z, c("POSIXt", "Date"))) {
					terra::time(b) = z
				} else {
					names(b) = paste0(third, z)
				}
			}
		} else if (length(x) == terra::nlyr(b))
			names(b) = names(x)
	} else {
		if (length(dim(x)) == 2) {
			b = raster::raster(nrows=dim(x)[ dxy[2] ], ncols=dim(x)[ dxy[1] ],
							   xmn = bb[1], xmx = bb[3], ymn = bb[2], ymx = bb[4],
							   crs = as(x_crs, "CRS"), vals = values)
		} else {
			third = setdiff(names(d), dxy)
			b = raster::brick(nrows=dim(x)[ dxy[2] ], ncols=dim(x)[ dxy[1] ],
							  xmn = bb[1], xmx = bb[3], ymn = bb[2], ymx = bb[4], nl = dim(x)[third],
							  crs = as(x_crs, "CRS"))
			raster::values(b) = values
			z = seq(d[[third]])
			if (!any(is.na(z))) {
				if (is.character(z))
					names(b) = z
				else
					b = raster::setZ(b, z)
			}
		}
	}
	return(b)
}

library(testthat)
library(stars)
context("gdal utils")


test_that('gdal_utils work', {
  skip_on_appveyor()

  fname = system.file("nc/tos_O1_2001-2002.nc", package = "stars")
  if (fname != "") {
   info = gdal_utils("info", fname, quiet = TRUE)
   sd2 = gdal_subdatasets(fname)[[4]]
   info = gdal_utils("info", sd2, quiet = TRUE)
   tf = tempfile()
   tf2 = tempfile()
   tf3 = tempfile()
   #tf = "foo"
   #gdal_utils("rasterize", points, tif) -> need a good example
   expect_true(gdal_utils("warp", sd2, tf, c("-t_srs", "+proj=utm +zone=11 +datum=WGS84")))
#  expect_true(gdal_utils("rasterize", sd2, tf)) # breaks on fedora-clang
   expect_true(gdal_utils("translate", sd2, tf))
   expect_true(gdal_utils("vectortranslate", sd2, tf2))
   expect_warning(gdal_utils("nearblack", sd2, tf))
  # create point geom:
   expect_true(gdal_utils("buildvrt", sd2, tf3))
#  expect_true(gdal_utils("demprocessing", sd2, tf, processing = "hillshade")) # breaks on fedora-clang
  }
  tf = tempfile()
  points = system.file("gpkg/nc.gpkg", package="sf")
  expect_true(gdal_utils("grid", points, tf))
})

# gdalwarp -t_srs '+proj=utm +zone=11 +datum=WGS84' -overwrite NETCDF:avhrr-only-v2.19810901.nc:anom utm11.tif
# becomes:
# gdal_utils("warp" "NETCDF:avhrr-only-v2.19810901.nc:anom", "utm11.tif", c("-t_srs", "+proj=utm +zone=11 +datum=WGS84"))

---
title: "8. NetCDF Proxy Workflows"
author: "David Blodgett"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{8. NetCDF Proxy Workflows}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

**For a better version of the stars vignettes see** https://r-spatial.github.io/stars/articles/

```{r, include = FALSE}
knitr::opts_chunk$set(echo = TRUE, collapse = TRUE, dev = "png")
suppressPackageStartupMessages(library(dplyr))
knitr::opts_chunk$set(fig.height = 4.5)
knitr::opts_chunk$set(fig.width = 6)
EVAL = suppressWarnings(require(starsdata, quietly = TRUE))
```

NetCDF data sources are available via more _and_ less granular files and/or OPeNDAP endpoints. This article demonstrates how `stars` enables discovery, access, and processing of NetCDF data across a wide range of such source-data organization schemes.

We'll start with some basics using datasets included with the `stars` installation. A call to `read_ncdf()`, for a dataset smaller than the default threshold, will just read in all the data. Below we read in and display the `reduced.nc` NetCDF file.

```{r}
library(stars)
f <- system.file("nc/reduced.nc", package = "stars")
(nc <- read_ncdf(f))
```

Let's assume `reduced.nc` was 10 years of hourly data, rather than 1 time step. It would be over 10GB rather than about 130KB and we would not be able to just read it all into memory. In this case, we need a way to read the file's metadata such that we could iterate over it in a way that meets the needs of our workflow objectives. This is where `proxy = TRUE` comes in. Below, we'll lower the option that controls whether `read_ncdf()` defaults to proxy _and_ use `proxy = TRUE` to show both ways of getting the same result.

```{r}
old_options <- options("stars.n_proxy" = 100)
(nc <- read_ncdf(f, proxy = TRUE))
options(old_options)
```

The above shows that we have a NetCDF sourced stars proxy derived from the `reduced.nc` file. We see it has four variables and their units are displayed. The normal `stars` `dimension(s)` are available and a `nc_request` object is also available. The `nc_request` object contains the information needed to make requests for data according to the dimensions of the NetCDF data source. With this information, we have what we need to request a chunk of data that is what we want and not too large.

```{r}
(nc <- read_ncdf(f, 
				 var = "sst", 
				 ncsub = cbind(start = c(90, 45, 1 , 1), 
				 			  count = c(90, 45, 1, 1))))

plot(nc)
```

The ability to view NetCDF metadata so we can make well formed requests against the data is useful, but the real power of a proxy object is that we can use it in a "lazy evaluation" coding style. That is, we can do virtual operations on the object, like subsetting with another dataset, prior to actually accessing the data volume. 

Lazy operations.

There are two kinds of lazy operations possible with `stars_proxy` objects. Some can be applied to the `stars_proxy` object itself without accessing underlying data. Others must be composed as a chain of calls that will be applied when data is actually required. 

Methods applied to a `stars_proxy` object:

- `[` - Nearly the same as stars_proxy
- `[[<-` - stars_proxy method works
- `print` - unique method for nc_proxy to facilitate unique workflows
- `dim` - stars_proxy method works
- `c` - stars_proxy method works
- `st_redimension` - Not sure what this entails but it might not make sense for nc_proxy.
- `st_mosaic` * Calls read_stars on assembled list. Not supported for now.
- `st_set_bbox`


Methods that add a call to the `call_list`.

- `[<-`
- `adrop`
- `aperm`
- `is.na`
- `split`
- `st_apply`
- `predict`
- `merge`
- `st_crop`
- `drop_levels`
- `Ops` (group generic for +, -, etc.)
- `Math` (group generic for abs, sqrt, tan, etc.)
- `filter`
- `mutate`
- `tansmute`
- `select`
- `rename`
- `pull`
- `slice` * hyperslabbing for NetCDF could be as above?
- `pull`
- `replace_na`

Methods that cause a `stars_proxy` object to be fetched and turned into a `stars` object.

- `as.data.frame`
- `plot`
- `st_as_stars`
- `aggregate`
- `st_dimensions<-` * https://github.com/r-spatial/stars/issues/494
- `hist`
- `st_downsample`
- `st_sample`
- `st_as_sf`
- `write_stars`

---
title: "3. stars tidyverse methods"
author: "Edzer Pebesma"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{3. stars tidyverse methods}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

**For a better version of the stars vignettes see** https://r-spatial.github.io/stars/articles/

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE, collapse = TRUE, dev = "png")
ev = TRUE
knitr::opts_chunk$set(fig.height = 4.5)
knitr::opts_chunk$set(fig.width = 6)
```

This vignette shows how some of the tidyverse verbs can be used on
`stars` objects.

The `stars` and `tidyverse` packages are loaded by
```{r}
library(stars)
library(dplyr)
```
Methods now available for class `stars` are
``` {r}
methods(class = "stars")
```

We will work with a three-band section of a landsat image:
```{r}
system.file("tif/L7_ETMs.tif", package = "stars") |>
	read_stars() -> x
x
```

## `slice`
`slice` slices a sub-array out of the cube; this is done by specifying
the dimension on which to act, and the slice number. 
```{r}
x |> slice(band, 6) -> x6
x6
```
It returns a lower-dimensional array if a single element is selected
along the slice dimension.

## `filter`
Similar to `slice`, `filter` selects on dimensions but evaluates their values
rather than their index: in
```{r}
x |> filter(x > 289000, x < 291000, band > 3) -> x7
x7
```
the subarray is created based on the x _coordinate_ values.

Note that `filter` converts the object to a `tbl_cube`, and uses
the `dplyr` `filter` method for `tbl_cube` objects. This has the
limitation that `stars` objects with rectilinear, curvilinear or
simple feature geometries cannot be handled. For such objects, using
regular `[` selection or using `st_crop` may be an alternative.

## `pull`

`pull` pulls out an array from a stars object:
```{r}
x |> pull(1) -> x8
class(x8)
dim(x8)
```

## `mutate`
```{r}
x |> mutate(band2 = 2 * L7_ETMs.tif) -> x2 
x2
```

## `select`

`select` selects an attribute, or a set of attributes:
```{r}
x2 |> select(band2) -> x9
x9
```

## `geom_stars`

`geom_raster` is a `ggplot2` geom function that accepts `stars`
objects as its `data` argument and

* sets up the raster or vector spatial coordinates as plot dimensions, and
the first attribute as the fill variable
* allows for downsampling (without choosing a suitable downsampling level)
* chooses between using `geom_raster`, `geom_rect` and `geom_sf` depending on
whether the geometry is regular, rectilinear or has vector geometries

An example use is
```{r}
library(ggplot2)
library(viridis)
ggplot() + 
  geom_stars(data = x) +
  coord_equal() +
  facet_wrap(~band) +
  theme_void() +
  scale_fill_viridis() +
  scale_x_discrete(expand = c(0, 0)) +
  scale_y_discrete(expand = c(0, 0))
```

---
title: "2. stars proxy objects"
author: "Edzer Pebesma"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{2. stars proxy objects}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

**For a better version of the stars vignettes see** https://r-spatial.github.io/stars/articles/

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE, collapse = TRUE, dev = "png")
set.seed(13579)
knitr::opts_chunk$set(fig.height = 4.5)
knitr::opts_chunk$set(fig.width = 6)
EVAL = x = suppressWarnings(require(starsdata, quietly = TRUE))
```

When your imagery or array data easily fits a couple of times in
R's working memory (RAM), consider yourself lucky. This document
was not written for you.  If your imagery is too large, or for
other reasons you want to work with smaller chunks of data than
the files in which they come, read on about your options.
First, we will discuss the low-level interface for this, then
the higher level, using stars proxy objects that delay all reading.

# Preamble: the starsdata package

To run all of the examples in this vignette, you must install a package with datasets that are too large (1 Gb) to be held in the `stars` package. They are in a [drat repo](https://github.com/eddelbuettel/drat), installation is done by
```{r eval=FALSE}
install.packages("starsdata", repos = "https://cran.uni-muenster.de/pebesma/")
# possibly after: options(timeout = 100)
# or from an alternative repository:
# install.packages("starsdata", repos = "http://pebesma.staff.ifgi.de", type = "source")
```

# Reading chunks, change resolution, select bands

`read_stars` has an argument called `RasterIO` which controls how
a GDAL dataset is being read. By default, all pixels and all bands
are read in memory. This can consume a lot of time and require
a lot of memory. Remember that your file may be compressed, and
that pixel values represented in the file by bytes are converted
to 8-byte doubles in R. 

The reason for using `RasterIO` for this is that the parameters we
use are directly mapped to the GDAL RasterIO function used (after
adapting the 1-based offset index in R to 0-based offset in C++).

## Reading a particular chunk 

<!-- We can read a (spatially) rectangular chunk of data by -->

An example of using `RasterIO` is
```{r}
library(stars)
tif = system.file("tif/L7_ETMs.tif", package = "stars")
rasterio = list(nXOff = 6, nYOff = 6, nXSize = 100, nYSize = 100, bands = c(1, 3, 4))
(x = read_stars(tif, RasterIO = rasterio))
dim(x)
```
Compare this to
```{r}
st_dimensions(read_stars(tif))
```
and we see that

* the `delta` values remain the same,
* the offset (x/y coordinates of origin) of the grid remain the same
* the `from` and `to` reflect the new area, and relate to the new `delta` values
* `dim(x)` reflects the new size, and
* only three bands were read

## Reading at a different resolution

Reading datasets at a lower (but also higher!) resolution can be
done by setting `nBufXSize` and `nBufYSize`

```{r}
rasterio = list(nXOff = 6, nYOff = 6, nXSize = 100, nYSize = 100,
                nBufXSize = 20, nBufYSize = 20, bands = c(1, 3, 4))
(x = read_stars(tif, RasterIO = rasterio))
```
and we see that in addition:

* the `delta` (raster cell size) values have increased a factor 5, because `nBufXSize` and `nBufYSize` were set to values a factor 5 smaller than `nXSize` and `nYSize`
* the offset coordinates of the grid are still the same
* the `from` and `to` reflect the new area, but relate to the new `delta` cell size values

We can also read at higher resolution; here we read a 3 x 3 area and blow it up to 100 x 100:
```{r}
rasterio = list(nXOff = 6, nYOff = 6, nXSize = 3, nYSize = 3,
   nBufXSize = 100, nBufYSize = 100, bands = 1)
x = read_stars(tif, RasterIO = rasterio)
dim(x)
plot(x)
```

The reason we "see" only three grid cells is that the default sampling method is "nearest neighbour". We can modify this by

```{r}
rasterio = list(nXOff = 6, nYOff = 6, nXSize = 3, nYSize = 3,
   nBufXSize = 100, nBufYSize = 100, bands = 1, resample = "cubic_spline")
x = read_stars(tif, RasterIO = rasterio)
dim(x)
plot(x)
```

The following methods are allowed for parameter `resample`:

| `resample` | method used                        |
|------------|------------------------------------|
|`nearest_neighbour`|   Nearest neighbour (default) |
|`bilinear` |      Bilinear (2x2 kernel) |
|`cubic`     |  Cubic Convolution Approximation (4x4 kernel) |
|`cubic_spline` |  Cubic B-Spline Approximation (4x4 kernel) |
|`lanczos` |  Lanczos windowed sinc interpolation (6x6 kernel) |
|`average` |  Average |
|`mode` |  Mode (selects the value which appears most often of all the sampled points) |
|`Gauss` |  Gauss blurring |

All these methods are implemented in GDAL; for what these methods exactly do, we refer to the GDAL documentation or source code.

# Stars proxy objects

Stars proxy objects take another approach: upon creation they contain no data at all, but only pointers to where the data can be read. Data is only read when it is needed, and only as much as is needed: if we plot a proxy objects, the data are read at the resolution of pixels on the screen, rather than at the native resolution, so that if we have e.g. a 10000 x 10000 Sentinel 2 (level 1C) image, we can open it by
```{r, eval=EVAL}
granule = system.file("sentinel/S2A_MSIL1C_20180220T105051_N0206_R051_T32ULE_20180221T134037.zip", package = "starsdata")
s2 = paste0("SENTINEL2_L1C:/vsizip/", granule, "/S2A_MSIL1C_20180220T105051_N0206_R051_T32ULE_20180221T134037.SAFE/MTD_MSIL1C.xml:10m:EPSG_32632")
(p = read_stars(s2, proxy = TRUE))
```
and this happens _instantly_, because no data is read. When we plot this object,
```{r,eval=EVAL}
system.time(plot(p))
```

This takes only around 1 second, since only those pixels are read that can be seen on the plot. If we read the entire image in memory first, as we would do with
```{r eval = FALSE}
p = read_stars(s2, proxy = FALSE)
```
then only the reading would take over a minute, and require 5 Gb memory.

## Methods for stars proxy objects

```{r}
methods(class = "stars_proxy")
```

## Select attributes

We can select attributes as with regular `stars` objects, by using the first argument to `[`:
```{r,eval=EVAL}
x = c("avhrr-only-v2.19810901.nc",
"avhrr-only-v2.19810902.nc",
"avhrr-only-v2.19810903.nc",
"avhrr-only-v2.19810904.nc",
"avhrr-only-v2.19810905.nc",
"avhrr-only-v2.19810906.nc",
"avhrr-only-v2.19810907.nc",
"avhrr-only-v2.19810908.nc",
"avhrr-only-v2.19810909.nc")
file_list = system.file(paste0("netcdf/", x), package = "starsdata")
y = read_stars(file_list, quiet = TRUE, proxy = TRUE)
names(y)
y["sst"]
```
Note that this selection limits the reading from 4 to 1 subdataset from all 9 NetCDF files.

## Select an area

Another possibility is to crop, or select a rectangular region based on a spatial object. This can be done by passing a `bbox` object, or an `sf`, `sfc` or `stars` object from which the bounding box will be taken. An example:

```{r,eval=EVAL}
bb = st_bbox(c(xmin = 10.125, ymin = 0.125, xmax = 70.125, ymax = 70.125))
ysub = y[bb]
st_dimensions(ysub)
class(ysub) # still no data here!!
plot(ysub, reset = FALSE) # plot reads the data, at resolution that is relevant
plot(st_as_sfc(bb), add = TRUE, lwd = .5, border = 'red')
```

## Lazy evaluation, changing evaluation order

Some other actions can be carried out on `stars_proxy` objects, but their effect is delayed until the data are actually needed (`plot`, `write_stars`).  For instance, range selections on dimensions other than shown above first need data, and can only then be carried out. Such functions are added to the object, in an attribute called `call_list`:

```{r,eval=EVAL}
yy = adrop(y)
yyy = yy[,1:10,1:10,]
class(yyy) # still no data
st_dimensions(yyy) # and dimensions not adjusted
attr(yyy, "call_list") # the name of object in the call (y) is replaced with x:
```

Doing this allows for optimizing the order in which operations are done. As an example, for `st_apply`, reading can be done sequentially over the dimensions over which the function is applied: 

* If for example a function is applied to each band (such as: compute band quantiles), bands can be read sequentially, and discarded after the quantiles have been computed. 
* If a time series function is applied to pixel time series and the result is plotted on a map, the time series function is only evaluated on the pixels actually plotted. This means that e.g. in
```{r eval = FALSE}
plot(st_apply(x, c("x", "y"), range))
```
the order of evaluation is reversed:  `plot` knows which pixels are going to be shown, and controls how `x` is downsampled *before* `st_apply` is carried out on this subset.


### Fetching the data

Fetching the data now involves reading the whole array and then evaluating the `call_list` on it, sequentially:
```{r,eval=EVAL}
(x = st_as_stars(yyy)) # read, adrop, subset
```

### Plotting with changed evaluation order

For the Sentinel 2 data, band 4 represents NIR and band 1 red, so we can compute NDVI by
```{r,eval=EVAL}
# S2 10m: band 4: near infrared, band 1: red.
#ndvi = function(x) (x[4] - x[1])/(x[4] + x[1])
ndvi = function(x1, x2, x3, x4) (x4 - x1)/(x4 + x1)
rm(x)
(s2.ndvi = st_apply(p, c("x", "y"), ndvi))
system.time(plot(s2.ndvi)) # read - compute ndvi - plot 
```

# Multi-resolution proxy objects

This sections shows some examples how `stars_proxy` objects deal with the situation where the different maps have dissimilar resolution. The assumptions here are:

* all maps need to have the same origin coordinates (typically upper-left corner) and CRS.
* the first map determines the "working" resolution, to which e.g. native or downsampled resolutions refer

We'll create four maps with cells size 1, 2 and 3:
```{r}
s1 = st_as_stars(matrix(1:16, 4))
s2 = st_as_stars(matrix(1:16, 4))
s3 = st_as_stars(matrix(1:16, 4))
attr(s1, "dimensions")$X1$offset = 0
attr(s1, "dimensions")$X2$offset = 4
attr(s2, "dimensions")$X1$offset = 0
attr(s2, "dimensions")$X2$offset = 4
attr(s3, "dimensions")$X1$offset = 0
attr(s3, "dimensions")$X2$offset = 4
attr(s1, "dimensions")$X1$delta =  1
attr(s1, "dimensions")$X2$delta = -1
attr(s2, "dimensions")$X1$delta =  2
attr(s2, "dimensions")$X2$delta = -2
attr(s3, "dimensions")$X1$delta =  3
attr(s3, "dimensions")$X2$delta = -3
plot(s1, axes = TRUE, text_values = TRUE, text_color = 'orange')
plot(s2, axes = TRUE, text_values = TRUE, text_color = 'orange')
plot(s3, axes = TRUE, text_values = TRUE, text_color = 'orange')
```

We created three rasters with identical cell values and dimensions, but different cell sizes, and hence extents. If we bind them in a single proxy object, with

```{r eval=TRUE}
fn1 = paste0(tempdir(), .Platform$file.sep, "img1.tif")
fn2 = paste0(tempdir(), .Platform$file.sep, "img2.tif")
fn3 = paste0(tempdir(), .Platform$file.sep, "img3.tif")
write_stars(s1, fn1)
write_stars(s2, fn2)
write_stars(s3, fn3) 
(r1 = read_stars(c(fn1, fn2, fn3), proxy = TRUE))
```

We see that **multi-resolution** is mentioned in the printed summary. When converting this to a `stars` object, the secondary rasters are resampled to the cellsize + extent of the first:

```{r eval=TRUE}
st_as_stars(r1) |>
  merge() |>
  plot(breaks = "equal", text_values = TRUE, text_color = 'orange', axes = TRUE)
```

If we do this for a sub-range, defined for the object resolutions, we get:
```{r eval=TRUE}
st_as_stars(r1[,2:4,2:4]) |>
  merge() |>
  plot(breaks = "equal", text_values = TRUE, text_color = 'orange', axes = TRUE)
```

We now create four maps, all over the same region ([0,4] x [0,4]), with different resolutions (cell size 1, 1/2 and 1/3): 
```{r}
s4 = st_as_stars(matrix(1: 16, 4))
s5 = st_as_stars(matrix(1: 64, 8))
s6 = st_as_stars(matrix(1:144,12))
attr(s4, "dimensions")$X1$offset = 0
attr(s4, "dimensions")$X2$offset = 4
attr(s5, "dimensions")$X1$offset = 0
attr(s5, "dimensions")$X2$offset = 4
attr(s6, "dimensions")$X1$offset = 0
attr(s6, "dimensions")$X2$offset = 4
attr(s4, "dimensions")$X1$delta =  1
attr(s4, "dimensions")$X2$delta = -1
attr(s5, "dimensions")$X1$delta =  1/2
attr(s5, "dimensions")$X2$delta = -1/2
attr(s6, "dimensions")$X1$delta =  1/3
attr(s6, "dimensions")$X2$delta = -1/3
plot(s4, axes = TRUE, text_values = TRUE, text_color = 'orange')
plot(s5, axes = TRUE, text_values = TRUE, text_color = 'orange')
plot(s6, axes = TRUE, text_values = TRUE, text_color = 'orange')
```

```{r eval=TRUE}
fn4 = paste0(tempdir(), .Platform$file.sep, "img4.tif")
fn5 = paste0(tempdir(), .Platform$file.sep, "img5.tif")
fn6 = paste0(tempdir(), .Platform$file.sep, "img6.tif")
write_stars(s4, fn4)
write_stars(s5, fn5)
write_stars(s6, fn6) 
(r2 = read_stars(c(fn4, fn5, fn6), proxy = TRUE))

st_as_stars(r2) |>
  merge() |>
  plot(breaks = "equal", text_values = TRUE, text_color = 'orange', axes = TRUE)
st_as_stars(r2[,2:4,2:4]) |>
  merge() |>
  plot(breaks = "equal", text_values = TRUE, text_color = 'orange', axes = TRUE)
```

Finally, an example where the first raster has the higher resolution:

```{r eval=TRUE}
(r3 = read_stars(c(fn6, fn5, fn4), proxy = TRUE))

st_as_stars(r3) |>
  merge() |>
  plot(breaks = "equal", text_values = TRUE, text_color = 'orange', axes = TRUE)
st_as_stars(r3[,2:6,3:6]) |>
  merge() |>
  plot(breaks = "equal", text_values = TRUE, text_color = 'orange', axes = TRUE)
```

---
title: "5. vector-raster conversions, reprojection, warping"
author: "Edzer Pebesma"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{5. vector-raster conversions, reprojection, warping}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

**For a better version of the stars vignettes see** https://r-spatial.github.io/stars/articles/

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE, collapse = TRUE, dev = "png")
suppressPackageStartupMessages(library(sf))
knitr::opts_chunk$set(fig.height = 4.5)
knitr::opts_chunk$set(fig.width = 6)
ev = TRUE
```

This vignette shows how `stars` object can be moved from vector
and raster representations.

# Rasterizing an `sf` vector object

```{r}
library(stars)
system.file("gpkg/nc.gpkg", package = "sf") |>
  read_sf() |>
  st_transform(32119) -> nc
nc$dens = nc$BIR79 / units::set_units(st_area(nc), km^2)
(nc.st = st_rasterize(nc["dens"], dx = 5000, dy = 5000))
plot(nc.st)
```

The algorithm used is the GDAL `rasterize` utility, all options of
this utility can be passed to `st_rasterize`. The geometry of the
final raster can be controlled by passing a target bounding box and
either the raster dimensions `nx` and `ny`, or pixel size by the
`dx` and `dy` parameters.

# Vectorizing a raster object to an `sf` object

`stars` objects can be converted into an `sf` object using
`st_as_sf`.  It has a number of options, depending on whether
pixels represent the point value at the pixel center, or small
square polygons with a single value.

We will work again with the landsat-7 6-band image, but will select
the first band and round the values:
```{r}
tif = system.file("tif/L7_ETMs.tif", package = "stars")
x = read_stars(tif)[, 1:50, 1:50, 1:2]
x[[1]] = round(x[[1]]/5)
```

## Polygonizing

In case raster cells reflect point values and we want to get
a vector representation of the whole field, we can draw contour
lines and export the contour sets (only available when the 
GDAL version is at least 2.4.0):
```{r eval=ev}
l =  st_contour(x, contour_lines = TRUE, breaks = 11:15)
plot(l[1], key.pos = 1, pal = sf.colors, lwd = 2, key.length = 0.8)
```

## Exporting to points

Alternatively, we can simply export all the pixels as points, and
get them either as a wide table with all bands per point, and no
replicated `POINT` geometries:
```{r}
st_as_sf(x, as_points = TRUE, merge = FALSE)
```
or as a long table with a single attribute and all points replicated:
```{r}
st_as_sf(x, as_points = TRUE, merge = FALSE, long = TRUE)
```
as we can see, an additional attribute `band` now indicates which band is concerned.

## Exporting to polygons

Alternatively, we can export to polygons and either get a single polygon per
pixel, as in
```{r}
st_as_sf(x[1], as_points = FALSE, merge = FALSE)
```
or merge polygons that have identical pixel values;
```{r}
p = st_as_sf(x, as_points = FALSE, merge = TRUE)
```
When plotted with boundaries, we see the resolved boundaries of areas with the
same pixel value:
```{r}
plot(p)
```

A further option `connect8` can be set to `TRUE` to use 8
connectedness, rather than the default 4 connectedness algorithm.
In both cases, the polygons returned will often be invalid according
to the simple feature standard, but can be made valid using
`lwgeom::st_make_valid`.

# Switching between vector and raster in `stars` objects

We can convert a raster dimension to a vector dimension while keeping other dimensions as they are in a `stars` object by
```{r}
x.sf = st_xy2sfc(x, as_points = TRUE)
x.sf
```
which also requires setting the `as_points` arguments as in `st_as_sf`.

# Reprojecting a raster

If we accept that curvilinear rasters are rasters too, and that
regular and rectilinear grids are special cases of curvilinear grids,
reprojecting a raster is no longer a "problem", it just recomputes
new coordinates for every raster cell, and generally results in a
curvilinear grid (that sometimes can be brought back to a regular
or rectilinear grid). If curvilinear grid cells are represented
by coordinates of the cell center, the actual shape of a grid cell
gets lost, and this may be a larger effect if grid cells are large
or if the transformation is stronger non-linear.

An example of the reprojection of the grid created above is
```{r}
nc.st |> st_transform("+proj=laea +lat_0=34 +lon_0=-60") -> nc.curv
nc.curv
plot(nc.curv, border = NA, graticule = TRUE)
```

where it should be noted that the dimensionality of the grid didn't
change: the same set of raster cells has been replotted in the new
CRS, but now in a curvilinear grid.

# Warping a raster

Warping a raster means creating a new _regular_ grid in a new CRS,
based on a (usually regular) grid in another CRS. We can do the
transformation of the previous section by first creating a target
grid:
```{r}
nc |> st_transform("+proj=laea +lat_0=34 +lon_0=-60") |> st_bbox() |>
	st_as_stars() -> newgrid
```
and then warping the old raster to the new
```{r}
nc.st |> st_warp(newgrid) -> nc.new
nc.new 
plot(nc.new)
```

This new object has a regular grid in the new CRS, aligned with
the new x- and y-axes.

---
title: "1. introduction"
author: "Edzer Pebesma"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{1. introduction}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
---

**For a better version of the stars vignettes see** https://r-spatial.github.io/stars/articles/

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE, collapse = TRUE, dev = "png")
ev = suppressWarnings(require(starsdata, quietly = TRUE))
knitr::opts_chunk$set(fig.height = 4.5)
knitr::opts_chunk$set(fig.width = 6)
```

Package `stars` provides infrastructure for _data cubes_, array
data with labeled dimensions, with emphasis on arrays where some
of the dimensions relate to time and/or space.

Spatial data cubes are arrays with one or more spatial dimensions.
Raster data cubes have at least two spatial dimensions that form the
raster tesselation. Vector data cubes have at least one spatial
dimension that may for instance reflect a polygon tesselation, or
a set of point locations. Conversions between the two (rasterization,
polygonization) are provided. Vector data are represented by simple
feature geometries (packages `sf`). Tidyverse methods are provided.

The `stars` package is loaded by
```{r}
library(stars)
```
Spatiotemporal arrays are stored in objects of class `stars`; 
methods for class `stars` currently available are

``` {r}
methods(class = "stars")
```
(tidyverse methods are only visible after loading package `tidyverse`).

# Reading a satellite image

We can read a satellite image through GDAL, e.g. from a GeoTIFF file in the package:
```{r}
tif = system.file("tif/L7_ETMs.tif", package = "stars")
x = read_stars(tif)
plot(x, axes = TRUE)
```

We see that the image is geographically referenced (has coordinate values along axes), and that the object returned (`x`) has three dimensions called `x`, `y` and `band`, and has one attribute:
```{r}
x
```

Each dimension has a name; the meaning of the fields of a single dimension are:

|*field* |*meaning*                                                   |
|--------|------------------------------------------------------------|
| from   | the origin index (1)                 |
| to     | the final index (dim(x)[i])          |
| offset | the start value for this dimension (pixel boundary), if regular |
| delta  | the step (pixel, cell) size for this dimension, if regular      |
| refsys | the reference system, or proj4string |
| point  | logical; whether cells refer to points, or intervals |
| values | the sequence of values for this dimension (e.g., geometries), if irregular |

This means that for an index i (starting at $i=1$) along a certain dimension, the corresponding dimension value (coordinate, time) is $\mbox{offset} + (i-1) \times \mbox{delta}$. This value then refers to the start (edge) of the cell or interval; in order to get the interval middle or cell centre, one needs to add half an offset. 

Dimension `band` is a simple sequence from 1 to 6. Since bands refer to colors, one could put their wavelength values in the `values` field.

For this particular dataset (and most other raster datasets), we see that delta for dimension `y` is negative: this means that consecutive array values have decreasing $y$ values: cell indexes increase from top to bottom, in the direction opposite to the $y$ axis.

`read_stars` reads all bands from a raster dataset, or optionally a subset of raster datasets, into a single `stars` array structure. While doing so, raster values (often UINT8 or UINT16) are converted to double (numeric) values, and scaled back to their original values if needed if the file encodes the scaling parameters.

The data structure `stars` is a generalization of the `tbl_cube` found in `cubelyr`; we can convert to that by
```{r eval=ev}
library(cubelyr)
as.tbl_cube(x)
```
but this will cause a loss of certain properties (cell size,
reference system, vector geometries)

## Switching attributes to dimensions and back

```{r}
(x.spl = split(x, "band"))
merge(x.spl)
```
We see that the newly created dimension lost its name, and the single attribute got a default name. We can set attribute names with `setNames`, and dimension names and values with `st_set_dimensions`:
```{r}
merge(x.spl) |>
  setNames(names(x)) |> 
  st_set_dimensions(3, values = paste0("band", 1:6)) |>
  st_set_dimensions(names = c("x", "y", "band"))
```

## Subsetting

Besides the `tidyverse` subsetting and selection operators explained
in [this vignette](stars3.html), we can also use `[` and `[[`.

Since `stars` objects are a list of `array`s with a metadata table
describing dimensions, list extraction (and assignment) works as expected:

```{r}
class(x[[1]])
dim(x[[1]])
x$two = 2 * x[[1]]
x
```
At this level, we can work with `array` objects directly.

The `stars` subset operator `[` works a bit different: its

* first argument selects attributes
* second argument selects the first dimension
* third argument selects the second dimension, etc

Thus,

```{r}
x["two", 1:10, , 2:4]
```
selects the second attribute, the first 10 columns (x-coordinate),
all rows, and bands 2-4.

Alternatively, when `[` is given a single argument of class `sf`,
`sfc` or `bbox`, `[` will work as a crop operator:
```{r}
circle = st_sfc(st_buffer(st_point(c(293749.5, 9115745)), 400), crs = st_crs(x))
plot(x[circle][, , , 1], reset = FALSE)
plot(circle, col = NA, border = 'red', add = TRUE, lwd = 2)
```

## Overviews
We can read rasters at a lower resolution when they contain so-called overviews. For this
GeoTIFF file, they were created with the `gdaladdo` utility, in particular
```
gdaladdo -r average L7_ETMs.tif  2 4 8 16
```
which adds coarse resolution versions by using the _average_
resampling method to compute values based on blocks of pixels.
These can be read by

```{r}
x1 = read_stars(tif, options = c("OVERVIEW_LEVEL=1"))
x2 = read_stars(tif, options = c("OVERVIEW_LEVEL=2"))
x3 = read_stars(tif, options = c("OVERVIEW_LEVEL=3"))
dim(x1)
dim(x2)
dim(x3)
par(mfrow = c(1, 3), mar = rep(0.2, 4))
image(x1[,,,1])
image(x2[,,,1])
image(x3[,,,1])
```

# Reading a raster time series: NetCDF

Another example is when we read raster time series model outputs in a NetCDF file, e.g. by
```{r eval=ev}
system.file("nc/bcsd_obs_1999.nc", package = "stars") |>
	read_stars() -> w
```

We see that

```{r eval=ev}
w
```

For this dataset we can see that

* variables have units associated (and a wrong unit, `C` is assigned to temperature)
* time is now a dimension, with proper units and time steps

Alternatively, this dataset can be read using `read_ncdf`, as in
```{r}
system.file("nc/bcsd_obs_1999.nc", package = "stars") |>
    read_ncdf()
```

The difference between `read_ncdf` and `read_stars` for NetCDF files
is that the former uses package RNetCDF to directly read the NetCDF
file, where the latter uses the GDAL driver for NetCDF files.

## Reading datasets from multiple files

Model data are often spread across many files. An example of a 0.25
degree grid, global daily sea surface temperature product is found
[here](https://psl.noaa.gov/data/gridded/data.noaa.oisst.v2.highres.html);
the subset from 1981 used below was downloaded from a NOAA ftp
site that is no longer available in this form. (ftp site used to
be eclipse.ncdc.noaa.gov/pub/OI-daily-v2/NetCDF/1981/AVHRR/).

We read the data by giving `read_stars` a vector with character names:
```{r eval=ev}
x = c(
"avhrr-only-v2.19810901.nc",
"avhrr-only-v2.19810902.nc",
"avhrr-only-v2.19810903.nc",
"avhrr-only-v2.19810904.nc",
"avhrr-only-v2.19810905.nc",
"avhrr-only-v2.19810906.nc",
"avhrr-only-v2.19810907.nc",
"avhrr-only-v2.19810908.nc",
"avhrr-only-v2.19810909.nc"
)
# see the second vignette:
# install.packages("starsdata", repos = "https://cran.uni-muenster.de/pebesma/")
file_list = system.file(paste0("netcdf/", x), package = "starsdata")
(y = read_stars(file_list, quiet = TRUE))
```

Next, we select sea surface temperature (`sst`), and drop the singular `zlev` (depth) dimension using `adrop`:
```{r eval=ev}
library(dplyr)
library(abind)
z <- y |> select(sst) |> adrop()
```

We can now graph the sea surface temperature (SST) using `ggplot`, which needs data in a long table form, and without units:
```{r eval=ev}
# convert POSIXct time to character, to please ggplot's facet_wrap()
z1 = st_set_dimensions(z, 3, values = as.character(st_get_dimension_values(z, 3)))
library(ggplot2)
library(viridis)
library(ggthemes)
ggplot() +  
  geom_stars(data = z1[1], alpha = 0.8, downsample = c(10, 10, 1)) + 
  facet_wrap("time") +
  scale_fill_viridis() +
  coord_equal() +
  theme_map() +
  theme(legend.position = "bottom") +
  theme(legend.key.width = unit(2, "cm"))
```

# Writing stars objects to disk

We can write a stars object to disk by using `write_stars`; this used the GDAL write engine. Writing NetCDF files without going through the GDAL interface is currently not supported.
`write_stars` currently writes only a single attribute:

```{r eval=ev}
write_stars(adrop(y[1]), "sst.tif")
```

See the explanation of `merge` above to see how multiple attributes
can be merged (folded) into a dimension.

# Cropping a raster's extent

Using a curvilinear grid, taken from the example of `read_ncdf`:
```{r}
prec_file = system.file("nc/test_stageiv_xyt.nc", package = "stars")
prec = read_ncdf(prec_file, curvilinear = c("lon", "lat"))
##plot(prec) ## gives error about unique breaks
## remove NAs, zeros, and give a large number
## of breaks (used for validating in detail)
qu_0_omit = function(x, ..., n = 22) {
  if (inherits(x, "units"))
    x = units::drop_units(na.omit(x))
  c(0, quantile(x[x > 0], seq(0, 1, length.out = n)))
}
library(dplyr) # loads slice generic
prec_slice = slice(prec, index = 17, along = "time")
plot(prec_slice, border = NA, breaks = qu_0_omit(prec_slice[[1]]), reset = FALSE)
nc = sf::read_sf(system.file("gpkg/nc.gpkg", package = "sf"), "nc.gpkg")
plot(st_geometry(nc), add = TRUE, reset = FALSE, col = NA, border = 'red')
```

We can now crop the grid to those cells falling in 
```{r}
nc = st_transform(nc, st_crs(prec_slice)) # datum transformation
plot(prec_slice[nc], border = NA, breaks = qu_0_omit(prec_slice[[1]]), reset = FALSE)
plot(st_geometry(nc), add = TRUE, reset = FALSE, col = NA, border = 'red')
```

The selection `prec_slice[nc]` essentially calls `st_crop(prec_slice, nc)` to get a cropped selection. What happened here is that all
cells not intersecting with North Carolina (sea) are set to `NA`
values. For regular grids, the extent of the resulting `stars`
object is also be reduced (cropped) by default; this can be
controlled with the `crop` parameter to `st_crop` and `[.stars`.

# Vector data cube example

Like `tbl_cube`, `stars` arrays have no limits to the number of dimensions they handle. An example is the origin-destination (OD) matrix, by time and travel mode.

## OD: space x space x travel mode x time x time

We create a 5-dimensional matrix of traffic between regions, by day, by time of day, and by travel mode. Having day and time of day each as dimension is an advantage when we want to compute patterns over the day, for a certain period.

```{r}
nc = st_read(system.file("gpkg/nc.gpkg", package="sf")) 
to = from = st_geometry(nc) # 100 polygons: O and D regions
mode = c("car", "bike", "foot") # travel mode
day = 1:100 # arbitrary
library(units)
units(day) = as_units("days since 2015-01-01")
hour = set_units(0:23, h) # hour of day
dims = st_dimensions(origin = from, destination = to, mode = mode, day = day, hour = hour)
(n = dim(dims))
traffic = array(rpois(prod(n), 10), dim = n) # simulated traffic counts
(st = st_as_stars(list(traffic = traffic),  dimensions = dims))
```

This array contains the simple feature geometries of origin and destination so that we can directly plot every slice without additional table joins. If we want to represent such an array as a `tbl_cube`, the simple feature geometry dimensions need to be replaced by indexes:
```{r eval=ev}
st |> as.tbl_cube()
```

The following demonstrates how we can use `dplyr` to filter travel mode `bike`, and compute mean bike traffic by hour of day:
```{r eval=ev}
b <- st |> 
  as.tbl_cube() |> 
  filter(mode == "bike") |> 
  group_by(hour) |>
  summarise(traffic = mean(traffic)) |> 
  as.data.frame()
require(ggforce) # for plotting a units variable
ggplot() +  
  geom_line(data = b, aes(x = hour, y = traffic))
```

# Extracting at point locations, aggregating over polygons

Data cube values at point location can be extracted by `st_extract`, an example is found
in [vignette 7](stars7.html)

Aggregates, such as mean, maximum or modal values can be obtained by `aggregate`. In this
example we use a categorical raster, and try to find the modal (most frequent) class within
two circular polygons:

```{r}
s = system.file("tif/lc.tif", package = "stars")
r = read_stars(s, proxy = FALSE) |> droplevels()
levels(r[[1]]) = abbreviate(levels(r[[1]]), 10) # shorten text labels
st_point(c(3190631, 3125)) |> st_sfc(crs = st_crs(r)) |> st_buffer(25000) -> pol1
st_point(c(3233847, 21027)) |> st_sfc(crs = st_crs(r)) |> st_buffer(10000) -> pol2
if (isTRUE(dev.capabilities()$rasterImage == "yes")) {
  plot(r, reset = FALSE, key.pos = 4)
  plot(c(pol1, pol2), col = NA, border = c('yellow', 'green'), lwd = 2, add = TRUE)
}
```

To find the modal value, we need a function that gives back the label
corresponding to the class which is most frequent, using `table`:
```{r}
f = function(x) { tb = table(x); names(tb)[which.max(tb)] }
```

We can then call `aggregate` on the raster map, and the set of the two
circular polygons `pol1` and `pol2`, and pass the function `f`:
```{r}
aggregate(r, c(pol1, pol2), f) |> st_as_sf()
```

---
title: "7. Statistical modelling with stars objects"
author: "Edzer Pebesma"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{7. Statistical modelling with stars objects}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

**For a better version of the stars vignettes see** https://r-spatial.github.io/stars/articles/

---
title: "6. How `raster` functions map to `stars` functions"
author: "Sébastien Rochette, Adriano Fantini, Edzer Pebesma"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{6. How raster functions map to stars functions}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

**For a better version of the stars vignettes see** https://r-spatial.github.io/stars/articles/

```{r, include = FALSE}
knitr::opts_chunk$set(echo = FALSE, collapse = TRUE, dev = "png")
suppressPackageStartupMessages(library(dplyr))
knitr::opts_chunk$set(fig.height = 4.5)
knitr::opts_chunk$set(fig.width = 6)
```

Here's an attempt at the table describing how `raster` functions map to `stars` functions, discussed in [issue #122](https://github.com/r-spatial/stars/issues/122). This table uses the functionality of the `raster` package as a template; it may be incomplete, imprecise or plain wrong, so take it with a pinch of salt. Any comment or correction is hugely appreciated, please contribute!

Some of the functions (`filter`, `slice`, `mutate`, `select`, `pull`, ...) are provided via `dplyr`, which must be loaded. See `?stars::dplyr`.

## COMMENT LEGEND

? = Not sure / unknown

\* = Not present, low priority

\# = Not present, high priority

NA = Not available by design

## Creating objects
```{r}
tribble(
  ~raster, ~stars, ~`Note/comment`,
  "{raster, stack, brick} (read)", "read_stars or read_stars(along = ...)","",
  "{stack, brick} (concatenate layers)", "c or c(along = ...)","",
  "subset", "{[ ] , slice, filter}","",
  "addLayer", "c() or c(along = ...)","",
  "dropLayer", "{[ ] , slice, filter}","",
  "unstack", "combine lapply and {[ ] , slice, filter}",""
) |> 
  knitr::kable()
```


## Changing spatial extent and/or resolution of objects
```{r}
tribble(
  ~raster, ~stars, ~`Note/comment`,
  "merge", "c", "#, currently only works for adjacent objects",
  "mosaic", "st_mosaic", "these are not identical, read the docs carefully",
  "crop", "filter, st_crop", "",
  "setExtent", "", "# maybe use st_warp?",
  "trim", "", "#",
  "aggregate", "aggregate", "WIP; raster's aggregate with fact=2 will not work, use st_warp in that case?",
  "disaggregate", "", "# use st_warp(use_gdal = TRUE)?",
  "resample", "{st_transform, st_warp}", "",
  "projectRaster", "{st_transform, st_warp}", "",
  "shift", "", "#, now use st_set_dimensions",
  "flip", "[] with reversed index", "#",
  "rotate", "", "*",
  "t", "", "NA"
) |> 
  knitr::kable()
```

## Cell based computation
```{r}
tribble(
  ~raster, ~stars, ~`Note/comment`,
  "calc", "st_apply", "",
  "overlay", "c(along = , ...) |> st_apply(...)", "",
  "cover", "[ ] <-", "",
  "mask", "[ ], st_crop", "when using an sf polygon as mask",
  "mask", "[ ]<- NA", "when using a stars object to mask",
  "cut", "cut", "",
  "subs", "", "",
  "reclassify", "mutate with case_when", "or forcats::fct_recode ?",
  "reclassify", "cut", "",
  "init", "[ ] <-", "",
  "stackApply", "{[ ] , slice, filter} |> st_apply(...)", "",
  "stackSelect", "", "",
) |> 
  knitr::kable()
```



## Spatial contextual computation
```{r}
tribble(
  ~raster, ~stars, ~`Note/comment`,
  "distance", "", "#",
  "gridDistance", "", "*",
  "distanceFromPoints", "", "#",
  "direction", "", "*",
  "focal", "f = st_apply(x1, 3, foc, w = matrix(1, 3, 3))", "See. [issue 176](https://github.com/r-spatial/stars/issues/176)",
  "localFun", "", "*",
  "boundaries", "st_as_sf(as_points=FALSE, merge=TRUE, connect8=TRUE)", "",
  "clump", "st_as_sf(r, merge = TRUE)", "`st_as_sf` returns polygons, `clump` a raster",
  "adjacent", "", "*",
  "area", "st_area", "",
  "terrain", "", "#",
  "Moran", "", ""
) |> 
  knitr::kable()
```


## Model predictions
```{r}
tribble(
  ~raster, ~stars, ~`Note/comment`,
  "predict", "predict", "",
  "interpolate", "gstat::idw, gstat::krige", "`st_warp` has raster-raster interpolations of `gdalwarp`"
) |> 
  knitr::kable()
```

## Data type conversion
```{r}
tribble(
  ~raster, ~stars, ~`Note/comment`,
  "rasterize", "st_as_stars", "",
  "rasterToPoints", "st_as_sf(as_points=TRUE)", "",
  "rasterToPolygons", "st_as_sf(as_points=FALSE, ...), st_polygonize", "",
  "rasterToContour", "st_contour", "requires GDAL >= 2.4.0",
  "rasterFromXYZ", "", "",
  "rasterFromCells", "", "",
) |> 
  knitr::kable()
```

## Summarizing
```{r}
tribble(
  ~raster, ~stars, ~`Note/comment`,
  "cellStats", "st_apply", "",
  "summary", "print, summary(as.vector(. |> pull()))", "",
  "freq", "table", "*",
  "crosstab", "", "",
  "unique", "unique(as.vector(. |> pull()))", "",
  "zonal", "", "*"
) |> 
  knitr::kable()
```


## Accessing values of objects
```{r}
tribble(
  ~raster, ~stars, ~`Note/comment`,
  "getValues", "{pull, [[ ]]}", "",
  "getValuesBlock", "{[ ] , slice, filter} |> pull()", "",
  "getValuesFocal", "{[ ] , slice, filter} |> pull()", "",
  "as.matrix", "[[ ]]", "currently behaves somewhat unexpectedly *", 
  "as.array", "[[ ]]", "currently behaves somewhat unexpectedly *", 
  "extract (by cell)", "{[ ] , slice, filter}", "",
  "extract (by polygon)", "x[sf_object]", "",
  "extract (by point)", "aggregate(stars_object, sf_object, function(x) x[1], as_points = FALSE)", "",
  "sampleRandom", "", "*", 
  "sampleRegular", "", "*",
  "minValue", "purrr::map(x, min)", "",
  "maxValue", "purrr::map(x, max)", "",
  "setMinMax", "", "," 
) |> 
  knitr::kable()
```


## Plotting
```{r}
tribble(
  ~raster, ~stars, ~`Note/comment`,
  "plot", "plot, geom_stars", "",
  "plotRGB", "plot(x, rgb =...)", "",
  "spplot", "-", "",
  "image", "image", "",
  "persp", "-", "",
  "contour", "(st_contour, then sf::plot)", "",
  "filledContour", "(same)", "",
  "text", "text", "",
  "`hist`", "`hist(x[[1]])`", "",
  "barplot", "", "",
  "density", "", "",
  "pairs", "", "",
  "boxplot", "", ""
) |> 
  knitr::kable()
```


## Getting and setting dimensions
```{r}
tribble(
  ~raster, ~stars, ~`Note/comment`,
  "ncol", "dim(x)[1]", "or use name instead of 1; cols may be the second dimension!", 
  "nrow", "dim(x)[2]", "or use name instead of 2; rows may be the first dimension!", 
  "ncell", "prod(dim(x))", "",
  "res", "st_dimensions", "can also not be a constant in case of rectilinear or curvilinear grids", 
  "nlayers", "-", "there is no concept of layers in stars", 
  "names", "names", "",
  "xres", "st_res(x)[1]", "may not be a constant in case of rectilinear or curvilinear grids", 
  "yres", "st_res(x)[2]", "may not be a constant in case of rectilinear or curvilinear grids", 
  "xmin", "st_bbox(x)[1]", "",
  "xmax", "st_bbox(x)[3]", "",
  "ymin", "st_bbox(x)[2]", "",
  "ymax", "st_bbox(x)[4]", "",
  "extent", "st_bbox(x)", "different ordering of numbers", 
  "origin", "-", "",
  "projection", "st_crs(x)", "",
  "isLonLat", "st_is_longlat(st_crs(x))", "",
  "filename", "", "stars_proxy objects carry file names where otherwise the array data is", 
  "bandnr", "", "stars has no general concept of bands", 
  "nbands", "dim(x)[3]", "may also be time; bands may also be in another dimension, or have another name", 
  "compareRaster", "all.equal(st_dimensions(x), st_dimensions(y))", "*", 
  "NAvalue", "", "-"
) |> 
  knitr::kable()
```


## Computing row, column, cell numbers and coordinates
```{r}
tribble(
  ~raster, ~stars, ~`Note/comment`,
  "xFromCol", "st_get_dimension_values(., 'x')[col]", "I am not sure how to generally get the x dimension name - here it is x",
  "yFromRow", "st_get_dimension_values(., 'y')[row]", "I am not sure how to generally get the y dimension name - here it is y", 
  "xFromCell", "", "",
  "yFromCell", "", "",
  "xyFromCell", "", "", 
  "colFromX", "", "*",
  "rowFromY", "", "*", 
  "rowColFromCell", "", "",
  "cellFromXY", "st_cells()", "",
  "cellFromRowCol", "", "",
  "cellsFromExtent", "", "",
  "coordinates", "st_coordinates", "",
  "validCell", "", "",
  "validCol", "col |> between(st_dimensions(.)$x$from, st_dimensions(.)$x$to)", "raster columns are not always named 'x'", 
  "validRow", "row |> between(st_dimensions(.)$y$from, st_dimensions(.)$y$to)", "raster columns are not always named 'y'", 
  "setValues", "[ ] <-", "",
  "writeRaster", "write_stars", "currently uses GDAL, somewhat limited", 
  "KML", "", ""
) |> 
  knitr::kable()
```


The format of this table follows the `raster-package` entry in the raster manual, found at `https://cran.r-project.org/web/packages/raster/raster.pdf`.

---
title: '4. stars data model'
author: "Edzer Pebesma"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{4. stars data model}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

**For a better version of the stars vignettes see** https://r-spatial.github.io/stars/articles/

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE, collapse = TRUE, dev = "png")
set.seed(13579)
knitr::opts_chunk$set(fig.height = 4.5)
knitr::opts_chunk$set(fig.width = 6)
```

This vignette explains the data model of `stars` objects, illustrated using artificial and real datasets.

## Stars objects

`stars` objects consist of 

* a (possibly empty) named list of arrays, each having named dimensions (`dim`) attribute
* an attribute called `dimensions` of class `dimensions` that carries dimension metadata
* a class name that includes `stars`

A `dimensions` object is a named list of `dimension` elements, each
describing the semantics a dimension of the data arrays (space,
time, type etc). In addition to that, a `dimensions` object has an
attribute called `raster` of class `stars_raster`, which is a named
list with three elements:

* `dimensions` length 2 character; the dimension names that constitute a spatial raster (or NA)
* `affine` length 2 numeric; the two affine parameters of the geotransform (or NA)
* `curvilinear` a boolean indicating whether a raster is a curvilinear raster (or NA)

The `affine` and `curvilinear` values are only relevant in case of
raster data, indicated by `dimensions` to have non-NA values.

A `dimension` object describes a _single_ dimension; it is a list with
named elements

* `from`: (numeric length 1): the start index of the array
* `to`: (numeric length 1): the end index of the array
* `offset`: (numeric length 1): the start coordinate (or time) value of the first pixel (i.e., a pixel/cell boundary)
* `delta`: (numeric length 1): the increment, or cell size
* `refsys`: (character, or `crs`): object describing the reference system; e.g. the PROJ string, or string `POSIXct` or `PCICt` (for 360 and 365 days/year calendars), or object of class `crs` (containing both EPSG code and proj4string)
* `point`: (logical length 1): boolean indicating whether cells/pixels refer to areas/periods, or to points/instances (may be NA)
* `values`: one of 
    * `NULL` (missing), 
    * a vector with coordinate values (numeric, `POSIXct`, `PCICt`, or `sfc`), 
    * an object of class `intervals` (a list with two vectors, `start` and `end`, with interval start- and end-values), or
    * a matrix with longitudes or latitudes for all cells (in case of curvilinear grids)

`from` and `to` will usually be 1 and the dimension size, but
`from` may be larger than 1 in case a sub-grid got was selected
(or cropped).

`offset` and `delta` only apply to _regularly_ discretized
dimensions, and are `NA` if this is not the case. If they are `NA`,
dimension values may be held in the `values` field.  Rectilinear and
curvilinear grids need grid values in `values` that can be either:

* for rectilinear grids: irregularly _spaced_ coordinate values, or coordinate _intervals_ of irregular width 
  (a rectilinear grid _can_ have one dimension that is regular), 
* for curvilinear grids: or a matrix with grid cell centre values for _all_ row/col combinations (usually in longitude or latitude).

Alternatively, `values` can contains a set of spatial geometries
encoded in an `sfc` vector ("list-column"), in which case we have a
[vector data cube](https://r-spatial.org/r/2022/09/12/vdc.html).

## Grid type

### Regular grids
With a very simple file created from a $4 \times 5$ matrix
```{r fig.width=4.5, fig.height=4}
suppressPackageStartupMessages(library(stars))
m = matrix(1:20, nrow = 5, ncol = 4)
dim(m) = c(x = 5, y = 4) # named dim
(s = st_as_stars(m))
```
we see that

* the rows (5) are mapped to the first dimension, the x-coordinate
* the columns (4) are mapped to the second dimension, the y-coordinate
* the `from` and `to` fields of each dimension define a range that corresponds to the array dimension:

```{r}
dim(s[[1]])
```
* offset and delta specify how increasing row and column index maps to x and y coordinate values respectively.

When we plot this object, using the `image` method for `stars` objects, 
```{r fig.width=4.5, fig.height=4}
image(s, text_values = TRUE, axes = TRUE)
```

we see that $(0,0)$ is the origin of the grid (grid corner), and $1$ the coordinate value increase from one index (row, col) to the next. It means that consecutive matrix columns represent grid lines, going from south to north. Grids defined this way are **regular**: grid cell size is constant everywhere.

Many actual grid datasets have y coordinates (grid rows) going from North to South (top to bottom); this is realised with a negative value for `delta`. We see that the grid origin $(0,0)$ did not change:
```{r fig.width=4.5, fig.height=4}
attr(s, "dimensions")[[2]]$delta = -1
image(s, text_values = TRUE, axes = TRUE)
```

An example is the GeoTIFF carried in the package, which, as probably all data sources read through GDAL, has a negative `delta` for the `y`-coordinate:
```{r}
tif = system.file("tif/L7_ETMs.tif", package = "stars")
st_dimensions(read_stars(tif))["y"]
```

### Raster attributes, rotated and sheared grids

Dimension tables of `stars` objects carry a `raster` attribute:
```{r}
str(attr(st_dimensions(s), "raster"))
```

which is a list that holds

* `dimensions`: character, the names of raster dimensions (if any), as opposed to e.g. spectral, temporal or other dimensions
* `affine`: numeric, the affine parameters
* `curvilinear`: a logical indicating whether the raster is curvilinear

These fields are needed at this level, because they describe properties of the array at a higher level than individual dimensions do: a pair of dimensions forms a raster, both `affine` and `curvilinear` describe how x and y _as a pair_ are derived from grid indexes (see below) when this cannot be done on a per-dimension basis.

With two affine parameters $a_1$ and $a_2$, $x$ and $y$ coordinates are derived from (1-based) grid indexes $i$ and $j$, grid offset values $o_x$ and $o_y$, and grid cell sizes $d_x$ and $d_y$ by

$$x = o_x + (i-1) d_x + (j-1) a_1$$
    
$$y = o_y + (i-1) a_2 + (j-1) d_y$$
Clearly, when $a_1=a_2=0$, $x$ and $y$ are entirely derived from their respective index, offset and cellsize.

Note that for integer indexes, the coordinates are that of the starting edge of a grid cell; to get the grid cell center of the top left grid cell (in case of a negative $d_y$), use $i=1.5$ and $j=1.5$.

We can rotate grids by setting $a_1$ and $a_2$ to a non-zero value:

```{r}
attr(attr(s, "dimensions"), "raster")$affine = c(0.1, 0.1)
plot(st_as_sf(s, as_points = FALSE), axes = TRUE, nbreaks = 20)
```

The rotation angle, in degrees, is

```{r}
atan2(0.1, 1) * 180 / pi
```

Sheared grids are obtained when the two rotation coefficients, $a_1$ and $a_2$, are unequal:
```{r}
attr(attr(s, "dimensions"), "raster")$affine = c(0.1, 0.2)
plot(st_as_sf(s, as_points = FALSE), axes = TRUE, nbreaks = 20)
```

Now, the y-axis and x-axis have different rotation in degrees of respectively
```{r}
atan2(c(0.1, 0.2), 1) * 180 / pi
```

## Rectilinear grids

[Rectilinear grids](https://en.wikipedia.org/wiki/Regular_grid) have orthogonal axes, but do not have congruent (equally sized and shaped) cells: each axis has its own irregular subdivision.

We can define a rectilinear grid by specifying the cell _boundaries_, meaning for every dimension we specify _one more_ value than the dimension size:
```{r}
x = c(0, 0.5, 1, 2, 4, 5)  # 6 numbers: boundaries!
y = c(0.3, 0.5, 1, 2, 2.2) # 5 numbers: boundaries!
(r = st_as_stars(list(m = m), dimensions = st_dimensions(x = x, y = y)))
st_bbox(r)
image(r, axes = TRUE, col = grey((1:20)/20))
```

Would we leave out the last value, than `stars` may come up with a _different_ cell boundary for the last cell, as this is now derived from the width of the one-but-last cell:
```{r}
x = c(0, 0.5, 1, 2, 4)  # 5 numbers: offsets only!
y = c(0.3, 0.5, 1, 2)   # 4 numbers: offsets only!
(r = st_as_stars(list(m = m), dimensions = st_dimensions(x = x, y = y)))
st_bbox(r)
```

This is not problematic if cells have a constant width, in which case the boundaries are reduced to an `offset` and `delta` value, irrespective whether an upper boundary is given:
```{r}
x = c(0, 1, 2, 3, 4)  # 5 numbers: offsets only!
y = c(0.5, 1, 1.5, 2)   # 4 numbers: offsets only!
(r = st_as_stars(list(m = m), dimensions = st_dimensions(x = x, y = y)))
st_bbox(r)
```

Alternatively, one can also set the _cell midpoints_ by specifying arguments `cell_midpoints` to the `st_dimensions` call:
```{r}
x = st_as_stars(matrix(1:9, 3, 3), 
                st_dimensions(x = c(1, 2, 3), y = c(2, 3, 10), cell_midpoints = TRUE))
```
When the dimension is regular, this results in `offset` being
shifted back with half a `delta`, or else in intervals derived from
the distances between cell centers.  This should obviously not be
done when cell boundaries are specified.

## Curvilinear grids

Curvilinear grids are grids whose grid lines are not straight. Rather than describing the curvature parametrically, the typical (HDF5 or NetCDF) files in which they are found have two raster layers with the longitudes and latitudes for every corresponding pixel of remaining layers. 

As an example, we will use a Sentinel 5P dataset available from package `starsdata`; this package can be installed with
```{r eval=FALSE}
install.packages("starsdata", repos = "https://cran.uni-muenster.de/pebesma/")
```

The dataset is found here:
```{r}
(s5p = system.file("sentinel5p/S5P_NRTI_L2__NO2____20180717T120113_20180717T120613_03932_01_010002_20180717T125231.nc", package = "starsdata"))
```

```{r echo=FALSE}
EVAL = s5p != ""
```
We can construct the curvilinear `stars` raster by calling `read_stars` on the right sub-array:
```{r eval=EVAL}
subs = gdal_subdatasets(s5p)
subs[[6]]
```

For this array, we can see the GDAL metadata under item `GEOLOCATION`:
```{r eval=EVAL}
gdal_metadata(subs[[6]], "GEOLOCATION")
```

which reveals where, in this dataset, the longitude and latitude arrays are kept.

```{r eval=EVAL}
nit.c = read_stars(subs[[6]]) 
threshold = units::set_units(9e+36, mol/m^2)
nit.c[[1]][nit.c[[1]] > threshold] = NA
nit.c
```

The curvilinear array has the actual arrays (raster layers, matrices)
with longitude and latitude values read in its dimension table. We
can plot this file:

```{r eval=EVAL}
plot(nit.c, breaks = "equal", reset = FALSE, axes = TRUE, as_points = TRUE, 
		 pch = 16,  logz = TRUE, key.length = 1)
maps::map('world', add = TRUE, col = 'red')
```

```{r eval=EVAL}
plot(nit.c, breaks = "equal", reset = FALSE, axes = TRUE, as_points = FALSE, 
		 border = NA, logz = TRUE, key.length = 1)
maps::map('world', add = TRUE, col = 'red')
```

We can downsample the data by
```{r eval=EVAL}
(nit.c_ds = stars:::st_downsample(nit.c, 8))
plot(nit.c_ds, breaks = "equal", reset = FALSE, axes = TRUE, as_points = TRUE, 
		 pch = 16, logz = TRUE, key.length = 1)
maps::map('world', add = TRUE, col = 'red')
```

which doesn't look nice, but plotting the cells as polygons looks better:
```{r eval=EVAL}
plot(nit.c_ds, breaks = "equal", reset = FALSE, axes = TRUE, as_points = FALSE, 
		 border = NA, logz = TRUE, key.length = 1)
maps::map('world', add = TRUE, col = 'red')
```

Another approach would be to warp the curvilinear grid to a regular grid,
e.g. by

```{r eval=EVAL}
w = st_warp(nit.c, crs = 4326, cellsize = 0.25)
plot(w)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidyverse.R
\name{geom_stars}
\alias{geom_stars}
\alias{theme_stars}
\title{ggplot geom for stars objects}
\usage{
geom_stars(
  mapping = NULL,
  data = NULL,
  ...,
  downsample = 0,
  sf = FALSE,
  na.action = na.pass
)

theme_stars(...)
}
\arguments{
\item{mapping}{see \link[ggplot2:geom_tile]{geom_raster}}

\item{data}{see \link[ggplot2:geom_tile]{geom_raster}}

\item{...}{see \link[ggplot2:geom_tile]{geom_raster}}

\item{downsample}{downsampling rate: e.g. 3 keeps rows and cols 1, 4, 7, 10 etc.; a value of 0 does not downsample; can be specified for each dimension, e.g. \code{c(5,5,0)} to downsample the first two dimensions but not the third.}

\item{sf}{logical; if \code{TRUE} rasters will be converted to polygons and plotted using \link[ggplot2:ggsf]{geom_sf}.}

\item{na.action}{function; if \code{NA} values need to be removed before plotting use the value \code{na.omit} here (only applies to objects with raster dimensions)}
}
\description{
ggplot geom for stars objects
}
\details{
\code{geom_stars} returns (a call to) either \link[ggplot2:geom_tile]{geom_raster}, \link[ggplot2]{geom_tile}, or \link[ggplot2:ggsf]{geom_sf}, depending on the raster or vector geometry; for the first to, an \link[ggplot2]{aes} call is constructed with the raster dimension names and the first array as fill variable. Further calls to \link[ggplot2:coord_fixed]{coord_equal} and \link[ggplot2]{facet_wrap} are needed to control aspect ratio and the layers to be plotted; see examples. If a \code{stars} array contains hex color values, and no \code{fill} parameter is given, the color values are used as fill color; see the example below.

If visual artefacts occur (Moiré-Effekt), then see the details section of \link{plot.stars}
}
\examples{
system.file("tif/L7_ETMs.tif", package = "stars") |> read_stars() -> x
if (require(ggplot2, quietly = TRUE)) {
  ggplot() + geom_stars(data = x) +
    coord_equal() +
    facet_wrap(~band) +
    theme_void() +
    scale_x_discrete(expand=c(0,0))+
    scale_y_discrete(expand=c(0,0))
  # plot rgb composite:
  st_as_stars(L7_ETMs)[,,,1:3] |> st_rgb() -> x # x contains colors as pixel values
  ggplot() + geom_stars(data = x)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/proxy.R, R/subset.R
\name{stars_subset}
\alias{stars_subset}
\alias{[<-.stars_proxy}
\alias{[.stars}
\alias{[<-.stars}
\alias{st_flip}
\title{subset stars objects}
\usage{
\method{[}{stars_proxy}(x, i, ..., downsample = 0) <- value

\method{[}{stars}(x, i = TRUE, ..., drop = FALSE, crop = !is_curvilinear(x))

\method{[}{stars}(x, i) <- value

st_flip(x, which = 1)
}
\arguments{
\item{x}{object of class \code{stars}}

\item{i}{first selector: integer, logical or character vector indicating attributes to select, or object of class \code{sf}, \code{sfc}, \code{bbox}, or \code{stars} used as spatial selector; see details}

\item{...}{further (logical or integer vector) selectors, matched by order, to select on individual dimensions}

\item{downsample}{downsampling rate used in case \code{i} is a \code{stars_proxy} object}

\item{value}{array of dimensions equal to those in \code{x}, or a vector or value that will be recycled to such an array}

\item{drop}{logical or character; if \code{TRUE}, degenerate dimensions (with only one value) are dropped; if \code{character} name(s) of the degenerate dimension(s) to drop}

\item{crop}{logical; if \code{TRUE} and parameter \code{i} is a spatial geometry (\code{sf} or \code{sfc}) object, the extent (bounding box) of the result is cropped to match the extent of \code{i} using \link{st_crop}. Cropping curvilinear grids is not supported.}

\item{which}{character or integer; dimension(s) to be flipped}
}
\value{
\code{st_flip} flips (reverts) the array values along the chosen dimension 
without(s) changing the dimension properties
}
\description{
subset stars objects
}
\details{
If \code{i} is an object of class \code{sf}, \code{sfc} or \code{bbox}, the spatial subset covering this geometry is selected, possibly followed by cropping the extent. Array values for which the cell centre is not inside the geometry are assigned \code{NA}. If \code{i} is of class \code{stars}, and attributes of \code{i} are \code{logical}, cells in \code{x} corresponding to \code{NA} or \code{FALSE} cells in \code{i} are assigned an \code{NA}. Dimension ranges containing negative values or \code{NA} may be partially supported. Character selectors are matched against the names of a dimension if it has names, otherwise to the dimension values.

in an assignment (or replacement form, \code{[<-}), argument \code{i} needs to be either (i) a \code{stars} object with logical attribute(s) that has dimensions matching (possibly after recycling) those of \code{x}, in which case the \code{TRUE} cells will be replaced and \code{i} and/or \code{value} will be recycled to the dimensions of the arrays in \code{x}, or (ii) a length-one integer or character vector indicating which array to replace, in which case \code{value} may be stars object or a vector or array (that will be recycled).
}
\examples{
tif = system.file("tif/L7_ETMs.tif", package = "stars")
x = read_stars(tif)
x[,,,1:3] # select bands
x[,1:100,100:200,] # select x and y by range
x["L7_ETMs.tif"] # select attribute
xy = structure(list(x = c(293253.999046018, 296400.196497684), y = c(9113801.64775462,
9111328.49619133)), names = c("x", "y"))
pts = st_as_sf(data.frame(do.call(cbind, xy)), coords = c("x", "y"), crs = st_crs(x))
image(x, axes = TRUE)
plot(st_as_sfc(st_bbox(pts)), col = NA, add = TRUE)
bb = st_bbox(pts)
(xx = x[bb])
image(xx)
plot(st_as_sfc(bb), add = TRUE, col = NA)
image(x)
pt = st_point(c(x = 290462.103109179, y = 9114202.32594085))
buf = st_buffer(st_sfc(pt, crs = st_crs(x)), 1500)
plot(buf, add = TRUE)

buf = st_sfc(st_polygon(list(st_buffer(pt, 1500)[[1]], st_buffer(pt, 1000)[[1]])),
   crs = st_crs(x))
image(x[buf])
plot(buf, add = TRUE, col = NA)
image(x[buf, crop=FALSE])
plot(buf, add = TRUE, col = NA)
# with i of class stars:
x[x > 75] # generates lots of NA's; pattern for each band
x[x[,,,1] > 75] # recycles a single band template for all bands
x = read_stars(tif)
# replace, using a logical stars selector: cuts all values above 90 to 90
x[x > 90] = 90
# replace a single attribute when there are more than one:
s = split(x)
names(s) = paste0("band", 1:6)
# rescale only band 1:
s[1] = s[1] * 0.75 
# rescale only attribute named "band2":
s["band2"] = s["band2"] * 0.85 
# create a new attribute from a numeric vector:
s["rnorm"] = rnorm(prod(dim(s))) 
s
lc = read_stars(system.file("tif/lc.tif", package = "stars"))
x = c(orig = lc, 
      flip_x = st_flip(lc, "x"), 
      flip_y = st_flip(lc, "y"), 
      flip_xy = st_flip(lc, c("x", "y")), 
      along = 3)
plot(x)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ops.R
\name{ops_stars}
\alias{ops_stars}
\alias{Ops.stars}
\alias{Math.stars}
\alias{Ops.stars_proxy}
\alias{Math.stars_proxy}
\title{S3 Ops Group Generic Functions for stars objects}
\usage{
\method{Ops}{stars}(e1, e2)

\method{Math}{stars}(x, ...)

\method{Ops}{stars_proxy}(e1, e2)

\method{Math}{stars_proxy}(x, ...)
}
\arguments{
\item{e1}{object of class \code{stars}}

\item{e2}{object of class \code{stars}}

\item{x}{object of class stars}

\item{...}{parameters passed on to the Math functions}
}
\value{
object of class \code{stars}
}
\description{
Ops functions for stars objects, including comparison, product and divide, add, subtract
}
\details{
if \code{e1} or \code{e2} is is a numeric vector, or \code{e2}
has less or smaller dimensions than \code{e1}, then \code{e2} is recycled
such that it fits \code{e1}, using usual R array recycling rules. The user
needs to make sure this is sensible; it may be needed to use \code{aperm}
to permutate dimensions first.
}
\examples{
tif = system.file("tif/L7_ETMs.tif", package = "stars")
x = read_stars(tif)
x * x
x / x
x + x
x + 10
all.equal(x * 10, 10 * x)
tif = system.file("tif/L7_ETMs.tif", package = "stars")
x = read_stars(tif)
a = sqrt(x)
b = log(x, base = 10)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/intervals.R
\name{make_intervals}
\alias{make_intervals}
\title{create an intervals object}
\usage{
make_intervals(start, end)
}
\arguments{
\item{start}{vector with start values, or 2-column matrix with start and end values in column 1 and 2, respectively}

\item{end}{vector with end values}
}
\description{
create an intervals object, assuming left-closed and right-open intervals
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\docType{data}
\name{bcsd_obs}
\alias{bcsd_obs}
\title{Monthly Gridded Meteorological Observations}
\format{
An object of class \code{stars_proxy} (inherits from \code{stars}) of dimension 81 x 33 x 12.
}
\usage{
bcsd_obs
}
\description{
These are the monthly observational data used for BCSD downscaling. See: <https://gdo-dcp.ucllnl.org/downscaled_cmip_projections/dcpInterface.html#About> for more information." ;
"Atmospheric Temperature, Air Temperature Atmosphere, Precipitation, Rain, Maximum Daily Temperature, Minimum  Daily Temperature" ;
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cf.R, R/cubble.R, R/OpenStreetMap.R,
%   R/proxy.R, R/rasterize.R, R/raster.R, R/sf.R, R/stars.R, R/xts.R
\name{st_as_stars}
\alias{st_as_stars}
\alias{st_as_stars.ncdfgeom}
\alias{st_as_stars.CFDataset}
\alias{st_as_stars.CFVariable}
\alias{st_as_stars.cubble_df}
\alias{st_as_stars.OpenStreetMap}
\alias{st_as_stars.stars_proxy}
\alias{st_as_stars.data.frame}
\alias{st_as_stars.Raster}
\alias{st_as_stars.SpatRaster}
\alias{st_as_stars.sf}
\alias{st_as_stars.list}
\alias{st_as_stars.default}
\alias{st_as_stars.stars}
\alias{st_as_stars.bbox}
\alias{st_as_stars.xts}
\title{convert objects into a stars object}
\usage{
\method{st_as_stars}{ncdfgeom}(.x, ..., sf_geometry = NA)

\method{st_as_stars}{CFDataset}(
  .x,
  ...,
  .var = .x$var_names[1],
  all_compatible = length(.var) == 1
)

\method{st_as_stars}{CFVariable}(.x, ...)

\method{st_as_stars}{cubble_df}(.x, ..., check_times = FALSE)

\method{st_as_stars}{OpenStreetMap}(.x, ..., as_col = FALSE)

\method{st_as_stars}{stars_proxy}(
  .x,
  ...,
  downsample = 0,
  url = attr(.x, "url"),
  envir = parent.frame()
)

\method{st_as_stars}{data.frame}(.x, ..., dims = coords, xy, y_decreasing = TRUE, coords = 1:2)

\method{st_as_stars}{Raster}(.x, ..., att = 1, ignore_file = FALSE)

\method{st_as_stars}{SpatRaster}(
  .x,
  ...,
  ignore_file = FALSE,
  as_attributes = all(terra::is.factor(.x))
)

\method{st_as_stars}{sf}(.x, ..., dims = attr(.x, "sf_column"))

st_as_stars(.x, ...)

\method{st_as_stars}{list}(.x, ..., dimensions = NULL)

\method{st_as_stars}{default}(.x = NULL, ..., raster = NULL)

\method{st_as_stars}{stars}(.x, ..., curvilinear = NULL, crs = st_crs("OGC:CRS84"))

\method{st_as_stars}{bbox}(
  .x,
  ...,
  nx,
  ny,
  dx = dy,
  dy = dx,
  xlim = .x[c("xmin", "xmax")],
  ylim = .x[c("ymin", "ymax")],
  values = 0,
  n = 64800,
  pretty = FALSE,
  inside = FALSE,
  nz,
  proxy = FALSE
)

\method{st_as_stars}{xts}(.x, ..., dimensions, name = "attr")
}
\arguments{
\item{.x}{object to convert}

\item{...}{in case \code{.x} is of class \code{bbox}, arguments passed on to 
\link{pretty}. In case \code{.x} is of class \code{nc_proxy}, arguments
passed on to \code{\link{read_ncdf}}.}

\item{sf_geometry}{sf data.frame with geometry and attributes to be added to stars object.
Must have same number of rows as timeseries instances.}

\item{.var}{Character vector with names of netCDF data variables to convert
into stars attributes. All data variables have to use the same set of axes.
If omitted, the first data variable will be used.}

\item{all_compatible}{Logical flag that indicates if all data variables with
the same axes as argument `var` uses should be converted into attributes.
Default is `TRUE` when only one data variable is specified, `FALSE`
otherwise.}

\item{check_times}{logical; should we check that the time stamps of all time series are identical?}

\item{as_col}{logical; return rgb numbers (FALSE) or (character) color values (TRUE)?}

\item{downsample}{integer: if larger than 0, downsample with this rate (number of pixels to skip in every row/column); if length 2, specifies downsampling rate in x and y.}

\item{url}{character; URL of the stars endpoint where the data reside}

\item{envir}{environment to resolve objects in}

\item{dims}{the column names or indices that form the cube dimensions}

\item{xy}{the x and y raster dimension names or indices; only takes effect after \code{dims} has been specified, see details}

\item{y_decreasing}{logical; if TRUE, (numeric) y values get a negative delta (decrease with increasing index)}

\item{coords}{same as dims, for symmetry with \link[sf]{st_as_sf}}

\item{att}{see \link[raster:factor]{factorValues}; column in the RasterLayer's attribute table}

\item{ignore_file}{logical; if \code{TRUE}, ignore the SpatRaster object file name}

\item{as_attributes}{logical; if \code{TRUE} and \code{.x} has more than one layer, load these as separate attributes rather than as a band or time dimension (only implemented for the case where \code{ignore_file} is \code{TRUE})}

\item{dimensions}{object of class dimensions}

\item{raster}{character; the names of the dimensions that denote raster dimensions}

\item{curvilinear}{only for creating curvilinear grids: named length 2 list holding longitude and latitude matrices or stars arrays, or the names of the corresponding attributes in \code{.x}; the names of this vector should correspond to raster dimensions the matrices are associated with; see Details.}

\item{crs}{object of class \code{crs} with the coordinate reference system of the values in \code{curvilinear}; see details}

\item{nx}{integer; number of cells in x direction; see details}

\item{ny}{integer; number of cells in y direction; see details}

\item{dx}{numeric or object of class units; cell size in x direction; see details}

\item{dy}{numeric or object of class units; cell size in y direction; see details}

\item{xlim}{length 2 numeric vector with extent (min, max) in x direction}

\item{ylim}{length 2 numeric vector with extent (min, max) in y direction}

\item{values}{value(s) to populate the raster values with}

\item{n}{the (approximate) target number of grid cells}

\item{pretty}{logical; should cell coordinates have \link{pretty} values?}

\item{inside}{logical; should all cells entirely fall inside the bbox, potentially not covering it completely (\code{TRUE}), or always cover the bbox (\code{FALSE}), or find a good approximation (\code{NA}, default)?}

\item{nz}{integer; number of cells in z direction; if missing no z-dimension is created.}

\item{proxy}{logical; should a \code{stars_proxy} object be created? (requires gdal_create binary when sf < 1.0-6)}

\item{name}{character; attribute name for array from an \code{xts} object}
}
\description{
convert objects into a stars object
}
\details{
For the \code{ncdfgeom} method: objects are point-timeseries with optional line or polygon geometry for each timeseries specified with the \code{sf_geometry} parameter. See \pkg{ncdfgeom} for more about this NetCDF-based format for geometry and timeseries.

If \code{xy} is not specified and the first two dimensions in \code{dims} are both numeric,
then it is set to these two dimensions.

The \code{st_as_stars} method for \code{sf} objects without any additional
arguments returns a one-dimensional data cube with a dimension for the simple features
geometries, and all remaining attributes as data cube attributes. When used with
further arguments, the method for \code{data.frame}s is called.

if \code{curvilinear} is a list with \code{stars} objects with longitude and latitude values, its coordinate reference system is typically not that of the latitude and longitude values. If \code{curvilinear} contains the names of two arrays in \code{.x}, then these are removed from the returned object.

For the \code{bbox} method: if \code{pretty} is \code{TRUE}, raster cells may extend the coordinate range of \code{.x} on all sides. If in addition to \code{nx} and \code{ny}, \code{dx} and \code{dy} are also missing, these are set to a single value computed as \code{sqrt(diff(xlim)*diff(ylim)/n)}. 

If \code{nx} and \code{ny} are missing and \code{values} is a matrix, the number of columns and rows of the matrix are taken.

Otherwise, if \code{nx} and \code{ny} are missing, they are computed as the (ceiling, floor, or rounded to integer value) of the ratio of the (x or y) range divided by (dx or dy), depending on the value of \code{inside}. Positive \code{dy} will be made negative. Further named arguments (\code{...}) are passed on to \code{pretty}. If \code{dx} or \code{dy} are \code{units} objects, their value is converted to the units of \code{st_crs(.x)} (only when sf >= 1.0-7).

for the \code{xts} methods, if \code{dimensions} are provided, time has to be the first dimension.
}
\examples{
if (require(plm, quietly = TRUE)) {
 data(Produc, package = "plm")
 st_as_stars(Produc)
}
if (require(dplyr, quietly = TRUE)) {
  # https://stackoverflow.com/questions/77368957/
spatial_dim <- st_sf(
  ID = 1:3,
  geometry = list(
    st_polygon(list(
      cbind(c(0, 1, 1, 0, 0), c(0, 0, 1, 1, 0))
    )),
    st_polygon(list(
      cbind(c(1, 2, 2, 1, 1), c(0, 0, 1, 1, 0))
    )),
    st_polygon(list(
      cbind(c(2, 3, 3, 2, 2), c(0, 0, 1, 1, 0))
    ))
  )
)
weekdays_dim <- data.frame(weekdays = c("Monday", "Tuesday", "Wednesday", 
    "Thursday", "Friday", "Saturday", "Sunday"))
hours_dim <- data.frame(hours = c("8am", "11am", "4pm", "11pm"))
sf_dta <- spatial_dim |>
  cross_join(weekdays_dim)|>
  cross_join(hours_dim) |>
  mutate(population = rnorm(n(), mean = 1000, sd = 200)) |>
  select(everything(), geometry)

st_as_stars(sf_dta, dims = c("weekdays", "hours", "geometry"))
}
demo(nc, echo=FALSE,ask=FALSE)
st_as_stars(nc)
st_as_stars(st_drop_geometry(nc), dims = "NAME")
data.frame(expand.grid(x=1:5, y = 1:5), z = rnorm(25)) |> st_as_stars()
nc = st_read(system.file("gpkg/nc.gpkg", package="sf"))
st_as_stars(nc)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{st_rgb}
\alias{st_rgb}
\title{reduce dimension to rgb (alpha) hex values}
\usage{
st_rgb(
  x,
  dimension = 3,
  use_alpha = dim(x)[dimension] == 4,
  maxColorValue = 255L,
  probs = c(0, 1),
  stretch = NULL
)
}
\arguments{
\item{x}{object of class \code{stars}}

\item{dimension}{dimension name or number to reduce}

\item{use_alpha}{logical; if TRUE, the fourth band will be used as alpha values}

\item{maxColorValue}{integer; maximum value for colors}

\item{probs}{probability values for quantiles used for stretching by "percent".}

\item{stretch}{logical or character; if \code{TRUE} or \code{"percent"},
each band is stretched to 0 ... maxColorValue by "percent clip" method using
probs values. If \code{"histogram"}, a "histogram equalization" is performed
(\code{probs} values are ignored). If stretch is \code{NULL} or \code{FALSE}, no stretching
is performed. Other character values are interpreted as "percent" and a message
will be printed.}
}
\description{
reduce dimension to rgb (alpha) hex values
}
\details{
the dimension's bands are mapped to red, green, blue, alpha; if a different
ordering is wanted, use \link{[.stars} to reorder a dimension, see examples.
Alternatively, you can use \link{plot.stars} with the \code{rgb} argument to create a three-band composition.
}
\examples{
tif = system.file("tif/L7_ETMs.tif", package = "stars")
x = read_stars(tif)
st_rgb(x[,,,3:1])
r = st_rgb(x[,,,c(6,5,4,3)], 3, use_alpha=TRUE) # now R=6,G=5,B=4,alpha=3
if (require(ggplot2)) {
 ggplot() + geom_stars(data = r) + scale_fill_identity()
}
r = st_rgb(x[,,,3:1],
		   probs = c(0.01, 0.99),
		   stretch = "percent")
plot(r)
r = st_rgb(x[,,,3:1],
		   probs = c(0.01, 0.99),
		   stretch = "histogram")
plot(r)
}
\seealso{
\link{st_apply}, \link[grDevices]{rgb}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stars.R
\name{st_res}
\alias{st_res}
\title{obtain (spatial) resolution of a stars object}
\usage{
st_res(x, all = FALSE, absolute = !all)
}
\arguments{
\item{x}{an object of class \code{stars}}

\item{all}{logical; if FALSE return a vector with the x/y raster resolution}

\item{absolute}{logical; only works when \code{all = FALSE}; if TRUE return absolute resolution values, if FALSE return \code{delta} values}
}
\value{
if \code{all = FALSE} a vector with x/y raster resolutions, otherwise a list with delta values
}
\description{
obtain resolution(s) of a stars object: by default only the (absolute) x/y raster dimensions, optionally all \code{delta} dimension parameters
}
\examples{
st_res(L7_ETMs)
st_res(L7_ETMs, absolute = FALSE)
st_res(L7_ETMs, all = TRUE)
if (require(starsdata)) {
  paste0("netcdf/", c("avhrr-only-v2.19810901.nc", 
    "avhrr-only-v2.19810902.nc",
    "avhrr-only-v2.19810903.nc",
    "avhrr-only-v2.19810904.nc")) |>
  system.file(package = "starsdata") |>
  read_stars(quiet = TRUE) -> x
  st_res(x) |> print()
  st_res(x, all = TRUE) |> print()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rotate.R
\name{st_rotate}
\alias{st_rotate}
\alias{st_rotate.stars}
\alias{st_rotate.sfc}
\alias{st_rotate.sf}
\title{Transform rotated pole long/lat regular grid to unrotated curvilinear grid}
\usage{
\method{st_rotate}{stars}(.x, lon0, lat0, north = TRUE, ...)

\method{st_rotate}{sfc}(.x, lon0, lat0, north = TRUE, ...)

\method{st_rotate}{sf}(.x, lon0, lat0, north = TRUE, ...)
}
\arguments{
\item{.x}{object of class \code{stars}}

\item{lon0}{longitude of the rotated pole in degrees}

\item{lat0}{latitude of the rotated pole in degrees}

\item{north}{logical; if \code{TRUE} the pole refers to the North pole, otherwise the South pole}

\item{...}{ignored}
}
\value{
curvilinear stars object with coordinates in regular long/lat (North pole at lat=90)
}
\description{
Transform rotated long/lat regular grid to unrotated curvilinear grid
}
\examples{
if (require("starsdata") && require("maps")) {
  # data downloaded from https://esgf-data.dkrz.de/search/cosmo-rea/
  nc = "netcdf/ts_EUR-6km_ECMWF-ERAINT_REA6_r1i1p1f1_COSMO_v1_mon_201801-201812.nc"
  f = system.file(nc, package = "starsdata")
  m = read_mdim(f, "ts")
  print(m)
  # NOTE this function is obsolete when reading m as
  # m = read_mdim(f, "ts", curvilinear = c("longitude", "latitude"))
  if (require(RNetCDF)) {
     x = open.nc(f)
     lon = att.get.nc(x, "rotated_latitude_longitude", "grid_north_pole_longitude")
     lat = att.get.nc(x, "rotated_latitude_longitude", "grid_north_pole_latitude")
     close.nc(x)
     print(c(lon = lon, lat = lat))
  } else {
     lon = -162
     lat = 39.25
  } 
  m1 = st_rotate(m, lon, lat)
  print(m1)
  h = function() maps::map(add = TRUE)
  plot(m1, downsample = c(10, 10, 5), axes = TRUE, hook = h, mfrow = c(1, 2)) 
    # curvilinear grid: downsample for plotting speed
  m2 = st_warp(m1, crs = st_crs("OGC:CRS84"), threshold = .1)
  plot(m2, hook = h, mfrow = c(3, 4)) # regular grid: plots fast
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ncproxy.R, R/plot.R, R/proxy.R
\name{plot}
\alias{plot}
\alias{plot.nc_proxy}
\alias{plot.stars}
\alias{image.stars}
\alias{plot.stars_proxy}
\title{plot stars object, with subplots for each level of first non-spatial dimension}
\usage{
\method{plot}{nc_proxy}(x, y, ..., downsample = get_downsample(dim(x)), max_times = 16)

\method{plot}{stars}(
  x,
  y,
  ...,
  join_zlim = TRUE,
  main = make_label(x, 1),
  axes = FALSE,
  downsample = TRUE,
  nbreaks = 11,
  breaks = "quantile",
  col = grey(1:(nbreaks - 1)/nbreaks),
  key.pos = get_key_pos(x, ...),
  key.width = kw_dflt(x, key.pos),
  key.length = 0.618,
  key.lab = ifelse(length(main) == 1, main, ""),
  reset = TRUE,
  box_col = NA,
  center_time = FALSE,
  hook = NULL,
  mfrow = NULL,
  compact = TRUE
)

\method{image}{stars}(
  x,
  ...,
  band = 1,
  attr = 1,
  asp = NULL,
  rgb = NULL,
  maxColorValue = ifelse(inherits(rgb, "data.frame"), 255, max(x[[attr]], na.rm = TRUE)),
  xlab = if (!axes) "" else names(d)[1],
  ylab = if (!axes) "" else names(d)[2],
  xlim = st_bbox(extent)$xlim,
  ylim = st_bbox(extent)$ylim,
  text_values = FALSE,
  text_color = "black",
  axes = FALSE,
  interpolate = FALSE,
  as_points = FALSE,
  key.pos = NULL,
  logz = FALSE,
  key.width = kw_dflt(x, key.pos),
  key.length = 0.618,
  add.geom = NULL,
  border = NA,
  useRaster = isTRUE(dev.capabilities()$rasterImage == "yes"),
  extent = x
)

\method{plot}{stars_proxy}(x, y, ..., downsample = get_downsample(dim(x)))
}
\arguments{
\item{x}{object of class \code{stars}}

\item{y}{ignored}

\item{...}{further arguments: for \code{plot}, passed on to \code{image.stars}; for \code{image}, passed on to \code{image.default} or \code{rasterImage}.}

\item{downsample}{logical or numeric; if \code{TRUE} will try to plot not many more pixels than actually are visible, if \code{FALSE}, no downsampling takes place, if numeric, the number of pixels/lines/bands etc that will be skipped; see Details.}

\item{max_times}{integer; maximum number of time steps to attempt to plot.}

\item{join_zlim}{logical; if \code{TRUE}, compute a single, joint zlim (color scale) for all subplots from \code{x}}

\item{main}{character; subplot title prefix; use \code{""} to get only time, use \code{NULL} to suppress subplot titles}

\item{axes}{logical; should axes and box be added to the plot?}

\item{nbreaks}{number of color breaks; should be one more than number of colors. If missing and \code{col} is specified, it is derived from that.}

\item{breaks}{numeric vector with actual color breaks, or a \code{style} name used in \link[classInt]{classIntervals}.}

\item{col}{colors to use for grid cells, or color palette function}

\item{key.pos}{numeric; side to plot a color key: 1 bottom, 2 left, 3 top, 4 right; set to \code{NULL} to omit key. Ignored if multiple columns are plotted in a single function call. Default depends on plot size, map aspect, and, if set, parameter \code{asp}. If it has length 2, the second value, ranging from 0 to 1, determines where the key is placed in the available space (default: 0.5, center).}

\item{key.width}{amount of space reserved for width of the key (labels); relative or absolute (using lcm)}

\item{key.length}{amount of space reserved for length of the key (labels); relative or absolute (using lcm)}

\item{key.lab}{character; label for color key in case of multiple subplots, use \code{""} to suppress}

\item{reset}{logical; if \code{FALSE}, keep the plot in a mode that allows adding further map elements; if \code{TRUE} restore original mode after plotting}

\item{box_col}{color for box around sub-plots; use \code{NA} to suppress plotting of boxes around sub-plots.}

\item{center_time}{logical; if \code{TRUE}, sub-plot titles will show the center of time intervals, otherwise their start}

\item{hook}{NULL or function; hook function that will be called on every sub-plot; see examples.}

\item{mfrow}{length-2 integer vector with nrows, ncolumns of a composite plot, to override the default layout}

\item{compact}{logical; place facets compactly (TRUE), or spread over the plotting device area?}

\item{band}{integer; which band (dimension) to plot}

\item{attr}{integer; which attribute to plot}

\item{asp}{numeric; aspect ratio of image}

\item{rgb}{integer; specify three bands to form an rgb composite. Experimental: rgb color table; see Details.}

\item{maxColorValue}{numeric; passed on to \link{rgb}}

\item{xlab}{character; x axis label}

\item{ylab}{character; y axis label}

\item{xlim}{x axis limits}

\item{ylim}{y axis limits}

\item{text_values}{logical; print values as text on image?}

\item{text_color}{character; color for printed text values}

\item{interpolate}{logical; when using \link{rasterImage} (rgb), should pixels be interpolated?}

\item{as_points}{logical; for curvilinear or sheared grids: parameter passed on to \link{st_as_sf}, determining whether raster cells will be plotted as symbols (fast, approximate) or small polygons (slow, exact)}

\item{logz}{logical; if \code{TRUE}, use log10-scale for the attribute variable. In that case, \code{breaks} and \code{at} need to be given as log10-values; see examples.}

\item{add.geom}{object of class \code{sfc}, or list with arguments to \code{plot}, that will be added to an image or sub-image}

\item{border}{color used for cell borders (only in case \code{x} is a curvilinear or rotated/sheared grid)}

\item{useRaster}{logical; use the rasterImage capabilities of the graphics device?}

\item{extent}{object which has a \code{st_bbox} method; sets the plotting extent}
}
\description{
plot stars object, with subplots for each level of first non-spatial dimension, and customization of legend key
}
\details{
when plotting a subsetted \code{stars_proxy} object, the default value for argument \code{downsample} will not be computed correctly, and has to be set manually.

Downsampling: a value for \code{downsample} of 0: no downsampling, 1: after every dimension value (pixel/line/band), one value is skipped (half of the original resolution), 2: after every dimension value, 2 values are skipped (one third of the original resolution), etc. If \code{downsample} is \code{TRUE} or a length 1 numeric vector, downsampling is only applied to the raster [x] and [y] dimensions.

To remove unused classes in a categorical raster, use the \link[base]{droplevels} function.

When bitmaps show visual artefacts (Moiré effects), make sure that device \link{png} is used rather than \code{ragg::agg_png} as the latter uses antialiasing for filled polygons which causes this; see also https://github.com/r-spatial/stars/issues/573 .

use of an rgb color table is experimental; see https://github.com/r-spatial/mapview/issues/208

when plotting a subsetted \code{stars_proxy} object, the default value for argument \code{downsample} will not be computed correctly, and has to be set manually.
}
\examples{
st_bbox(L7_ETMs) |> st_as_sfc() |> st_centroid() |> st_coordinates() -> pt
hook1 = function() {
    text(pt[,"X"], pt[,"Y"], "foo", col = 'orange', cex = 2)
}
plot(L7_ETMs, hook = hook1)
x = st_set_dimensions(L7_ETMs, 3, paste0("B_", 1:6))
hook2 = function(..., row, col, nr, nrow, ncol, value, bbox) {
   str = paste0("row ", row, "/", nrow, ", col ", col, "/", ncol, "\nnr: ", nr, " value: ", value)
   bbox |> st_as_sfc() |> st_centroid() |> st_coordinates() -> pt
   text(pt[,"X"], pt[,"Y"], str, col = 'red', cex = 2)
}
plot(x, hook = hook2, col = grey(c(.2,.25,.3,.35)))
if (isTRUE(dev.capabilities()$rasterImage == "yes")) {
  lc = read_stars(system.file("tif/lc.tif", package = "stars"))
  levels(lc[[1]]) = abbreviate(levels(lc[[1]]), 6) # so it's not only legend
  plot(lc, key.pos=4)
}
tif = system.file("tif/L7_ETMs.tif", package = "stars")
x = read_stars(tif)
image(x, col = grey((3:9)/10))
if (isTRUE(dev.capabilities()$rasterImage == "yes")) {
  image(x, rgb = c(1,3,5)) # false color composite
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mdim.R, R/proxy.R, R/subset.R
\name{st_crop}
\alias{st_crop}
\alias{st_crop.mdim}
\alias{st_crop.stars_proxy}
\alias{st_crop.stars}
\title{crop a stars object}
\usage{
\method{st_crop}{mdim}(x, y, ...)

\method{st_crop}{stars_proxy}(
  x,
  y,
  ...,
  crop = TRUE,
  epsilon = sqrt(.Machine$double.eps),
  collect = TRUE
)

\method{st_crop}{stars}(
  x,
  y,
  ...,
  crop = TRUE,
  epsilon = sqrt(.Machine$double.eps),
  as_points = all(st_dimension(y) == 2, na.rm = TRUE),
  normalize = FALSE
)
}
\arguments{
\item{x}{object of class \code{stars}}

\item{y}{object of class \code{sf}, \code{sfc} or \code{bbox}; see Details below.}

\item{...}{ignored}

\item{crop}{logical; if \code{TRUE}, the spatial extent of the returned object is cropped to still cover \code{obj}, if \code{FALSE}, the extent remains the same but cells outside \code{y} are given \code{NA} values.}

\item{epsilon}{numeric; factor to shrink the bounding box of \code{y} towards its center before cropping.}

\item{collect}{logical; if \code{TRUE}, repeat cropping on \code{stars} object, i.e. after data has been read}

\item{as_points}{logical; only relevant if \code{y} is of class \code{sf} or \code{sfc}: if \code{FALSE}, treat \code{x} as a set of points, else as a set of small polygons. Default: \code{TRUE} if \code{y} is two-dimensional, else \code{FALSE}; see Details}

\item{normalize}{logical; if \code{TRUE} then pass the cropped object to \code{\link[sf]{st_normalize}} before returning. This typically changes the `offset` field and resets the `from` field to 1, and changes the bounding box of the returned object accordingly.}
}
\description{
crop a stars object
}
\details{
for raster \code{x}, \code{st_crop} selects cells that intersect with \code{y}. 
For intersection, are raster cells interpreted as points or as small polygons? 
If \code{y} is of class \code{stars}, \code{x} raster cells are interpreted as points; if \code{y} is of class \code{bbox}, \code{x} cells are interpreted as cells (small polygons). Otherwise, if \code{as_points} is not given, cells are interpreted as points if \code{y} has a two-dimensional geometry.
}
\examples{
l7 = read_stars(system.file("tif/L7_ETMs.tif", package = "stars"))
d = st_dimensions(l7)

# area around cells 3:10 (x) and 4:11 (y):
offset = c(d[["x"]]$offset, d[["y"]]$offset)
res = c(d[["x"]]$delta, d[["y"]]$delta)
bb = st_bbox(c(xmin = offset[1] + 2 * res[1],
	ymin = offset[2] + 11 * res[2],
	xmax = offset[1] + 10 * res[1],
	ymax = offset[2] +  3 * res[2]), crs = st_crs(l7))
l7[bb]
# equivalent:
st_crop(l7, bb)

plot(l7[,1:13,1:13,1], reset = FALSE)
image(l7[bb,,,1], add = TRUE, col = sf.colors())
plot(st_as_sfc(bb), add = TRUE, border = 'green', lwd = 2)

# slightly smaller bbox:
bb = st_bbox(c(xmin = offset[1] + 2.1 * res[1],
	ymin = offset[2] + 10.9 * res[2],
	xmax = offset[1] +  9.9 * res[1],
	ymax = offset[2] +  3.1 * res[2]), crs = st_crs(l7))
l7[bb]

plot(l7[,1:13,1:13,1], reset = FALSE)
image(l7[bb,,,1], add = TRUE, col = sf.colors())
plot(st_as_sfc(bb), add = TRUE, border = 'green', lwd = 2)

# slightly larger bbox:
bb = st_bbox(c(xmin = offset[1] + 1.9 * res[1],
	ymin = offset[2] + 11.1 * res[2],
	xmax = offset[1] + 10.1 * res[1],
	ymax = offset[2] +  2.9 * res[2]), crs = st_crs(l7))
l7[bb]

plot(l7[,1:13,1:13,1], reset = FALSE)
image(l7[bb,,,1], add = TRUE, col = sf.colors())
plot(st_as_sfc(bb), add = TRUE, border = 'green', lwd = 2)

# half a cell size larger bbox:
bb = st_bbox(c(xmin = offset[1] + 1.49 * res[1],
	ymin = offset[2] + 11.51 * res[2],
	xmax = offset[1] + 10.51 * res[1],
	ymax = offset[2] +  2.49 * res[2]), crs = st_crs(l7))
l7[bb]

plot(l7[,1:13,1:13,1], reset = FALSE)
image(l7[bb,,,1], add = TRUE, col = sf.colors())
plot(st_as_sfc(bb), add = TRUE, border = 'green', lwd = 2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transform.R
\name{st_transform}
\alias{st_transform}
\alias{st_transform.stars}
\alias{st_transform_proj.stars}
\title{transform geometries in stars objects to a new coordinate reference system, without warping}
\usage{
\method{st_transform}{stars}(x, crs, ...)

st_transform_proj.stars(x, crs, ...)
}
\arguments{
\item{x}{object of class \code{stars}, with either raster or simple feature geometries}

\item{crs}{object of class \code{crs} with target crs}

\item{...}{ignored}
}
\description{
transform geometries in stars objects to a new coordinate reference system, without warping
}
\details{
For simple feature dimensions, \link[sf]{st_transform} is called, leading to lossless transformation. For gridded spatial data, a curvilinear grid with transformed grid cell (centers) is returned, which is also lossless. To convert this to a regular grid in the new \code{CRS}, use \link{st_warp} (which is in general lossy).

If array values contain geometries and an array as a whole is of class `sfc` and 
has a non-missing CRS, array geometries are also transformed.
}
\examples{
geomatrix = system.file("tif/geomatrix.tif", package = "stars")
(x = read_stars(geomatrix))
new = st_crs('OGC:CRS84')
y = st_transform(x, new)
plot(st_transform(st_as_sfc(st_bbox(x)), new), col = NA, border = 'red')
plot(st_as_sfc(y, as_points=FALSE), col = NA, border = 'green', axes = TRUE, add = TRUE)
image(y, col = heat.colors(12), add = TRUE)
plot(st_as_sfc(y, as_points=TRUE), pch=3, cex=.5, col = 'blue', add = TRUE)
plot(st_transform(st_as_sfc(x, as_points=FALSE), new), add = TRUE)
}
\seealso{
\link{st_warp}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stars.R
\name{merge}
\alias{merge}
\alias{split.stars}
\alias{split}
\alias{merge.stars}
\title{merge or split stars object}
\usage{
\method{split}{stars}(x, f = length(dim(x)), drop = TRUE, ...)

\method{merge}{stars}(x, y, ..., name = "attributes")
}
\arguments{
\item{x}{object of class \code{stars}}

\item{f}{the name or index of the dimension to split; by default the last dimension}

\item{drop}{ignored}

\item{...}{if defined, the first unnamed argument is used for dimension values, if not defined, attribute names are used for dimension values}

\item{y}{needs to be missing}

\item{name}{name for the new dimension}
}
\value{
merge merges attributes of a stars object into a new dimension; split splits a dimension over attributes
}
\description{
merge attributes into a dimension, or split a dimension over attributes
}
\details{
split.stars works on the first attribute, and will give an error when more than one attribute is present
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tile.R
\name{st_tile}
\alias{st_tile}
\title{Specify parameters to load raster in blocks}
\usage{
st_tile(img_rows, img_cols, x_window, y_window, overlap = 0)
}
\arguments{
\item{img_rows}{number of input raster rows (integer)}

\item{img_cols}{number of input raster columns (integer)}

\item{x_window}{number of rows in block (integer)}

\item{y_window}{number of columns in block (integer)}

\item{overlap}{number of overlapping pixels (integer)}
}
\value{
matrix with specified \code{nXOff}, \code{nYOff}, \code{nXsize},
and \code{nYSize} parameters for every block
}
\description{
Helper function for specifying the block parameters (\code{nXOff},
\code{nYOff}, \code{nXsize}, and \code{nYSize}) required by \code{RasterIO}
argument in \link{read_stars}
}
\examples{
\dontrun{
tif = system.file("tif/L7_ETMs.tif", package = "stars")
r = read_stars(tif, proxy = TRUE)
tiles = st_tile(nrow(r), ncol(r), 256, 256)
for (i in seq_len(nrow(tiles))) {
  tile = read_stars(tif, proxy = FALSE, RasterIO = tiles[i, ])
  # write tiles to separate files
  write_stars(tile, dsn = paste0(i, ".tif"))
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read.R
\name{read_stars}
\alias{read_stars}
\title{read raster/array dataset from file or connection}
\usage{
read_stars(
  .x,
  sub = TRUE,
  ...,
  options = character(0),
  driver = character(0),
  quiet = FALSE,
  NA_value = NA_real_,
  along = NA_integer_,
  RasterIO = list(),
  proxy = getOption("stars.n_proxy") \%||\% 1e+08,
  curvilinear = character(0),
  normalize_path = TRUE,
  RAT = character(0),
  tolerance = 1e-10,
  exclude = "",
  shorten = TRUE
)
}
\arguments{
\item{.x}{character vector with name(s) of file(s) or data source(s) to be read, or a function that returns such a vector}

\item{sub}{character, integer or logical; name, index or indicator of sub-dataset(s) to be read}

\item{...}{passed on to \link{st_as_stars} if \code{curvilinear} was set}

\item{options}{character; opening options}

\item{driver}{character; driver to use for opening file. To override fixing for subdatasets and autodetect them as well, use \code{NULL}.}

\item{quiet}{logical; print progress output?}

\item{NA_value}{numeric value to be used for conversion into NA values; by default this is read from the input file}

\item{along}{length-one character or integer, or list; determines how several arrays are combined, see Details.}

\item{RasterIO}{list with named parameters for GDAL's RasterIO, to further control the extent, resolution and bands to be read from the data source; see details.}

\item{proxy}{logical; if \code{TRUE}, an object of class \code{stars_proxy} is read which contains array
metadata only; if \code{FALSE} the full array data is read in memory. Always \code{FALSE} for curvilinear girds.
If set to a number, defaults to \code{TRUE} when the number of cells to be read is larger than that number.}

\item{curvilinear}{length two character vector with names of subdatasets holding longitude and latitude values for all raster cells, or named length 2 list holding longitude and latitude matrices; the names of this list should correspond to raster dimensions referred to}

\item{normalize_path}{logical; if \code{FALSE}, suppress a call to \link{normalizePath} on \code{.x}}

\item{RAT}{character; raster attribute table column name to use as factor levels}

\item{tolerance}{numeric; passed on to \link{all.equal} for comparing dimension parameters.}

\item{exclude}{character; vector with category value(s) to exclude}

\item{shorten}{logical or character; if \code{TRUE} and \code{length(.x) > 1}, remove common start and end parts of array names; if character a new prefix}
}
\value{
object of class \code{stars}
}
\description{
read raster/array dataset from file or connection
}
\details{
In case \code{.x} contains multiple files, they will all be read and combined with \link{c.stars}. Along which dimension, or how should objects be merged? If \code{along} is set to \code{NA} it will merge arrays as new attributes if all objects have identical dimensions, or else try to merge along time if a dimension called \code{time} indicates different time stamps. A single name (or positive value) for \code{along} will merge along that dimension, or create a new one if it does not already exist. If the arrays should be arranged along one of more dimensions with values (e.g. time stamps), a named list can passed to \code{along} to specify them; see example.

\code{RasterIO} is a list with zero or more of the following named arguments:
\code{nXOff}, \code{nYOff} (both 1-based: the first row/col has offset value 1),
\code{nXSize}, \code{nYSize}, \code{nBufXSize}, \code{nBufYSize}, \code{bands}, \code{resample}.
See \url{https://gdal.org/en/latest/doxygen/classGDALDataset.html} for their meaning;
\code{bands} is an integer vector containing the band numbers to be read (1-based: first band is 1).
Note that if \code{nBufXSize} or \code{nBufYSize} are specified for downsampling an image,
resulting in an adjusted geotransform. \code{resample} reflects the resampling method and
has to be one of: "nearest_neighbour" (the default),
"bilinear", "cubic", "cubic_spline", "lanczos", "average", "mode", or "Gauss".

Data that are read into memory (\code{proxy=FALSE}) are read into a numeric (double) array, except for categorical variables which are read into an numeric (integer) array of class \code{factor}.
}
\examples{
tif = system.file("tif/L7_ETMs.tif", package = "stars")
(x1 = read_stars(tif))
(x2 = read_stars(c(tif, tif)))
(x3 = read_stars(c(tif, tif), along = "band"))
(x4 = read_stars(c(tif, tif), along = "new_dimensions")) # create 4-dimensional array
x1o = read_stars(tif, options = "OVERVIEW_LEVEL=1")
t1 = as.Date("2018-07-31")
# along is a named list indicating two dimensions:
read_stars(c(tif, tif, tif, tif), along = list(foo = c("bar1", "bar2"), time = c(t1, t1+2)))

m = matrix(1:120, nrow = 12, ncol = 10)
dim(m) = c(x = 10, y = 12) # named dim
st = st_as_stars(m)
attr(st, "dimensions")$y$delta = -1
attr(st, "dimensions")$y$offset = 12
st
tmp = tempfile(fileext = ".tif")
write_stars(st, tmp)
(red <- read_stars(tmp))
read_stars(tmp, RasterIO = list(nXOff = 1, nYOff = 1, nXSize = 10, nYSize = 12,
   nBufXSize = 2, nBufYSize = 2))[[1]]
(red <- read_stars(tmp, RasterIO = list(nXOff = 1, nYOff = 1, nXSize = 10, nYSize = 12,
   nBufXSize = 2, nBufYSize = 2)))
red[[1]] # cell values of subsample grid:
\dontrun{
  plot(st, reset = FALSE, axes = TRUE, ylim = c(-.1,12.1), xlim = c(-.1,10.1),
    main = "nBufXSize & nBufYSize demo", text_values = TRUE)
  plot(st_as_sfc(red, as_points = TRUE), add = TRUE, col = 'red', pch = 16)
  plot(st_as_sfc(st_as_stars(st), as_points = FALSE), add = TRUE, border = 'grey')
  plot(st_as_sfc(red, as_points = FALSE), add = TRUE, border = 'green', lwd = 2)
}
file.remove(tmp)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom.R
\name{st_intersects.stars}
\alias{st_intersects.stars}
\title{spatial intersect predicate for stars and sfc object}
\usage{
\method{st_intersects}{stars}(x, y, sparse = TRUE, ..., as_points = NA, transpose = FALSE)
}
\arguments{
\item{x}{object of class stars}

\item{y}{object that has an `st_geometry` method: of class `sf` or `sfc`, or `stars` object with an `sfc` dimension}

\item{sparse}{logical; if TRUE, return the a sparse logical matrix (object of class `sgbp`), if FALSE, return a logical matrix}

\item{...}{ignored, or passed on to `st_intersects.sf` for curvilinear grids}

\item{as_points}{logical, should grid cells be considered as points (TRUE) or polygons (FALSE)? Default: FALSE and warning emitted}

\item{transpose}{logical; should the transpose of the `sgbp` object be returned?}
}
\value{
`sgbp` object if sparse = TRUE, logical matrix otherwise
}
\description{
spatial intersect predicate for stars and sfc object
}
\details{
curvilinear grids are always converted to polygons, so points on grid boundaries may intersect with two cells touched; for other grids each cell boundary or corner belongs only to one cell.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/raster.R
\name{as}
\alias{as}
\alias{coerce,stars,Raster-method}
\alias{coerce,stars_proxy,Raster-method}
\alias{coerce,stars,Terra-method}
\alias{coerce,stars_proxy,Terra-method}
\title{Coerce stars object into a RasterLayer, RasterBrick, or SpatRaster}
\value{
RasterLayer or RasterBrick

SpatRaster
}
\description{
Coerce stars object into a RasterLayer, RasterBrick, or SpatRaster
}
\details{
If the stars object has more than three dimensions, all dimensions higher than the third will be collapsed into the third dimensions. If the stars object has only an x/y raster but multiple attributes, these are merged first, then put in a raster brick or SpatRaster.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\docType{data}
\name{stars_sentinel2}
\alias{stars_sentinel2}
\title{Sentinel-2 sample tile}
\format{
An object of class \code{stars_proxy} (inherits from \code{stars}) of dimension 10980 x 10980 x 4.
}
\usage{
stars_sentinel2
}
\description{
Sentinel-2 sample tile, downloaded from <https://scihub.copernicus.eu/>
reads the four 10-m bands: B2 (490 nm), B3 (560 nm), B4 (665 nm) and B8 (842 nm)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/factors.R
\name{cut_stars}
\alias{cut_stars}
\alias{cut.array}
\alias{cut.matrix}
\alias{cut.stars}
\title{cut methods for stars objects}
\usage{
\method{cut}{array}(x, breaks, ...)

\method{cut}{matrix}(x, breaks, ...)

\method{cut}{stars}(x, breaks, ...)
}
\arguments{
\item{x}{see \link[base]{cut}}

\item{breaks}{see \link[base]{cut}}

\item{...}{see \link[base]{cut}}
}
\value{
an array or matrix with a \code{levels} attribute; see details
}
\description{
cut methods for stars objects
}
\details{
R's \code{factor} only works for vectors, not for arrays or matrices. This is a work-around (or hack?) to keep the factor levels generated by \code{cut} and use them in plots.
}
\examples{
tif = system.file("tif/L7_ETMs.tif", package = "stars")
x = read_stars(tif)
cut(x, c(0, 50, 100, 255))
cut(x[,,,1], c(0, 50, 100, 255))
plot(cut(x[,,,1], c(0, 50, 100, 255)))
tif = system.file("tif/L7_ETMs.tif", package = "stars")
x1 = read_stars(tif)
(x1_cut = cut(x1, breaks = c(0, 50, 100, Inf)))  # shows factor in summary
plot(x1_cut[,,,c(3,6)]) # propagates through [ and plot
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ops.R
\name{st_apply}
\alias{st_apply}
\alias{st_apply.stars}
\title{st_apply apply a function to one or more array dimensions}
\usage{
\method{st_apply}{stars}(
  X,
  MARGIN,
  FUN,
  ...,
  CLUSTER = NULL,
  PROGRESS = FALSE,
  FUTURE = FALSE,
  rename = TRUE,
  .fname,
  single_arg = has_single_arg(FUN, list(...)) || can_single_arg(FUN),
  keep = FALSE
)
}
\arguments{
\item{X}{object of class \code{stars}}

\item{MARGIN}{see \link[base]{apply}; index number(s) or name(s) of the dimensions over which \code{FUN} will be applied}

\item{FUN}{see \link[base]{apply} and see Details.}

\item{...}{arguments passed on to \code{FUN}}

\item{CLUSTER}{cluster to use for parallel apply; see \link[parallel]{makeCluster}}

\item{PROGRESS}{logical; if \code{TRUE}, use \code{pbapply::pbapply} to show progress bar}

\item{FUTURE}{logical;if \code{TRUE}, use \code{future.apply::future_apply}}

\item{rename}{logical; if \code{TRUE} and \code{X} has only one attribute and 
\code{FUN} is a simple function name, rename the attribute of the returned object 
to the function name}

\item{.fname}{function name for the new attribute name (if one or more 
dimensions are reduced) or the new dimension (if a new dimension is created); 
if missing, the name of \code{FUN} is used}

\item{single_arg}{logical; if \code{TRUE}, FUN takes a single argument (like \code{fn_ndvi1} below), 
if \code{FALSE} FUN takes multiple arguments (like \code{fn_ndvi2} below).}

\item{keep}{logical; if \code{TRUE}, preserve dimension metadata (e.g. time stamps)}
}
\value{
object of class \code{stars} with accordingly reduced number of dimensions; 
in case \code{FUN} returns more than one value, a new dimension is created carrying 
the name of the function used; see the examples. Following the logic of 
\link[base]{apply}, This new dimension is put before the
other dimensions; use \link{aperm} to rearrange this, see last example.
}
\description{
st_apply apply a function to array dimensions: aggregate over space, time, or something else
}
\details{
FUN is a function which either operates on a single object, which will 
be the data of each iteration step over dimensions MARGIN, or a function that 
has as many arguments as there are elements in such an object. See the NDVI 
examples below. The second form can be VERY much faster e.g. when a trivial 
function is not being called for every pixel, but only once (example).

The heuristics for the default of \code{single_arg} work often, but not always; try
setting this to the right value when \code{st_apply} gives an error.
}
\examples{
tif = system.file("tif/L7_ETMs.tif", package = "stars")
x = read_stars(tif)
st_apply(x, 1:2, mean) # mean band value for each pixel
st_apply(x, c("x", "y"), mean) # equivalent to the above
st_apply(x, 3, mean)   # mean of all pixels for each band
\dontrun{
 st_apply(x, "band", mean) # equivalent to the above
 st_apply(x, 1:2, range) # min and max band value for each pixel
 fn_ndvi1 = function(x) (x[4]-x[3])/(x[4]+x[3]) # ONE argument: will be called for each pixel
 fn_ndvi2 = function(red,nir) (nir-red)/(nir+red) # n arguments: will be called only once
 ndvi1 = st_apply(x, 1:2, fn_ndvi1)
   # note that we can select bands 3 and 4 in the first argument:
 ndvi2 = st_apply(x[,,,3:4], 1:2, fn_ndvi2) 
 all.equal(ndvi1, ndvi2)
 # compute the (spatial) variance of each band; https://github.com/r-spatial/stars/issues/430
 st_apply(x, 3, function(x) var(as.vector(x))) # as.vector is required!
 # to get a progress bar also in non-interactive mode, specify:
 if (require(pbapply)) { # install it, if FALSE
   pboptions(type = "timer")
 }
 st_apply(x, 1:2, range) # dimension "range" is first; rearrange by:
 st_apply(x, 1:2, range) |> aperm(c(2,3,1))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gt.R
\name{st_geotransform}
\alias{st_geotransform}
\alias{st_geotransform<-}
\alias{st_geotransform<-.stars}
\title{get or set the geotransform, or rotation matrix}
\usage{
st_geotransform(x, ...)

st_geotransform(x) <- value

\method{st_geotransform}{stars}(x) <- value
}
\arguments{
\item{x}{object of class stars or dimensions}

\item{...}{ignored}

\item{value}{length 6 numeric vector, or 2 x 2 (scaled) rotation matrix}
}
\description{
get or set the geotransform, or rotation matrix
}
\examples{
# using the "classical" rotation matrix, see https://en.wikipedia.org/wiki/Rotation_matrix :
rot = function(theta, dxdy = c(1., -1.)) {
   th = theta / 180 * pi
   matrix(c(cos(th), sin(th), -sin(th), cos(th)), 2, 2) \%*\% 
		   matrix(c(dxdy[2], 0, 0, dxdy[1]), 2, 2)
}
l = st_downsample(st_as_stars(L7_ETMs), 9) # save time in plotting
st_geotransform(l) = rot(20, c(28.5, 28.5)) # clockwise, 20 degrees, scale by cell size
plot(l[,,,1])
m = rot(20, c(1, 2))
g = expand.grid(x = 0:4, y = 0:4)
plot(g[1:2], asp = 1)
text(g[,1], g[,2], labels = seq_along(g[,1]), pos = 4)
g = t(m \%*\% t(as.matrix(g)))
points(g, col = 'red')
text(g[,1], g[,2], labels = seq_along(g[,1]), pos = 4, col = 'red')

m = matrix(1:20, 4)
s0 = st_as_stars(m)
s = s0
# dy > 0, clockwise rotation:
st_geotransform(s) = rot(10, c(1,1))
plot(s0, reset = FALSE)
plot(s, add = TRUE)
# dy < 0, counter clockwise rotation, + expansion in x-direction:
layout(1)
s0 = st_as_stars(st_bbox(s0), dx = 1)
s0$values = 1:20
s0
plot(s0, reset = FALSE)
s = s0
st_geotransform(s) = rot(10, c(2,1))
plot(s, add = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sf.R
\name{st_contour}
\alias{st_contour}
\title{Compute or plot contour lines or sets}
\usage{
st_contour(
  x,
  na.rm = TRUE,
  contour_lines = FALSE,
  breaks = pretty(na.omit(as.vector(x[[1]])))
)
}
\arguments{
\item{x}{object of class \code{stars}}

\item{na.rm}{logical; should missing valued cells be removed, or also be converted to features?}

\item{contour_lines}{logical; if \code{FALSE}, polygons are returned (contour sets), otherwise contour lines}

\item{breaks}{numerical; values at which to "draw" contour levels}
}
\description{
Compute contour lines or sets
}
\details{
this function requires GDAL >= 2.4.0
}
\seealso{
for polygonizing rasters following grid boundaries, see \link{st_as_sf} with arguments \code{as_points=FALSE} and \code{merge=TRUE}; \link{contour} plots contour lines using R's native algorithm (which also plots contour levels)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stars.R
\name{st_raster_type}
\alias{st_raster_type}
\title{get the raster type (if any) of a stars object}
\usage{
st_raster_type(x, dimension = character(0))
}
\arguments{
\item{x}{object of class \code{stars}}

\item{dimension}{optional: numbers or names of dimension(s) to get per-dimension type}
}
\value{
if \code{dimension} is not specified, return the spatial raster type: 
one of \code{NA} (if the object does not have raster dimensions), 
\code{"curvilinear"}, \code{"rectilinear"}, \code{"affine"}, or \code{"regular"}.
In case dimension(s) are specified, return one of \code{"regular"}, \code{"rectilinear"}
(irregular but numeric), or \code{"discrete"} (anything else).
}
\description{
get the raster type (if any) of a stars object
}
\details{
categories \code{"curvilinear"} and \code{"affine"} only refer to
the relationship between a pair of spatial (raster) dimensions.
}
\examples{
tif = system.file("tif/L7_ETMs.tif", package = "stars")
x = read_stars(tif)
st_raster_type(x)
st_raster_type(x, 1:3)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom.R
\name{st_join.stars}
\alias{st_join.stars}
\title{Spatially join a stars and an `sf` object}
\usage{
\method{st_join}{stars}(
  x,
  y,
  join = st_intersects,
  ...,
  what = "left1",
  as_points = NA,
  warn = TRUE
)
}
\arguments{
\item{x}{object of class stars}

\item{y}{object of class sf, or one that can be coerced into that by \link{st_as_sf}}

\item{join}{the join function, which should return an sgbp object; see details}

\item{...}{arguments that will be passed on to the join function}

\item{what}{"left1", "right" or "inner"; see details}

\item{as_points}{logical; controls whether grid cells in \code{x} will be treated as points, or as cell areas; the \link{st_intersects.stars} method by default will derive this from \code{x}'s metadata, or else assume areas.}

\item{warn}{logical; if TRUE, warn on 1-to-many matches when \code{what} is \code{"left1"}}
}
\value{
If what is "left1", an object of class stars with the (first) value of y at spatial instances of x
}
\description{
Spatially join a stars and an `sf` object
}
\details{
When there is more than one match to a single x value, the first matching record from y is taken (and if \code{warn} is TRUE a warning is raised). If what is "inner", an object of class \code{sf} with all matching records of x and y.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dimensions.R, R/stars.R
\name{print_stars}
\alias{print_stars}
\alias{as.data.frame.dimensions}
\alias{print.dimensions}
\alias{print.stars}
\title{print stars or dimensions object}
\usage{
\method{as.data.frame}{dimensions}(
  x,
  ...,
  digits = max(3, getOption("digits") - 3),
  usetz = TRUE,
  stars_crs = getOption("stars.crs") \%||\% 28,
  all = FALSE
)

\method{print}{dimensions}(x, ...)

\method{print}{stars}(x, ..., n = 1e+05, abbrev = 30)
}
\arguments{
\item{x}{object of class stars or of class dimensions}

\item{...}{passed on to \code{as.data.frame.dimensions}}

\item{digits}{number of digits to print numbers}

\item{usetz}{logical; used to format \code{POSIXct} values}

\item{stars_crs}{maximum width of string for CRS objects}

\item{all}{logical; if \code{TRUE} print also fields entirely filled with \code{NA} or \code{NULL}}

\item{n}{when prod(dim(x)) > 10 * n, the first n cells are used for attribute summary statistics}

\item{abbrev}{number of characters to abbreviate attribute names to}
}
\description{
print stars or dimensions object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\docType{data}
\name{L7_ETMs}
\alias{L7_ETMs}
\title{Landsat-7 bands for a selected region around Olinda, BR}
\format{
An object of class \code{stars_proxy} (inherits from \code{stars}) of dimension 349 x 352 x 6.
}
\usage{
L7_ETMs
}
\description{
Probably containing the six 30 m bands:
\itemize{
 \item Band 1 Visible (0.45 - 0.52 µm) 30 m
 \item Band 2 Visible (0.52 - 0.60 µm) 30 m
 \item Band 3 Visible (0.63 - 0.69 µm) 30 m
 \item Band 4 Near-Infrared (0.77 - 0.90 µm) 30 m
 \item Band 5 Short-wave Infrared (1.55 - 1.75 µm) 30 m
 \item Band 7 Mid-Infrared (2.08 - 2.35 µm) 30 m
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/proxy.R, R/subset.R
\name{st_normalize}
\alias{st_normalize}
\alias{st_normalize.stars_proxy}
\alias{st_normalize.stars}
\title{Move raster offset such that grid start at row/col (1,1)}
\usage{
\method{st_normalize}{stars_proxy}(x, domain = c(0, 0, 1, 1), ...)

\method{st_normalize}{stars}(x, domain = c(0, 0, 1, 1), ...)
}
\arguments{
\item{x}{object of class `stars` or `stars_proxy`}

\item{domain}{ignored}

\item{...}{ignored}
}
\description{
Move raster offset such that grid start at row/col (1,1)
}
\details{
the `stars_proxy` method does nothing, but raises an error when a raster does not start at cell (1,1)

the `stars` method "moves" the `offset` of the raster dimensions, such that the `from`
field becomes `(1,1)` for the raster dimensions; as a side effect, it moves the raster
dimensions `x` and `y` to first and second position.  If the `from` field is already `(1,1)` or the
object does not have raster dimensions it does nothing.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dimensions.R
\name{expand_dimensions}
\alias{expand_dimensions}
\alias{expand_dimensions.dimensions}
\title{expand the dimension values into a list}
\usage{
expand_dimensions(x, ...)

\method{expand_dimensions}{dimensions}(x, ..., max = FALSE, center = NA)
}
\arguments{
\item{x}{object of class `stars` or `dimensions`}

\item{...}{ignored}

\item{max}{logical; if `TRUE` return the max (end) values of the dimensions intervals}

\item{center}{logical; if `TRUE` return the center values of intervals, otherwise return offset (start) of intervals; if `NA` (default) return centers for x/y dimensions, offsets for all others}
}
\description{
expand the dimension values into a list
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidyverse.R
\name{dplyr}
\alias{dplyr}
\alias{filter.stars}
\alias{filter.stars_proxy}
\alias{mutate.stars}
\alias{mutate.stars_proxy}
\alias{transmute.stars}
\alias{transmute.stars_proxy}
\alias{select.stars}
\alias{select.stars_proxy}
\alias{rename.stars}
\alias{rename.stars_proxy}
\alias{pull.stars}
\alias{pull.stars_proxy}
\alias{as.tbl_cube.stars}
\alias{slice.stars}
\alias{slice.stars_proxy}
\alias{replace_na.stars}
\alias{replace_na.stars_proxy}
\title{dplyr verbs for stars objects}
\usage{
filter.stars(.data, ...)

filter.stars_proxy(.data, ...)

mutate.stars(.data, ...)

mutate.stars_proxy(.data, ...)

transmute.stars(.data, ...)

transmute.stars_proxy(.data, ...)

select.stars(.data, ...)

select.stars_proxy(.data, ...)

rename.stars(.data, ...)

rename.stars_proxy(.data, ...)

pull.stars(.data, var = -1)

pull.stars_proxy(.data, ...)

as.tbl_cube.stars(x, ...)

slice.stars(.data, along, index, ..., drop = length(index) == 1)

slice.stars_proxy(.data, along, index, ...)

replace_na.stars(data, replace, ...)

replace_na.stars_proxy(data, ...)
}
\arguments{
\item{.data}{object of class \code{stars}}

\item{...}{see \link[dplyr]{filter}}

\item{var}{see \link[dplyr]{pull}}

\item{x}{object of class \code{stars}}

\item{along}{name or index of dimension to which the slice should be applied}

\item{index}{integer value(s) for this index}

\item{drop}{logical; drop dimensions that only have a single index?}

\item{data}{data set to work on}

\item{replace}{see \link[tidyr]{replace_na}: list with variable=value pairs, where value is the replacement value for NA's}
}
\description{
dplyr verbs for stars objects; package dplyr needs to be loaded before these methods can be used for stars objects.
}
\examples{
tif = system.file("tif/L7_ETMs.tif", package = "stars")
x1 = read_stars(tif)
if (require(dplyr, quietly = TRUE)) {
 x1 |> slice("band", 2:3)
 x1 |> slice("x", 50:100)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sf.R
\name{st_as_sf}
\alias{st_as_sf}
\alias{st_as_sfc.stars}
\alias{st_as_sf.stars}
\alias{st_as_sf.stars_proxy}
\title{Convert stars object into an sf object}
\usage{
\method{st_as_sfc}{stars}(x, ..., as_points, which = seq_len(prod(dim(x)[1:2])))

\method{st_as_sf}{stars}(
  x,
  ...,
  as_points = FALSE,
  merge = FALSE,
  na.rm = TRUE,
  use_integer = is.logical(x[[1]]) || is.integer(x[[1]]),
  long = FALSE,
  connect8 = FALSE
)

\method{st_as_sf}{stars_proxy}(x, ..., downsample = 0)
}
\arguments{
\item{x}{object of class \code{stars}}

\item{...}{ignored}

\item{as_points}{logical; should cells be converted to points or to polygons? See details.}

\item{which}{linear index of cells to keep (this argument is not recommended to be used)}

\item{merge}{logical; if \code{TRUE}, cells with identical values are merged (using \code{GDAL_Polygonize} or \code{GDAL_FPolygonize}); if \code{FALSE}, a polygon for each raster cell is returned; see details}

\item{na.rm}{logical; should missing valued cells be removed, or also be converted to features?}

\item{use_integer}{(relevant only if \code{merge} is \code{TRUE}): if \code{TRUE}, before polygonizing values are rounded to 32-bits signed integer values (GDALPolygonize), otherwise they are converted to 32-bit floating point values (GDALFPolygonize).}

\item{long}{logical; if \code{TRUE}, return a long table form \code{sf}, with geometries and other dimensions recycled}

\item{connect8}{logical; if \code{TRUE}, use 8 connectedness. Otherwise the 4 connectedness algorithm will be applied.}

\item{downsample}{see \link{st_as_stars}}
}
\description{
Convert stars object into an sf object
}
\details{
If \code{merge} is \code{TRUE}, only the first attribute is converted into an \code{sf} object. If \code{na.rm} is \code{FALSE}, areas with \code{NA} values are also written out as polygons. Note that the resulting polygons are typically invalid, and use \link[sf:valid]{st_make_valid} to create valid polygons out of them.
}
\examples{
tif = system.file("tif/L7_ETMs.tif", package = "stars")
x = read_stars(tif)
x = x[,1:100,1:100,6] # subset of a band with lower values in it
x[[1]][x[[1]] < 30] = NA # set lower values to NA
x[[1]] = x[[1]] < 100 # make the rest binary
x
(p = st_as_sf(x)) # removes NA areas
(p = st_as_sf(x[,,,1], merge = TRUE)) # glues polygons together
all(st_is_valid(p)) # not all valid, see details
plot(p, axes = TRUE)
(p = st_as_sf(x, na.rm = FALSE, merge = TRUE)) # includes polygons with NA values
plot(p, axes = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stars.R, R/tidyverse.R
\name{st_coordinates}
\alias{st_coordinates}
\alias{st_coordinates.stars}
\alias{as.data.frame.stars}
\alias{as_tibble.stars}
\title{retrieve coordinates for raster or vector cube cells}
\usage{
\method{st_coordinates}{stars}(x, ..., add_max = FALSE, center = TRUE)

\method{as.data.frame}{stars}(x, ..., add_max = FALSE, center = NA, add_coordinates = TRUE)

as_tibble.stars(.x, ..., add_max = FALSE, center = NA)
}
\arguments{
\item{x}{object of class \code{stars}}

\item{...}{ignored}

\item{add_max}{logical; if \code{TRUE}, dimensions are given with a min (x) and max (x_max) value}

\item{center}{logical; (only if \code{add_max} is FALSE): should grid cell center coordinates be returned (TRUE) or offset values (FALSE)? \code{center} can be a named logical vector or list to specify values for each dimension.}

\item{add_coordinates}{logical; if `TRUE`, columns with dimension values preceed the array values, 
otherwise they are omitted}

\item{.x}{object to be converted to a tibble}
}
\description{
retrieve coordinates for raster or vector cube cells
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stars.R
\name{st_dim_to_attr}
\alias{st_dim_to_attr}
\title{create an array with dimension values}
\usage{
st_dim_to_attr(x, which = seq_along(dim(x)))
}
\arguments{
\item{x}{object of class \code{stars}}

\item{which}{integer; indices of the dimensions to address (default: all)}
}
\value{
\code{stars} object with dimension values as attributes
}
\description{
create an array with dimension values
}
\examples{
tif = system.file("tif/L7_ETMs.tif", package = "stars")
x1 = read_stars(tif)
(x = st_dim_to_attr(x1))
plot(x)
(x = st_dim_to_attr(x1, 2:3))
plot(x)
(x= st_dim_to_attr(x1, 3))
plot(x)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prcomp.R
\name{prcomp}
\alias{prcomp}
\alias{prcomp.stars_proxy}
\alias{prcomp.stars}
\title{Principal components of stars object}
\usage{
\method{prcomp}{stars_proxy}(x, ..., downsample = 0)

\method{prcomp}{stars}(x, ..., quiet = FALSE)
}
\arguments{
\item{x}{object of class `stars` or `stars_proxy`}

\item{...}{see \link[stats]{prcomp}}

\item{downsample}{see \link{st_as_stars}}

\item{quiet}{logical; if `TRUE`, suppress message that PCs will be computed on last dimension; see details}
}
\value{
object of class `prcomp`, see \link[stats]{prcomp}
}
\description{
Compute principal components of stars object
}
\details{
if `x` has only one attribute, principal components will be computed in the space of the last dimension of `x`
to predict PC scores into a `stars` object, use \link{predict.stars}; see example below
}
\examples{
l7 = split(st_as_stars(L7_ETMs), 3) # use bands as features
plot(prcomp(l7))
plot(merge(predict(l7, model = prcomp(l7))))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dimensions.R
\name{st_dimensions}
\alias{st_dimensions}
\alias{st_dimensions.stars}
\alias{st_dimensions<-}
\alias{st_dimensions<-.stars}
\alias{st_dimensions<-.stars_proxy}
\alias{st_dimensions<-.list}
\alias{st_dimensions.array}
\alias{st_dimensions.default}
\alias{st_set_dimensions}
\alias{st_get_dimension_values}
\title{get dimensions from stars object}
\usage{
st_dimensions(.x, ...)

\method{st_dimensions}{stars}(.x, ...)

st_dimensions(x) <- value

\method{st_dimensions}{stars}(x) <- value

\method{st_dimensions}{stars_proxy}(x) <- value

\method{st_dimensions}{list}(x) <- value

\method{st_dimensions}{array}(.x, ...)

\method{st_dimensions}{default}(
  .x,
  ...,
  .raster,
  affine = c(0, 0),
  cell_midpoints = FALSE,
  point = FALSE
)

st_set_dimensions(
  .x,
  which,
  values = NULL,
  point = NULL,
  names = NULL,
  xy,
  ...
)

st_get_dimension_values(.x, which, ..., where = NA, max = FALSE, center = NA)
}
\arguments{
\item{.x}{object to retrieve dimensions information from}

\item{...}{further arguments}

\item{x}{object of class \code{dimensions}}

\item{value}{new object of class \code{dimensions}, with matching dimensions}

\item{.raster}{length 2 character array with names (if any) of the raster dimensions}

\item{affine}{numeric; specify parameters of the affine transformation}

\item{cell_midpoints}{logical; if \code{TRUE} AND the dimension values are strictly regular, the values are interpreted as the cell midpoint values rather than the cell offset values when calculating offset (i.e., the half-cell-size correction is applied); can have a value for each dimension, or else is recycled}

\item{point}{logical; does the pixel value (measure) refer to a point (location) value or to an pixel (area) summary value?}

\item{which}{integer or character; index or name of the dimension to be changed}

\item{values}{values for this dimension (e.g. \code{sfc} list-column), or length-1 \code{dimensions} object; setting special value \code{NULL} removes dimension values, for instance to remove curvilinear raster coordinates}

\item{names}{character; vector with new names for all dimensions, or with the single new name for the dimension indicated by \code{which}}

\item{xy}{length-2 character vector; (new) names for the \code{x} and \code{y} raster dimensions}

\item{where}{character, one of 'start', 'center' or 'end'. Set to NA (default) to ignore and use \code{max} and \code{center} explictly.  This argument provides a convenient alternative to setting \code{max} and \code{center}.}

\item{max}{logical; if \code{TRUE} return the end, rather than the beginning of an interval}

\item{center}{logical; if \code{TRUE} return the center of an interval; if \code{NA} return the center for raster dimensions, and the start of intervals in other cases}
}
\value{
the \code{dimensions} attribute of \code{x}, of class \code{dimensions}
}
\description{
get dimensions from stars object
}
\details{
dimensions can be specified in two ways. The simplest is to pass a vector with numeric values for a numeric dimension, or character values for a categorical dimension. Parameter \code{cell_midpoints} is used to specify whether numeric values refer to the offset (start) of a dimension interval (default), or to the center; the center case is only available for regular dimensions. For rectilinear numeric dimensions, one can specify either a vector with cell borders (start values), or a data.frame with two columns named "start" and "end", with the respective interval start and end values. In the first case, the end values are computed from the start values by assuming the last two intervals have equal width.
}
\examples{
x = read_stars(system.file("tif/L7_ETMs.tif", package = "stars"))
# Landsat 7 ETM+ band semantics: https://landsat.gsfc.nasa.gov/the-enhanced-thematic-mapper-plus/
# set bands to values 1,2,3,4,5,7:
(x1 = st_set_dimensions(x, "band", values = c(1,2,3,4,5,7), names = "band_number", point = TRUE))
# set band values as bandwidth
rbind(c(0.45,0.515), c(0.525,0.605), c(0.63,0.69), c(0.775,0.90), c(1.55,1.75), c(2.08,2.35)) |>
  units::set_units("um") -> bw # or: units::set_units(µm) -> bw
# set bandwidth midpoint:
(x2 = st_set_dimensions(x, "band", values = 0.5 * (bw[,1]+bw[,2]), 
   names = "bandwidth_midpoint", point = TRUE))
# set bandwidth intervals:
(x3 = st_set_dimensions(x, "band", values = make_intervals(bw), names = "bandwidth"))
m = matrix(1:20, nrow = 5, ncol = 4)
dim(m) = c(x = 5, y = 4) # named dim
(s = st_as_stars(m))
st_get_dimension_values(s, 'x', where = "start")
st_get_dimension_values(s, 'x', center = FALSE)
st_get_dimension_values(s, 'x', where = "center")
st_get_dimension_values(s, 'x', center = TRUE)
st_get_dimension_values(s, 'x', where = "end")
st_get_dimension_values(s, 'x', max = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/warp.R
\name{st_warp}
\alias{st_warp}
\title{Warp (resample) grids in stars objects to a new grid, possibly in an new coordinate reference system}
\usage{
st_warp(
  src,
  dest,
  ...,
  crs = NA_crs_,
  cellsize = NA_real_,
  segments = 100,
  use_gdal = FALSE,
  options = character(0),
  no_data_value = NA_real_,
  debug = FALSE,
  method = "near",
  threshold = NA_real_
)
}
\arguments{
\item{src}{object of class \code{stars} with source raster}

\item{dest}{object of class \code{stars} with target raster geometry}

\item{...}{ignored}

\item{crs}{coordinate reference system for destination grid, only used when \code{dest} is missing}

\item{cellsize}{length 1 or 2 numeric; cellsize in target coordinate reference system units}

\item{segments}{(total) number of segments for segmentizing the bounding box before transforming to the new crs}

\item{use_gdal}{logical; if \code{TRUE}, use gdal's warp or warper, through \link[sf]{gdal_utils}}

\item{options}{character vector with options, passed on to gdalwarp}

\item{no_data_value}{value used by gdalwarp for no_data (NA) when writing to temporary file; 
not setting this when \code{use_gdal} is \code{TRUE} leads to a warning}

\item{debug}{logical; if \code{TRUE}, do not remove the temporary gdalwarp destination file, and print its name}

\item{method}{character; see details for options; methods other than \code{near} only work when \code{use_gdal=TRUE}}

\item{threshold}{numeric; distance threshold for warping curvilinear grids: new cells at distances larger than threshold are assigned NA values.}
}
\description{
Warp (resample) grids in stars objects to a new grid, possibly in an new coordinate reference system
}
\details{
\code{method} should be one of \code{near}, \code{bilinear}, \code{cubic}, \code{cubicspline}, \code{lanczos}, \code{average}, \code{mode}, \code{max}, \code{min}, \code{med}, \code{q1} or \code{q3}; see https://github.com/r-spatial/stars/issues/109

For gridded spatial data (dimensions \code{x} and \code{y}), see figure; the existing grid is transformed into a regular grid defined by \code{dest}, possibly in a new coordinate reference system. If \code{dest} is not specified, but \code{crs} is, the procedure used to choose a target grid is similar to that of \link[raster]{projectRaster}. This entails: (i) the envelope (bounding box polygon) is transformed into the new crs, possibly after segmentation (red box); (ii) a grid is formed in this new crs, touching the transformed envelope on its East and North side, with (if cellsize is not given) a cellsize similar to the cell size of \code{src}, with an extent that at least covers \code{x}; (iii) for each cell center of this new grid, the matching grid cell of \code{x} is used; if there is no match, an \code{NA} value is used.
}
\examples{
geomatrix = system.file("tif/geomatrix.tif", package = "stars")
(x = read_stars(geomatrix))
new_crs = st_crs('OGC:CRS84')
y = st_warp(x, crs = new_crs)
plot(st_transform(st_as_sfc(st_bbox(x)), new_crs), col = NA, border = 'red')
plot(st_as_sfc(y, as_points=FALSE), col = NA, border = 'green', axes = TRUE, add = TRUE)
image(y, add = TRUE, nbreaks = 6)
plot(st_as_sfc(y, as_points=TRUE), pch=3, cex=.5, col = 'blue', add = TRUE)
plot(st_transform(st_as_sfc(x, as_points=FALSE), new_crs), add = TRUE)
# warp 0-360 raster to -180-180 raster:
r = read_stars(system.file("nc/reduced.nc", package = "stars"))
r |> st_set_crs('OGC:CRS84') |> st_warp(st_as_stars(st_bbox(), dx = 2)) -> s
plot(r, axes = TRUE) # no CRS set, so no degree symbols in labels
plot(s, axes = TRUE)
# downsample raster (90 to 270 m)
r = read_stars(system.file("tif/olinda_dem_utm25s.tif", package = "stars"))
r270 = st_as_stars(st_bbox(r), dx = 270)
r270 = st_warp(r, r270)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write.R
\name{write_stars}
\alias{write_stars}
\alias{write_stars.stars}
\alias{write_stars.stars_proxy}
\alias{detect.driver}
\title{write stars object to gdal dataset (typically: to file)}
\usage{
write_stars(obj, dsn, layer, ...)

\method{write_stars}{stars}(
  obj,
  dsn,
  layer = 1,
  ...,
  driver = detect.driver(dsn),
  options = character(0),
  type = if (is.factor(obj[[1]]) && length(levels(obj[[1]])) < 256) "Byte" else "Float32",
  NA_value = NA_real_,
  update = FALSE,
  normalize_path = TRUE,
  scale_offset = c(1, 0)
)

\method{write_stars}{stars_proxy}(
  obj,
  dsn,
  layer = 1,
  ...,
  driver = detect.driver(dsn),
  options = character(0),
  scale_offset = c(1, 0),
  type = "Float32",
  NA_value = NA_real_,
  chunk_size = c(dim(obj)[1], floor(2.5e+07/dim(obj)[1])),
  progress = TRUE
)

detect.driver(filename)
}
\arguments{
\item{obj}{object of class \code{stars}}

\item{dsn}{gdal dataset (file) name}

\item{layer}{attribute name; if missing, the first attribute is written}

\item{...}{passed on to \link[sf:gdal]{gdal_write}}

\item{driver}{driver driver name; see \link[sf]{st_drivers}}

\item{options}{character vector with dataset creation options, passed on to GDAL}

\item{type}{character; output binary type, one of: \code{Byte} for eight bit unsigned integer, \code{UInt16} for sixteen bit unsigned integer, \code{Int16} for sixteen bit signed integer, \code{UInt32} for thirty two bit unsigned integer, \code{Int32} for thirty two bit signed integer, \code{Float32} for thirty two bit floating point, \code{Float64} for sixty four bit floating point.}

\item{NA_value}{non-NA value that should represent R's \code{NA} value in the target raster file; if set to \code{NA}, it will be ignored.}

\item{update}{logical; if \code{TRUE}, an existing file is being updated}

\item{normalize_path}{logical; see \link{read_stars}}

\item{scale_offset}{length 2 numeric vector with scale, offset values: raw values computed by raw = (value - offset) / scale are written to dsn; scale and offset values are written to dsn or else a warning is raised}

\item{chunk_size}{length two integer vector with the number of pixels (x, y) used in the read/write loop; see details.}

\item{progress}{logical; if \code{TRUE}, a progress bar is shown}

\item{filename}{character; used for guessing driver short name based on file 
extension; see examples}
}
\description{
write stars object to gdal dataset (typically: to file)
}
\details{
\code{write_stars} first creates the target file, then updates it sequentially by writing blocks of \code{chunk_size}.

in case \code{obj} is a multi-file \code{stars_proxy} object, all files are written as layers into the output file \code{dsn}
}
\examples{
detect.driver("L7_ETMs.tif")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mosaic.R
\name{st_mosaic}
\alias{st_mosaic}
\alias{st_mosaic.stars}
\alias{st_mosaic.character}
\alias{st_mosaic.stars_proxy}
\title{build mosaic (composite) of several spatially disjoint stars objects}
\usage{
st_mosaic(.x, ...)

\method{st_mosaic}{stars}(
  .x,
  ...,
  dst = tempfile(fileext = file_ext),
  options = c("-vrtnodata", "-9999", "-srcnodata", "nan"),
  file_ext = ".tif"
)

\method{st_mosaic}{character}(
  .x,
  ...,
  dst = tempfile(fileext = file_ext),
  options = c("-vrtnodata", "-9999"),
  file_ext = ".tif"
)

\method{st_mosaic}{stars_proxy}(
  .x,
  ...,
  dst = tempfile(fileext = file_ext),
  options = c("-vrtnodata", "-9999"),
  file_ext = ".tif"
)
}
\arguments{
\item{.x}{object of class stars, or character vector with input dataset names}

\item{...}{further input stars objects}

\item{dst}{character; destination file name; this will be a VRT file with references to the source file(s), see details}

\item{options}{character; options to the gdalbuildvrt command}

\item{file_ext}{character; file extension, determining the format used to write to (".tif" implies GeoTIFF)}
}
\value{
the stars method returns a stars object with the composite of the input; the \code{character} method returns the file name of the file with the mosaic; see also the GDAL documentation of \code{gdalbuildvrt}
}
\description{
build mosaic (composite) of several spatially disjoint stars objects
}
\details{
the gdal function buildvrt builds a mosaic of input images; these input images can be multi-band, but not higher-dimensional data cubes or stars objects with multiple attributes; note that for the `stars` method, the `dst` file may contain references to temporary files that are going to be removed at termination of the R session.

uses \link[sf]{gdal_utils} to internally call \code{buildvrt}; no executables external to R are called.
}
\examples{
x = read_stars(system.file("tif/L7_ETMs.tif", package = "stars"))
x1 = x[,100:200,100:200,]
x2 = x[,150:300,150:300,]
plot(st_mosaic(x1, x2))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/proxy.R, R/stars.R
\name{c.stars}
\alias{c.stars}
\alias{c.stars_proxy}
\title{combine multiple stars objects, or combine multiple attributes in a single stars object into a single array}
\usage{
\method{c}{stars_proxy}(
  ...,
  along = NA_integer_,
  along_crs = FALSE,
  try_hard = FALSE,
  nms = names(list(...)),
  tolerance = sqrt(.Machine$double.eps)
)

\method{c}{stars}(
  ...,
  along = NA_integer_,
  try_hard = FALSE,
  nms = names(list(...)),
  tolerance = sqrt(.Machine$double.eps)
)
}
\arguments{
\item{...}{object(s) of class \code{star}: in case of multiple arguments, these are combined into a single stars object, in case of a single argument, its attributes are combined into a single attribute. In case of multiple objects, all objects should have the same dimensionality.}

\item{along}{integer; see \link{read_stars}}

\item{along_crs}{logical; if \code{TRUE}, combine arrays along a CRS dimension}

\item{try_hard}{logical; if \code{TRUE} and some arrays have different dimensions, combine those that dimensions matching to the first array}

\item{nms}{character; vector with array names}

\item{tolerance}{numeric; values used in \link{all.equal} to compare dimension values
combine those that dimensions matching to the first array}
}
\value{
a single \code{stars} object with merged (binded) arrays.
}
\description{
combine multiple stars objects, or combine multiple attributes in a single stars object into a single array
}
\details{
An error is raised when attempting to combine arrays with different
measurement units into a single array. If this was intentded, \code{drop_units} 
can be used to remove units of a \code{stars} object before merging.
}
\examples{
tif = system.file("tif/L7_ETMs.tif", package = "stars")
x = read_stars(tif)
(new = c(x, x))
c(new) # collapses two arrays into one with an additional dimension
c(x, x, along = 3)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract.R
\name{st_extract}
\alias{st_extract}
\alias{st_extract.stars}
\title{Extract cell values at point locations}
\usage{
st_extract(x, ...)

\method{st_extract}{stars}(
  x,
  at,
  ...,
  bilinear = FALSE,
  time_column = attr(at, "time_column") \%||\% attr(at, "time_col"),
  interpolate_time = bilinear,
  FUN = mean,
  resampling = c("nearest", "bilinear", "cubic", "cubicspline"),
  sfc_attribute = NULL
)
}
\arguments{
\item{x}{object of class \code{stars} or \code{stars_proxy}}

\item{...}{passed on to \link{aggregate.stars} when geometries are not exclusively POINT geometries}

\item{at}{object of class \code{sf} or \code{sfc} with geometries, or two-column matrix with coordinate points in rows, indicating where to extract values of \code{x}, or a \code{stars} object with geometry and temporal dimensions (vector data cube)}

\item{bilinear}{logical; use bilinear interpolation rather than nearest neighbour?}

\item{time_column}{character or integer; name or index of a column with time or date values that will be matched to values of the first temporal dimension (matching classes \code{POSIXct}, \code{POSIXt}, \code{Date}, or \code{CFTime}), in \code{x}, after which this dimension is reduced. This is useful to extract data cube values along a trajectory; see https://github.com/r-spatial/stars/issues/352 .}

\item{interpolate_time}{logical; should time be interpolated? if FALSE, time instances are matched using the coinciding or the last preceding time in the data cube.}

\item{FUN}{function used to aggregate pixel values when geometries of \code{at} intersect with more than one pixel}

\item{resampling}{character; resampling method; for method cubic or cubicspline,
`stars_proxy` objects should be used and GDAL should have version >= 3.10.0}

\item{sfc_attribute}{character; if \code{at} is of class \code{stars} should the aggregation be performed
for the attribute geometry rather than the dimension geometry? If \code{NULL} (default),
the aggregation is performed at the dimension geometries, else the name of the attribute geometry to perform the aggregation on.
If the given attribute geometry does not exist, the aggregation defaults to the dimension geometry.}
}
\value{
if \code{at} is of class \code{matrix}, a matrix with extracted values is returned; 
if \code{at} is of class \code{stars} and a temporal dimension was passed to \code{time_column},
a \code{stars} object with the original \code{at} dimensions
and the extracted values as attributes.
otherwise: if \code{x} has more dimensions than only x and y (raster), an 
object of class \code{stars} with POINT geometries replacing x and y raster
dimensions, if this is not the case, an object of \code{sf} with extracted values.
}
\description{
Extract cell values at point locations
}
\details{
points outside the raster are returned as \code{NA} values. For
large sets of points for which extraction is needed, passing a matrix as
to \code{at} may be much faster than passing an \code{sf} or \code{sfc} object.
}
\examples{
tif = system.file("tif/L7_ETMs.tif", package = "stars")
r = read_stars(tif)
pnt = st_sample(st_as_sfc(st_bbox(r)), 10)
st_extract(r, pnt)
st_extract(r, pnt) |> st_as_sf()
st_extract(r[,,,1], pnt)
st_extract(r, st_coordinates(pnt)) # "at" is a matrix: return a matrix
# Extraction on non-POINT geometries
poly = st_buffer(pnt, 1000)
st_extract(r, poly)

# Extraction with time matching
rdate = c(r, r*2, along = "date")
dates = c(Sys.Date()-1, Sys.Date())
rdate = st_set_dimensions(rdate, "date", values = c(dates))

pntsf = st_sf(date = dates, geometry = pnt)
st_extract(split(rdate, "band"), pntsf) # POINT geometries

polysf = st_buffer(pntsf, 1000)
st_extract(split(rdate, "band"), polysf, time_column = "date") # POLYGON geometries

vdc = st_sf(rdm = rnorm(20), polygons = st_buffer(st_sample(st_bbox(pnt), 20), 500),
			geometry = rep(pnt, 2), date = rep(dates, each = 10)) |> 
	st_as_stars(dims = c("geometry", "date"))

(vdc_new = st_extract(split(rdate, "band"), vdc)) # stars vector data cube
merge(vdc_new, name = "band")

### Extraction applied to the geometries inside the vector data cube (cell values)
(vdc_new2 = st_extract(split(rdate, "band"), vdc,
					   sfc_attribute = "polygons")) # stars vector data cube
merge(vdc_new2, name = "band")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/proxy.R, R/stars.R
\name{predict.stars}
\alias{predict.stars}
\alias{predict.stars_proxy}
\title{Predict values, given a model object, for a stars or stars_proxy object}
\usage{
\method{predict}{stars_proxy}(object, model, ...)

\method{predict}{stars}(object, model, ..., drop_dimensions = FALSE)
}
\arguments{
\item{object}{object of class `stars`}

\item{model}{model object of a class that has a predict method; check with `methods(class = class(object))`}

\item{...}{arguments passed on to this predict method}

\item{drop_dimensions}{logical; if `TRUE`, remove dimensions (coordinates etc) from `data.frame` with predictors}
}
\description{
Predict values, given a model object, for a stars or stars_proxy object
}
\details{
separate predictors in object need to be separate attributes in object; 
in case they are e.g. in a band dimension, use `split(object)`
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rasterize.R
\name{st_rasterize}
\alias{st_rasterize}
\title{rasterize simple feature geometries}
\usage{
st_rasterize(
  sf,
  template = guess_raster(sf, ...) \%||\% st_as_stars(st_bbox(sf), values = NA_real_,
    ...),
  file = tempfile(),
  driver = "GTiff",
  options = character(0),
  align = FALSE,
  proxy = FALSE,
  ...
)
}
\arguments{
\item{sf}{object of class \code{sf}}

\item{template}{optional; stars object with desired target geometry, or target geometry alignment if \code{align=TRUE}; see details}

\item{file}{temporary file name}

\item{driver}{driver for temporary file}

\item{options}{character; options vector for \code{GDALRasterize}}

\item{align}{logical; if \code{TRUE}, \code{template} is only used for the geometry _alignment_, 
informing target resolution and offset}

\item{proxy}{logical; should a proxy object be returned?}

\item{...}{arguments passed on to \link{st_as_stars}}
}
\description{
rasterize simple feature geometries
}
\details{
if `template` is a `stars` object, non-NA cells that are not covered by `sf` receive the value in `template`; see also argument `align`.
}
\examples{
demo(nc, echo = FALSE, ask = FALSE)
(x = st_rasterize(nc)) # default grid:
plot(x, axes = TRUE)
# a bit more customized grid:
(x = st_rasterize(nc, st_as_stars(st_bbox(nc), nx = 100, ny = 50, values = NA_real_)))
plot(x, axes = TRUE)
(ls = st_sf(a = 1:2, st_sfc(st_linestring(rbind(c(0.1, 0), c(1.1, 1))),
   st_linestring(rbind(c(0, 0.05), c(1, 0.05))))))
(grd = st_as_stars(st_bbox(ls), nx = 10, ny = 10, xlim = c(0, 1.0), ylim = c(0, 1),
   values = NA_real_))
# Only the left-top corner is part of the grid cell:
sf_extSoftVersion()["GDAL"]
plot(st_rasterize(ls, grd), axes = TRUE, reset = FALSE) # ALL_TOUCHED=FALSE; 
plot(ls, add = TRUE, col = "red")
plot(st_rasterize(ls, grd, options = "ALL_TOUCHED=TRUE"), axes = TRUE, reset = FALSE)
plot(ls, add = TRUE, col = "red")
# add lines to existing 0 values, summing values in case of multiple lines:
(grd = st_as_stars(st_bbox(ls), nx = 10, ny = 10, xlim = c(0, 1.0), ylim = c(0, 1), values = 0))
r = st_rasterize(ls, grd, options = c("MERGE_ALG=ADD", "ALL_TOUCHED=TRUE"))
plot(r, axes = TRUE, reset = FALSE)
plot(ls, add = TRUE, col = "red")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stars.R
\name{st_set_bbox}
\alias{st_set_bbox}
\title{set bounding box parameters of regular grid}
\usage{
st_set_bbox(x, value, ...)
}
\arguments{
\item{x}{object of class dimensions, stars or stars_proxy}

\item{value}{object of class bbox}

\item{...}{ignored}
}
\description{
set bounding box parameters of regular grid
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/proxy.R, R/stars.R
\name{redimension}
\alias{redimension}
\alias{st_redimension.stars_proxy}
\alias{st_redimension}
\alias{st_redimension.stars}
\title{redimension array, or collapse attributes into a new dimension}
\usage{
\method{st_redimension}{stars_proxy}(
  x,
  new_dims = st_dimensions(x),
  along = list(new_dim = names(x)),
  ...
)

st_redimension(x, new_dims, along, ...)

\method{st_redimension}{stars}(
  x,
  new_dims = st_dimensions(x),
  along = setNames(list(names(x)), name),
  ...,
  name = "new_dim"
)
}
\arguments{
\item{x}{object of class \code{stars}}

\item{new_dims}{target dimensions: either a `dimensions` object or an integer vector with the dimensions' sizes}

\item{along}{named list with new dimension name and values}

\item{...}{ignored}

\item{name}{character name of the new dimension}
}
\description{
redimension array, or collapse attributes into a new dimension
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mdim.R
\name{mdim}
\alias{mdim}
\alias{read_mdim}
\alias{write_mdim}
\title{Read or write data using GDAL's multidimensional array API}
\usage{
read_mdim(
  filename,
  variable = character(0),
  ...,
  options = character(0),
  raster = NULL,
  offset = integer(0),
  count = integer(0),
  step = integer(0),
  proxy = FALSE,
  debug = FALSE,
  bounds = TRUE,
  curvilinear = NA,
  normalize_path = TRUE
)

write_mdim(
  x,
  filename,
  driver = detect.driver(filename),
  ...,
  root_group_options = character(0),
  options = character(0),
  as_float = TRUE,
  normalize_path = TRUE
)
}
\arguments{
\item{filename}{name of the source or destination file or data source}

\item{variable}{name of the array to be read; if `"?"`, a list of array names is returned, with group name as list element names.}

\item{...}{ignored}

\item{options}{character; driver specific options regarding the opening (read_mdim) or creation (write_mdim) of the dataset}

\item{raster}{names of the raster variables (default: first two dimensions)}

\item{offset}{integer; zero-based offset for each dimension (pixels) of sub-array to read, defaults to 0 for each dimension(requires sf >= 1.0-9)}

\item{count}{integer; size for each dimension (pixels) of sub-array to read (default: read all); a value of NA will read the corresponding dimension entirely; counts are relative to the step size (requires sf >= 1.0-9)}

\item{step}{integer; step size for each dimension (pixels) of sub-array to read; defaults to 1 for each dimension (requires sf >= 1.0-9)}

\item{proxy}{logical; return proxy object?}

\item{debug}{logical; print debug info?}

\item{bounds}{logical or character: if \code{TRUE} tries to infer from "bounds" attribute; if character, 
named vector of the form \code{c(longitude="lon_bnds", latitude="lat_bnds")} with names dimension names}

\item{curvilinear}{control reading curvilinear (geolocation) coordinate arrays; if \code{NA} try reading the x/y dimension names; if character, defines the arrays to read; if \code{FALSE} do not try; see also \link{read_stars}}

\item{normalize_path}{logical; if \code{FALSE}, suppress a call to \link{normalizePath} on \code{filename}}

\item{x}{stars object}

\item{driver}{character; driver name}

\item{root_group_options}{character; driver specific options regarding the creation of the root group}

\item{as_float}{logical; if \code{TRUE} write 4-byte floating point numbers, if \code{FALSE} write 8-byte doubles}
}
\description{
Read or write data using GDAL's multidimensional array API
}
\details{
it is assumed that the first two dimensions are easting and northing
}
\examples{
set.seed(135)
m = matrix(runif(10), 2, 5)
names(dim(m)) = c("stations", "time")
times = as.Date("2022-05-01") + 1:5
pts = st_as_sfc(c("POINT(0 1)", "POINT(3 5)"))
s = st_as_stars(list(Precipitation = m)) |>
 st_set_dimensions(1, values = pts) |>
 st_set_dimensions(2, values = times)
nc = tempfile(fileext=".nc")
if (compareVersion(sf_extSoftVersion()["GDAL"], "3.4.0") > -1) {
  write_mdim(s, nc)
  # try ncdump on the generated file
  print(read_mdim(nc))
}
}
\seealso{
\link[sf]{gdal_utils}, in particular util \code{mdiminfo} to query properties of a file or data source containing arrays
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{contour.stars}
\alias{contour.stars}
\title{plot contours of a stars object}
\usage{
\method{contour}{stars}(x, ...)
}
\arguments{
\item{x}{object of class \code{stars}}

\item{...}{other parameters passed on to \link[graphics]{contour}}
}
\description{
plot contours of a stars object
}
\details{
this uses the R internal contour algorithm, which (by default) plots contours; \link{st_contour} uses the GDAL contour algorithm that returns contours as simple features.
}
\examples{
d = st_dimensions(x = 1:ncol(volcano), y = 1:nrow(volcano))
r = st_as_stars(t(volcano))
r = st_set_dimensions(r, 1, offset = 0, delta = 1)
r = st_set_dimensions(r, 2, offset = 0, delta = -1)
plot(r, reset = FALSE)
contour(r, add = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ops.R
\docType{methods}
\name{\%in\%,stars-method}
\alias{\%in\%,stars-method}
\title{evaluate whether cube values are in a given set}
\usage{
\S4method{\%in\%}{stars}(x, table)
}
\arguments{
\item{x}{data cube value}

\item{table}{values of the set}
}
\description{
evaluate whether cube values are in a given set
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ncdf.R
\name{read_ncdf}
\alias{read_ncdf}
\title{Read NetCDF into stars object}
\usage{
read_ncdf(
  .x,
  ...,
  var = NULL,
  ncsub = NULL,
  curvilinear = character(0),
  eps = sqrt(.Machine$double.eps),
  ignore_bounds = FALSE,
  make_time = TRUE,
  make_units = TRUE,
  proxy = NULL,
  downsample = 0
)
}
\arguments{
\item{.x}{NetCDF file or source as a character vector or an nc_proxy object.}

\item{...}{ignored}

\item{var}{variable name or names (they must be on matching grids)}

\item{ncsub}{matrix of start, count columns (see Details)}

\item{curvilinear}{length two character named vector with names of variables holding
longitude and latitude values for all raster cells. `stars` attempts to figure out appropriate
curvilinear coordinates if they are not supplied.}

\item{eps}{numeric; dimension value increases are considered identical when they differ less than \code{eps}}

\item{ignore_bounds}{logical; should bounds values for dimensions, if present, be ignored?}

\item{make_time}{if \code{TRUE} (the default), an attempt is made to provide
a CFTime class for any "time" dimension.}

\item{make_units}{if \code{TRUE} (the default), an attempt is made to set the units property of each variable}

\item{proxy}{logical; if \code{TRUE}, an object of class \code{stars_proxy} is read which contains array
metadata only; if \code{FALSE} the full array data is read in memory. If not set, defaults to \code{TRUE}
when the number of cells to be read is larger than \code{options(stars.n_proxy)}, or to 1e8 if that option was not set.}

\item{downsample}{integer; number of cells to omit between samples along each dimension. 
e.g. \code{c(1,1,2)} would return every other cell in x and y and every third cell
in the third dimension (z or t). If 0, no downsampling is applied. Note that this transformation
is applied AFTER NetCDF data are read using st_downsample. As such, if proxy=TRUE, this 
option is ignored.}
}
\description{
Read data from a file (or source) using the NetCDF library directly.
}
\details{
The following logic is applied to coordinates. If any coordinate axes have
regularly spaced coordinate variables they are reduced to the
offset/delta form with 'affine = c(0, 0)', otherwise the values of the coordinates
are stored and used to define a rectilinear grid.

If the data has two or more dimensions and the first two are regular
they are nominated as the 'raster' for plotting.

If the \code{curvilinear} argument is used it specifies the 2D arrays
containing coordinate values for the first two dimensions of the data read. It is currently
assumed that the coordinates are 2D and that they relate to the first two dimensions in
that order.

Any time dimension in the source is captured in a 'CFTime' object, unless the
\code{make_time} argument is false. In that case, the numeric offsets 
representing time are returned instead.


If \code{var} is not set the first set of variables on a shared grid is used.

\code{start} and \code{count} columns of ncsub must correspond to the variable dimension (nrows)
and be valid index using \code{\link[RNetCDF]{var.get.nc}} convention (start is 1-based). If the count value
is \code{NA} then all steps are included. Axis order must match that of the variable/s being read.
}
\examples{
f <- system.file("nc/reduced.nc", package = "stars")
if (require(ncmeta, quietly = TRUE)) {
 read_ncdf(f)
 read_ncdf(f, var = c("anom"))
 read_ncdf(f, ncsub = cbind(start = c(1, 1, 1, 1), count = c(10, 12, 1, 1)))
}

\donttest{
if (require(ncmeta, quietly = TRUE)) {
 #' precipitation data in a curvilinear NetCDF
 prec_file = system.file("nc/test_stageiv_xyt.nc", package = "stars")
 prec = read_ncdf(prec_file, curvilinear = c("lon", "lat"), ignore_bounds = TRUE)
}

##plot(prec) ## gives error about unique breaks
## remove NAs, zeros, and give a large number
## of breaks (used for validating in detail)
qu_0_omit = function(x, ..., n = 22) {
  x = units::drop_units(na.omit(x))
  c(0, quantile(x[x > 0], seq(0, 1, length.out = n)))
}
if (require(dplyr, quietly = TRUE)) {
  prec_slice = slice(prec, index = 17, along = "time")
  plot(prec_slice, border = NA, breaks = qu_0_omit(prec_slice[[1]]), reset = FALSE)
  nc = sf::read_sf(system.file("gpkg/nc.gpkg", package = "sf"), "nc.gpkg")
  plot(st_geometry(nc), add = TRUE, reset = FALSE, col = NA)
}
} # donttest
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stars.R
\name{st_cells}
\alias{st_cells}
\title{return the cell index corresponding to the location of a set of points}
\usage{
st_cells(x, sf)
}
\arguments{
\item{x}{object of class \code{stars}}

\item{sf}{object of class \code{sf} or \code{sfc}}
}
\description{
If the object has been cropped without normalization, then the indices return
are relative to the original uncropped extent.  See \code{\link{st_crop}}
}
\examples{
set.seed(1345)
st_bbox(L7_ETMs) |> 
  st_as_sfc() |> 
  st_sample(10) -> pts 
(x <- st_cells(L7_ETMs, pts))
# get the pixel values (first band only):
st_as_stars(L7_ETMs)[[1]][x]
# get pixel values for all bands:
st_as_stars(L7_ETMs) |> split() |> sapply(`[`, x)
# compare with st_extract():
st_as_stars(L7_ETMs) |> split() |> st_extract(pts)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sf.R
\name{st_xy2sfc}
\alias{st_xy2sfc}
\alias{st_sfc2xy}
\title{replace x y raster dimensions with simple feature geometry list (points, or polygons = rasterize) and vice versa}
\usage{
st_xy2sfc(x, as_points, ..., na.rm = TRUE)

st_sfc2xy(x, ...)
}
\arguments{
\item{x}{object of class \code{stars}, or of class \code{sf}}

\item{as_points}{logical; if \code{TRUE}, generate points at cell centers, else generate polygons}

\item{...}{for `st_xy2sfc`: arguments passed on to \code{st_as_sfc}, for `st_sfc2xy` arguments passed on to \link{st_as_stars.data.frame}}

\item{na.rm}{logical; omit (remove) cells which are entirely missing valued (across other dimensions)?}
}
\value{
`st_xy2sfc` returns an object of class \code{stars} with x and y raster dimensions replaced by a single sfc geometry list column containing either points, or polygons. Adjacent cells with identical values are not merged; see \code{st_rasterize} for this.

`st_sfc2xy` returns an object of class \code{stars} with a POINT list replaced by X and Y raster dimensions. This only works when the points are distributed over a regular or rectilinear grid.
}
\description{
replace x y raster dimensions with simple feature geometry list (points, or polygons = rasterize) and vice versa
}
\details{
`st_xy2sfc` replaces x y raster dimensions with simple feature geometry list (points, or polygons = rasterize)

`st_sfc2xy` replaces POINT simple feature geometry list with an x y raster
}
\examples{
(reduced_nc = read_ncdf(system.file("nc/reduced.nc", package = "stars")))
(x = stars::st_xy2sfc(reduced_nc, as_points = TRUE, na.rm = FALSE))
# roundtrip:
st_sfc2xy(x, dims=c("X", "Y","zlev","time"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sample.R
\name{st_downsample}
\alias{st_downsample}
\alias{st_downsample.stars}
\alias{st_downsample.stars_proxy}
\title{downsample stars or stars_proxy objects}
\usage{
st_downsample(x, n, ...)

\method{st_downsample}{stars}(x, n, ..., offset = 0, FUN)

\method{st_downsample}{stars_proxy}(x, n, ...)
}
\arguments{
\item{x}{object of class stars or stars_proxy}

\item{n}{integer; for each dimension the number of pixels/lines/bands etc that will be skipped; see Details.}

\item{...}{arguments passed on to \code{FUN} (e.g., \code{na.rm = TRUE} to ignore missing values if FUN is \code{mean})}

\item{offset}{integer; offset(s) for downsampling, in pixels, starting at the offset of 
each dimension; should be smaller or equal to \code{n}}

\item{FUN}{function; if given, downsampling will apply FUN to each of the the subtiles}
}
\description{
downsample a stars or stars_proxy object either by skipping rows, columns and bands,
or by computing a single value (e.g. the mean) from the sub-tiles involved
}
\details{
If all n == 0, no downsampling takes place; if it is 1, every second row/column/band
is skipped, if it is 2, every second+third row/column/band are skipped, etc.

Downsampling a \code{stars_proxy} object returns a \code{stars} object, is
equivalent to calling \code{st_as_stars(x, downsample = 2)}, and only downsamples
the first two (x and y) dimensions.

Downsampled regular rasters keep their dimension offsets, have a cell size (delta) that
is n[i]+1 times larger, and may result in a (slightly) different extent.

Note that terra's \link[terra]{aggregate} with \code{fact=2} corresponds to
\code{st_downsample(x, n = 1, FUN = mean)}: \code{fact} is one larger than \code{n}.
}
\examples{
(m = matrix(1:121, 11, 11))
(s = st_as_stars(m))
st_downsample(s, 1)
st_downsample(s, 1)[[1]]
st_downsample(s, 1, offset = 1)
st_downsample(s, 1, offset = 1)[[1]]
st_downsample(s, 1, offset = c(0,1))
st_downsample(s, 1, offset = c(0,1))[[1]]
st_downsample(s, 1, FUN = mean)
st_downsample(s, 1, FUN = mean)[[1]]
st_downsample(s, 1, offset = 1, FUN = mean)
st_downsample(s, 1, offset = c(0,1), FUN = mean)[[1]]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aggregate.R
\name{aggregate.stars}
\alias{aggregate.stars}
\alias{aggregate}
\title{spatially or temporally aggregate stars object}
\usage{
\method{aggregate}{stars}(
  x,
  by,
  FUN,
  ...,
  drop = FALSE,
  join = st_intersects,
  as_points = any(st_dimension(by) == 2, na.rm = TRUE),
  rightmost.closed = FALSE,
  left.open = FALSE,
  exact = FALSE,
  weights = NULL,
  transform = NULL
)
}
\arguments{
\item{x}{object of class \code{stars} with information to be aggregated}

\item{by}{object of class \code{sf} or \code{sfc} for spatial aggregation, for temporal aggregation a vector with time values (\code{Date}, \code{POSIXct}, or \code{CFTime}) that is interpreted as a sequence of left-closed, right-open time intervals or a string like "months", "5 days" or the like (see \link{cut.POSIXt}), or a function that cuts time into intervals; if by is an object of class \code{stars}, it is converted to sfc by \code{st_as_sfc(by, as_points = FALSE)} thus ignoring its time component. Note: each pixel is assigned to only a single group (in the order the groups occur) so non-overlapping spatial features and temporal windows are recommended.}

\item{FUN}{aggregation function, such as \code{mean}}

\item{...}{arguments passed on to \code{FUN}, such as \code{na.rm=TRUE}}

\item{drop}{logical; ignored}

\item{join}{function; function used to find matches of \code{x} to \code{by}}

\item{as_points}{see \link[stars]{st_as_sf}: shall raster pixels be taken as points, or small square polygons?}

\item{rightmost.closed}{see \link{findInterval}}

\item{left.open}{logical; used for time intervals, see \link{findInterval} and \link{cut.POSIXt}}

\item{exact}{logical; if \code{TRUE}, use \link[exactextractr]{coverage_fraction} to compute exact overlap fractions of polygons with raster cells}

\item{weights}{single-layer \code{SpatRaster} or \code{stars} object on the same grid as \code{x} with secondary per-cell weights, e.g. population or cropland area. Only used when \code{exact = TRUE}, and cells with \code{NA} weight are treated as zero weight. Not supported for \code{stars_proxy} objects}

\item{transform}{function or one-sided formula, evaluated by \code{rlang::as_function}, applied to cell values before aggregation; e.g. \code{~ .x^2} or \code{~ pmax(0, .x - 10) - pmax(0, .x - 30)}. Units on \code{x} are dropped when set. Note: any transform that derives parameters from the values it is given is applied separately to each attribute (and to each chunk for \code{stars_proxy} objects), so its results may not be comparable across attributes or chunks, e.g. \code{splines::bs(.x)} places knots at quantiles of its input, so fix such parameters explicitly, as in \code{splines::bs(.x, knots = ..., Boundary.knots = ...)}}
}
\description{
spatially or temporally aggregate stars object, returning a data cube with lower spatial or temporal resolution
}
\examples{
# aggregate time dimension in format Date
tif = system.file("tif/L7_ETMs.tif", package = "stars")
t1 = as.Date("2018-07-31")
x = read_stars(c(tif, tif, tif, tif), along = list(time = c(t1, t1+1, t1+2, t1+3)))[,1:30,1:30]
st_get_dimension_values(x, "time")
x_agg_time = aggregate(x, by = t1 + c(0, 2, 4), FUN = max) 

# aggregate time dimension in format Date - interval
by_t = "2 days"
x_agg_time2 = aggregate(x, by = by_t, FUN = max) 
st_get_dimension_values(x_agg_time2, "time")
#TBD:
#x_agg_time - x_agg_time2

# aggregate time dimension in format POSIXct
x = st_set_dimensions(x, 4, values = as.POSIXct(c("2018-07-31", 
                                                  "2018-08-01", 
                                                  "2018-08-02", 
                                                  "2018-08-03")), 
                      names = "time")
by_t = as.POSIXct(c("2018-07-31", "2018-08-02"))
x_agg_posix = aggregate(x, by = by_t, FUN = max)
st_get_dimension_values(x_agg_posix, "time")
#TBD:
# x_agg_time - x_agg_posix
aggregate(x, "2 days", mean)
if (require(ncmeta, quietly = TRUE)) {
 # Spatial aggregation, see https://github.com/r-spatial/stars/issues/299
 prec_file = system.file("nc/test_stageiv_xyt.nc", package = "stars")
 prec = read_ncdf(prec_file, curvilinear = c("lon", "lat"))
 prec_slice = dplyr::slice(prec, index = 17, along = "time")
 nc = sf::read_sf(system.file("gpkg/nc.gpkg", package = "sf"), "nc.gpkg")
 nc = st_transform(nc, st_crs(prec_slice))
 agg = aggregate(prec_slice, st_geometry(nc), mean)
 plot(agg)
}

# example of using a function for "by": aggregate by month-of-year
d = c(10, 10, 150)
a = array(rnorm(prod(d)), d) # pure noise
times = Sys.Date() + seq(1, 2000, length.out = d[3])
m = as.numeric(format(times, "\%m"))
signal = rep(sin(m / 12 * pi), each = prod(d[1:2])) # yearly period
s = (st_as_stars(a) + signal) |>
      st_set_dimensions(3, values = times)
f = function(x, format = "\%B") {
	  months = format(as.Date(paste0("01-", 1:12, "-1970")), format)
	  factor(format(x, format), levels = months)
}
agg = aggregate(s, f, mean)
plot(agg)

\donttest{
# exact = TRUE with secondary weights: population-weighted mean
# population density per municipality
if (requireNamespace("exactextractr", quietly = TRUE) &&
    requireNamespace("terra", quietly = TRUE)) {
  dens = read_stars(system.file("sao_miguel/gpw_v411_2020_density_2020.tif", 
     package = "exactextractr"))
  pop = read_stars(system.file("sao_miguel/gpw_v411_2020_count_2020.tif", 
     package = "exactextractr"))
  conc = sf::read_sf(system.file("sao_miguel/concelhos.gpkg", 
     package = "exactextractr"))
  aggregate(dens, conc, mean, exact = TRUE, weights = pop, na.rm = TRUE)
}
} # donttest
}
\seealso{
\link[sf]{aggregate}, \link[sf]{st_interpolate_aw}, \link{st_extract}, https://github.com/r-spatial/stars/issues/317
}
